Require Import Machine.
Require Import world_model.
Require Import msl.msl_standard.
Require Import WorldLogic.
Require Import SfLib.
Require Import LibTactics.
Require Import ContextSemantics.
Require Import StepStarCs.

Require Import MSetPositive.
Require Import FMapPositive.
Require Import MSetFacts.
Require Import FMapFacts.
Module PSetFacts := WFactsOn PositiveSet.E PositiveSet.
Module PMapFacts := FMapFacts.WFacts_fun PositiveMap.E PositiveMap.

(*Reserved Notation "HGM '|-h' {{ P }} t {{ Q }}" (no associativity, at level 90, t at level 0).*)

Section Definitions.
  Import Machine.Notation.
  Import world_model.Notation.
  Open Scope forces_pred.
  Open Scope entails_pred.

  Definition htriple (P:pred world) (t:term) (Q:pred world) :=
    frame_halted_assert P t Q.


(*


  Definition env_typing := PositiveSet.t.
  Definition ch_typing := PositiveMap.t (value->pred world).

  Definition extend_env (G:env_typing) (x:var) : env_typing := 
    PositiveSet.add x G.
  Fixpoint extend_ch (H:ch_typing) (c:cname) (R:value->pred world) : ch_typing :=
    PositiveMap.add c R H.

  (* WTF Coq? Inductive notation is buggy as hell. Should be able to use the same notation... *)
  Reserved Notation "HGM '|-h' << P >> t << Q >>" (no associativity, at level 90, t at level 0).

  Definition env_In := PositiveSet.In.
  Definition ch_In := @PositiveMap.In (value->pred world).

  Definition locals_In (x:ident) (l:locals) :=
    LocalsMap.map_share x l <> Share.bot /\ LocalsMap.map_val x l <> None.

  Definition expr_free_vars (e:expr) : env_typing.
    admit.
  Defined.

  Lemma expr_free_vars_correct: forall e w,
     ((exists v, eval e v w) -> forall x, env_In x (expr_free_vars e) -> locals_In x (w_env w))
  /\ (forall x, env_In x (expr_free_vars e) -> locals_In x (w_env w)) -> exists v, eval e v w.
  Admitted.

  Definition well_formed_expr (G: env_typing) (e:expr) : Prop :=
    PositiveSet.Subset (expr_free_vars e) G.

  Lemma well_formed_expr_correct: forall G e,
    well_formed_expr G e ->
    forall w, ((forall x, env_In x G -> locals_In x (w_env w)) -> exists v, eval e v w).
  Proof.
    intros.
    apply expr_free_vars_correct.
    split; intros; firstorder.
  Qed.
*)

End Definitions.

Delimit Scope htriple_scope with htriple.

Module Notation.
  Notation "  '|-h' {{ P }} t {{ Q }}" := (htriple P%pred t%term Q%pred)
   (no associativity, at level 90, t at level 0) : htriple_scope.

  Module U.
    Notation "├ᴴ ⦃ P ⦄ t ⦃ Q ⦄" := (htriple P%pred t%term Q%pred)
     (no associativity, at level 89, t at level 0, format " '├ᴴ' ⦃ P ⦄  t  ⦃ Q ⦄") : htriple_scope.
  End U.
End Notation.

(*
Module Morphisms.
  Require Import Setoid.
  Require Export Morphisms.
  Export SetoidTactics.
  Require Export SetoidClass.
  Import WorldLogic.Tactics.
  Import Bisimulation.Morphisms.

  Add Parametric Morphism : (htriple)
  with signature equiv ==> equiv ==> equiv
  as union_m.
    intros; destruct H; destruct H0; split. 
    apply or_elim.
    apply or_intro1; auto.
    apply or_intro2; auto.
    apply or_elim.
    apply or_intro1; auto.
    apply or_intro2; auto.
  Qed.

(*
  Add Morphism htriple_soundness : htriple_m.
    split; intros.
    eapply hConsequence; eauto; try firstorder firstorder.
    setoid_symmetry; auto.

    eapply hConsequence; eauto; try firstorder firstorder.
  Qed.
*)

  Add Morphism ch_tm_typing_valid : ch_tm_typing_valid_m.
  Admitted.    

  Add Morphism ch_wtm_typing_valid : ch_wtm_typing_valid_m.
    intros b x y H. induction b.
    simpl; intuition; [setoid_rewrite<- H | setoid_rewrite H]; auto.
    simpl; intuition; first [apply IHb | setoid_rewrite<- H | setoid_rewrite H]; auto.
    simpl; intuition; first [apply IHb1 | apply IHb2]; auto.
  Qed.

  Add Morphism ch_ctx_typing_valid : ch_ctx_typing_valid_m.
    intros C x y H. induction  C.
    simpl; split; auto.
    simpl; intuition; first [apply IHC | setoid_rewrite<- H | setoid_rewrite H]; auto.
    simpl; intuition; first [apply IHC | setoid_rewrite<- H | setoid_rewrite H]; auto.
    simpl; intuition; first [apply IHC | setoid_rewrite<- H | setoid_rewrite H]; auto.
  Qed.

(*  Add Morphism world_plus with signature equiv ==> equiv ==> equiv  as world_plus_m.*)
  Add Morphism well_formed : well_formed_m.
    unfold well_formed; intros.
    split; intros.
    destruct H1 as [ww[?[?[??]]]].
    exists ww; intuition.
    setoid_rewrite H2. setoid_rewrite H0. solvew.
    setoid_rewrite<- H; auto.
    destruct H1 as [ww[?[?[??]]]].
    exists ww; intuition.
    setoid_rewrite H2. setoid_rewrite H0. solvew.
    setoid_rewrite H; auto.
  Qed.


End Morphisms.
*)

Section WellFormed.
  Import WorldLogic.
  Import Machine.Notation.
  Import world_model.Notation.
  Import Notation.
  Import ContextSemantics.Notation.
  Import WorldLogic.Morphisms.
  Import WorldLogic.Notation.
  Import WorldLogic.Notation.U.
  Import WorldLogic.Tactics.
  Import channel_sa.
  Open Scope join_scope.
  Open Scope htriple_scope.
  Open Scope forces_pred.
  Open Scope entails_pred.
  Open Scope context_scope.

  Definition total_channels (ch:channel_sa.channels cname) := forall c,
    match ch c with
    | Some (exist ce _) => ce_share ce=Share.top
    | None => True
    end.
  Definition total_env (l: locals) :=
    forall x, LocalsMap.map_share x l = Share.top \/ LocalsMap.map_share x l = Share.bot.
  Definition total_heap (h: heap) :=
    forall x, MemMap.map_share x h = Share.top \/ MemMap.map_share x h = Share.bot.

  Definition total_world (w:world) : Prop :=
     total_channels (w_prod w)
  /\ total_channels (w_cons w)
  /\ total_env (w_env w)
  /\ total_heap (w_heap w)
  /\ forall c, w_prod w c=None <-> w_cons w c=None.

  Lemma total_channel_incr: forall c ch,
    total_channels ch <-> total_channels (channels_incr c ch).
  Proof.
    unfold total_channels, channels_incr; split; intros.
    destruct (cname_eq c c0). subst c0.
    specialize (H c).
    case_eq (ch c); intros.
    rewrite H0 in H.
    destruct p as [[[p n]]].
    simpl in *.
    destruct (cname_eq c c); intuition.
    simpl.
    destruct (dec_share_identity p); auto.
    simpl.
    rewrite H0 in *; auto.
    destruct (ch c) as [[[[p' n']]] | ].
    simpl in *.
    destruct (cname_eq c c0). intuition.
    specialize (H c0).
    destruct (ch c0) as [[[[p'' n'']]] | ]; auto.
    simpl.
    specialize (H c0).
    destruct (ch c0) as [[[[p'' n'']]] | ]; auto.

    destruct (cname_eq c c0). subst c0.
    specialize (H c).
    case_eq (ch c); intros; auto.
    rewrite H0 in H.
    destruct p as [[[p n]]].
    simpl in *.
    destruct (cname_eq c c); intuition.
    simpl in H.
    destruct (dec_share_identity p); auto.
    specialize (H c0).
    destruct (ch c) as [[[[p' n']]] | ]; auto.
    simpl in H.
    destruct (cname_eq c c0); intuition.
  Qed.
  Lemma channels_incr_None: forall c c' ch,
    ch c' = None <-> channels_incr c ch c' = None.
  Proof.
    unfold channels_incr. intros c c' ch.
    case_eq (ch c). intros p ?.
    destruct p. simpl. destruct (cname_eq c c'). subst c'.
    rewrite H. split; intros; discriminate.
    split; auto.
    simpl; split; auto.
  Qed.
  Lemma total_world_prod_incr: forall c w,
    total_world w <-> total_world (w_prod_incr c w).
  Proof.
    unfold total_world; split; intros;
      intuition; destruct w; simpl in *; auto.
    apply total_channel_incr; auto. auto.
    apply H4.
    apply channels_incr_None in H3; auto.
    apply channels_incr_None; apply H4; auto.
    apply total_channel_incr in H0; auto.
    apply H4. apply channels_incr_None; auto.
    eapply channels_incr_None; apply H4; auto.
  Qed.
  Lemma total_world_cons_incr: forall c w,
    total_world w <-> total_world (w_cons_incr c w).
  Proof.
    unfold total_world; split; intros;
      intuition; destruct w; simpl in *; auto.
    apply total_channel_incr; auto.
    apply channels_incr_None; apply H4; auto.
    apply H4. eapply channels_incr_None; eauto.
    apply total_channel_incr in H; auto.
    eapply channels_incr_None; apply H4; auto.
    apply H4. apply channels_incr_None; auto.
  Qed.
  Lemma total_world_new_channel: forall c w H1 H2,
    total_world w ->
    total_world (w_add_new_channel c w H1 H2).
  Proof.
    unfold total_world, total_channels, w_add_new_channel, cons_channel; split; intros.
    simpl in *. intuition.
    rewrite w_prod_wupd_prod.
    destruct (positive_eq_dec c c0). subst c0.
    clear H0 H. destruct (dec_share_identity Share.top); auto.
    apply H0.
    rewrite w_prod_wupd_prod.
    rewrite w_cons_wupd_prod.
    rewrite w_cons_wupd_cons.
    rewrite w_env_wupd_prod.
    rewrite w_env_wupd_cons.
    rewrite w_heap_wupd_prod.
    rewrite w_heap_wupd_cons.
    intuition.
    simpl.
    destruct (positive_eq_dec c c0).
    destruct (dec_share_identity Share.top); auto.
    apply H.
    simpl in *.
    destruct (positive_eq_dec c c0). subst c0.
    destruct (dec_share_identity Share.top); auto.
    discriminate.
    apply H6; auto.
    simpl in *.
    destruct (positive_eq_dec c c0). subst c0.
    destruct (dec_share_identity Share.top); auto.
    discriminate.
    apply H6; auto.
  Qed.
  Lemma total_world_upd_var: forall x v w,
    total_world w <-> total_world (w_env_set w x v).
  Proof.
  Admitted.

  Lemma can_recv_upd_var: forall w x v c,
    can_recv c w <-> can_recv c (w_env_set w x v).
  Proof.
  Admitted.

(*
  Lemma total_world_preservation_step: forall ww s b s' b' C,
    total_world ww ->
    ww ||= world_state s (+) world_context C (+) world_wterm b ->
    step s b s' b' ->
    exists ww',
      ww' ||= s' (+) C (+) b' /\
      total_world ww'.
  Proof.
    intros.
    induction H1; subst; try solve [exists ww; intuition].

    (* st_send *)
    exists (w_prod_incr c ww); intuition.
    apply world_join_equiv in H6.
    simplw.
    apply transitivity with (w_prod_incr c w_w' (+) (s (+) C)).
    apply world_prod_incr; auto.
    rewrite H0.
    solvew.
    rewrite H6. solvew.
    apply total_world_prod_incr; auto.

    (* st_recv *)
    exists (w_cons_incr c (wupd_var ww x v)); intuition.
    apply world_join_equiv in H4.
    simplw.
    patternw (w_cons_incr c (wupd_var w_w' x v)).
    apply world_cons_incr.
    apply can_recv_upd_var.
    destruct H2 as [p[n[w1[w2[?[??]]]]]].
    apply world_join_equiv in H2.
    rewrite H2 in H4.
    joinablew_ex w2 w' H4.
    destruct H8 as [w2'].
    exists p; exists n; exists w1; exists w2'.
    intuition.
    apply world_join_equiv.
    rewrite H4. rewrite H8. solvew.
    simplw.
    apply world_upd_var.
    apply world_join_equiv in H4.
    clear -H1 H4.
    destruct H4 as [_ [? _]]. unfold fst,snd in H.
    apply LocalsMap.map_join_char with (a:=x) in H.
    destruct H as [? _].
    rewrite H1 in H.
    apply join_com in H.
    destruct H.
    rewrite<- H0.
    apply Share.lub_top.
    rewrite H0.
    rewrite H4.
    transitivity (w (+) C (+) (s_pop_ch s c (+) w')).
    rewrite world_plus_state_pop; eauto.
    solvew.
    solvew.
    apply total_world_cons_incr; auto.
    apply total_world_upd_var; auto.

    (* st_newch *)
Admitted.

assert (w_prod ww c = None).
clear -H5 H1 H0.
destruct H0 as [?[? _]].
specialize (H c).
specialize (H0 c).
specialize (H5 c).
rewrite H1 in H5.
case_eq (w_prod ww c); intros; auto.
elimtype False.
rewrite H2 in H.
destruct p as [[[p n] Hid1] Hid2].
simpl in H. subst p.
simpl in *.
destruct x as .
apply H5.

    eapply well_formed_state_push; eauto.


    exists ()
  Admitted.
*)

  Lemma total_world_preservation_cstep: forall ww s C b s' C' b',
    total_world ww ->
    World ww == Join_state s (x) Join_context C (x) Join_wterm b ->
    cstep s C b s' C' b' ->
    exists ww',
      World ww' == Join_state s' (x) C' (x) b' /\
      total_world ww'.
  Proof.
  Admitted.

  Lemma total_world_s_iter: forall s s',
    s_iter s s' ->
    Join_state s == Join_state s'.
Admitted.

  Lemma total_world_preservation: forall ww s C b s' C' b',
    total_world ww ->
    World ww == Join_state s (x) Join_context C (x) Join_wterm b ->
    cstep_star s C b s' C' b' ->
    exists ww',
      World ww' == Join_state s' (x) C' (x) b' /\
      total_world ww'.
  Proof.
    intros.
    revert ww H0 H.
    induction H1; intros.
    exists ww; auto.
    edestruct total_world_preservation_cstep as [ww'[??]]; eauto.
    setoid_rewrite total_world_s_iter in H0; eauto.
  Qed.

  Lemma total_worlds_dec (C:context) (b:wterm) : {w | total_world w /\ (World w==C (x) b)}+{forall w, ~total_world w \/ ~(World w == C (x) b)}.
  Admitted.


  Import List.
  Definition resource_invariant_holds (R:value->pred world) (wcl:list world) (q:list value) :=
    Forall2 (fun w v => w |= R v) wcl q.

  Definition well_formed_state (s:state) (w:world) := forall c m n,
    pendpoint (w_prod w c) Share.top m ->
    pendpoint (w_cons w c) Share.top n ->
    match s_get_ch s c with
      | Some (R,wcl,q) => m >= n /\ length q = m-n /\ length wcl = m-n /\ resource_invariant_holds R wcl q
      | None => False
      end.

(*
  Fixpoint ch_tm_typing_valid (t:term) (H:ch_typing) : Prop := match t with
    | tm_send c e => ch_In c H
    | tm_recv x c => ch_In c H
    | tm_endch c => ch_In c H
    | tm_while_max e z t' => ch_tm_typing_valid t' H
    | tm_if e t1 t2 => (ch_tm_typing_valid t1 H) /\ (ch_tm_typing_valid t2 H)
    | tm_seq t1 t2 => (ch_tm_typing_valid t1 H) /\ (ch_tm_typing_valid t2 H)
    | tm_par P1 t1 P2 t2 => (ch_tm_typing_valid t1 H) /\ (ch_tm_typing_valid t2 H)
    | tm_newch c RI t => (ch_tm_typing_valid t H)
    | tm_skip => True
    | tm_assert _ => True
    end.

  Fixpoint ch_wtm_typing_valid (t:wterm) (H:ch_typing) : Prop := match t with
    | wtm_term _ t => ch_tm_typing_valid t H
    | wtm_seq b1 t2 => (ch_wtm_typing_valid b1 H) /\ (ch_tm_typing_valid t2 H)
    | wtm_par b1 b2 => (ch_wtm_typing_valid b1 H) /\ (ch_wtm_typing_valid b2 H)
    end.

  Fixpoint ch_ctx_typing_valid (C:context) (H:ch_typing) := match C with
    | cx_hole => True
    | cx_seq C1 t2 => (ch_ctx_typing_valid C1 H) /\ (ch_tm_typing_valid t2 H)
    | cx_par1 C1 b2 => (ch_ctx_typing_valid C1 H) /\ (ch_wtm_typing_valid b2 H)
    | cx_par2 b1 C2 => (ch_wtm_typing_valid b1 H) /\ (ch_ctx_typing_valid C2 H)
  end.

  Lemma ch_typing_noninterference: forall C b (H:ch_typing),
    ch_ctx_typing_valid C H ->
    ch_wtm_typing_valid b H ->
    forall s s' C' b',
      cstep_star s C b s' C' b' ->
      (forall c, ~ch_In c H -> s_get_ch s c = s_get_ch s' c).
  Admitted.

  Lemma ch_typing_noninterference': forall C b (H:ch_typing),
    ch_ctx_typing_valid C H ->
    ch_wtm_typing_valid b H ->
    forall s s' s'' C' b',
      cstep_star s C b s' C' b' ->
      (forall c, ch_In c H -> s_get_ch s'' c = s_get_ch s c) ->
      exists s''',
        cstep_star s'' C b s''' C' b' /\
        (forall c, ch_In c H -> s_get_ch s''' c = s_get_ch s' c).
  Admitted.

  Lemma ch_typing_preservation1: forall s C b s' C' b' (H:ch_typing),
    ch_ctx_typing_valid C H ->
    cstep_star s C b s' C' b' ->
    ch_ctx_typing_valid C' H.
  Admitted.

  Lemma ch_typing_preservation2: forall s C b s' C' b' (H:ch_typing),
    ch_wtm_typing_valid b H ->
    cstep_star s C b s' C' b' ->
    ch_wtm_typing_valid b' H.
  Admitted.

  Lemma ch_ctx_typing_composed: forall (H:ch_typing) C1 C2,
    ch_ctx_typing_valid C1 H ->
    ch_ctx_typing_valid C2 H ->
    ch_ctx_typing_valid (compose_contexts C1 C2) H.
  Proof.
    intros; induction C1; destruct C2;
      try solve [simpl in *; intuition].
  Qed.

  Lemma ch_ctx_wtm_typing_compose: forall (H:ch_typing) C b,
    ch_ctx_typing_valid C H ->
    ch_wtm_typing_valid b H ->
    ch_wtm_typing_valid C[b] H.
  Proof.
    intros; induction C;
      try solve [simpl in *; intuition].
  Qed.


  Lemma ch_typing_decompose1: forall (H:ch_typing) C b,
    ch_wtm_typing_valid C[b] H ->
    ch_ctx_typing_valid C H.
  Proof.
Admitted.

  Lemma ch_typing_decompose2: forall (H:ch_typing) C b,
    ch_wtm_typing_valid C[b] H ->
    ch_wtm_typing_valid b H.
  Proof.
Admitted.

  Lemma ch_ctx_typing_composed_split: forall H C1 C2,
    ch_ctx_typing_valid (compose_contexts C1 C2) H <->
    (ch_ctx_typing_valid C1 H /\ ch_ctx_typing_valid C2 H).
  Proof. intros; induction C1; simpl in *; intuition. Qed.

  Definition well_formed (H:ch_typing) (s:state) (C:context) (X:JoinPred) :=
    exists ww, total_world ww /\ World ww == s (x) C (x) X
    /\ ch_ctx_typing_valid C H (* context C only uses channels in H *)
    /\ well_formed_state s ww. (* the total send/recv endpoint counts correlate with the number of items in each channel's queue *)



Lemma prod_prod_incr_diff: forall c c' w,
  c <> c' ->
  w_prod (w_prod_incr c w) c' = w_prod w c'.
Proof. intros; destruct w; simpl. apply channels_incr_diff; auto. Qed.

Lemma cons_cons_incr_diff: forall c c' w,
  c <> c' ->
  w_cons (w_cons_incr c w) c' = w_cons w c'.
Proof. intros; destruct w; simpl. apply channels_incr_diff; auto. Qed.

  Lemma well_formed_state_push: forall ww s c R w wl v q,
    w |= R v ->
    s_get_ch s c = Some (R, wl, q) ->
    well_formed_state s ww ->
    well_formed_state (s_push_ch s c w v) (w_prod_incr c ww).
  Proof.
    unfold well_formed_state in *; intros.
    rewrite cons_prod_incr in H3.
    destruct (cname_eq c c0). subst c0.
    apply pendpoint_prod_decr in H2.
    destruct H2.
    specialize (H1 c (m-1) n H2 H3).
    rewrite H0 in H1.
    erewrite s_ch_get_push_same; eauto.
    destruct H1 as [?[?[??]]].
    split. clear -H1. omega.
    split. rewrite app_length. simpl. rewrite H5. clear -H4 H1. omega.
    split. rewrite app_length. simpl. rewrite H6. clear -H4 H1. omega.
    apply Forall2_app; auto.
    rewrite s_ch_get_push_diff; auto.
    apply H1; auto.
    rewrite prod_prod_incr_diff in H2; auto.
  Qed.


  Lemma well_formed_state_pop: forall ww s c R w wl v q,
    s_get_ch s c = Some (R, w :: wl, v :: q) ->
    well_formed_state s ww ->
    well_formed_state (s_pop_ch s c) (w_cons_incr c ww).
  Proof.
    unfold well_formed_state; intros.
    rewrite prod_cons_incr in H1.
    destruct (cname_eq c c0). subst c0.
    apply pendpoint_cons_decr in H2.
    destruct H2.
    specialize (H0 c m (n-1) H1 H2).
    rewrite H in H0.
    destruct H0 as [?[?[??]]].
    erewrite s_ch_get_pop_same; eauto.
    simpl in *.
    split. clear -H4 H0 H3 H5. omega.
    split. clear -H4 H0 H3 H5. omega.
    split. clear -H4 H0 H3 H5. omega.
    inversion H6; auto.
    rewrite s_ch_get_pop_diff; auto.
    apply H0; auto.
    rewrite cons_cons_incr_diff in H2; auto.
  Qed.
    
  Lemma well_formed_state_upd_var: forall s ww x v,
    well_formed_state s ww ->
    well_formed_state s (wupd_var ww x v).
  Proof.
  Admitted.


  Lemma well_formed_state_new_channel: forall c w s R (H1: w_prod w c = None) (H2: w_cons w c = None),
    well_formed_state s w ->
    well_formed_state (s_new_ch s c R) (w_add_new_channel c w H1 H2).
  Proof.
    unfold well_formed_state; intros.
    destruct (cname_eq c c0). subst c0.
    rewrite s_ch_get_newch_same.
    simpl in *.
    unfold cons_channel in *.
    simpl in *.    
    destruct (positive_eq_dec c c).
    destruct (dec_share_identity Share.top).
      elimtype False; apply top_share_nonidentity; auto.
    simpl in *.
    destruct H0 as [_ ?]; destruct H3 as [_ ?]. subst m n.
    intuition.
    constructor.
    intuition.
    simpl in *.
    unfold cons_channel in *.
    destruct (positive_eq_dec c c0). subst. intuition.
    rewrite s_ch_get_newch_diff.
    apply H; auto.
    auto.
  Qed.

Lemma join_endpoint_cons_channel: forall c ch (H: ch c = None),
  join (just_endpoint c (ce_make_top 0)) ch (cons_channel c (ce_make_top 0) ch H).
Proof.
  simpl. intros. intro c'.
  unfold just_endpoint, cons_channel.
  destruct (positive_eq_dec c c'). subst c'.
  simpl.
  destruct (dec_share_identity Share.top); auto.
  elimtype False; apply top_share_nonidentity; auto.
  generalize (eq_refl ch).
  generalize ch at 1 3.
  intro ch'; intros.
  assert (ch' c = None). subst; auto.
  rewrite H1.
  replace (cons_channel_obligation_1 (ce_make_top 0) ch H eq_refl n) with (just_endpoint_obligation_1 (ce_make_top 0) (eq_refl c) n).
  replace (@cons_channel_obligation_2 c (ce_make_top 0) ch H c
             (@eq_refl positive c) n) with (@just_endpoint_obligation_2 c (ce_make_top 0) c (@eq_refl positive c) n).
  apply lj_None2.
  extensionality.
  simpl in *.
  elimtype False; apply top_share_nonidentity; apply endpoint_share_identity in x; auto.
  extensionality.
  simpl in *.
  elimtype False; apply top_share_nonidentity; auto.
  constructor.
Qed.

Lemma  world_add_new_channel: forall c X z (H1: w_prod z c = None) (H2: w_cons z c = None),
  World z == X -> World (w_add_new_channel c z H1 H2) == (new_channel c) (x) X.
Proof.
  intros.
  apply Join_equiv_equiv'.
  unfold w_add_new_channel, new_channel.
  split; repeat intro.
  exists (new_channel c); exists z; intuition.
  simpl in H0.
  subst w; unfold new_channel.
  destruct z.
  split; try split; unfold fst, snd; intuition.
  apply join_endpoint_cons_channel.
  apply join_endpoint_cons_channel.
  simpl.
  apply LocalsMap.empty_map_join.
  apply MemMap.empty_map_join.
  reflexivity.
  apply Forces_Forces'_equiv.
  apply Forces_equiv; auto.

  simpl.
  destruct H0 as [a1[a2[?[??]]]].
  simpl in H3.
  subst a1.
  replace a2 with z in *.
  Focus 2.
  apply Forces_eq with X; auto.
  apply Forces_equiv; auto.
  apply Forces_Forces'_equiv; auto.

  destruct w.
  destruct z as [z_p z_c z_e z_h].
  f_equal.
  destruct H0 as [[? _]_]. simpl in H0.
  extensionality c'.
  specialize (H0 c').
  simpl in H0.

  inversion H0; subst; clear H0.
  unfold just_endpoint in H5.
  unfold cons_channel.
  destruct (positive_eq_dec c c'); auto.
  simpl in *.
  destruct (dec_share_identity Share.top).
  elimtype False; apply top_share_nonidentity; auto.
  discriminate.
  unfold just_endpoint in *.
  unfold cons_channel.
  destruct (positive_eq_dec c c'); auto. subst c'.
  simpl in *.
  destruct (dec_share_identity Share.top).
  elimtype False; apply top_share_nonidentity; auto.
  f_equal. 
  apply exist_ext.
  apply exist_ext.
  auto.
  unfold just_endpoint, cons_channel in *.
  destruct (positive_eq_dec c c'); auto. subst c'.
  simpl in *.
  destruct (dec_share_identity Share.top).
  elimtype False; apply top_share_nonidentity; auto.
Admitted.

  Lemma s_iter_channels_eq: forall s s',
    s_iter s s' -> s_c s = s_c s'.
  Proof.
    intros [?] [?]; intros.
    inverts H.
    reflexivity.
  Qed.

  Lemma s_iter_world_equiv: forall s s',
    s_iter s s' -> Join_equiv s s'.
  Proof.
    intros.
    apply Join_equiv_equiv'.
    apply s_iter_channels_eq in H.
    split; intro; simpl; intros; subst; unfold join_state in *.
    rewrite<- H; auto.
    rewrite H; auto.
  Qed.  

  Lemma channels_eq_s_get_ch: forall s s',
    s_c s = s_c s' -> s_get_ch s = s_get_ch s'.
  Proof. unfold s_get_ch; intros; rewrite H; auto. Qed.

  Lemma channels_eq_well_formed_state: forall s s' w,
    s_c s = s_c s' -> well_formed_state s w -> well_formed_state s' w.
  Proof.
    unfold well_formed_state; intros.
    rewrite channels_eq_s_get_ch with s' s; auto.
    apply H0; auto.
  Qed.

  Lemma well_formed_preservation_step: forall s C (b:wterm) s' b' H,
    well_formed H s C b-> step s b s' b' -> well_formed H s' C b'.
  Proof.
    unfold well_formed; intros.
    setoid_rewrite<- Join_gwemp2 at 2 in H0.
    setoid_rewrite<- Join_gwemp2 at 2.
    remember (World gwemp) as X.
    clear HeqX.
    assert (s (x) C (x) b (x) X == s (x) C (x) b (x) X). reflexivity.
    assert (s' (x) C (x) b' (x) X == s' (x) C (x) b' (x) X). reflexivity.
    revert H3.
    generalize (s' (x) C (x) b' (x) X) at 1 3 as Y.
    intro Y.
    revert H2 H0.
    generalize (s (x) C (x) b (x) X) at 1 3 as Z.
    revert X Y.
    step_cases (induction H1) Case; intros X Y Z HZ HH HY; subst;
      try solve [assert (H5: Y==Z); [solve [setoid_rewrite HZ; setoid_rewrite HY; solvew] |];
      setoid_rewrite H5 in HY; setoid_rewrite H5;
      destruct HH as [ww[?[?[??]]]]; exists ww; splits; auto].

    Case "st_send".
    destruct HH as [ww[?[?[??]]]].
    exists (w_prod_incr c ww); intuition.
    apply total_world_prod_incr; auto.
    unfold gjoin in *.
    apply Join_join in H5.
    transitivity ((w_prod_incr c w_w') (x) (s (x) C (x) X)).
    apply Join_prod_incr; auto.
    rewrite H6.
    rewrite HZ; solvew.
    rewrite H5. rewrite HY; solvew.
    eapply well_formed_state_push; eauto.

    Case "st_recv".
    destruct HH as [ww[?[?[??]]]].
    exists (w_cons_incr c (wupd_var ww x v)); intuition.
    apply total_world_cons_incr; auto.
    apply total_world_upd_var; auto.
    apply Join_join in H3.
    rewrite HY.
    simplw.
    patternw (w_cons_incr c (wupd_var w_w' x v)).
    apply Join_cons_incr.
    apply can_recv_upd_var. auto.
    apply can_recv_sub with w; auto.
    exists w'. eapply Join_join; auto.

    apply Join_upd_var.
    apply Join_join in H3.
    clear -H0 H3.
    destruct H3 as [_ [? _]]. unfold fst,snd in H.
    apply LocalsMap.map_join_char with (a:=x) in H.
    destruct H as [? _].
    rewrite H0 in H.
    apply join_com in H.
    destruct H.
    rewrite<- H1.
    apply Share.lub_top.
    rewrite H6.
    rewrite H3.
    transitivity (w (x) C (x) (s_pop_ch s c (x) w' (x) X)).
    rewrite Join_state_pop; eauto.
    rewrite HZ; solvew.
    solvew.
    eapply well_formed_state_pop; eauto.
    eapply well_formed_state_upd_var; eauto.


    Case "st_newch".
    destruct HH as [ww[?[?[??]]]].
    assert (w_prod ww c = None /\ w_cons ww c = None).
      clear -H5 H1 H0.
      destruct H1 as [?[?[_[_ ?]]]].
      specialize (H c).
      specialize (H1 c).
      specialize (H5 c).
      rewrite H0 in H5.
      case_eq (w_prod ww c); intros; auto.
      elimtype False.
      case_eq (w_cons ww c); intros.
      rewrite H3 in H. rewrite H4 in H1.
      destruct p as [[[p n]]].
      destruct p0 as [[[p' n']]].
      simpl in *. subst p p'.
      apply (H5 n n').
      rewrite H3. simpl; auto.
      rewrite H4. simpl; auto.
      apply H2 in H4. rewrite H3 in H4. discriminate.
      split; auto. apply H2; auto.
    destruct H6.
    exists (w_add_new_channel c ww H6 H7).
    intuition.
    apply total_world_new_channel; auto.
    rewrite HY. simplw.
    apply Join_join in H2.
    rewrite H2.
    patternw (new_channel c).
    apply world_add_new_channel.
    rewrite H3. rewrite HZ. solvew.
    apply well_formed_state_new_channel; auto.

    Case "st_endch".
admit.

    Case "st_par".
    apply Join_join in H2.
    assert (H5: Y==Z).
      setoid_rewrite HZ; setoid_rewrite HY; simplw; setoid_rewrite H2; solvew.
    setoid_rewrite H5 in HY; setoid_rewrite H5.
    auto.

    Case "st_seq1".
    simplw in HZ; simplw in HY.
    apply IHstep with (X:=X) (Z:=Z); auto.

    Case "st_par1".
    simplw in HZ; simplw in HY.
    apply IHstep with (X:=b2 (x) X) (Z:=Z); auto.
    rewrite HZ; solvew.
    rewrite HY; solvew.

    Case "st_par2".
    simplw in HZ; simplw in HY.
    apply IHstep with (X:=b1 (x) X) (Z:=Z); auto.
    rewrite HZ; solvew.
    rewrite HY; solvew.

    Case "st_par_join".
    simplw in HZ; simplw in HY.
    apply Join_join in H0.
    setoid_rewrite H0 in HY.
    setoid_rewrite HY.
    simplw in HY; simplw in HZ.
    setoid_replace Z with Y in HH.
    setoid_rewrite HY in HH. simplw.
    apply HH.
    setoid_rewrite HY.
    setoid_rewrite HZ.
    solvew.
  Qed.


  Lemma well_formed_preservation_cstep: forall s C (b:wterm) s' C' b' H,
    well_formed H s C b-> cstep s C b s' C' b' -> well_formed H s' C' b'.
  Proof.
    unfold well_formed; intros.
    destruct H0 as [ww[?[?[??]]]].
    replace C with (compose_contexts cx_hole C) in H3.
      2: simpl; auto.
    replace C with (compose_contexts cx_hole C) in H2.
      2: simpl; auto.
    replace C' with (compose_contexts cx_hole C').
      2: simpl; auto.
    remember cx_hole as CC.
    clear HeqCC.
    revert CC H2 H3.
    induction H1; intros.

    destruct IHcstep with CC as [ww'[?[?[??]]]]; auto.
    rewrite H2. solvew.
    apply ch_ctx_typing_composed_split in H3.
    apply ch_ctx_typing_composed_split.
    intuition. apply H6.
    exists ww'; do 2 (split; auto).
    rewrite H6. solvew.
    split; auto.
    apply ch_ctx_typing_composed_split in H7.
    apply ch_ctx_typing_composed_split in H3.
    apply ch_ctx_typing_composed_split.
    simpl in *; intuition.

    destruct IHcstep with (cx_par1 CC b2) as [ww'[?[?[??]]]]; auto.
    rewrite H2. solvew.
    apply ch_ctx_typing_composed_split in H3.
    apply ch_ctx_typing_composed_split.
    simpl in *; intuition.
    exists ww'; do 2 (split; auto).
    rewrite H6. solvew.
    split; auto.
    apply ch_ctx_typing_composed_split in H7.
    apply ch_ctx_typing_composed_split in H3.
    apply ch_ctx_typing_composed_split.
    simpl in *; intuition.

    destruct IHcstep with (cx_par2 b1 CC) as [ww'[?[?[??]]]]; auto.
    rewrite H2. solvew.
    apply ch_ctx_typing_composed_split in H3.
    apply ch_ctx_typing_composed_split.
    simpl in *; intuition.
    exists ww'; do 2 (split; auto).
    rewrite H6. solvew.
    split; auto.
    apply ch_ctx_typing_composed_split in H7.
    apply ch_ctx_typing_composed_split in H3.
    apply ch_ctx_typing_composed_split.
    simpl in *; intuition.

(*
Check ch_wtm_typing_valid.
assert (exists H', ch_wtm_typing_valid t1 H')
    assert (well_formed (H) s (compose_contexts CC (cx_par1 C t1)) t2)
ch_wtm_typing_valid
*)
Admitted.

  Lemma well_formed_preservation_s_iter: forall s C (b:wterm) s' H,
    well_formed H s C b-> s_iter s s' -> well_formed H s' C b.
  Proof.
Admitted.


  Lemma well_formed_preservation: forall s C (b:wterm) s' C' b' H,
    well_formed H s C b-> cstep_star s C b s' C' b' -> well_formed H s' C' b'.
  Proof.
    intros. induction H1; auto.
    apply IHcstep_star; eapply well_formed_preservation_cstep; eauto.
    apply IHcstep_star; eapply well_formed_preservation_s_iter; eauto.
  Qed.

(* Lemma well_formed_seq: forall s C b t,
  well_formed D s C b -> well_formed D s C (b;t).
Admitted. *)


Lemma well_formed_state_nonempty: forall s w w1 w2 c (n m:nat) p,
  well_formed_state s w ->
  gjoin w1 w2 w ->
  has_endpoint c Share.top n (w_cons w2) ->
  has_endpoint c p m (w_prod w2) ->
  m >= n ->
  match s_get_ch s c with
    | Some (_,wcl,q) => length q >= m-n /\ length wcl >= m-n
    | None => False
    end.
Admitted.


*)
End WellFormed.


Module POrdProps := FMapFacts.OrdProperties PositiveMap.

Section Soundness.
  Import Machine.Notation.
  Import world_model.Notation.
  Import Notation.
  Import WorldLogic.Notation.
  Open Scope join_scope.
  Open Scope forces_pred.
  Open Scope htriple_scope.

  Set Implicit Arguments.

(*
  Definition exists_fresh_ch (H:ch_typing) : { c | ~PositiveMap.In c H}.
    case_eq (POrdProps.max_elt H); intros.
    destruct p.
    eapply POrdProps.max_elt_Above in H0.
    exists (k~1)%positive.
    intro.
    assert (PositiveMap.In (k~1)%positive (PositiveMap.remove k H)).
      apply PMapFacts.remove_in_iff.
      split; auto.
      clear -k; intro H; revert H; induction k; intros; inversion H; auto.
    apply H0 in H2.
    apply POrdProps.ME.lt_le in H2.
    apply H2.
    clear -k.
    induction k; simpl; auto.
    apply POrdProps.max_elt_Empty in H0.
    exists xH.
    intro.
    destruct H1 as [v].
    apply H0 in H1.
    auto.
  Qed.

  Definition env_typing_correct (w:world) (G:env_typing) :=
    forall x, env_In x G -> LocalsMap.map_share x (w_env w) = Share.top.

  Definition htriple_soundness (DG:prod ch_typing env_typing) P t Q :=
    forall F w w' s C s' C',
      well_formed (fst DG) s C (w:t)%wterm ->
      env_typing_correct w (snd DG) ->
      w |= gsep_con P F ->
      cstep_star s C (w:t) s' C' (w':tm_skip) ->
      (env_typing_correct w' (snd DG) /\ w' |= gsep_con Q F).

  Lemma htriple_typing: forall H G P t Q,
    htriple_soundness (H,G) P t Q -> ch_tm_typing_valid t H.
  Admitted.
*)

(*
  Theorem htriple_sound: forall H G P t Q,
    (H,G) |-h {{P}} t {{Q}} -> htriple_soundness (H,G) P t Q /\ ch_tm_typing_valid t H.
Admitted.
*)

(*
  Theorem htriple_preservation: forall H G P Q s C w t s' C' w' b',
      htriple_soundness (H,G) P t Q ->
      well_formed H s C (w:t)%wterm ->
      cstep_star s C (w:t) s' C' b' ->
      w' ||= b' ->
      htriple_soundness (H,G) (exactly w') (just_term b') Q.
  Proof.
  Admitted.

  Lemma env_typing_preservation: forall D G w P t Q s C s' C' b' w',
    env_typing_correct w G -> w |= P ->
    htriple_soundness (D,G) P t Q ->
    cstep_star s C (w:t) s' C' b' ->
    w' ||= b' ->
    env_typing_correct w' G.
Admitted.
*)

End Soundness.


Section Lemmas.
  Import Machine.Notation.
  Import world_model.Notation.
  Import world_model.Notation.U.
  Import WorldLogic.Notation.
  Import WorldLogic.Morphisms.
  Import WorldLogic.Tactics.
  Import Notation.
  Open Scope htriple_scope.
  Open Scope forces_pred.
  Open Scope entails_pred.
  Open Scope pred.
  Open Scope join_scope.


  Lemma halted_assert_pred_rewrite: forall s C b Q Q',
    Q'==Q ->
    halted_assert s C b Q ->
    halted_assert s C b Q'.
  Proof.
    unfold halted_assert.
    intros.
    apply H.
    eapply H0; eauto.
  Qed.

  Lemma htriple_pred_rewrite: forall P P' t Q Q',
    P'==P -> Q'==Q ->
    htriple P t Q ->
    htriple P' t Q'.
  Proof.
    unfold htriple,frame_halted_assert.
    intros.
    eapply halted_assert_pred_rewrite; eauto.
    setoid_rewrite H0.
    reflexivity.
    apply H1.
    setoid_rewrite<- H.
    auto.
  Qed.

  Lemma H_Frame: forall P t Q F,
    frame_halted_assert P t Q ->
    frame_halted_assert (gsep_con P F) t (gsep_con Q F).
  Proof.
    unfold frame_halted_assert.
    intros.
    eapply halted_assert_pred_rewrite.
    setoid_rewrite<- gstar_assoc.
    reflexivity.
    eapply H.
    setoid_rewrite gstar_assoc.
    auto.
  Qed.

  Lemma H_Consequence: forall P P' Q Q' t,
    P' |-- P -> Q |-- Q' ->
    frame_halted_assert P t Q ->
    frame_halted_assert P' t Q'.
  Proof.
    unfold frame_halted_assert, halted_assert.
    intros.
    eapply gstar_intro; eauto.
    eapply H1; eauto.
    eapply gstar_intro; eauto.
  Qed.


(*  
  Lemma htriple_simpl: forall H G P t Q w w' s C s' C',
    (H,G) |-h {{P}} t {{Q}} ->
    well_formed H s C (w:t)%wterm ->
    env_typing_correct w G ->
    w |= P ->
    cstep_star s C (w:t) s' C' (w':tm_skip) ->
    (env_typing_correct w' G /\ w' |= Q).
  Proof.
    intros.
    setoid_rewrite<- gstar_unit2.
    eapply H0; eauto.
    rewrite gstar_unit2.
    auto.
  Qed.
*)
(*
  Lemma htriple_skip_same_pred: forall H G P1 P2,
    (H,G) |-h {{P1}} tm_skip {{P2}} -> P1 |-- P2.
  Proof.
    intros. 
    remember (tm_sterm tm_skip) as t.
    setoid_rewrite<- star_unit2.
    remember emp as F. clear HeqF.
    revert F Heqt.
    induction H0; intros; subst; try discriminate.
    setoid_rewrite<- star_assoc.
    apply IHhtriple; auto.
    apply derives_cut with (P*F)%pred;
    auto with seplog_hints;
    apply derives_cut with (Q*F)%pred;
    auto with seplog_hints.
    auto with seplog_hints.
  Qed.

  Lemma htriple_seq_inversion: forall H G P1 P3 t1 t2,
    (H,G) |-h {{P1}} tm_seq t1 t2 {{P3}} ->
    exists P2,
      (H,G) |-h {{P1}} t1 {{P2}} /\ ((H,G) |-h {{P2}} t2 {{P3}}).
  Proof.
    intros.
    remember (tm_seq t1 t2) as t.
    revert t1 t2 Heqt.
    induction H0; intros; subst; try discriminate; subst.
    edestruct IHhtriple as [P2[??]]; eauto;
      exists (P2*F); split; constructor; auto.
    edestruct IHhtriple as [P2[??]]; eauto.
      exists (P2); split. econstructor; eauto; apply refl.
      econstructor; eauto; apply refl.
    inversion Heqt; subst t0 t3.
    exists P2; auto.
  Qed.
*)


(*
  Lemma htriple_post_true: forall P t,
    htriple P t pTrue.
  Admitted. *)

End Lemmas.


