Require Import Machine.
(*Require Import world_model.*)
Require Import msl.base.
Require Import SfLib.
Require Import LibTactics.
Import State.PMapFacts.
Require Import BinPos.

(*Require Import Omega.*)

Section Definitions.
  Import Machine.Notation.
  Open Scope wterm_scope.


  Fixpoint term_free_iter_var (t:term) (i:positive) : Prop :=
    match t with
    | tm_while_max _ (i_stall j _ _) t1 => i <> j /\ term_free_iter_var t1 i
    | tm_while_max e z t1 => term_free_iter_var t1 i
    | tm_if e t1 t2 => term_free_iter_var t1 i /\ term_free_iter_var t2 i
    | tm_seq t1 t2 => term_free_iter_var t1 i /\ term_free_iter_var t2 i
    | tm_par P1 t1 P2 t2 => term_free_iter_var t1 i /\ term_free_iter_var t2 i
    | tm_local e f => forall x, term_free_iter_var (f x) i
    | _ => True
    end.

  Fixpoint wterm_free_iter_var (b:wterm) (i:positive) : Prop :=
    match b with
    | w:t => term_free_iter_var t i
    | b1 || b2 => wterm_free_iter_var b1 i /\ wterm_free_iter_var b2 i
    | b1;; t2 => wterm_free_iter_var b1 i /\ term_free_iter_var t2 i
    end.

  Definition ss_iter_star (s s': state) :=
    s_iter_star (supd_iter empty_state (s_i s)) (supd_iter empty_state (s_i s')).

End Definitions.

Section Lemmas.
  Import Notation.
  Import Machine.Notation.

  Lemma s_iter_snoc: forall s s' s'',
    s_iter_star s s' -> s_iter s' s'' -> s_iter_star s s''.
  Proof.
    intros.
    induction H.
    eapply s_iter_cons; eauto. apply s_iter_nil.
    eapply s_iter_cons; eauto.
  Qed.

  Lemma ss_iter_cons: forall s s' s'',
    s_iter s s' ->
    ss_iter_star s' s'' ->
    ss_iter_star s s''.
  Proof.
    intros.
    eapply s_iter_cons; eauto.
    destruct s; destruct s'.
    inverts H.
    simpl in *.
    replace (makeState empty_ch_state (FMapPositive.PositiveMap.add i iter_go s_i) ResSet.empty) with
      (s_iter_set (makeState empty_ch_state s_i ResSet.empty) i iter_go).
    constructor.
    unfold s_iter_get in *. simpl; auto.
    reflexivity.
  Qed.

  Lemma ss_iter_snoc: forall s s' s'',
    ss_iter_star s s' ->
    s_iter s' s'' ->
    ss_iter_star s s''.
  Proof.
    intros.
    eapply s_iter_snoc; eauto.
    destruct s; destruct s'.
    inverts H0.
    simpl in *.
    replace ({|s_c:=empty_ch_state;s_i:=(FMapPositive.PositiveMap.add i iter_go s_i0);s_all_res:=ResSet.empty|}) with
      (s_iter_set (makeState empty_ch_state s_i0 ResSet.empty) i iter_go).
    constructor.
    unfold s_iter_get in *. simpl; auto.
    reflexivity.
  Qed.

  Lemma step_s_iter: forall s b s' b',
    step s b s' b' -> s_i s = s_i s'.
  Proof.
    intros.
    induction H; subst; auto.
    rewrite s_push_ch_respects_s_i; reflexivity.
    rewrite s_pop_ch_respects_s_i; reflexivity.
    rewrite s_new_ch_respects_s_i; reflexivity.
    rewrite s_rm_ch_respects_s_i; reflexivity.
  Qed.


  Lemma iter_decrs_stall1: forall i n z1 z',
    iteration_decrs (i_stall i n z1) z' -> exists n', z'=i_stall i n' z1 /\ n>=n'.
  Proof.
    intros.
    remember (i_stall i n z1) as z.
    revert Heqz.
    induction H; intros; subst; eauto.
    destruct IHiteration_decrs as [n'[??]]; auto.
    subst z'.
    inverts H0.
    exists n0; split; auto. omega.
  Qed.

  Lemma iter_decrs_stall2: forall i z (n':nat) z1,
    iteration_decrs z (i_stall i n' z1) -> exists (n:nat), z=i_stall i n z1 /\ n>=n'.
  Proof.
    intros.
    remember (i_stall i n' z1) as z'.
    revert n' Heqz'.
    induction H; intros; subst; eauto.
    inverts H0.
    edestruct IHiteration_decrs as [n[??]]; eauto.
    subst z.
    exists n; split; auto. omega.
  Qed.


Ltac eq_unspec f H := match goal with
 [ H: ?a = ?b |- _ ] => let H' := fresh "H" in
    assert (H': f a = f b); [f_equal; apply H | clear H; rename H' into H]
end.

Ltac eq_spec H x := match goal with
 [ H: ?a = ?b |- _ ] => let H' := fresh "H" in
    assert (H': a x = b x); [rewrite H; reflexivity | clear H; rename H' into H]
end.

  Lemma step_respecting_s_i: forall s b s' b',
    step s b s' b' -> s_i s = s_i s'.
  Proof.
    intros.
    induction H; subst; auto.
    symmetry; apply s_push_ch_respects_s_i.
    symmetry; apply s_pop_ch_respects_s_i.
    symmetry; apply s_new_ch_respects_s_i.
    symmetry; apply s_rm_ch_respects_s_i.
  Qed.

  Lemma step_s_iter_conflict: forall s b s' b',
    s_iter s s' -> step s b s' b' -> False.
  Proof.
    intros.
    apply step_respecting_s_i in H0.
    eapply s_iter_change_s_i; eauto.
  Qed.


  Lemma s_iter_go_preservation: forall s s' i,
    s_iter_get s i = Some iter_go ->
    s_iter s s' ->
    s_iter_get s' i = Some iter_go.
  Proof. intros. inverts H0. destruct (Pos.eq_dec i0 i); subst. false; rewrite H1 in H; discriminate. rewrite s_iter_get_set_neq; auto. Qed.

  Lemma step_go_preservation: forall s b s' b' i,
    s_iter_get s i = Some iter_go ->
    step s b s' b' ->
    s_iter_get s' i = Some iter_go.
  Proof.
    intros.
    unfold s_iter_get in *.
    erewrite<- step_s_iter; eauto.
  Qed.

  Lemma iteration_decr_i_plus: forall (z1 z1' z2: iterations),
    iteration_decr z1 z1' ->
    iteration_decr (i_plus z1 z2) (i_plus z1' z2).
  Proof.
    intros.
    inverts H.
    simpl; destruct z2; constructor.
    constructor.
  Qed.

  Lemma FMapPositive_add_add_eq: forall i t (s: FMapPositive.PositiveMap.t t) n n',
    FMapPositive.PositiveMap.add i n (FMapPositive.PositiveMap.add i n' s) = FMapPositive.PositiveMap.add i n s.
  Proof. 
    intros.
    revert i.
    induction s using FMapPositive.PositiveMap.tree_ind; intros.
    induction i; simpl; try rewrite IHi; reflexivity.
    induction i; simpl; try rewrite IHs2; try rewrite IHs1; reflexivity.
  Qed.

  Lemma s_iter_star_s_c: forall sch1 sch2 s s' wS1 wS2,
    s_iter_star (makeState sch1 s wS1) (makeState sch1 s' wS1) ->
    s_iter_star (makeState sch2 s wS2) (makeState sch2 s' wS2).
  Proof.
    intros.
    remember (makeState sch1 s wS1) as ss.
    remember (makeState sch1 s' wS1) as ss'.
    revert s s' sch1 sch2 wS1 wS2 Heqss Heqss'.
    induction H; intros; subst.
    inverts Heqss'.
    apply s_iter_nil.
    inverts H.
    unfold s_iter_get, s_iter_set in *; simpl in *.
    eapply s_iter_cons; eauto.
    eapply s_iter_go'; eauto.
  Qed.

  Lemma step_elim_free_iter_var: forall s i n b ss' b',
    step (s_iter_set s i n) b ss' b' ->
    wterm_free_iter_var b i ->
    exists s', ss' = s_iter_set s' i n /\ step s b s' b'.
  Proof.
    intros.
    remember (s_iter_set s i n) as ss.
    revert s Heqss.
    step_cases (induction H) Case HH; intros; try subst;
      first [Case "st_newch" (* internal error on eauto *)
       | try solve [exists s; split; eauto]].

    simpl in H0. destruct H0.
    rewrite s_iter_get_set_neq in H; auto.
    exists s; split; auto.

    rewrite s_get_ch_iter_set in H2.
    exists (s_push_ch s c w' v); split; eauto.
    apply s_push_ch_iter_set.

    rewrite s_get_ch_iter_set in H6.
    exists (s_pop_ch s c); split; eauto.
    apply s_pop_ch_iter_set.

    rewrite s_get_ch_iter_set in H3.
    exists s; split; eauto.

    exists (s_new_ch s (fresh_channel s w) R); split.
    rewrite fresh_channel_iter_set.
    apply s_new_ch_iter_set.
    rewrite fresh_channel_iter_set.
    eapply HH with (c:=FreshChannels.fc_c (fresh_channel (s_iter_set s i n) w)); eauto.
    rewrite fresh_channel_iter_set; reflexivity.
    rewrite fresh_channel_iter_set; reflexivity.

    rewrite s_get_ch_iter_set in H3.
    exists (s_rm_ch s c); split; eauto.
    apply s_rm_ch_iter_set.

    edestruct IHstep as [s''[??]]; eauto. apply H0.
    edestruct IHstep as [s''[??]]; eauto. apply H0.
    edestruct IHstep as [s''[??]]; eauto. apply H0.
  Qed.

  (* standard variable freshness that is a pain to prove *)
  Definition s_iter_free_vars (s:state) : {free_vars | forall i, In i free_vars <-> s_iter_get s i <> None}. admit. Qed.
  Definition s_iter_term_free_vars (t:term) : {free_vars | forall i, ~In i free_vars <-> term_free_iter_var t i}.
    induction t; try solve [exists (@nil positive); simpl; split; intro; auto].

    destruct IHt as [free_vars H].
    destruct i.
    eexists (p::free_vars); simpl; split; intro.
    intuition. eapply H; auto.
    intuition. eapply H; eauto.
    eexists free_vars; simpl; split; intro; apply H; auto.

    destruct IHt1 as [free_vars1 H1].
    destruct IHt2 as [free_vars2 H2].
    eexists (free_vars1++free_vars2); simpl; split; intro.
      rewrite in_app_iff in H; split; [apply H1 | apply H2]; intuition.
      destruct H as [H3 H4]; apply H1 in H3; apply H2 in H4; rewrite in_app_iff; intuition.

    rename t into f.
(* need to place restrictions on what f can do! this is morally true, but technically false. *)
admit. 

    destruct IHt1 as [free_vars1 H1].
    destruct IHt2 as [free_vars2 H2].
    eexists (free_vars1++free_vars2); simpl; split; intro.
      rewrite in_app_iff in H; split; [apply H1 | apply H2]; intuition.
      destruct H as [H3 H4]; apply H1 in H3; apply H2 in H4; rewrite in_app_iff; intuition.

    destruct IHt1 as [free_vars1 H1].
    destruct IHt2 as [free_vars2 H2].
    eexists (free_vars1++free_vars2); simpl; split; intro.
      rewrite in_app_iff in H; split; [apply H1 | apply H2]; intuition.
      destruct H as [H3 H4]; apply H1 in H3; apply H2 in H4; rewrite in_app_iff; intuition.
  Defined.

  Definition s_iter_exists_fresh (s:state) : {i | s_iter_get s i = None}. admit. Qed. 
  Definition s_iter_wterm_wterm_exists_fresh (s:state) (b1 b2:wterm): {i | s_iter_get s i = None /\ wterm_free_iter_var b1 i /\ wterm_free_iter_var b2 i}. admit. Qed. 

  Lemma ss_iter_step_cons : forall s b s' b' s'',
    step s b s' b' ->
    ss_iter_star s' s'' ->
    ss_iter_star s s''.
  Proof.
    intros.
    unfold ss_iter_star in *.
    rewrite (step_s_iter s b s' b') in *; eauto.
  Qed.

  Lemma s_iter_set_eq: forall s s' i n,
    ss_iter_star s s' ->
    s_iter_set s i n = s_iter_set s' i n -> s_iter_star s s'.
  Proof.
    intros.
    unfold ss_iter_star in H.
    unfold s_iter_set, s_iter_get in *.
    inverts H0.
    destruct s; destruct s'; simpl in *.
    remember (makeState empty_ch_state s_i ResSet.empty) as ss.
    remember (makeState empty_ch_state s_i0 ResSet.empty) as ss'.
    revert s_i s_i0 Heqss Heqss' H2.
    induction H; intros; subst.
    inverts Heqss'; auto. inverts H2. apply s_iter_nil.
    inverts H. inverts H2.
    unfold s_iter_get in *. simpl in *.
    eapply s_iter_cons.
    constructor. unfold s_iter_get. simpl; eauto.
    unfold s_iter_set in *. simpl in *.
    destruct (Pos.eq_dec i0 i); subst.
    apply IHs_iter_star; auto.
    f_equal.
    rewrite<- H4.
    apply FMapPositive_add_add_eq.
    eapply s_iter_star_s_c; eauto.
  Qed.

  Lemma s_iter_star_go_preservation: forall s s' i,
    s_iter_get s i = Some iter_go ->
    s_iter_star s s' -> s_iter_get s' i = Some iter_go.
  Proof.
    intros. induction H0; auto. apply IHs_iter_star. eapply s_iter_go_preservation; eauto.
  Qed.

  Lemma s_iter_set_nop: forall s i a,
    s_iter_get s i = Some a ->
    s_iter_set s i a = s.
  Proof.
    intros.
    unfold s_iter_get,s_iter_set in *.
    destruct s; simpl in *.
    f_equal.
    apply FMapPositive.PositiveMapAdditionalFacts.gsident; auto.
  Qed.

  Lemma s_iter_normalize: forall i j s s',
    j <> i ->
    s_iter_get s' j = Some iter_go ->
    s_iter_star s s' ->
    s_iter_star (s_iter_set s j iter_go) s'.
  Proof.
    intros.
    induction H1; intros.
    rewrite s_iter_set_nop; eauto.
    apply s_iter_nil.
    inverts H1.
    destruct (Pos.eq_dec i0 j); subst; auto.
    spec IHs_iter_star. auto.
    eapply s_iter_cons; eauto.
    rewrite s_iter_set_set_neq'; auto.
    constructor.
    rewrite s_iter_get_set_neq; auto.
  Qed.

  Lemma s_iter_normalize': forall i j s s' s_,
    j <> i ->
    s_iter_get s' j = Some iter_go ->
    s_iter_star s s' ->
    s_ = s_iter_set s j iter_go ->
    s_iter_star s_ s'.
  Proof.
    intros; subst; eapply s_iter_normalize; eauto.
  Qed.

  Lemma s_iter_None_preservation: forall i s s',
    s_iter_get s i = None ->
    s_iter s s' ->
    s_iter_get s' i = None.
  Proof.
    intros.
    inverts H0.
    destruct (Pos.eq_dec i0 i); subst.
    rewrite H in H1; discriminate.
    rewrite s_iter_get_set_neq; auto.
  Qed.

  Lemma s_iter_star_None_preservation: forall i s s',
    s_iter_get s i = None ->
    s_iter_star s s' ->
    s_iter_get s' i = None.
  Proof.
    intros.
    induction H0; intros; auto.
    apply IHs_iter_star.
    eapply s_iter_None_preservation; eauto.
  Qed.

  Lemma s_iter_step_None_preservation: forall i s b s' b',
    s_iter_get s i = None ->
    step s b s' b' ->
    s_iter_get s' i = None.
  Proof.
    intros.
    unfold s_iter_get in *.
    erewrite<- step_s_iter; eauto.
  Qed.

  Lemma ss_iter_star_go_preservation: forall i s s',
    s_iter_get s i = Some iter_go ->
    ss_iter_star s s' -> s_iter_get s' i = Some iter_go.
  Proof.
    intros.
    unfold ss_iter_star in *.
    remember (supd_iter empty_state (s_i s)) as ss.
    remember (supd_iter empty_state (s_i s')) as ss'.
    revert s s' Heqss Heqss' H.
    induction H0; unfold s_iter_get in *; intros; subst; simpl in *.
    inverts Heqss'. auto.
    inverts H.
    destruct (Pos.eq_dec i0 i); subst.
    false. unfold s_iter_get in *. simpl in *. rewrite H1 in H2; discriminate.
    unfold s_iter_set, s_iter_get in *. simpl in *.
    destruct s0; simpl in *.
    destruct s'0; simpl in *.
    spec IHs_iter_star (makeState empty_ch_state (FMapPositive.PositiveMap.add i0 iter_go s_i) ResSet.empty).
    spec IHs_iter_star (makeState empty_ch_state s_i0 ResSet.empty).
    spec IHs_iter_star.
    reflexivity.
    spec IHs_iter_star.
    reflexivity.
    apply IHs_iter_star.
    simpl.
    rewrite PMapFacts.add_neq_o; auto.
  Qed.

  Lemma s_iter_set_None: forall s s' n i,
    s_iter_get s i = None ->
    s_iter s s' ->
    s_iter (s_iter_set s i n) (s_iter_set s' i n).
  Proof.
    intros.
    inverts H0.
    destruct (Pos.eq_dec i i0); subst.
    false.
    rewrite s_iter_set_set_neq'; auto.
    constructor.
    rewrite s_iter_get_set_neq; auto.
  Qed.

  Lemma s_iter_star_step_eq: forall s b s' b',
    step s b s' b' -> s_iter_star s s' -> s' = s.
  Proof.
    intros.
    apply step_s_iter in H.
    destruct s'; destruct s; simpl in *; subst.
    f_equal; auto.    
    apply s_iter_star_respects_s_c in H0; auto.
    apply s_iter_star_respects_s_all_res in H0; auto.
  Qed.

  Lemma iteration_decr_iter_plus2: forall z1 z1' z2,
    iteration_decr z1 z1' ->
    iteration_decr (iter_plus2 z1 z2) (iter_plus2 z1' z2).
  Proof.
    intros.
    inverts H.
    simpl.
    destruct z2; constructor.
    constructor.
    constructor.
  Qed.



  Lemma iteration_decr_dec: forall i:iteration_expr,
    {i' | iteration_decr i i'}+{i=i_nat 0}+{xz | match xz with pair x z => i=i_stall x 0 z end}.
  Proof.
    intros.
    destruct i.
    destruct n; eauto.
    right; exists (p,i); auto. 
    left; left; eexists; constructor.
    destruct i.
    destruct n; eauto.
    left; left; eexists; constructor.
    left; left; eexists; constructor.
  Qed.


End Lemmas.