Require Import msl.msl_standard.
Require Import SfLib.
Require Import LibTactics.

Generalizable Variables A.

Section Definitions.
(*  Class sepalgd (A:Type) {sa: sepalg A} :=
    { join_dec: forall (x y:A), {z| @join _ sa x y z}+{forall z, ~@join _ sa x y z}}.
*)
  Definition sa_join_dec {A:Type} (sa: sepalg A):= forall (x y:A), {z| @join A sa x y z}+{forall z, ~@join A sa x y z}.

  Inductive JoinPred {A:Type} : Type :=
  | Join: forall (sa:sepalg A), JoinPred->JoinPred-> JoinPred
  | World : A->JoinPred
  | Pred: forall
      (f: A->Prop)
      (f_dec: forall w, {f w}+{~f w})
      (f_sat: {w | f w}+{forall w, ~f w})
      (f_fun: forall w1 w2, f w1 -> f w2 -> w1=w2),
      JoinPred.

  Fixpoint Forces' {A:Type} (c:A) (X : JoinPred) := match X with
    | Join sa X Y => exists a, exists b, @join _ sa a b c /\ Forces' a X /\ Forces' b Y
    | World a => a=c
    | Pred f _ _ _ => f c
  end.

  Inductive Forces {A:Type} (w:A) : @JoinPred A->Prop :=
   | ForcesJoin: forall sa wx wy X Y,
      Forces wx X -> Forces wy Y ->
      @join A sa wx wy w ->
      Forces w (Join sa X Y)
  | ForcesWorld: Forces w (World w)
  | ForcesPred: forall (f:A->Prop) f_dec f_sat f_fun, f w -> Forces w (Pred f f_dec f_sat f_fun).

  Lemma Forces_Forces'_equiv: forall A (w:A) X,
    Forces w X <-> Forces' w X.
  Proof.
    split; intros.
    induction H; simpl; eauto.
    revert w H; induction X; intros; simpl in *; eauto.
    destruct H as [a[b[?[??]]]]; econstructor; eauto.
    subst; constructor.
    constructor; auto.
  Qed.

  Program Definition Identity {A:Type} {sa:sepalg A} (X:@JoinPred A) : @JoinPred A := Pred (fun w => identity w /\ Forces w X)  _ _ _ .
  Next Obligation. admit. Qed.
  Next Obligation. admit. Qed.
  Next Obligation. admit. Qed.

(*  Definition Derives {A:Type} (X Y:JoinPred) : Prop := forall a:A, Forces a X -> Forces a Y.*)
  Definition Join_equiv {A:Type} (X Y: @JoinPred A) := forall w, Forces w X <-> Forces w Y.
  Definition Join_equiv' {A:Type} (X Y: @JoinPred A) := forall w, Forces' w X <-> Forces' w Y.

  Definition Joins {A:Type} (P: A->Prop) := exists w, P w.

  Definition Forces_app {A:Type} X (w:A) := Forces w X.
  Coercion Forces_app : JoinPred >-> Funclass.
  Hint Unfold Forces_app.

(*  Definition Joins {A:Type} (X: @JoinPred A) := {w:A | Forces w X}.
  Definition Joins_equiv {A:Type} (X Y: @JoinPred A) := {w:A | Forces w X /\ Forces w Y}.
*)


(*
  Lemma Forces'_dec {A:Type}: forall X, {a:A | Forces' a X}+{forall a:A, ~(Forces' a X)}.
  Proof.
    induction X; intros.
    destruct IHX1 as [[a H1] | H1].
    destruct IHX2 as [[b H2] | H2].
    destruct (@join_dec _ _ sa a b) as [[ab H12] | H12].
    left; exists ab; auto; eapply Join_elim; eauto.
    right; intros. intro.
      destruct H as [a_[b_[?[??]]]].
      replace a_ with a in *; [| eapply Forces_fun; eauto].
      replace b_ with b in *; [| eapply Forces_fun; eauto].
      apply H12 in H; auto.
    right; intros. intro.
      destruct H as [a_[b_[_[_ ?]]]].
      apply H2 in H; auto.
    right; intros. intro.
      destruct H as [a_[_[_[? _]]]].
      apply H1 in H; auto.

    left; exists a; eauto.
    auto.
  Qed.

  Lemma Forces_dec {A:Type}: forall X, {a:A | Forces a X}+{forall a:A, ~(Forces a X)}.
  Proof.
    induction X; intros.
    destruct IHX1 as [[a H1] | H1].
    destruct IHX2 as [[b H2] | H2].
    destruct (@join_dec _ _ sa a b) as [[ab H12] | H12].
    left; exists ab; auto; eapply Join_elim; eauto.
    right; intros. intro.
      destruct H as [a_[b_[?[??]]]].
      replace a_ with a in *; [| eapply Forces_fun; eauto].
      replace b_ with b in *; [| eapply Forces_fun; eauto].
      apply H12 in H; auto.
    right; intros. intro.
      destruct H as [a_[b_[_[_ ?]]]].
      apply H2 in H; auto.
    right; intros. intro.
      destruct H as [a_[_[_[? _]]]].
      apply H1 in H; auto.

    left; exists a; eauto.
    auto.
  Qed.
*)

End Definitions.

Delimit Scope join_scope with join.
Delimit Scope join_helper_scope with join_helper.
Bind Scope join_scope with join.
Bind Scope join_scope with JoinPred.

Module Notation.
  Notation "X (+) Y" := (@Join _ _ X%join Y%join) (at level 50, left associativity) : join_scope.

  Notation "X '==' Y" := (fun w => Forces w X%join /\ Forces w Y%join) (at level 70, no associativity) : join_helper_scope.
  Notation "'{[' P ']}'" := (@Joins _ P%join_helper) (at level 0, X at level 68, Y at level 69) : join_scope.

  Notation " w ||= X " := (Forces w X) (at level 70, no associativity) : join_scope.
  Module U.
    Notation "X ⨁ Y" := (@Join _ _ X%join Y%join) (at level 50, left associativity) : join_scope.
    Notation " w ⊫ X " := (Forces w X) (at level 70, no associativity) : join_scope.
  End U.

End Notation.

Module Morphisms.
  Require Export Morphisms.
  Require Export SetoidTactics.
  Require Export Setoid.
  Require Export SetoidClass.
  Import Notation.
  Open Scope join_scope.

  Hint Unfold Join_equiv.

  Program Instance JoinPredSetoid {A:Type} : Setoid (@JoinPred A) :=
      { equiv:= @Join_equiv A}.
  Solve Obligations using firstorder.

  Lemma Join_equiv_equiv': forall {A:Type} X Y,
    X == Y <-> @Join_equiv' A X Y.
  Proof.
    unfold Join_equiv, Join_equiv'; split; intros;
      split; intros; apply Forces_Forces'_equiv; apply H; apply Forces_Forces'_equiv; auto.
  Qed.


  Lemma Join_equiv_equiv {A:Type}: @Join_equiv A = @equiv _ Morphisms.JoinPredSetoid. auto. Qed.

  Add Parametric Morphism {A}: (@Forces A)
  with signature eq ==> equiv ==> equiv
  as Forces_m. firstorder. Qed.

  Add Parametric Morphism {A} {sa:sepalg A}: (@Join A sa)
  with signature equiv ==> equiv ==> equiv
  as Join_m. 
    intros; unfold equiv in *.
    apply Join_equiv_equiv'.
    apply Join_equiv_equiv' in H.
    apply Join_equiv_equiv' in H0.
    firstorder.
  Qed.

  Program Instance Joins1_m A :
    Proper (equiv ==> iff)%signature (fun X => @Joins A (fun w => w||=X)).
  Next Obligation.
    intros X Y; split; intros.
    destruct H0 as [w ?]. exists w. setoid_rewrite<- H. auto.
    destruct H0 as [w ?]. exists w. setoid_rewrite H. auto.
  Qed.

  Program Instance Joins2_m A :
    Proper (equiv ==> equiv ==> iff)%signature (fun X Y => @Joins A (fun w => w||=X /\ w||=Y)).
  Next Obligation.
    intros X1 X2 ? Y1 Y2 ?; split; intros H1; destruct H1 as [w[??]].
    exists w; split; [setoid_rewrite<- H | setoid_rewrite<- H0]; auto.
    exists w; split; [setoid_rewrite H | setoid_rewrite H0]; auto.
  Qed.

  Definition pair_equiv {A} (a b: (@JoinPred A)*(@JoinPred A)) := fst a==fst b /\ snd a==snd b.
  Program Instance JoinPredProdSetoid {A} : Setoid (@prod (@JoinPred A) (@JoinPred A)) :=
      { equiv:= @pair_equiv A}.
  Solve Obligations using firstorder.

  Program Instance Joins3_m A :
    Proper (equiv ==> iff)%signature (fun (P: (@JoinPred A)*(@JoinPred A)) => @Joins A (fun w => let (X,Y):=P in w||=X /\ w||=Y)).
  Next Obligation.
    intros [X1 X2] [Y1 Y2] [? ?]; split; intros; simpl in *.
    setoid_rewrite<- H; setoid_rewrite<- H0; auto.
    setoid_rewrite H; setoid_rewrite H0; auto.
  Qed.

End Morphisms.


Section JoinLogic.
  Import Notation.
  Import Notation.U.
  Import Morphisms.
  Open Scope join_scope.

  Hint Unfold Forces'.
  Hint Resolve @join_com.

  Lemma Forces_join: forall A (sa: sepalg A) (a b ab:A),
    @join A sa a b ab <-> ab ||= World a (+) World b.
  Proof.
    split; intros.
    apply Forces_Forces'_equiv; simpl; eauto.
    apply Forces_Forces'_equiv in H.
    destruct H as [x[y[?[??]]]]; simpl in *.
    subst; auto.
  Qed.

  Lemma Forces_eq: forall A (w1 w2:A) X,
    w1 ||= X -> w2 ||= X -> w1=w2.
  Proof.
    intros.
    revert w2 H0.
    induction H; intros.
    inverts H2.
    erewrite IHForces1 in *; eauto.
    erewrite IHForces2 in *; eauto.
    eapply join_eq; eauto.
    inverts H0; auto.
    inverts H0; eauto.
  Qed.

  Lemma Forces_join': forall A (sa: sepalg A) (a b ab: A) (X Y: JoinPred),
    a ||= X -> b ||= Y ->
    (join a b ab <-> ab ||= X (+) Y).
  Proof.
    split; intros.
    econstructor; eauto.
    inverts H1.
    replace a with wx in *.
    2: eapply Forces_eq; eauto.
    replace b with wy in *.
    2: eapply Forces_eq; eauto.
    auto.
  Qed.

  Lemma Joins1: forall A (X Y: @JoinPred A),
    {[ X==Y ]} -> {[ X ]}.
  Proof.
    intros A X Y [w[??]]; exists w; auto. 
  Qed.

  Lemma Joins2: forall A (X Y: @JoinPred A),
    {[ X==Y ]} -> {[ Y ]}.
  Proof.
    intros A X Y [w[??]]; exists w; auto. 
  Qed.

  Lemma Joins_joins: forall A (sa: sepalg A) (a b ab:A),
    @join A sa a b ab <-> {[ World ab == World a (+) World b ]}.
  Proof.
    split; intros.
    exists ab; split.
    constructor.
    apply Forces_Forces'_equiv; simpl; eauto.
    destruct H as [ab_[??]].
    inverts H.
    apply Forces_join; auto.
  Qed.

  Lemma Joins_Forces: forall A (sa: sepalg A) (w:A) X,
    w ||= X <-> {[ World w == X ]}.
  Proof.
    split; intros.
    exists w; split; auto. constructor.
    destruct H as [w_[??]].
    inverts H; auto.
  Qed.

  Lemma Joins_Join: forall A (X Y: @JoinPred A),
    {[ X==Y ]} -> X == Y.
  Proof.
    intros A X Y [w[??]]; split; intro;
      assert (w0=w); subst; auto;
      eapply Forces_eq; eauto.
  Qed.

  Lemma Join_joins': forall A (sa: sepalg A) (a b ab: A) (X Y: JoinPred),
    a ||= X -> b ||= Y ->
    (join a b ab <-> {[ World ab == X (+) Y ]}).
  Proof.
    split; intros.
    apply Joins_Forces; auto; eapply Forces_join'; eauto; constructor.
    eapply Forces_join'; eauto. eapply Joins_Forces; auto.
  Qed.

  Lemma Join_com: forall A (sa: sepalg A) (X Y: @JoinPred A),
    X (+) Y == Y (+) X.
  Proof.
    split; intros; inverts H; econstructor; eauto.
  Qed.

  Lemma Join_assoc: forall A (sa: sepalg A) (X Y Z: @JoinPred A),
    X (+) (Y (+) Z) == (X (+) Y) (+) Z.
  Proof.
    split; intros.
    inverts H. inverts H4.
    edestruct (@join_assoc _ sa) as [ab[??]];
      try (apply @join_com; eauto).
    econstructor; eauto.
    econstructor; eauto.
    inverts H. inverts H3.
    edestruct (@join_assoc _ sa) as [ab[??]]; eauto.
    econstructor; eauto.
    econstructor; eauto.
  Qed.

  Lemma Join_bot2 : forall A (sa: sepalg A) X (y:A),
    (forall a, unit_for y a) ->
    X (+) World y == X.
  Proof.
    split; intros.
    inverts H0. inverts H5.
    replace wx with w in *; auto.
    eapply join_eq; eauto.
    apply join_com.
    eapply H.
    econstructor; eauto.
    econstructor; eauto.
    apply join_com. apply H.
  Qed.

  Lemma Join_bot1 : forall A (sa: sepalg A) X (y:A),
    (forall a, unit_for y a) ->
    World y (+) X == X.
  Proof.
    intros; setoid_rewrite Join_com; apply Join_bot2; auto.
  Qed.

  Lemma Join_swap1: forall A (sa: sepalg A) X Y Z,
    ((X (+) Y) (+) Z) == ((X (+) Z) (+) Y).
  Proof.
    intros.
    setoid_rewrite<- Join_assoc at 2.
    setoid_rewrite Join_com at 4.
    setoid_rewrite Join_assoc at 1.
    reflexivity.
  Qed.

  Lemma Join_swap2: forall A (sa: sepalg A) X Y Z,
    X (+) (Y (+) Z) == Y (+) (X (+) Z).
  Proof.
    intros.
    setoid_rewrite Join_assoc at 2.
    setoid_rewrite Join_com at 4.
    setoid_rewrite<- Join_assoc at 1.
    reflexivity.
  Qed.

  Lemma Forces_equiv: forall A (c:A) X,
    World c == X <-> c ||= X.
  Proof.
    split; intros.
    destruct H with c as [? _]. apply H0. constructor.
    split; intros.
    inverts H0; auto.
    replace w with c in *.
    constructor.
    eapply Forces_eq; eauto.
  Qed.

  Lemma Forces_canc1: forall A (sa: sepalg A) w X Y Z,
    w ||= X (+) Y -> w ||= X (+) Z -> Y == Z.
  Proof.
    intros.
    inverts H. inverts H0.
    apply Forces_equiv in H3.
    setoid_rewrite<- H3 in H4.
    inverts H4.
    assert (wy0=wy); subst.
      apply join_com in H8; apply join_com in H6.
      eapply join_canc; eauto.
    apply Forces_equiv in H7.
    apply Forces_equiv in H5.
    setoid_rewrite H5 in H7.
    auto.
  Qed.

  Lemma Forces_canc2: forall A (sa: sepalg A) w X Y Z,
    w ||= X (+) Y -> w ||= Z (+) Y -> X == Z.
  Proof.
    intros.
    setoid_rewrite Join_com in H.
    setoid_rewrite Join_com in H0.
    eapply Forces_canc1; eauto.
  Qed.

  Lemma Joins_canc1: forall A (sa: sepalg A) X Y Z,
    {[X (+) Y ==  X (+) Z]} -> {[Y == Z]}.
  Proof.
    intros.
    destruct H as [w[??]].
    inverts H.
    exists wy.
    assert (Y==Z).
      eapply Forces_canc1; eauto. econstructor; eauto.
    setoid_rewrite<- H.
    split; auto.
  Qed.

  Lemma Joins_canc2: forall A (sa: sepalg A) X Y Z,
    {[X (+) Y == Z (+) Y]} -> {[X == Z]}.
  Proof.
    intros.
    setoid_rewrite Join_com in H.
    eapply Joins_canc1; eauto.
  Qed.

  Lemma Join_join: forall A (sa: sepalg A) (a b ab:A),
    @join A sa a b ab <-> World ab == World a (+) World b.
  Proof.
    split; intros.
    apply Forces_equiv; apply Forces_join; auto.
    apply Forces_join; apply Forces_equiv; auto.
  Qed.

  Lemma Join_eq: forall A (w1 w2:A) X,
    World w1 == X -> World w2 == X -> w1=w2.
  Proof.
    intros.
    apply Forces_eq with X;
      apply Forces_equiv; auto.
  Qed.

  Lemma Join_join': forall A (sa: sepalg A) (a b ab: A) (X Y: JoinPred),
    World a == X -> World b == Y ->
    (join a b ab <-> World ab == X (+) Y).
  Proof.
    split; intros.
    apply Forces_equiv.
    eapply Forces_join'; eauto;
      apply Forces_equiv; auto.
    eapply Forces_join'; eauto;
      apply Forces_equiv; eauto.
  Qed.

  Lemma Join_canc1: forall A (sa: sepalg A) w X Y Z,
    World w == X (+) Y -> World w == X (+) Z -> Y == Z.
  Proof.
    intros; eapply Forces_canc1; apply Forces_equiv; eauto.
  Qed.

  Lemma Join_canc2: forall A (sa: sepalg A) w X Y Z,
    World w == X (+) Y -> World w == Z (+) Y -> X == Z.
  Proof.
    intros; eapply Forces_canc2; apply Forces_equiv; eauto.
  Qed.

  Lemma Join_split: forall A (sa: sepalg A) ab X Y,
    World ab == X (+) Y ->
    exists a b, World a == X /\ World b == Y.
  Proof.
    intros.
    apply Forces_equiv in H. inverts H.
    exists wx wy; split; apply Forces_equiv; auto.
  Qed.

  Lemma Join_equiv_inv: forall A X Y,
    X == Y <-> forall (w:A), World w == X <-> World w == Y.
  Proof.
    split; intros.
    split; intros.
    setoid_rewrite<- H; auto.
    setoid_rewrite H; auto.
    split; intros; apply Forces_equiv; apply H; apply Forces_equiv; auto.
  Qed.

  Lemma Join_elim_unit_for: forall (A:Type) (sa:sepalg A) w (X: @JoinPred A),
    World w == World w (+) X -> forall wx, World wx == X -> unit_for wx w.
  Proof.
    intros.
    eapply Join_join' with (a:=w) (X:=World w) (Y:=X) in H; eauto.
    reflexivity.
  Qed.

  Lemma Join_elim_identity: forall (A:Type) (sa:sepalg A) w (X: @JoinPred A),
    World w == World w (+) X -> X == Identity X.
  Proof.
    intros.
    split; intros.
    econstructor.
    split; auto.
    apply unit_identity with w.
    eapply Join_elim_unit_for; eauto.
    apply Forces_equiv; auto.
    inverts H0.
    clear f_fun0.
    destruct H2.
    auto.
  Qed.

  Lemma Joins_elim_identity: forall (A:Type) (sa:sepalg A) (X Y: @JoinPred A),
    {[ X == X (+) Y ]} -> Y == Identity Y.
  Proof.
    intros.
    destruct H as [w[??]].
    eapply Join_elim_identity; eauto.
    apply Forces_equiv.
    apply Forces_equiv in H.
    setoid_rewrite H.
    auto.
  Qed.

  Lemma Identity_split: forall (A:Type) (sa:sepalg A) (X Y: @JoinPred A),
    Identity (X(+)Y) == Identity X (+) Identity Y.
  Admitted.

  Lemma Identity_split1: forall (A:Type) (sa:sepalg A) (X Y: @JoinPred A),
    Identity (X(+)Y) == Identity X.
  Admitted.

  Lemma Identity_splitT: forall (A:Type) (sa:sepalg A) (X Y: @JoinPred A),
    Identity (X(+)Y) == Identity Y.
  Admitted.

  Lemma Join_Identity2: forall (A:Type) (sa:sepalg A) (X Y: @JoinPred A),
    X (+) Identity Y == X.
  Admitted.

  Lemma Join_Identity1: forall (A:Type) (sa:sepalg A) (X Y: @JoinPred A),
    Identity X (+) Y == Y.
  Proof.
    intros. setoid_rewrite Join_com. apply Join_Identity2.
  Qed.


End JoinLogic.

Module Tactics.
  Require Import SetoidClass.
  Import Morphisms.

  Tactic Notation "patternw" constr(w) :=
  let do_patternw w :=
    repeat (try (setoid_rewrite (Join_com _ _ _ w));
          try (setoid_rewrite (Join_swap2 _ _ _ w _));
          try (setoid_rewrite<- (Join_assoc _ _ w _ _)))
  in let rec sub_patternw w := match w with
    | Join _ ?X ?Y => sub_patternw Y; sub_patternw X; setoid_rewrite Join_assoc
    | ?w => do_patternw w
  end
  in sub_patternw w.

  Tactic Notation "patternw" constr(w) "in" hyp(H) :=
  let do_patternw w :=
    repeat (try (setoid_rewrite (Join_com _ _ _ w) in H);
          try (setoid_rewrite (Join_swap2 _ _ _ w _) in H);
          try (setoid_rewrite<- (Join_assoc _ _ w _ _) in H))
  in let rec sub_patternw w := match w with
    | Join _ ?X ?Y => sub_patternw Y; sub_patternw X; setoid_rewrite Join_assoc in H
    | ?w => do_patternw w
  end
  in sub_patternw w.

  Tactic Notation "elimw" :=
  let do_elim w := patternw w; progress (apply Morphisms.Join_m; try reflexivity)
  in let rec find_elim Z := match Z with
(*    | Join_context ?w => do_elim w
    | Join_wterm ?w => do_elim w
    | Join_state ?w => do_elim w
    | World ?w => do_elim w
    | JGWorld ?w => do_elim w *)
    | Join _ ?X ?Y => first [find_elim X | find_elim Y]
    | ?w => do_elim w
    end
  in
  match goal with
  | [  |- equiv ?X  _ ] => find_elim X
  end.

  Tactic Notation "join_Join" "in" "*" :=
    repeat match reverse goal with
      | [H : join _ _ _ |- _] => apply Join_join in H
    end.

  Ltac joinablew_ex w1 w2 H :=
    let H0 := fresh "H" in assert(H0:= H);
    patternw w2 in H0;
    patternw w1 in H0;
    setoid_rewrite Join_assoc in H0;
    apply Join_split in H0.

  Module Tests.
    Import Notation Notation.U.
    Open Scope join_scope.
    Lemma test_patternw1: forall {A} {sa:sepalg A} X Y Z,
      X (+) Y (+) Z == (Z(+)X)(+)Y.
    Proof. intros. patternw Y. patternw Z. reflexivity. Qed.

    Lemma test_patternw2: forall {A} {sa:sepalg A} X Y Z W,
      X (+) Y (+) Z == W(+)Z -> Z (+) (X (+) Y) == Z (+) W.
    Proof. intros. patternw Z in H. auto. Qed.

    Lemma test_patternw3: forall {T} {sa:sepalg T} A B C D E,
      A(+)B(+)C(+)D(+)E == B(+)C(+)E(+)D(+)A.
    Proof. intros; patternw ((C(+)B)(+)(A(+)E)); reflexivity. Qed.

    Lemma test_elimw1: forall {A} {sa:sepalg A} X Y Z W,
      Y(+)Z == W -> X (+) (Y (+) Z) == X (+) W.
    Proof. intros. elimw. exact H. Qed.

    Lemma test_join_Join: forall {A} {sa:sepalg A} a b c ab abc,
      join a b ab -> True -> join ab c abc -> abc ||= World a (+) (World b (+) World c).
    Proof. intros; join_Join in *; setoid_rewrite Join_assoc; setoid_rewrite<- H; apply Forces_equiv; auto. Qed.

  End Tests.

End Tactics.
