Require Import Machine.
Require Import ContextSemantics.
Require Import Bisimulation.
Require Import world_model.
Require Import msl.msl_standard.
Require Import SfLib.
Require Import LibTactics.
Require Import IterationBounds.
Require Import IterationSubstitution.
Require Import StepStar.

  Definition s_iter_schema_exists_fresh2 s f1 f2 :
       {i : positive | s_iter_get s i = None /\ free_iter_var f1 i /\ free_iter_var f2 i}.
Admitted.

Import world_model.Notation.
Import Bisimulation.Notation.
Import Bisimulation.Morphisms.
Import world_model.Notation.


Section Definitions.
(*  Import ContextSemantics.Notation.*)

  Open Scope forces_pred.
  Import Machine.Notation.

  Import IterationSubstitution.Notation.

  Definition bisim_split_prop1 (P: pred world) (f fh:iter_schema):=
    forall s w z1, w|=P -> bisimilar s (w: f (iter_plus2 z1 i_inf)) (w: fh z1;; f i_inf).
  Definition bisim_split_prop2 (P: pred world) (f:iter_schema) (n:nat):=
    forall s w (x y: nat), w|=P -> bisimilar s (w: f (x + n + y)) (w: f x;; f y).

  Definition bisim_split_prop' (P: pred world) (f :iter_schema) (n:nat):=
    forall s w z1 z2, w|=P -> bisimilar s (w: f (iter_plus2 z1 (i_plus n z2))) (w: f z1;; f z2).

  Definition bisim_split_prop (P: pred world) (f fh:iter_schema) (n:nat):=
    bisim_split_prop1 P f fh /\ bisim_split_prop2 P fh n.

  Lemma bisim_split_prop_impl: forall P f n,
    bisim_split_prop' P f n -> bisim_split_prop P f f n.
  Proof.
    intros.
    split; repeat intro.
    change i_inf with (i_plus n i_inf) at 1.
    apply H; auto.
    rewrite<- plus_assoc.
    change (iter_const (x+(n+y))) with (iter_plus2 x (i_plus n y)).
    apply H; auto.
  Qed.

  Definition equiv_schema_nat P (f: iter_schema) x (g: iter_schema) y :=
    x > 0 /\ y > 0 /\
    forall s w, w|=P -> bisimilar s (w: f x) (w: g y).

  Definition equiv_schema_nat' P (f g: iter_schema) := exists x y,
    equiv_schema_nat P f x g y.

  Definition bisim_termination_condition_prop P (f fh:iter_schema) Q :=
    (forall s w, w|=P -> bisimilar s (w:fh i_inf;; tm_assert Q) (w:fh i_inf)) /\
(*    (forall s w, w|=P -> bisimilar s (w:f i_inf; tm_assert Q) (w:f i_inf)) /\*)
    (forall s w, w|=P -> bisimilar s (w: tm_assert Q;; f i_inf) (w: tm_assert Q)).


End Definitions.


Module Notation.
  Export IterationSubstitution.Notation.
End Notation.

Import Machine.Notation.


Section Lemmas.
  Import Notation.
  Open Scope forces_pred.


  Lemma iter_factor: forall k x n, x > 0 -> exists c d, k + n + d = (c+1)*x + c*n.
  Proof.
    intros.
    assert (k + n < k * n + k * x + n * n + n * x + x).
      induction n; ring_simplify; try omega.
      destruct x. false; omega.
      clear H.
      induction x. omega. 
      eapply transitivity; eauto;
        ring_simplify; omega.
    exists (k+n) ((k + n + 1) * x + (k + n) * n - k - n).
    ring_simplify.
    omega.
  Qed.


  Lemma bisim_schema_inf_split1: forall P (f fh:iter_schema),
    bisim_split_prop1 P f fh ->
    (forall (z:iterations) s w, w|=P -> bisimilar s (w: f i_inf) (w: fh z;; f i_inf)).
  Proof.
    intros.
    assert (iter_const i_inf = iter_plus2 z i_inf).
      rewrite iter_plus2_iterations; rewrite i_plus_inf; auto.
    rewrite H1 at 1.
    apply H; auto.
  Qed.

(*
  Lemma bisim_schema_inf_split2: forall P (f:iter_schema) n,
    bisim_split_prop P f n ->
    forall (z:iterations) s w, w|=P -> bisimilar s (w: f i_inf) (w: f i_inf; f z).
  Proof.
    intros.
    assert (iter_const i_inf = iter_plus2 i_inf (i_plus n z)).
      rewrite iter_plus2_iterations; reflexivity.
    rewrite H1 at 1.
    apply H; auto.
  Qed.
*)

  Lemma bisim_schema_inf_inf_split: forall P (f fh:iter_schema) s w,
    w |= P ->
    bisim_split_prop1 P f fh ->
    bisimilar s (w: f i_inf) (w: fh i_inf;; f i_inf).
  Proof.
    intros.
    assert (iter_const i_inf = iter_plus2 i_inf i_inf).
      reflexivity.
    rewrite H1 at 1.
    apply H0; auto.
  Qed.

  Lemma bisim_schema_stall_inf_split: forall P (f fh:iter_schema) s w i x,
    w |= P ->
    bisim_split_prop1 P f fh ->
    bisimilar s (w: f (i_stall i x i_inf)) (w: fh (i_stall i x 0);; f i_inf).
  Proof.
    intros.
    assert (i_stall i x i_inf = iter_plus2 (i_stall i x 0) i_inf).
      reflexivity.
    rewrite H1 at 1.
    apply H0; auto.
  Qed.

  Lemma bisim_schema_sum_split: forall (f:iter_schema) P (n: nat),
    bisim_split_prop2 P f n ->
    forall (k1 k2:nat) s w, w|=P -> bisimilar s (w: f (k1+n+k2)) (w: f k1;; f k2).
  Proof.
    intros.
    apply H; auto.
  Qed.

  Lemma bisim_schema_factor_back: forall (f:iter_schema) P (c x n: nat),
    bisim_split_prop2 P f n ->
    forall s w, w|=P -> bisimilar s (w: f (((S c)+1)*x+(S c)*n)) (w: f ((c+1)*x+c*n);; f x).
  Proof.
    intros.
    replace ((S c + 1) * x + S c * n) with (((c + 1) * x + c * n) + n + x).
    eapply bisim_schema_sum_split; eauto.
    ring_simplify; omega.
  Qed.

  Lemma halted_assert_multiples: forall (f:iter_schema) (x n:nat) P,
    (forall s C w, w|=P -> halted_assert s C (w:f x) P) ->
    bisim_split_prop2 P f n ->
    (forall s C w c, w|=P -> halted_assert s C (w: f ((c + 1) * x + c * n)) P).
  Proof.
    intros.
    intro; intros.
    revert s' C' w' H2.
    induction c; intros.
    simpl in *.
    replace (x+0+0) with x in H2; try omega.
    eapply H; eauto.
    assert (bisimilar s (w : f ((S c + 1) * x + S c * n)) (w : f ((c + 1) * x + c * n);; f x)).
      eapply bisim_schema_factor_back; eauto.
    eapply bisim_halted_converge1 in H3; eauto.
    edestruct cstep_star_seq_inversion as [[s'0[C'0[w'0[??]]]] | [bt[??]]]; eauto.
    eapply H.
    eapply IHc; eauto.
    eauto.
    inverts H5.
    eexists; eauto.
  Qed.  

  Lemma halted_assert_inf: forall (f:iter_schema) (x n:nat) P,
    (forall s C w, w|=P -> halted_assert s C (w:f x) P) ->
    x > 0 ->
    bisim_split_prop2 P f n ->
    (forall s C w, w|=P -> halted_assert s C (w: f i_inf) P).
  Proof.
    intros.
    intro; intros.
    edestruct cstep_star_schema_inf_nat_iter as [N[b'[?[??]]]]; eauto.
    apply subst_iter_skip in H4.
    subst b'.
    destruct (iter_factor N x n) as [c[d ?]]; auto.
    spec H6 (n+d) N.
    spec H6; auto.
    simpl in H6.
    replace (N+(n+d)) with (N+n+d) in H6.
    rewrite H4 in H6.
    eapply halted_assert_multiples with (w:=w) (f:=f) (x:=x) (n:=n) (c:=c); eauto.
    omega.
  Qed.

  Lemma bisim_termination_condition_prop_f_fh: forall P (x n:nat) (f fh: iter_schema) Q,
    (forall s C w, w|=P -> halted_assert s C (w:fh x) P) ->
    x > 0 ->
    bisim_split_prop P f fh n ->
    bisim_termination_condition_prop P f fh Q ->
    (forall s w, w|=P -> bisimilar s (w:f i_inf) (w:fh i_inf)).
  Proof.
    intros.
    eapply bisim_trans.
    eapply bisim_schema_inf_split1 with (z:=i_inf); eauto.
    apply H1.
    eapply bisim_trans.
    apply bisim_seq1.
    apply bisim_symm.
    apply H2; auto.
    eapply bisim_trans.
    setoid_rewrite<- bisim_Cs_seq_assoc.
    apply bisim_seq2.
    intros.
    setoid_rewrite bisim_Cs_tm_seq_seq at 1.
    apply H2.
    eapply halted_assert_inf with (f:=fh) (x:=x); eauto.
    apply H1.
    apply H2; auto.
  Qed.

  Lemma bisim_termination_condition_prop_f: forall P (x n:nat) (f fh: iter_schema) Q,
    (forall s C w, w|=P -> halted_assert s C (w:fh x) P) ->
    x > 0 ->
    bisim_split_prop P f fh n ->
    bisim_termination_condition_prop P f fh Q ->
    (forall s w, w|=P -> bisimilar s (w:f i_inf;; tm_assert Q) (w:f i_inf)).
  Proof.
    intros.
    eapply bisim_trans.
    apply bisim_seq1.
    eapply bisim_termination_condition_prop_f_fh; eauto.
    eapply bisim_trans.
    apply H2; auto.
    apply bisim_symm.
    eapply bisim_termination_condition_prop_f_fh; eauto.
  Qed.


  Lemma bisim_equiv_schema_multiples: forall s w P c (f g: iter_schema) (n m x y:nat),
    w|=P ->
    (forall s C w, w|=P -> halted_assert s C (w:f x) P) ->
    bisim_split_prop2 P f n ->
    bisim_split_prop2 P g m ->
    equiv_schema_nat P f x g y ->
    bisimilar s (w : f ((c + 1) * x + c * n)) (w : g ((c + 1) * y + c * m)).
  Proof.
    intros.
    induction c.
    simpl.
    replace (x+0+0) with x.
    replace (y+0+0) with y.
    apply H3; auto.
    omega. omega.
    eapply bisim_trans.
    eapply bisim_schema_factor_back; eauto.
    eapply bisim_trans.
    apply bisim_seq; intros; eauto.
    apply H3; auto.
    eapply halted_assert_multiples with (f:=f); eauto.
    apply bisim_symm.
    eapply bisim_schema_factor_back; eauto.
  Qed.

  Lemma bisim_schema_factor: forall (f g:iter_schema) P k (n m x y: nat),
    (forall s C w, w|=P -> halted_assert s C (w:f x) P) ->
    bisim_split_prop2 P f n ->
    bisim_split_prop2 P g m ->
    equiv_schema_nat P f x g y ->
    exists (k' k'':nat), forall s w, w|=P -> bisimilar s (w: f (k+n+k')) (w: g k'').
  Proof.
    intros.
    destruct H2 as [?[_ ?]].
    destruct (iter_factor k x n) as [c[d ?]]; auto.
    exists d ((c+1)*y+c*m).
    intros.
    rewrite H4.
    clear H4.
    induction c.
    simpl.
    replace (x+0+0) with x; try omega.
    replace (y+0+0) with y; try omega.
    apply H3; auto.
    eapply bisim_trans.
    eapply bisim_schema_factor_back; eauto.
    eapply bisim_trans.
    apply bisim_seq; eauto.
    intros.
    apply H3.
    eapply halted_assert_multiples; eauto.
    apply bisim_symm.
    eapply bisim_schema_factor_back; eauto.
  Qed.

  Lemma bisim_iter_schema_go_stall: forall i s w f n z,
    s_iter_get s i = Some iter_go ->
    free_iter_var f i ->
    bisimilar s (w: f (i_stall i n z)) (w: f (i_plus n z)).
  Proof.
    intros.
    revert z n w s H0 H.
    induction f; simpl; intros.

    apply bisim_refl.

    apply bisim_trans with (w: tm_while_max e (i_stall i (n0-n) (i_minus z (n-n0))) (f (i_plus n0 z))).
    eapply bisim_while with (P:=TT).
    simpl; auto.
    intros; intro; auto.
    intros; intro; auto.
    intros; eauto.
    eapply IHf; auto.
    eapply cstep_star_s_iter_preservation; eauto.
    eapply bisim_trans.
    apply bisim_while_go_stall; auto.
    replace (iter_const (i_plus (n0-n) (i_minus z (n-n0)))) with (iter_const (i_minus (i_plus n0 z) n)).
    simpl; destruct z; apply bisim_refl.
    destruct z; simpl; do 2 f_equal; omega.

    eapply bisim_while with (P:=TT).
    simpl; auto.
    intros; intro; auto.
    intros; intro; auto.
    intros; eapply IHf.
    destruct i0; try destruct H0; auto.
    eapply cstep_star_s_iter_preservation; eauto.

    destruct H0.
    eapply bisim_if; intros.
    eapply IHf1; eauto.
    eapply IHf2; eauto.

    destruct H0.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    eapply bisim_seq; intros.
    eapply IHf1; eauto.
    eapply IHf2; eauto.
    eapply cstep_star_s_iter_preservation; eauto.

    destruct H0.
    eapply bisim_tm_par; intros.
    eapply IHf1; eauto.
    eapply step_star_go_preservation; eauto.
    eapply IHf2; eauto.
    eapply step_star_go_preservation; eauto.

    rename i0 into f.
    eapply bisim_tm_local; eauto.
    reflexivity.
  Qed.

  Lemma bisim_witer_schema_go_stall: forall i s f n z,
    s_iter_get s i = Some iter_go ->
    wfree_iter_var f i ->
    bisimilar s (f (i_stall i n z)) (f (i_plus n z)).
  Proof.
    intros.
    revert z n s H0 H.
    induction f; simpl; intros.

    apply bisim_iter_schema_go_stall; auto.

    destruct H0.
    apply bisim_seq; eauto.
    intros.
    apply bisim_iter_schema_go_stall; auto.
    eapply cstep_star_s_iter_preservation; eauto.

    destruct H0.
    apply bisim_par; eauto.
  Qed.

  Lemma step_star_schema_split_front: forall s w P Q (f fh:iter_schema) s' bb',
    bisim_split_prop1 P f fh ->
    (forall s C w, w|=P -> halted_assert s C (w:fh i_inf) P) ->
    bisim_termination_condition_prop P f fh Q ->
    w|=P ->
    step_star s (w: f i_inf) s' bb' ->
    exists (X:nat), forall x, x>=X -> exists b',
      step_star s (w: fh x) s' b' /\
      bisimilar s' bb' (b';; f i_inf).
  Proof.
    introv Hf_split Hf_P [Hb_f_Q Hb_Q_f2] HPw Hss_fi.
    destruct (s_iter_schema_exists_fresh2 s f fh) as [i[Hsi_None [Hfree_f Hfree_fh]]].
    edestruct step_star_schema_inf_nat_iter as [X[fb'[Hb_fi_fxi[Hfree_fb' Hss_f_fb']]]]; eauto.
    subst bb'.

    assert (Hf_free_i: forall (z:iterations), term_free_iter_var (f z) i).
      intros. apply term_free_iter_var_subst; auto. introv HH; inverts HH.
    assert (Hf_free_i': forall w (z:iterations), wterm_free_iter_var (w : f z) i).
      intros. apply term_free_iter_var_subst; auto. introv HH; inverts HH.
    assert (Hfh_free_i: forall (z:iterations), term_free_iter_var (fh z) i).
      intros. apply term_free_iter_var_subst; auto. introv HH; inverts HH.
    assert (Hfh_free_i': forall w (z:iterations), wterm_free_iter_var (w : fh z) i).
      intros. apply term_free_iter_var_subst; auto. introv HH; inverts HH.

    remember (s_iter_set s i iter_stall) as s0.
    remember (s_iter_set s' i iter_stall) as s0'.

    exists X; intros x HxX.
    cut (step_star s0 (w:f (i_stall i x i_inf)) s0' (fb' (i_stall i x i_inf))).
    intros.
    cut (bisimilar s0 (w:f (i_stall i x i_inf)) (w:fh (i_stall i x 0);; f i_inf)).
    intros.
    edestruct step_star_bisim1 as [bbb'[Hss_ff_bbb' Hb_fb'_bbb']]; eauto.
    edestruct step_star_seq_inversion as [[s'0[w'[Hss_fstall_skip Hss_f_bbb']]] | [bx'[H12 H13]]].
    apply Hss_ff_bbb'.
    generalize (step_star_s_iter_derived_paused _ _ _ _ i Hss_f_bbb').
    intros Hs'0i_stall; spec Hs'0i_stall.
    subst; rewrite s_iter_get_set_eq; reflexivity.

    (* w: f[x,i] -> w': skip, so w': f[inf] ~ bbb' ~ bb' *)
    cut (bisimilar s'0 (w':f i_inf) (w':tm_skip)).
    intros Hb_f_skip.
(*    cut (bisimilar s0' bbb' (w':tm_skip)).
    intros Hb_bbb'_skip.*)
    cut (step_star s0 (w: fh x) s'0 (w':tm_skip)).
    intros Hss_fx_skip.
    exists (w':tm_skip); split.
    eapply step_star_elim_free_iter_var with (n:=iter_stall); eauto.
    eapply ss_iter_star_step_star; eauto.
    subst s0 s0'.
    eapply step_star_app; eauto.
    apply s_iter_step_star; auto.
    eapply s_iter_star_bisim_halted_step_star; eauto. eexists; eauto.
    apply bisim_intro_free_iter_var with (i:=i) (ii:=iter_go).
    eapply s_iter_step_star_None_preservation; eauto.
    eapply wterm_free_iter_var_subst; eauto.
    intros; discriminate.
    simpl; split; auto.
    change (iter_const i_inf) with (iter_const (i_plus x i_inf)).
    eapply bisim_trans.
    apply bisim_symm.
    apply bisim_witer_schema_go_stall with (i:=i); auto.
    rewrite s_iter_get_set_eq; reflexivity.
    apply bisim_s_iter with (s:=s0').
    subst s0'.
    rewrite<- s_iter_set_set_eq' with (m:=iter_go) (n:= iter_stall).
    constructor.
    rewrite s_iter_get_set_eq; reflexivity.
    eauto.
    setoid_rewrite bisim_Cs_seq_skip1'.
    eapply bisim_trans; eauto.
    simpl.
    subst s0'.
    apply bisim_symm.
    apply bisim_symm in Hb_f_skip.
    eapply step_star_halted_bisim2; eauto. eexists; eauto.
    (*  step_star s0 (w: f x) s0' (w':tm_skip)  *)
    replace (iter_const x) with (iter_const (i_plus x 0)).
    apply step_star_stalled_skip with (i:=i); eauto.
    rewrite Hs'0i_stall. intro; discriminate.
    simpl. do 2 f_equal; omega.

    (*  bisimilar C'0 s'0 (w' : f i_inf) (w' : tm_skip)  *)
    apply step_star_stalled_skip_inf in Hss_fstall_skip.
    assert (w' |= P).
      eapply Hf_P; eauto.
      apply step_star_add_context_cstep_star with (C:=cx_hole); eauto.
    cut (w' |= gsep_con Q TT). intros ?.
    setoid_rewrite<- bisim_Cs_seq_skip1' at 1.
    setoid_rewrite<- bisim_Cs_assert_skip with (P:=Q); eauto.
    eapply step_star_assert_skip_resolve; eauto.
    spec Hb_f_Q (s_iter_set s0 i iter_go) w HPw.
    eapply bisim_halted_converge2' in Hb_f_Q; eauto.
    apply step_star_add_iter_go; eauto.
    eexists; eauto.
    rewrite Hs'0i_stall; intro; discriminate.

    (*  Case: f[x]->bx' and bbb'=bx';f[i_inf]  *)
    subst bbb'.
    assert (H15: bisimilar (s_iter_set s0 i iter_go) (w : fh (i_stall i x 0)) (w : fh x)).
      setoid_replace (iter_const x) with (i_plus x 0); auto.
      apply bisim_iter_schema_go_stall; auto.
      rewrite s_iter_get_set_eq; auto.
      simpl; f_equal; auto.
    edestruct step_star_bisim1 with (s:=s_iter_set s0 i iter_go) (bb:=w:fh x) as [bbbb'[??]]; eauto.
    apply step_star_add_iter_go; eauto.
    exists bbbb'; split.
    rewrite Heqs0 in H1.
    rewrite Heqs0' in H1.
    do 2 rewrite s_iter_set_set_eq' in H1.
    apply step_star_elim_free_iter_var with (i:=i) (n:=iter_go); auto.
    eapply ss_iter_star_step_star; eauto.
    cut (bisimilar (s_iter_set s' i iter_go) (fb' i_inf) (bbbb';; f i_inf)).
      intros ?.
    apply bisim_intro_free_iter_var with (i:=i) (ii:=iter_go); eauto.
    eapply s_iter_step_star_None_preservation; eauto.
    eapply wterm_free_iter_var_step_star_preservation; eauto.
    split; auto.
    eapply wterm_free_iter_var_step_star_preservation; eauto.
    (* bisimilar (s_iter_set s' i iter_go) (fb' i_inf) (bbbb'; f i_inf) *)
    apply bisim_trans with (bx';;f i_inf).
    apply bisim_trans with (fb' (i_stall i x i_inf)).
    apply bisim_symm.
    setoid_replace (iter_const i_inf) with (i_plus x i_inf); auto.
    apply bisim_witer_schema_go_stall; auto.
    rewrite s_iter_get_set_eq; auto.
    apply bisim_s_iter with (s:=s0').
    subst s0'.
    setoid_rewrite<- s_iter_set_set_eq' with (n:=iter_stall) at 2.
    constructor.
    rewrite s_iter_get_set_eq; auto.
    auto.
    subst s0'.
    rewrite s_iter_set_set_eq' in H2; auto.
    apply bisim_seq1; subst; auto.

    (*  bisimilar C s0 (w : f (i_stall i x i_inf)) (w : f (i_stall i x 0); f i_inf)  *)
    eapply bisim_schema_stall_inf_split; eauto.

    (*  step_star s0 (w : f (i_stall i x i_inf)) s0' (b' (i_stall i x i_inf))  *)
    subst s0 s0'.
    apply step_star_add_iter; eauto.
    replace (i_stall i x i_inf) with (iter_plus1 x (i_stall i 0 i_inf)); eauto.
    simpl. f_equal. omega.
  Qed.


  Lemma halted_assert_bisim: forall s C ba bb P,
    halted_assert s C ba P ->
    bisimilar s ba bb ->
    halted_assert s C bb P.
  Proof.
    intros.
    intro; intros.
    eapply H.
    eapply bisim_halted_converge1; eauto.
    apply bisim_symm; auto.
    eexists; eauto.
  Qed.

  Lemma equiv_schema_nat_halted_assert: forall P f x g y,
    equiv_schema_nat P f x g y ->
    (forall s C w, w|=P -> halted_assert s C (w:f x) P) ->
    (forall s C w, w|=P -> halted_assert s C (w:g y) P).
  Proof.
    intros.
    eapply halted_assert_bisim; eauto.
    apply H; auto.
  Qed.

  Lemma bisim_termination_condition_prop_conclusion': forall P f fh Q s w s' w',
    bisim_termination_condition_prop P f fh Q ->
    w |= P ->
    step_star s (w: fh i_inf) s' (w':tm_skip) ->
    w' |= gsep_con Q TT.
  Proof.
    intros.
    destruct H as [? _].
    spec H s w H0.
    eapply step_star_assert_skip_resolve; eauto.
    eapply bisim_halted_converge2'; eauto.
    eexists; eauto.
  Qed.

  Lemma bisim_termination_condition_prop_conclusion: forall P (f:iter_schema) Q s w s' w',
    (forall s w, w|=P -> bisimilar s (w:f i_inf;; tm_assert Q) (w:f i_inf)) ->
    w |= P ->
    step_star s (w: f i_inf) s' (w':tm_skip) ->
    w' |= gsep_con Q TT.
  Proof.
    intros.
    spec H s w H0.
    eapply step_star_assert_skip_resolve; eauto.
    eapply bisim_halted_converge2'; eauto.
    eexists; eauto.
  Qed.

  Lemma ge_mult: forall x y,
    x > 0 ->
    y*x >= y.
  Proof.
    intros.
    destruct x. inverts H. clear H.
    induction y; auto.
    simpl in *. omega.
  Qed.

  Lemma ge_plus: forall x y z,
    x >= z ->
    x + y >= z.
  Proof.
    intros. omega.
  Qed.

  Lemma halted_bisim_assert_term: forall s b Q s' w',
    bisimilar s b (b;; tm_assert Q) ->
    step_star s b s' (w': tm_skip) ->
    w' |= gsep_con Q TT.
  Proof.
    intros.
    apply step_star_assert_skip_resolve with (s:=s) (b:=b) (s':=s').
    eapply bisim_halted_converge1'; eauto.
    eexists; eauto.
  Qed.

  Lemma bisim_seq_term_cond: forall (f fh: iter_schema) (fx fn:nat) P Q,
    fx > 0 ->
    bisim_split_prop2 P fh fn ->
    (forall s C w, w|=P -> halted_assert s C (w : fh fx) P) ->
    (forall s w, w|=P -> bisimilar s (w : fh i_inf;; tm_assert Q) (w : fh i_inf)) ->
    (forall s w, w|=P -> bisimilar s (w : tm_assert Q;; f i_inf) (w : tm_assert Q)) ->
    (forall s w, w|=P -> bisimilar s (w: fh i_inf;; f i_inf) (w: fh i_inf)).
  Proof.
    intros.
    eapply bisim_trans.
    apply bisim_seq1.
    apply bisim_symm; eapply H2; eauto.
    eapply bisim_trans with (w: fh i_inf;; tm_assert Q).
    setoid_rewrite<- bisim_Cs_seq_assoc.
    apply bisim_seq2.
    intros.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply H3.
    eapply halted_assert_inf; eauto.
    apply H2; auto.
  Qed.

  Lemma loop_coterm: forall (f g fh gh: iter_schema) s w s' b' (fx fn gy gm:nat) P Q,
    halted b' ->
    w |= P ->
    (forall s C w, P w -> halted_assert s C (w : fh fx) P) ->
    bisim_split_prop P f fh fn ->
    bisim_split_prop P g gh gm ->
    equiv_schema_nat P fh fx gh gy ->
    bisim_termination_condition_prop P f fh Q ->
    bisim_termination_condition_prop P g gh Q ->
    step_star s (w: f i_inf) s' b' ->
    step_star s (w: g i_inf) s' b'.
  Proof.
    introv [w' ?] HwP HfP Hf_split Hg_split Hbisim_xy Hf_term_cond Hg_term_cond ?.
    subst b'.
    assert (bisimilar s (w: f i_inf) (w: fh i_inf)).
      eapply bisim_trans.
      eapply bisim_schema_inf_split1 with (z:=i_inf); eauto. apply Hf_split.
      destructs Hf_term_cond.
      eapply bisim_seq_term_cond with (fx:=fx); eauto.
      apply Hbisim_xy.
      apply Hf_split.
    edestruct step_star_schema_inf_nat as [n ?].
      eapply bisim_halted_converge1'; eauto. eexists; eauto.
    edestruct bisim_schema_factor with (f:=fh) (g:=gh) (k:=n) as [k'[k'' ?]]; eauto.
      apply Hf_split. apply Hg_split.
    spec H2 s w HwP.
    eapply bisim_halted_converge2' with (bb:=w: gh k'';; g i_inf); eauto.
    eapply step_star_seq_join.
    eapply bisim_halted_converge1' with (ba:=w: fh(n+fn+k')); eauto.
    eapply H1. omega.
    eexists; eauto.
    eapply bisim_halted_converge1' with (ba:=w':tm_assert Q); eauto.
    apply step_single. apply st_assert; eauto.
    eapply halted_bisim_assert_term; eauto.
    apply bisim_symm.
    eapply bisim_termination_condition_prop_f; eauto. apply Hbisim_xy.
    setoid_rewrite<- bisim_Cs_seq_skip1' at 2.
    setoid_rewrite<- bisim_Cs_assert_skip; eauto.
    apply bisim_symm.
    apply Hg_term_cond.
    eapply halted_assert_multiples with (f:=fh) (c:=n); eauto.
      apply Hf_split.
      apply step_star_add_context_cstep_star with (C:=cx_hole).
      eapply H1. assert (fx > 0). apply Hbisim_xy. ring_simplify.
      do 2 eapply ge_plus; auto. apply ge_mult; auto.
    eapply halted_bisim_assert_term; eauto.
    apply bisim_symm.
    eapply bisim_termination_condition_prop_f; eauto. apply Hbisim_xy.
    eexists; eauto.
    eapply bisim_schema_inf_split1; eauto.
      apply Hg_split.
    eexists; eauto.
  Qed.

  Theorem bisim_loop: forall P Q (f fh g gh:iter_schema) (n m x y: nat),
    (forall s C w, w|=P -> halted_assert s C (w:fh x) P) ->
    bisim_split_prop P f fh n ->
    bisim_split_prop P g gh m ->
    equiv_schema_nat P fh x gh y ->
    bisim_termination_condition_prop P f fh Q ->
    bisim_termination_condition_prop P g gh Q ->
    (forall s w, w|=P -> bisimilar s (w:f i_inf) (w: g i_inf)).
  Proof.
    intros.
    assert (forall s C w, w|=P -> halted_assert s C (w:gh y) P).
      eapply equiv_schema_nat_halted_assert; eauto.
    exists (fun s ab bb => (exists cb, exists (f fh g gh:iter_schema), exists (n m x y:nat),
        (forall s C w, w|=P -> halted_assert s C (w:fh x) P) /\
        (forall s C w, w|=P -> halted_assert s C (w:gh y) P) /\
        (forall C, halted_assert s C cb P) /\
        bisim_split_prop P f fh n /\
        bisim_split_prop P g gh m /\
        equiv_schema_nat P fh x gh y /\
        bisim_termination_condition_prop P f fh Q /\
        bisim_termination_condition_prop P g gh Q /\
        bisimilar s ab (cb;; f i_inf) /\ bisimilar s bb (cb;; g i_inf))
      \/ bb=ab).
    split.
    2: left; exists (w: tm_skip) f fh g gh n; exists m x y; splits; auto;
        [ unfold halted_assert; intros C ? ? ? H9; apply cstep_star_skip_resolve in H9; inverts H9; auto
        | splits; auto; (setoid_rewrite bisim_Cs_seq_skip1; [apply bisim_refl | reflexivity]) ].
    apply bisimulation_symm_equiv.
    clear s w f fh g gh n m x y H H0 H1 H2 H3 H4 H5 H6.
    intros s ba bb [[bc[f[fh[g[gh[fn[gm[fx[gy[Hf_P[Hg_P[Hcb_P[Hf_split[Hg_split[Hbisim_xy[Hf_term_cond[Hg_term_cond [??]]]]]]]]]]]]]]]]]] | ?].
    splits.

    split; intros; eauto.
    eapply bisim_halted_converge2'; eauto.
    apply bisim_halted1 in H; auto.
    edestruct step_star_seq_inversion as [[s'[w'[??]]]|[bc'_[??]]]; eauto.
    eapply step_star_seq_join; eauto.
    eapply loop_coterm with (f:=f) (fh:=fh) (g:=g) (gh:=gh); eauto.
    eapply Hcb_P; eauto.
    apply step_star_add_context_cstep_star with (C:=cx_hole). eauto.
    false; subst ba; destruct H1 as [w ?]; discriminate.

    (* unstall some iter var *)
    intros s' ?.
    left; exists bc f fh g gh; exists fn gm fx gy; splits; eauto.
    intros; intro; intros; eapply Hcb_P.
    eapply cstep_iteration; eauto.
    splits; eauto; eapply bisim_s_iter; eauto.

    (* stepping context *)
    intros b s' b' ?.
    left; exists bc f fh g gh; exists fn gm fx gy; splits; eauto.
    intros; intro; intros; eapply Hcb_P with (C:=cx_par1 C b).
    eapply cstep_context_cstep_star.
    apply cst_par1_step2; eauto.
    apply cstep_star_par_cx_par1.
    apply cstep_star_par1_step; eauto.
    splits; eauto; eapply bisim_context; eauto.

    (* program steps *)
    intros s'' ba'' ?.
    edestruct step_bisim1 as [bc''[??]]; eauto.
    edestruct step_star_seq_inversion as [[s'[w'[??]]]|[bc'[??]]]; eauto.
    (* bc terminates *)
    clear H2.
    edestruct step_star_schema_split_front with (w:=w') (f:=f) as [X ?]; eauto.
      apply Hf_split.
    eapply halted_assert_inf; eauto. apply Hbisim_xy.
      apply Hf_split.
    eapply Hcb_P; eauto.
    apply step_star_add_context_cstep_star with (C:=cx_hole); eauto.
    edestruct (iter_factor X fx fn) as [c[d ?]]. apply Hbisim_xy.
    spec H2 ((c + 1) * fx + c * fn).
    destruct H2 as [bc'[??]].
    rewrite<- H6; omega.
    assert (bisimilar s bb ((bc;; fh ((c+1)*fx + c*fn));; g i_inf)).
      eapply bisim_trans; eauto.
      setoid_rewrite<- bisim_Cs_seq_assoc.
      apply bisim_seq2.
      intros.
      assert (w'0|=P).
        eapply Hcb_P; eauto.
      setoid_rewrite bisim_Cs_tm_seq_seq.
      apply bisim_symm.
      eapply bisim_trans.
      apply bisim_seq1.
      eapply bisim_equiv_schema_multiples; eauto.
        apply Hf_split. apply Hg_split.
      apply bisim_symm.
      eapply bisim_schema_inf_split1; eauto.
        apply Hg_split.
    edestruct step_star_bisim2 with (ba:=bb) as [bb''[??]].
      eapply step_star_seq_app with (t:=g i_inf).
      eapply step_star_seq_join; eauto.
      auto.
    exists bb''; splits; auto.
      left; exists bc' f fh g gh; exists fn gm fx gy; splits; intros; auto.
      intro; intros.
      eapply halted_assert_multiples with (c:=c) (x:=fx) (n:=fn) (w:=w'); eauto.
        apply Hf_split.
      eapply Hcb_P; eauto. apply step_star_add_context_cstep_star with (C:=cx_hole); eauto.
      eapply cstep_star_app; eauto.
      apply step_star_add_context_cstep_star; eauto.
      splits; auto.
      eapply bisim_trans; eauto.

    (* bc doesn't terminate *)
    subst bc''. clear H2.
    edestruct step_star_bisim2 with (ba:=bb) (bb:=bc;;g i_inf) as [bb''[??]]; eauto.
      apply step_star_seq_app; eauto.
    exists bb''; splits; eauto.
      left; exists bc' f fh g gh; exists fn gm fx gy; splits; auto.
      intros C'' s'''  C''' w''' ?.
      eapply Hcb_P; eauto.
      eapply cstep_star_app; eauto.
      apply step_star_add_context_cstep_star; eauto.

    (* R is symmetric*)
    left; exists bc g gh f fh; exists gm fn gy fx; splits; auto.
    edestruct Hbisim_xy as [?[??]].
    splits; auto.
    intros.
    apply bisim_symm; auto.

    (* ab = bb *)
    subst; splits; auto.
    split; intros; auto. apply step_nil.
    intros s' ba' ?.
    exists ba'; splits; auto. apply step_single; auto.
  Qed.

  Corollary bisim_loop_simpl1: forall P Q (f g: iter_schema) (n m x y: nat),
    (forall s C w, w|=P -> halted_assert s C (w:f x) P) ->
    bisim_split_prop P f f n ->
    bisim_split_prop P g g m ->
    equiv_schema_nat P f x g y ->
    bisim_termination_condition_prop P f f Q ->
    bisim_termination_condition_prop P g g Q ->
    (forall s w, w|=P -> bisimilar s (w:f i_inf) (w: g i_inf)).
  Proof.
    intros.
    eapply bisim_loop; eauto.
  Qed.

  Open Scope entails_pred.

  Lemma is_bool_dec: forall e,
    (is_bool e == pIsTrue e || pIsFalse e)%pred.
  Proof.
    split; repeat intro.
    destruct H as [[|n]]; simpl in *; eauto.
    destruct H; simpl in *; eauto.
    destruct H as [n[??]]; eauto.
  Qed.

  Corollary bisim_loop_simpl2: forall P (e:expr) (f fh: iter_schema) (n x: nat),
    x > 0 ->
    P |-- is_bool e ->
    (forall s C w, w|=P -> halted_assert s C (w:fh x) P) ->
    (forall s w, P w -> bisimilar s (w : tm_assert (pIsFalse e);; fh x) (w : tm_assert (pIsFalse e))) ->
    bisim_split_prop P f fh n ->
    bisim_termination_condition_prop P f fh (pIsFalse e) ->
    (forall s w, w|=P -> bisimilar s (w:f i_inf) (w: tm_while_max e i_inf (fh x))).
  Proof.
    intros.
    pose (istm_loop_var e 0 (fh x)) as g.
    change (tm_while_max e i_inf (fh x)) with (g i_inf).
    eapply bisim_loop with (gh:=g) (Q:=pIsFalse e) (m:=0) (y:=1); eauto.
    split.
    repeat intro.
    simpl.
    do 2 rewrite iter_expr_minus_zero.
    setoid_rewrite<- bisim_Cs_while_split_iter_plus2.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_refl.
    repeat intro.
    simpl.
    repeat rewrite NPeano.Nat.sub_0_r.
    rewrite plus_0_r.
    change (iter_const (x0+y)) with (iter_plus2 x0 y).
    setoid_rewrite<- bisim_Cs_while_split_iter_plus2.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_refl.
    unfold equiv_schema_nat.
    splits; auto.
    intros.
    simpl.
    setoid_rewrite bisim_Cs_while_one_if.
    generalize H6; intro H6'.
    apply H0 in H6'.
    setoid_rewrite is_bool_dec in H6'.
    destruct H6'.
    setoid_rewrite bisim_Cs_if_true; auto.
    apply bisim_refl.
    setoid_rewrite bisim_Cs_if_false; auto.
    setoid_rewrite<- bisim_Cs_seq_skip1' at 1.
    setoid_rewrite<- bisim_Cs_assert_skip with (P:=pIsFalse e) (F:=exactly gwemp).
    auto.
    setoid_rewrite gstar_unit2; auto.
    setoid_rewrite gstar_unit2; auto.

    splits; intros; simpl.
    setoid_rewrite bisim_Cs_while_inf_post_false' at 2.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_refl.
    setoid_rewrite<- bisim_Cs_tm_seq_seq.
    setoid_rewrite bisim_Cs_assert_false_while_nop.
    apply bisim_refl.
  Qed.


End Lemmas.





