Require Import Omega.
Require Import List.
Require Import FMapPositive.
Require Import FMapFacts.
Require Import world_model.
Require Import msl.msl_standard.
Require Import LibTactics.
Require Import SfLib.
Require Export State.
(*Module PMapFacts := FMapFacts.WFacts_fun PositiveMap.E PositiveMap.*)


Section Definitions.

  Definition w_env_set' w x v := match v with
  | Some v => w_env_set w x v
  | None => w
  end.

  Program Definition pred_subst P x y : pred world :=
    (fun w => P (w_env_set' w x (w_env_get w y))).
  Solve Obligations using discriminate.

Inductive term: Type :=
| tm_while_max: expr->iteration_expr->term->term
| tm_if: expr->term->term->term
| tm_skip: term
| tm_local: expr->(var->term)->term
| tm_freelocal: var->term
| tm_send: expr->expr->term
| tm_recv: var->expr->term
| tm_newch: var->(value->pred world)->term
| tm_endch: expr->term
| tm_assign: var->expr->term
| tm_load: var->expr->term
| tm_store: expr->expr->term
| tm_assert: pred world->term
| tm_seq: term->term->term
| tm_par: pred world->term->pred world->term->term.


(*
  (* replace var x with var y *)
  Fixpoint term_subst t x y : term :=
    let y' := expr_var y in
  match t with
  | tm_while_max e z t1 => tm_while_max (expr_subst e x y') z (term_subst t1 x y)
  | tm_if e t1 t2 => tm_if (expr_subst e x y') (term_subst t1 x y) (term_subst t2 x y)
  | tm_skip => tm_skip
  | tm_seq t1 t2 => tm_seq (term_subst t1 x y) (term_subst t2 x y)
  | tm_par P1 t1 P2 t2 => tm_par (pred_subst P1 x y) (term_subst t1 x y) (pred_subst P2 x y) (term_subst t2 x y)
  | tm_send c e => tm_send (expr_subst c x y') (expr_subst e x y')
  | tm_recv z c => if Pos.eq_dec z x
      then tm_recv y (expr_subst c x y')
      else tm_recv z (expr_subst c x y')
  | tm_newch z R => if Pos.eq_dec z x
      then tm_newch y (fun v => pred_subst (R v) x y)
      else tm_newch z (fun v => pred_subst (R v) x y)
  | tm_endch e => tm_endch (expr_subst e x y')
  | tm_assert P => tm_assert (pred_subst P x y)
  | tm_assign z e => if Pos.eq_dec z x
      then tm_assign y (expr_subst e x y')
      else tm_assign z (expr_subst e x y')
  | tm_load z e => if Pos.eq_dec z x
      then tm_load y (expr_subst e x y')
      else tm_load z (expr_subst e x y')
  | tm_store e1 e2 => tm_store (expr_subst e1 x y') (expr_subst e2 x y')
  | tm_local e f => tm_local e (fun z => if Pos.eq_dec z x then f z else term_subst (f z) x y)
  | tm_freelocal z => if Pos.eq_dec z x then tm_freelocal x else tm_freelocal z
  end.

  Definition term_var_not_free x f :=
    forall y, f y = term_subst (f y) x y.

  Definition exists_term_var_not_free f :=
    { x:var | forall y, (y >= x)%positive -> term_var_not_free y f}.

  Lemma expr_exists_var_not_free: forall (e:expr),
    { x:var | forall y, (y>=x)%positive -> not_free_in_expr y e}.
Admitted.
*)


(*
  Inductive term_var_abs (f: var->term) : Prop :=
  | tva_seq: forall f1 f2,
      (forall x, f x = tm_seq (f1 x) (f2 x)) ->
      term_var_abs f1 ->
      term_var_abs f2 ->
      term_var_abs f
  | tva_while_max: forall f1 e z,
      (forall x, f x = tm_while_max e z (f1 x)) ->
      term_var_abs f1 ->
      term_var_abs f
  | tva_assign: forall fe,
      (forall x, f x = tm_assign x (fe x)) ->
      term_var_abs f
  | tva_assign': forall fe y,
      (forall x, f x = tm_assign y (fe x) /\ x<>y) ->
      term_var_abs f
  | tva_skip:
      (forall x, f x = tm_skip) ->
      term_var_abs f
.

 (* Definition term_var_abs (f: var->term) :=
    { x | forall y, f y = term_subst (f x) x y}.
*) *)

(*
Ltac tva :=
match goal with
| [ |- term_var_abs (fun x => tm_seq _ _) ] => eapply tva_seq; intros; try reflexivity
| [ |- term_var_abs (fun x => tm_while_max _ _ _) ] => eapply tva_while_max; intros; try reflexivity
| [ |- term_var_abs (fun x => tm_assign _ _) ] => eapply tva_assign; intros; try reflexivity
| [ |- term_var_abs (fun x => tm_skip) ] => eapply tva_skip; intros; try reflexivity
end.
tva.
tva.
tva.
tva.
eapply tva_assign; intros.
*)


Definition tm_while e t := tm_while_max e i_inf t.
Definition tm_ift e t := tm_if e t tm_skip.

Hint Unfold tm_while tm_ift.

Inductive wterm : Type :=
| wtm_term: world->term->wterm
| wtm_seq: wterm->term->wterm
| wtm_par: wterm->wterm->wterm
.

Fixpoint join_wterms b (w:world) := match b with
| wtm_term w' t' => w'=w
| wtm_seq b t => join_wterms b w
| wtm_par b1 b2 => exists w1, exists w2,
    join_wterms b1 w1 /\ join_wterms b2 w2 /\ gjoin w1 w2 w
end.

Lemma join_wterms_eq: forall b w1 w2,
  join_wterms b w1 -> join_wterms b w2 -> w1=w2.
Proof.
  intros.
  revert w1 w2 H H0.
  induction b; intros.
  simpl in *.
  subst; auto.
  subst; auto.
  destruct H as [w1'[w2'[?[??]]]].
  destruct H0 as [w1''[w2''[?[??]]]].
  replace w1'' with w1' in *; eauto.
  replace w2'' with w2' in *; eauto.
  eapply  gjoin_eq; eauto.
Qed.

Lemma join_wterms_dec b : {w | join_wterms b w}+{forall w, ~join_wterms b w}.
induction  b.
left; exists w; simpl; auto.
simpl; auto.

destruct IHb1. destruct IHb2.
destruct s as [w1 H1].
destruct s0 as [w2 H2].
destruct (gjoin_dec w1 w2).
destruct s as [w12 H12].
left; exists w12; simpl; exists w1; exists w2; auto.

right.
repeat intro.
destruct H as [w1'[w2'[?[??]]]].
replace w1' with w1 in *; [| eapply join_wterms_eq; eauto].
replace w2' with w2 in *; [| eapply join_wterms_eq; eauto].
apply n in H3; auto.

right; repeat intro.
destruct H as [w1'[w2'[?[??]]]].
apply n in H0; auto.

right; repeat intro.
destruct H as [w1'[w2'[?[??]]]].
apply n in H; auto.
Qed.

Fixpoint join_wterms' b (w:world) := match b with
| wtm_term w' t' => w'=w
| wtm_seq b t => join_wterms' b w
| wtm_par b1 b2 => match join_wterms_dec b1, join_wterms_dec b2 with
   | inleft (exist w1 _), inleft (exist w2 _) => match gjoin_dec w1 w2 with
      | inleft (exist w' _) => w' = w
      | _ => False
      end
   | _,_ => False
   end
end.

Lemma join_wterms_equiv: forall b w,
  join_wterms b w <-> join_wterms' b w.
Proof.
  split; intros.
  revert w H.
  induction b; intros; subst; auto.

  apply IHb; apply H.

  simpl.
  destruct H as [w1[w2[?[??]]]].
  destruct (join_wterms_dec b1).
  destruct (join_wterms_dec b2).
  destruct s as [w1' Hw1].
  replace w1' with w1 in *; [| eapply join_wterms_eq; eauto].
  destruct s0 as [w2' Hw2].
  replace w2' with w2 in *; [| eapply join_wterms_eq; eauto].
  destruct (gjoin_dec w1 w2); auto.
  destruct s as [w12 H12].
  eapply @join_eq with (H:=gworld_sa); eauto.
  apply n in H1; auto.
  elimtype False.
  apply n in H0; auto.
  apply n in H; auto.

  induction b; intros; simpl in *; subst; auto.
  destruct (join_wterms_dec b1).
  destruct (join_wterms_dec b2).
  destruct s as [w1 H1].
  destruct s0 as [w2 H2].
  destruct (gjoin_dec w1 w2).
  destruct s as [w12 H12].
  subst; exists w1; exists w2; auto.
  contradiction.
  destruct s as [w1 H1]; contradiction.
  contradiction.
Qed.

Reserved Notation "s1 / t1 --> s2 / t2" (no associativity, at level 30, t1 at next level, s2 at next level, t2 at next level).


End Definitions.


Delimit Scope term_scope with term.
Bind Scope term_scope with term.
Delimit Scope wterm_scope with wterm.
Bind Scope wterm_scope with wterm.

Module Notation.
  Import world_model.Notation.

  Notation "w : t" := (wtm_term w t%term) (left associativity, at level 20, t at level 48, format "w :  t") : wterm_scope.
  Notation "b1 ;; t2" := (wtm_seq b1%wterm t2%term) (left associativity, at level 49, format "b1 ;;  t2") : wterm_scope.
  Notation "t1 ;;; t2" := (tm_seq t1%term t2%term) (left associativity, at level 49, format "t1 ;;;  t2") : term_scope.
  Notation "b1 || b2" := (wtm_par b1%wterm b2%wterm) (left associativity, at level 50, b2 at next level, format "b1  ||  b2") : wterm_scope.

  Notation " x ::= e " := (tm_assign x e%expr) (no associativity, at level 48, format "x ::=  e") : term_scope.
  Notation " e1 <- e2" := (tm_store e1%expr e2%expr) (no associativity, at level 48, format "e1 <-  e2") : term_scope.
  Notation " x ::= [ e ] " := (tm_load x e%expr) (no associativity, at level 48, format "x ::=  [ e ]", e at level 0) : term_scope.
  Notation " x ::= 'newch' R " :=
    (tm_newch x R%pred) (no associativity, at level 48, format "x ::=  newch  R") : term_scope.
  Notation "'local' x := e 'in' t" := (tm_local e (fun x => tm_seq t%term (tm_freelocal x)))
    (left associativity, at level 70, e at level 0, x at level 0, format "'local'  x :=  e  'in'  t") : term_scope.
(*  Notation "t [ y / x ]" := (term_subst t x y)
     (no associativity, at level 9, format "t [ y / x ]") : term_scope.
*)

  Arguments Scope tm_assert [ pred ].
  Arguments Scope tm_assign [ _ expr_scope ].
  Arguments Scope tm_load [ _ expr_scope ].
  Arguments Scope tm_store [ expr_scope expr_scope ].
  Arguments Scope tm_while_max [ expr_scope _ term_scope ].
  Arguments Scope tm_while [ expr_scope term_scope ].
  Arguments Scope tm_ift [ expr_scope term_scope ].
  Arguments Scope tm_seq [ term_scope term_scope ].
  Arguments Scope tm_par [ pred term_scope pred term_scope ].


  Module U.
    Notation "∞" := i_inf : wterm_scope.
    Notation "w ⦂ t" := (wtm_term w t%term) (no associativity, at level 20, format "w ⦂  t") : wterm_scope.
    Notation "b1 ⨾ t2" := (wtm_seq b1%wterm t2%term) (no associativity, at level 50, format "b1 ⨾  t2") : wterm_scope.
    Notation "b1 ∥ b2" := (wtm_par b1%wterm b2%wterm) (left associativity, at level 50, b2 at next level, format "b1 ∥ b2") : wterm_scope.
  
(* ː ∥ ∣ ∶ ⦂ ⨟ ⨾ ⫼⫴   *)
  End U.

  Module TestNotations.
    Open Scope wterm_scope.

    Inductive test_colon_equals_compat : Prop := .

    Let a_nat := nat.

    Check (let x:= 5 in x).
    Definition aaa := 6.


    Lemma test1: forall w t1 t2,
      w: t1;; t2 = (w: t1);; t2.
    Proof. reflexivity. Qed.
    Lemma test2: forall (b1:wterm) (t2 t3: term),
      b1;; t2;; t3 = (b1;; t2);; t3.
    Proof. reflexivity. Qed.
    Lemma test3: forall (b1 b2:wterm) (t3 t4: term),
      b1;; t3 || b2;; t4 = (b1;; t3) || (b2;; t4).
    Proof. reflexivity. Qed.
    Lemma test4: forall w1 w2 t1 t2 t3 t4,
      w1: t1;; t3 || w2: t2;; t4 = (w1: t1;; t3) || (w2: t2;; t4).
    Proof. reflexivity. Qed.
    Lemma test5: forall w e t,
      w: tm_while e t = w: tm_while_max e i_inf t.
    Proof. reflexivity. Qed.
    Definition abc:= nat. (* ':=' is not broken! *)
    Lemma test_list: [1,2,3,4] = 1::2::3::4::nil. reflexivity. Qed.
    Check (forall w t b, w: t || b = w: tm_skip)%wterm.
    Goal (forall (t1 t2: term),
      t1;;;t2 = tm_seq t1 t2)%term.
    Proof. intros; reflexivity. Qed.
    Goal (forall (e:expr) (x:ident),
      tm_while e (( x::= [e] )) = tm_skip)%term. Abort.
    Goal (forall (x:ident) (e:expr),
      x::= [e] = tm_load x e)%term.
    Proof. intros; reflexivity. Qed.
    Goal (forall (x:ident) (e:expr),
      tm_load x e = x::= [e])%term.
    Proof. intros; reflexivity. Qed.
    Goal (forall (x:ident) (e:expr),
      x::= e = tm_assign x e)%term.
    Proof. intros; reflexivity. Qed.
    Goal (forall (x:ident) (e:expr),
      tm_assign x e = x::= e)%term.
    Proof. intros; reflexivity. Qed.
    Goal (forall (e1 e2:expr),
      e1<- e2 = tm_store e1 e2)%term.
    Proof. intros; reflexivity. Qed.
    Goal (forall (e1 e2:expr),
      tm_store e1 e2 = e1<- e2)%term.
    Proof. intros; reflexivity. Qed.
    Goal (forall (x:ident) (e:expr),
      expr_plus x 1 <- e;;; x::= e = tm_seq (tm_store (expr_plus x 1) e) (tm_assign x e))%term.
    Proof. intros; reflexivity. Qed.
    Goal (forall (x:ident) (e:expr),
      x::= [e];;; x::= e = tm_seq (tm_load x e) (tm_assign x e))%term.
    Proof. intros; reflexivity. Qed.
    Goal (forall (x:ident),
      x ::= newch (fun _=> emp) =
      tm_newch x (fun _=> emp))%term.
    Proof. intros; reflexivity. Qed.
    Goal (forall (x:var) (e e': expr),
      local x := e in tm_load x e' = tm_local e (fun x => tm_load x e';;; tm_freelocal x))%term.
    Proof. reflexivity. Qed.

    Goal (forall (x:var) (e e': expr),
      local x := e in tm_load x e';;;tm_load x e' = tm_local e (fun y => tm_load y e';;; tm_load y e';;; tm_freelocal y))%term.
    Proof. reflexivity. Qed.

    Goal (forall (x:var) (e e': expr),
      local x := e in x::= [e'] = tm_local e (fun x => tm_load x e';;; tm_freelocal x))%term.
    Proof. reflexivity. Qed.

    Goal (forall w (t:term) b,
      w: t || b = wtm_par (w: t) b)%wterm.
    Proof. reflexivity. Qed.

    Goal (forall w (x:var) (e e': expr) b,
      w: (local x := e in tm_load x e') || b = wtm_par (w: tm_local e (fun x => tm_load x e';;; tm_freelocal x)) b)%wterm.
    Proof. reflexivity. Qed.

  End TestNotations.

End Notation.


Section OperationalSemantics.
  Import Notation.
  Import world_model.Notation.
  Open Scope forces_pred.


  Lemma term_eq_dec: forall (t1 t2: term), {t1=t2}+{t1<>t2}.
  Proof.
  Admitted.

  Lemma wterm_eq_dec: forall (t1 t2: wterm), {t1=t2}+{t1<>t2}.
  Proof.
  Admitted.

  Definition halted b:= exists w, b = (w:tm_skip)%wterm.
  Lemma halted_dec b : {halted b}+{~halted b}.
  Proof. 
   destruct b.
    destruct (term_eq_dec t tm_skip); subst.
    left; exists w; auto.
    right; intro. destruct H. inversion H.
    apply n; auto.
    right; intro; destruct H; discriminate.
    right; intro; destruct H; discriminate.
  Qed.

Inductive step (s:state) : wterm->state->wterm->Prop :=
(* while loops *)
| st_while_true: forall w e t z z',
    evalTrue e w ->
    iteration_decr z z' ->
    step s (w: tm_while_max e z t) s (w: t;; (tm_while_max e z' t))
| st_while_false: forall w e t z,
    evalFalse e w ->
    step s (w: tm_while_max e z t) s (w: tm_skip)
| st_while_go: forall w e t i z,
    s_iter_get s i = Some iter_go ->
    step s (w: tm_while_max e (i_stall i 0 z) t) s (w: tm_while_max e z t)
| st_while_limit: forall w e t,
    step s (w: tm_while_max e 0 t) s (w: tm_skip)
(* if statements *)
| st_if_true: forall w e t1 t2,
    evalTrue e w ->
    step s (w: tm_if e t1 t2) s (w: t1)
| st_if_false: forall w e t1 t2,
    evalFalse e w ->
    step s (w: tm_if e t1 t2) s (w: t2)
(* thread communicateion *)
| st_send: forall w w' w_w' wl s' ec c e v R q,
    eval ec (cname_value c) w_w' ->
    eval e v w_w' ->
    s_get_ch s c = Some (R,wl,q) ->
    can_send c w_w' ->
    w' |= R v ->
    s' = (s_push_ch s c w' v) ->
    gjoin w w' (w_prod_incr c w_w') ->
    step s (w_w': tm_send ec e) s' (w: tm_skip)
| st_recv: forall w w' w_w' wl s' ec c x v q R,
    eval ec (cname_value c) w ->
    LocalsMap.map_share x (w_env w) = Share.top ->
    can_recv c w ->
    w' |= R v ->
    gjoin w w' w_w' ->
    s' = (s_pop_ch s c) ->
    s_get_ch s c = Some (R,w'::wl,v::q) ->
    step s (w: tm_recv x ec) s' (w_cons_incr c (w_env_set w_w' x v): tm_skip)
| st_recv_blocked: forall w ec c x R,
    eval ec (cname_value c) w ->
    LocalsMap.map_share x (w_env w) = Share.top ->
    can_recv c w ->
    s_get_ch s c = Some (R,nil,nil) ->
    step s (w: tm_recv x ec) s (w: tm_recv x ec)
| st_newch: forall w x c w' R s',
    c = FreshChannels.fc_c (fresh_channel s w) ->
    w' = w_env_set (FreshChannels.fc_w (fresh_channel s w)) x (cname_value c) ->
    s' = s_new_ch s (fresh_channel s w) R ->
    step s (w: tm_newch x R) s' (w': tm_skip)
| st_endch: forall ec c w w' w_w' s' R,
    eval ec (cname_value c) w ->
    total_send_recv c w' ->
    gjoin w w' w_w' -> 
    s_get_ch s c = Some(R,nil,nil) ->
    s' = s_rm_ch s c ->
    step s (w_w': tm_endch ec) s' (w:tm_skip)
(* miscellaneous *)
| st_assert: forall w P,
    w |= gsep_con P TT ->
    step s (w: tm_assert P) s (w: tm_skip)
| st_assign: forall w x e v w',
    eval e v w ->
    w_env_share w x = Share.top ->
    w' = w_env_set w x v ->
    step s (w: tm_assign x e) s (w': tm_skip)
| st_load: forall w x e addr v w',
    eval e (addr_value addr) w ->
    w_heap_get w addr = Some v ->
    w' = w_env_set w x v ->   
    step s (w: tm_load x e) s (w': tm_skip)
| st_store: forall w e_dest dest_addr e_value v w',
    eval e_value v w ->
    eval e_dest (addr_value dest_addr) w ->
    w_heap_share w dest_addr = Share.top ->
    w' = w_heap_set w dest_addr v ->   
    step s (w: tm_store e_dest e_value) s (w': tm_skip)
(*| st_local: forall w x e v t w' y fresh_y,
    eval e v w ->
    fresh_y = fresh_var w v ->
    y = FreshLocals.fl_x fresh_y ->
    w' = FreshLocals.fl_w fresh_y ->
    step s (w: tm_local x e t) s (w': term_subst t x y;; tm_freelocal y)
*)
| st_local: forall w e v f w' x,
    eval e v w ->
    gjoin w (new_local x v) w' ->
    step s (w: tm_local e f) s (w': f x)
| st_freelocal: forall w x w' v,
    gjoin w' (new_local x v) w ->
    step s (w: tm_freelocal x) s (w': tm_skip)
(* Stepping at term level *)
| st_seq: forall w t1 t2,
    step s (w: tm_seq t1 t2) s (w: t1;; t2)
| st_par: forall w1 w2 w (P1 P2: pred world) t1 t2,
    w1 |= gsep_con P1 TT ->
    w2 |= gsep_con P2 TT ->
    gjoin w1 w2 w ->
    step s (w: tm_par P1 t1 P2 t2) s (w1:t1 || w2:t2)
(* Stepping at wterm level *)
| st_seq1: forall b1 s' b1' t2,
    step s b1 s' b1' ->
    step s (b1;; t2) s' (b1';; t2)
| st_seq_skip: forall w t2,
    step s (w: tm_skip;; t2) s (w: t2)
| st_par1: forall s' b1 b2 b1',
    step s b1 s' b1' ->
    step s (b1 || b2) s' (b1' || b2)
| st_par2: forall s' b1 b2 b2',
    step s b2 s' b2' ->
    step s (b1 || b2) s' (b1 || b2')
| st_par_join: forall w1 w2 w,
    gjoin w1 w2 w ->
    step s (w1:tm_skip || w2:tm_skip) s (w:tm_skip)
.





(*
where "s1 / t1 --> s2 / t2" := (step s1 (t1%wterm) s2 (t2%wterm)).
where "s1 / t1 --> s2 / t2" := (step s1 (t1%term) s2 (t2%term)).
*)

Inductive step_star s b : state->wterm->Prop :=
| step_nil:
    step_star s b s b
| step_cons: forall s' s'' b' b'',
    step s b s' b' ->
    step_star s' b' s'' b'' ->
    step_star s b s'' b''
| st_iteration: forall s' s'' b'',
    (* collapse or increment the known value of iteration variable z *)
    s_iter s s' ->
    step_star s' b s'' b'' ->
    step_star s b s'' b''
.

Inductive step_star_sans_s_iter s b : state->wterm->Prop :=
| step_nil_:
    step_star_sans_s_iter s b s b
| step_cons_: forall s' s'' b' b'',
    step s b s' b' ->
    step_star_sans_s_iter s' b' s'' b'' ->
    step_star_sans_s_iter s b s'' b''
.

End OperationalSemantics.

 Tactic Notation "step_cases" tactic(first) ident(c) :=
  first;
  [ Case_aux c "st_while_true" | Case_aux c "st_while_false"|
   Case_aux c "st_while_go" | Case_aux c "st_while_limit" |
   Case_aux c "st_if_true" | Case_aux c "st_if_false" | Case_aux c "st_send" | Case_aux c "st_recv" |
   Case_aux c "st_recv_blocked" | Case_aux c "st_newch" | Case_aux c "st_endch" |
   Case_aux c "st_assert" |
   Case_aux c "st_assign" |
   Case_aux c "st_load" |
   Case_aux c "st_store" |
   Case_aux c "st_local" | Case_aux c "st_freelocal" |
   Case_aux c "st_seq" |
   Case_aux c "st_par" | Case_aux c "st_seq1" | Case_aux c "st_seq_skip" | Case_aux c "st_par1" |
   Case_aux c "st_par2" | Case_aux c "st_par_join" ].

Tactic Notation "step_cases" tactic(first) ident(c) ident(H) :=
  first;
  [ Case_aux c "st_while_true"; pose st_while_true as H  | Case_aux c "st_while_false"; pose st_while_false as H |
   Case_aux c "st_while_go"; pose st_while_go as H | Case_aux c "st_while_limit"; pose st_while_limit as H |
   Case_aux c "st_if_true"; pose st_if_true as H | Case_aux c "st_if_false"; pose st_if_false as H | Case_aux c "st_send"; pose st_send as H | Case_aux c "st_recv"; pose st_recv as H |
   Case_aux c "st_recv_blocked"; pose st_recv_blocked as H | Case_aux c "st_newch"; pose st_newch as H | Case_aux c "st_endch"; pose st_endch as H |
   Case_aux c "st_assert"; pose st_assert as H |
   Case_aux c "st_assign"; pose st_assign as H |
   Case_aux c "st_load"; pose st_load as H | Case_aux c "st_store"; pose st_store as H |
   Case_aux c "st_local"; pose st_local as H | Case_aux c "st_freelocal"; pose st_freelocal as H |
   Case_aux c "st_seq"; pose st_seq as H |
   Case_aux c "st_par"; pose st_par as H | Case_aux c "st_seq1"; pose st_seq1 as H | Case_aux c "st_seq_skip"; pose st_seq_skip as H | Case_aux c "st_par1"; pose st_par1 as H |
   Case_aux c "st_par2"; pose st_par2 as H | Case_aux c "st_par_join"; pose st_par_join as H ].


  Tactic Notation "match_left" ident(b) :=
    let x:= fresh b in
    match goal with
    | [ |- context [ match ?a with inleft _ => _ | inright _ => _ end] ] => let H:=fresh "H" in (destruct a as [ [x] | H ]; [ | false; eapply H; eauto ])
    end.
  Tactic Notation "match_gjoin_left_destruct_exactly" constr(a) constr(w1) constr(w2) constr(w) :=
      match a with
      | gjoin_dec w1 w2 =>
          let w' := fresh "w" in let H':= fresh "H" in
            destruct a as [ [w' H'] | H' ];
            [ assert (w'=w); [eapply (@join_eq world gworld_sa w1 w2); eauto | subst w']
            | false; eapply H'; eauto]
      | gjoin_dec w2 w1 =>
          let w' := fresh "w" in let H':= fresh "H" in
            destruct a as [ [w' H'] | H' ];
            [ assert (w'=w); [eapply (@join_eq world gworld_sa w1 w2); eauto | subst w']
            | false; eapply H'; eapply (@join_com world gworld_sa); eauto]
      end.

  Tactic Notation "match_gjoin_left_exactly" constr(w1) constr(w2) constr(w) :=
    match goal with
    | [ |- context [ match ?a with inleft _ => _ | inright _ => _ end ] ] =>
      match_gjoin_left_destruct_exactly a w1 w2 w
    | [ HH: context [ match ?a with inleft _ => _ | inright _ => _ end ] |- _ ] => 
      match_gjoin_left_destruct_exactly a w1 w2 w
    end.

  Tactic Notation "match_gjoin_left":=
    match reverse goal with
    | H: @join world gworld_sa ?w1 ?w2 ?w |- _  => match_gjoin_left_exactly w1 w2 w
    | H: gjoin ?w1 ?w2 ?w |- _  => match_gjoin_left_exactly w1 w2 w
    end.


Section Misc.
  Import Notation.
  Import world_model.Notation.
  Import State.Morphisms.
  Bind Scope wterm_scope with wterm.
  Open Scope forces_pred.
  Require Import MSetPositive.

Fixpoint term_size (t:term) := match t with
| tm_while_max e z t' => 1 + (term_size t')
| tm_if e t1 t2 => 1 + (term_size t1) + (term_size t2)
| tm_skip => 1
| tm_seq t1 t2 => 1 + (term_size t1) + (term_size t2)
| tm_par P1 t1 P2 t2 => 1 + (term_size t1) + (term_size t2)
| tm_send c e => 1
| tm_recv x c => 1
| tm_newch _ _ => 1
| tm_endch _ => 1
| tm_assert _ => 1
| tm_assign _ _ => 1
| tm_load _ _ => 1
| tm_store _ _ => 1
| tm_local _ _ => 1
| tm_freelocal _ => 1
end.

Fixpoint wterm_size (b:wterm) := match b with
| wtm_term _ t => term_size t
| wtm_seq b1 t2 => 1 + (wterm_size b1) + (term_size t2)
| wtm_par b1 b2 => 1 + (wterm_size b1) + (wterm_size b2)
end.

Lemma term_size_intro: forall t1 t2,
  t1 = t2 -> term_size t1 = term_size t2.
Proof.
  intros; rewrite H; auto.
Qed.

Lemma wterm_size_intro: forall t1 t2,
  t1 = t2 -> wterm_size t1 = wterm_size t2.
Proof.
  intros; rewrite H; auto.
Qed.

Lemma term_size_different: forall t1 t2,
  term_size t1 <> term_size t2 -> t1 <> t2.
Proof.
  intros.
  intro.
  subst.
  apply H.
  auto.
Qed.

Lemma wterm_size_different: forall t1 t2,
  wterm_size t1 <> wterm_size t2 -> t1 <> t2.
Proof.
  intros.
  intro.
  subst.
  apply H.
  auto.
Qed.

Ltac absurd_rec_def :=
  match goal with
  | [ H : (?t = ?E) |- _ ] => match E with
    | context[t] => elimtype False; apply term_size_intro in H; simpl in H; omega
    end
  | [ H : (?E = ?t) |- _ ] => match E with
    | context[t] => elimtype False; apply term_size_intro in H; simpl in H; omega
    end
  end.




  Lemma step_seq_inversion: forall s b t s' b',
    step s (b;; t) s' b' ->
    (exists b'0, step s b s' b'0 /\ b'=b'0;; t)%wterm
    \/ (exists w, b=w:tm_skip /\ b'=w:t /\ s'=s)%wterm.
  Proof.
    intros; inverts H; eauto; left; exists b; splits; eauto; apply st_iteration; auto.
  Qed.


  Lemma step_commute_front_s_iter: forall s b s' b' s'',
    step s b s' b' -> s_iter s' s'' ->
    exists s0, s_iter s s0 /\ step s0 b s'' b'.
  Proof.
    intros.
    step_cases (induction H) Case HH;
      try solve [
        eexists; split; [eauto | ]; eapply HH; clear HH; eauto
      | destruct IHstep as [s0[??]]; auto; eexists; split; eauto; eapply HH; clear HH; eauto].

    eexists; split; eauto; eapply HH; clear HH; eauto.
    inverts H0.
    destruct (Pos.eq_dec i i0); subst.
    false.
    rewrite s_iter_get_set_neq; auto.

    subst.
    edestruct s_push_ch_s_iter_commute1 as [s'0[??]]; eauto; subst.
    eexists; split; eauto; eapply HH; clear HH; eauto.
    erewrite s_get_ch_s_iter_noninterference; eauto.

    subst.
    edestruct s_pop_ch_s_iter_commute1 as [s'0[??]]; eauto; subst.
    eexists; split; eauto; eapply HH; clear HH; eauto.
    erewrite s_get_ch_s_iter_noninterference; eauto.

    eexists; split; eauto; eapply HH; clear HH; eauto.
    erewrite s_get_ch_s_iter_noninterference; eauto.

    subst.
    edestruct s_new_ch_s_iter_commute1 as [s'0[??]]; [eauto|]; subst.
    exists s'0; split; auto.
    subst.
    eapply HH with (c:= FreshChannels.fc_c (fresh_channel s w)).
    apply fresh_channel_cname_total_world_eq.
    intros. setoid_rewrite s_get_ch_s_iter_noninterference with (s:=s); auto; reflexivity.
    apply s_iter_is_alloced_ch; auto.
    f_equal.
    apply fresh_channel_world_total_world_eq.
    intros. setoid_rewrite s_get_ch_s_iter_noninterference with (s:=s); auto; reflexivity.
    apply s_iter_is_alloced_ch; auto.
    reflexivity.

    subst.
    edestruct s_rm_ch_s_iter_commute1 as [s'0[??]]; eauto; subst.
    eexists; split; eauto; eapply HH; clear HH; eauto.
    erewrite s_get_ch_s_iter_noninterference; eauto.
  Qed.

  Lemma step_commute_front_s_iter_star: forall s b s' b' s'',
    step s b s' b' -> s_iter_star s' s'' ->
    exists s0, s_iter_star s s0 /\ step s0 b s'' b'.
  Proof.
    intros.
    revert s b b' H.
    induction H0; intros.
    eexists; split; eauto; apply s_iter_nil.
    edestruct step_commute_front_s_iter as [s0'[??]]; eauto.
    edestruct IHs_iter_star as [s0''[??]]; eauto.
    eexists; split; eauto.
    eapply s_iter_cons; eauto.
  Qed.

  Lemma step_star_commute_front_s_iter_star: forall s b s' b',
    step_star s b s' b' ->
    exists s0,
      s_iter_star s s0 /\ step_star_sans_s_iter s0 b s' b'.
  Proof.
    intros.
    induction H.
    eexists; split. apply s_iter_nil. apply step_nil_.
    edestruct IHstep_star as [s0[??]].
    edestruct step_commute_front_s_iter_star as [s0'[??]]; eauto.
    eexists; split; eauto.
    eapply step_cons_; eauto.
    edestruct IHstep_star as [s0[??]].
    eexists; split; eauto.
    eapply s_iter_cons; eauto.
  Qed.

  Definition single_stepping_wterm (b: wterm):= forall s s' b',
    step s b s' b' -> (s'=s/\b'=b) \/ exists w', b'=(w':tm_skip)%wterm.

  Definition steps_respect_state (b: wterm):= forall s s' b',
    step_star s b s' b' -> s'=s.


  Definition simple_step_wterm b:= forall s s' b',
    step s b s' b' -> s'=s /\ (b'=b \/ exists w', b'=w':tm_skip)%wterm.
  Definition state_independent_step_wterm b:= forall s b',
    step s b s b' -> forall s', step s' b s' b'.




(*
Lemma step_star_blocking_no_resolution: forall s w (t:sterm) s' tw',
  step s (w:t) s (w:t) ->
  step_star s (w:t) s' tw' ->
  tw' = (w:t)%wterm.
Proof.
  intros.
  remember (w: t)%wterm as tw.
  revert w t Heqtw H.
  induction H0; intros; subst; auto.
  edestruct step_sterm_no_par as [w'' [t1'' [? | [t2'' ?]]]].
  apply H.
  subst t'.
  edestruct sterm_step_deterministic.
  apply H1. apply H.
  subst s'.
  inversion H3; subst w'' t1''.
  eapply IHstep_star; eauto.

  subst t'.
  edestruct sterm_step_deterministic.
  apply H1. apply H.
  subst s'.
  inversion H3.
Qed.
*)

(*
Lemma step_join_wterms_preservation: forall s b s' b' w,
  step s b s' b' -> join_wterms b w -> exists w', join_wterms b' w'.
Admitted.

Lemma step_star_join_wterms_preservation: forall s b s' b' w,
  step_star s b s' b' -> join_wterms b w -> exists w', join_wterms b' w'.
Proof.
  intros.
  revert w H0.
  induction H; intros; eauto.
  edestruct step_join_wterms_preservation; eauto.
Qed.
*)

Fixpoint extend_world b w :=
  match b with
  | wtm_term w' t => match gjoin_dec w w' with
      | inleft (exist w'' _) => wtm_term w'' t
      | inright _ => b
      end
  | wtm_seq b' t => wtm_seq (extend_world b' w) t
  | wtm_par b1 b2 => wtm_par (extend_world b1 w) b2
end.

Fixpoint just_term b : term :=
  match b with
  | wtm_term w' t => t
  | wtm_seq b' t => tm_seq (just_term b') t
  | wtm_par b1 b2 => match join_wterms_dec b1, join_wterms_dec b2 with
      | inleft (exist w1 _), inleft (exist w2 _) => tm_par (exactly w1) (just_term b1) (exactly w2) (just_term b2)
      | _,_ =>tm_par FF (just_term b1) FF (just_term b2)
      end
  end.


  Lemma join_wterms_extend_world1: forall b wb wF w,
    join_wterms b wb ->
    gjoin wb wF w ->
    join_wterms (extend_world b wF) w.
  Proof.
    induction b; intros.
    simpl in *.
    subst wb.
    match_gjoin_left.
    reflexivity.
    simpl in *.
    eapply IHb; eauto.
    simpl in *.
    destruct H as [w1[w2[?[??]]]].
    edestruct @join_assoc with (H:=gworld_sa) (a:=w2) (b:=w1) as [w1F[??]]; eauto.
    exists w1F w2; splits; auto.
    eapply IHb1; eauto.
    eapply @join_com; eauto.
  Qed.

  Lemma join_wterms_extend_world2: forall b wF w,
    join_wterms (extend_world b wF) w ->
    exists wb, join_wterms b wb /\ (gjoin wb wF w \/ extend_world b wF=b).
  Proof.
    induction b; intros; simpl in *.
    destruct gjoin_dec as [[ww]|?].
    simpl in H; subst w0.
    exists w; split; auto.
    left.
    eapply @join_com; eauto.
    simpl in H; subst w0.
    exists w; split; auto.
    (**)
    edestruct IHb as [wb[??]]; eauto.
    exists wb; split; auto.
    destruct H1; auto.
    right; f_equal. auto.
    (**)
    destruct H as [w1[w2[?[??]]]].
    clear IHb2.
    edestruct IHb1 as [wb[?[?|?]]]; eauto.
    edestruct @join_assoc with (H:=gworld_sa) (a:=wF) (b:=wb) as [wb2[??]]; eauto.
    exists wb2; split; eauto.
    left; eapply @join_com; eauto.
    rewrite H3 in H.
    assert (w1=wb); try subst w1.
      eapply join_wterms_eq; eauto.
    exists w; split; eauto.
    right; f_equal; auto.
  Qed.    


  Lemma extend_world_compose: forall b wF1 wF2 wF wb wbF,
    gjoin wF1 wF2 wF ->
    join_wterms b wb ->
    gjoin wb wF wbF ->
    extend_world (extend_world b wF1) wF2 = extend_world b wF.
  Proof.
    induction b; intros.
    simpl in *.
    subst w.
    match_gjoin_left.
    simpl.
    edestruct @join_assoc with (H:=gworld_sa) (a:=wF2) (b:=wF1) as [wbF1[??]]; eauto.
    match_gjoin_left.
    simpl.
    match_gjoin_left.
    reflexivity.
    (**)
    simpl in *.
    f_equal; eauto.
    (**)
    simpl in *.
    f_equal.
    destruct H0 as [w1[w2[?[??]]]].
    edestruct @join_assoc with (H:=gworld_sa) (a:=w2) (b:=w1) as [w1F[??]]; eauto.
  Qed.

  Lemma extend_world_simpl: forall w1 w2 w t,
    gjoin w1 w2 w ->
    extend_world (w1: t)%wterm w2 = (w: t)%wterm.
  Proof.
    intros.
    simpl.
    apply join_com in H.
    destruct (gjoin_dec w2 w1) as [[?]|?].
    f_equal.
    eapply @join_eq with (H:=gworld_sa); eauto.
    false; eapply n; eauto.
  Qed.

  Lemma extend_world_term_id: forall w t wF,
    ~gjoins w wF ->
    (extend_world (w: t) wF = w: t)%wterm.
  Proof.
    intros.
    simpl.
    destruct gjoin_dec as [[wwF] | ?]; auto.
    false; eapply H; eexists; eauto.
  Qed.


(*
Lemma join_wupd_var: forall w1 w2 w x v,
  w1 |= defined x -> join w1 w2 w ->
  join (wupd_var w1 x v) w2 (wupd_var w x v).
Proof.
  intros.
  destruct H as [v'].
  destruct H as [?[?[??]]].
  destruct H as [?[??]].
  unfold wupd_var.

  destruct (LocalsMap.map_upd x v (w_env w1)).
  destruct (LocalsMap.map_upd x v (w_env w)).
  split; simpl; try apply H0.
  split; simpl; try apply H0.
Admitted.  


*)


Lemma step_spin_no_change: forall s b s',
  step s b s' b -> s'=s.
Proof.
  intros.
  remember b as b1 in H at 1.
  remember b as b2 in H at 1.
  revert b Heqb1 Heqb2.
  induction H; intros; try subst; try subst b s'; auto; inverts Heqb2; eauto.
Qed.



Lemma join_wterms_par: forall b1 b1' b2 b2' w1 w2 w,
  join_wterms (b1||b2)%wterm w ->
  join_wterms b1 w1 -> join_wterms b1' w1 ->
  join_wterms b2 w2 -> join_wterms b2' w2 ->
  join_wterms (b1'||b2')%wterm w.
Admitted.

(*
Lemma joins_wupd_var: forall w1 w2 x v,
  w1 |= defined x ->
  joins w1 w2 -> joins (wupd_var w1 x v) w2.
Proof.
  intros.
  destruct H0 as [w].
  exists (wupd_var w x v).
  apply join_wupd_var; auto.
Qed.

Lemma joins_wupd_var': forall w1 w2 x v,
  LocalsMap.map_share x (w_env w1) = Share.top ->
  joins w1 w2 -> joins (wupd_var w1 x v) w2.
Proof.
  intros.
  destruct H0 as [w].
  exists (wupd_var w x v).
  apply join_wupd_var'; auto.
Qed.

Lemma joins_wupd_var'': forall w1 w2 x v,
  joins (wupd_var w1 x v) w2 -> joins w1 w2.
Proof.
Admitted.
Lemma joins_cons_incr: forall w1 w2 c,
  joins (w_cons_incr c w1) w2 <-> joins w1 w2.
Proof.
Admitted.
*)

End Misc.

Ltac assert_undef H :=
  first [ generalize H; false | idtac H "was undefined"].


Ltac absurd_rec_def :=
  match goal with
  | [ H : (?t = ?E) |- _ ] => match E with
    | context[t] => elimtype False; apply term_size_intro in H; simpl in H; omega
    end
  | [ H : (?E = ?t) |- _ ] => match E with
    | context[t] => elimtype False; apply term_size_intro in H; simpl in H; omega
    end
  end.





