Require Import Machine.
Require Import StepStar.
Require Import ContextSemantics.
Require Import world_model.
Require Import msl.msl_standard.
Require Import SfLib.
Require Import LibTactics.
Require Import WorldLogic.
Require Import StepStarCs.
Require Import HoareLogic.
Require Import IterationBounds.
Require Export Bisimulation.

Section refactor.
  Import Machine.Notation.
  Import ContextSemantics.Notation.
  Open Scope context_scope.
  Open Scope wterm_scope.

  Lemma step_star_step_star_sans_s_iter: forall s b s' b',
    step_star_sans_s_iter s b s' b' -> step_star s b s' b'.
  Proof.
    induction 1; econstructor; eauto.
  Qed.

  Lemma step_star_sans_s_iter_respects_s_i: forall s b s' b',
    step_star_sans_s_iter s b s' b' -> s_i s = s_i s'.
  Proof.
    intros.
    induction H; auto.
    apply step_s_iter in H.
    etransitivity; eauto.
  Qed.

  Lemma step_star_remove_par2_skip: forall s b1 w2 s' b1' w2',
    step_star s (b1 || w2:tm_skip) s' (b1' || w2':tm_skip) ->
    step_star s b1 s' b1' /\ w2'=w2.
  Proof.
    intros.
    edestruct step_star_par_skip_resolve as [b1'_[?[?|?]]]; eauto.
    inverts H1; auto.
    destruct H1 as [?[?[?[??]]]]; false.
  Qed.

  Lemma step_star_commute_front_s_iter: forall s b s' b' s'',
    step_star s b s' b' -> s_iter s' s'' ->
    exists s'0,
      s_iter s s'0 /\ step_star s'0 b s'' b'.
  Proof.
    intros.
    induction H.
    eexists; split; eauto; apply step_nil.
    destruct IHstep_star as [s'0[??]]; auto.
    edestruct step_commute_front_s_iter as [s'00[??]]; eauto.
    exists s'00; split; auto.
      eapply step_cons; eauto.
    exists s'; split; auto.
    eapply step_star_iter_snoc; eauto.
  Qed.

  Lemma step_star_sans_s_iter_commute_s_iter1: forall s b s' b' s'',
    step_star_sans_s_iter s b s' b' -> s_iter s' s'' ->
    exists s0,
      s_iter s s0 /\ step_star_sans_s_iter s0 b s'' b'.
  Proof.
    intros.
    revert s'' H0.
    induction H; intros.
    exists s''; split; auto; apply step_nil_.
    edestruct IHstep_star_sans_s_iter as [s0'[??]]; eauto.
    edestruct step_commute_front_s_iter as [s0[??]]; eauto.
    exists s0; split; auto.
    eapply step_cons_; eauto.
  Qed.

  Lemma s_iter_get_eq_s_i: forall s s' i,
    s_i s = s_i s' -> s_iter_get s i = s_iter_get s' i.
  Proof.
    intros. destruct s; destruct s'; simpl in *; subst; auto.
  Qed.

  Lemma s_iter_change': forall s s',
    s_iter s s' -> s_i s <> s_i s'.
  Proof.
    intros.
    inverts H.
    intro.
    apply s_iter_get_eq_s_i with (i:=i) in H.
    rewrite s_iter_get_set_eq in H.
    false.
  Qed.

  Lemma s_iter_star_app: forall s s' s'',
    s_iter_star s s' -> s_iter_star s' s'' -> s_iter_star s s''.
  Proof.
    intros.
    induction H; auto.
    eapply s_iter_cons; eauto.
  Qed.

  Lemma s_iter_False: forall s s',
    s_iter s s' -> s_iter s' s -> False.
  Proof.
    intros.
    inverts H; inverts H0.
    destruct (Pos.eq_dec i i0).
    subst i0.
    repeat rewrite s_iter_set_set_eq in H.
    rewrite s_iter_get_set_eq in H.
    discriminate.
    rewrite s_iter_set_set_neq' in H; auto.
    rewrite s_iter_set_set_eq' in H.
    rewrite s_iter_get_set_eq in H.
    discriminate.
  Qed.

  Lemma s_iter_stall_pred: forall i s s',
    s_iter s s' ->
    s_iter_get s' i = Some iter_stall ->
    s_iter_get s i = Some iter_stall.
  Proof.
    intros.
    inverts H.
    destruct (Pos.eq_dec i i0).
    subst i0.
    auto.
    rewrite s_iter_get_set_neq in H0; auto.
  Qed.

  Lemma s_iter_star_change: forall s s',
    s_iter_star s s' ->
    s' = s \/ (exists i, s_iter_get s i = Some iter_stall /\ s_iter_get s' i = Some iter_go).
  Proof.
    intros.
    induction H; auto.
    destruct IHs_iter_star as [?|[i[??]]].
    subst s''.
    inverts H.
    right.
    exists i; split; auto.
    rewrite s_iter_get_set_eq; auto.
    right.
    exists i; split; auto.
    eapply s_iter_stall_pred; eauto.
  Qed.


  Lemma s_iter_star_antisymm: forall s s',
    s_iter_star s s' ->
    s_iter_star s' s ->
    s'=s.
  Proof.
    intros.
    edestruct s_iter_star_change with (s:=s) as [? | [i[??]] ]; eauto.
    eapply s_iter_star_go_preservation in H2; eauto.
    false.
  Qed.

  Lemma step_star_respecting_s_i_s_iter: forall s0 s b s' b',
    s_i s0 = s_i s' ->
    s_iter s0 s ->
    step_star s b s' b' ->
    False.
  Proof.
    intros.
    inverts H0.
    apply step_star_go_preservation with (i:=i) in H1.
    apply s_iter_get_eq_s_i with (i:=i) in H.
    false.
    rewrite s_iter_get_set_eq; auto.
  Qed.

  Lemma step_star_s_iter_step_star: forall s b s' b',
    step_star s b s' b' -> s_i s = s_i s' ->
    step_star_sans_s_iter s b s' b'.
  Proof.
    intros.
    induction H.
    apply step_nil_.
    eapply step_cons_; eauto.
    apply IHstep_star.
    apply step_s_iter in H.
    rewrite<-H; rewrite H0; reflexivity.
    false; clear IHstep_star.
    eapply step_star_respecting_s_i_s_iter; eauto.
  Qed.

  Lemma step_star_sans_s_iter_app: forall s b s' b' s'' b'',
    step_star_sans_s_iter s b s' b' ->
    step_star_sans_s_iter s' b' s'' b'' ->
    step_star_sans_s_iter s b s'' b''.
  Proof.
    intros.
    induction H; auto.
    eapply step_cons_; eauto.
  Qed.


  Lemma gjoins_gwemp2: forall w, gjoins gwemp w.
  Proof. eexists; apply gjoin_bot2. Qed.
  Lemma gjoins_gwemp1: forall w, gjoins w gwemp.
  Proof. eexists; apply gjoin_bot1. Qed.

End refactor.

Module PermuteTactics.
(* largely taken from http://permalink.gmane.org/gmane.science.mathematics.logic.coq.club/4211
by Jesse A. Tov,  23 Nov 06:02 
*)

  Require Import Permutation.
Ltac focus_env_rec X L :=
  match L with
    | X           => constr:(nil ++ L ++ nil)
    | X ++ ?L2    => constr:(nil ++ X ++ L2)
    | ?L1 ++ X    => constr:(L1 ++ X ++ nil)
    | _ ++ X ++ _ => L
    | ?L1 ++ ?L2  =>
      match L1 with
        | context [X] => match focus_env_rec X L1 with
                           ?L11 ++ X ++ ?L12 =>
                           constr:(L11 ++ X ++ (L12 ++ L2))
                         end
        | _           => match focus_env_rec X L2 with
                           ?L21 ++ X ++ ?L22 =>
                           constr:((L1 ++ L21) ++ X ++ L22)
                         end
      end
  end.

(** Re-associates a list in the conclusion to focus on some sublist X. *)
Tactic Notation "focus_env" constr(X) :=
  match goal with
    [ |- context [?L] ] =>
    let L' := focus_env_rec X L in
    change L with L'
  end.

(** The number of appended components in a (right-associated) list. *)
Ltac app_length L :=
  match L with
    | _ ++ ?L' => let len' := app_length L' in constr:(S len')
    | _ => constr:1
  end.

Lemma Permutation_add_front_mid {A} {C E F1 F2: list A} :
  Permutation E (F1 ++ F2) ->
  Permutation (C ++ E) (F1 ++ C ++ F2).
Proof.
  induction C; simpl; eauto using Permutation_cons_app.
Qed.

Lemma Permutation_swap_left {A} {E1 E2 F: list A} :
  Permutation (E1 ++ E2) F ->
  Permutation (E2 ++ E1) F.
Proof. intros. setoid_rewrite Permutation_app_swap; auto. Qed.

Lemma Permutation_swap_right {A} {E F1 F2: list A} :
  Permutation E (F1 ++ F2) ->
  Permutation E (F2 ++ F1).
Proof. intros. setoid_rewrite Permutation_app_swap; auto. Qed.

Lemma Permutation_add_mid {A: Type} {G E1 E2 F1 F2: list A} :
  Permutation (E1 ++ E2) (F1 ++ F2) ->
  Permutation (E1 ++ G ++ E2) (F1 ++ G ++ F2).
Proof.
  intros.
  erewrite<- Permutation_add_front_mid.
  2: symmetry; eauto.
  apply Permutation_add_front_mid.
  reflexivity.
Qed.


Ltac rotate_left  := apply Permutation_swap_left; repeat rewrite app_ass.
Ltac rotate_right := apply Permutation_swap_right; repeat rewrite app_ass.

(** Temporarily hide some subterm of the conclusion while performing
    the given tactic. *)
Tactic Notation "hiding" constr(C) "[" tactic(tac) "]" :=
  let temp := fresh "temp" in
    set (temp := C); tac; subst temp.

Tactic Notation "hiding" "left" "[" tactic(tac) "]" :=
  match goal with
    [ |- Permutation ?E _ ] => hiding E [ tac ]
  end.

Tactic Notation "hiding" "right" "[" tactic(tac) "]" :=
  match goal with
    [ |- Permutation _ ?F ] => hiding F [ tac ]
  end.


Lemma cons_app: forall {A} (x y:A) l, x::y::l = [x]++(y::l).
Proof. intros. reflexivity. Qed.


Ltac permut4 :=
  repeat rewrite cons_app; 
  let rec loop i  :=
    match goal with
      | [ |- Permutation (?C ++ ?E) ?F ] =>
        first
          [ hiding (C ++ E) [ focus_env C ];
            apply Permutation_add_front_mid
          | rotate_left ];
        match i with
          | S ?i' => loop i'
          | _     => idtac
        end
      | _ => idtac
    end in
  match goal with [ |- Permutation ?E _ ] =>
    let n := app_length E in
      loop n
  end; try reflexivity.

Ltac permut3 :=
  repeat rewrite cons_app; 
  let rec loop i j m :=
    match goal with
      | [ |- Permutation (?C ++ _) (?C ++ _) ] =>
        apply Permutation_app_head;
        match constr:(i, m) with
          | (S ?i', S ?m') => loop i' m' m'
          | _              => idtac
        end
      | [ |- Permutation (_ ++ _) (_ ++ _) ] =>
        match j with
          | S ?j' => rotate_right; loop i j' m
          | _     =>
            match i with
              | S ?i' => rotate_left; loop i' m m
              | _     => idtac
            end
        end
      | _ => idtac
    end in
  match goal with [ |- Permutation ?E ?F ] =>
    let n := app_length E in
    let m := app_length F in
    loop n m m
  end; try reflexivity.

Ltac permut2 :=
  repeat rewrite cons_app; 
  repeat match goal with
    [ |- Permutation ?E _ ] =>
    match E with context [?X] =>
      hiding E [ focus_env X ];
      focus_env X;
      apply Permutation_add_mid
    end end; try reflexivity.
End PermuteTactics.


Module JoinList.
  Import Permutation.
  Export PermuteTactics.
  

  Fixpoint join_list wl w := match wl with
  | w1::wl => exists w', join_list wl w' /\ gjoin w1 w' w
  | nil => w=gwemp
  end.

  Lemma join_list_app: forall w1 w2 w wl1 wl2,
    join_list wl1 w1 ->
    join_list wl2 w2 ->
    gjoin w1 w2 w ->
    join_list (wl1++wl2) w.
  Proof.
    intros.
    revert w1 w2 w wl2 H H0 H1.
    induction wl1; intros; simpl in *; subst.
    apply gworld_unit_identity in H1. subst; auto.
    destruct H as [w'[??]].
    edestruct gjoin_assoc with (a:=a) (b:=w') as [?[??]]; eauto.
  Qed.

  Lemma join_list_cons: forall w1 w2 w wl,
    join_list wl w2 ->
    gjoin w1 w2 w ->
    join_list (w1::wl) w.
  Proof.
    intros.
    exists w2; split; auto.
  Qed.

  Lemma join_list_cons_gwemp: forall w wl,
    join_list (gwemp::wl) w <-> join_list wl w.
  Proof.
    split; intros.
    destruct H as [w1[??]].
    apply gworld_unit_identity in H0; subst; auto.
    exists w; split; auto.
    apply gjoin_bot2.
  Qed.


  Lemma join_list_perm: forall w wl1 wl2,
    join_list wl1 w ->
    Permutation wl1 wl2 ->
    join_list wl2 w.
  Proof.
    introv H0 H.
    revert w H0.
    induction H; intros; auto.
    (**)
    destruct H0 as [w1[??]].
    exists w1; auto.
    destruct H0 as [w1[[w2[??]]?]].
    edestruct gjoin_assoc with (a:=x) (b:=w2) as [?[??]]; eauto.
    apply gjoin_com; eauto.
    exists x0; split; auto.
    exists w2; split; auto.
    apply gjoin_com; auto.
  Qed.


  Lemma join_list_app_inv: forall w wl1 wl2,
    join_list (wl1++wl2) w ->
    exists w1 w2,
      join_list wl1 w1 /\ join_list wl2 w2 /\ gjoin w1 w2 w.
  Proof.
    intros.
    revert wl1 wl2 H.
    induction wl1; intros; simpl in *; subst.
    exists gwemp w; splits; auto.
    apply gjoin_bot2.
    destruct H as [w'[??]].
    edestruct IHwl1 with (wl2:= a::wl2) as [w1[w2[?[??]]]]; eauto.
    apply join_list_perm with (wl1:=a:: (wl1++wl2)).
    exists w'; split; auto.
    apply Permutation_middle.
    destruct H2 as [wa[??]].
    edestruct gjoin_assoc with (a:=wa) (b:=a) (c:=w1) as [?[??]]; try apply gjoin_com; eauto.
    exists x wa; splits; auto.
    exists w1; split; auto.
    apply gjoin_com; auto.
  Qed.

  Lemma join_list_eq: forall wl w w',
    join_list wl w -> join_list wl w' -> w' = w.
  Proof.
    intros. revert w w' H H0.
    induction wl; intros.
    congruence.
    destruct H as [?[??]].
    destruct H0 as [?[??]].
    replace x0 with x in *; auto.
    eapply gjoin_eq; eauto.
  Qed.

  Lemma join_list_group: forall wl1 wl2 w1 w,
    join_list wl1 w1 ->
    (join_list (wl1 ++ wl2) w <-> join_list (w1::wl2) w).
  Proof.
    split; intros.

    edestruct join_list_app_inv as [w1'[w2[?[??]]]]; eauto.
    replace w1' with w1 in *.
    exists w2; split; auto.
    eapply join_list_eq; eauto.

    destruct H0 as [?[??]].
    eapply join_list_app; eauto.
  Qed.

Require Import Setoid.
Require Export Morphisms.
Export SetoidTactics.
Require Export SetoidClass.
Import Permutation.

  Add Parametric Morphism: join_list
  with signature (@Permutation world) ==> eq ==> equiv
  as join_list_m.
    split; intros.
    eapply join_list_perm; eauto.
    symmetry in H.
    eapply join_list_perm; eauto.
  Qed.

  Lemma join_list_group1: forall w1 wl1 wl2 wl wl' w,
    join_list wl1 w1 ->
    Permutation wl (wl1++wl2) ->
    Permutation wl' (w1::wl2) ->
    join_list wl w -> join_list wl' w.
  Proof.
    intros.
    setoid_rewrite H0 in H2.
    setoid_rewrite H1.
    apply-> join_list_group; eauto.
  Qed.
  Lemma join_list_group2: forall w1 wl1 wl2 wl wl' w,
    join_list wl1 w1 ->
    Permutation wl (wl1++wl2) ->
    Permutation wl' (w1::wl2) ->
    join_list wl' w -> join_list wl w.
  Proof.
    intros.
    setoid_rewrite H0.
    setoid_rewrite H1 in H2.
    apply<- join_list_group; eauto.
  Qed.

  Lemma join_list_single: forall w1 w2,
    w1=w2 <-> join_list ([w1]) w2.
  Proof. split; intros; subst. exists gwemp. split. reflexivity. apply gjoin_bot1. destruct H as [w2'[??]]. simpl in H; subst. apply gjoin_com in H0. apply gworld_unit_identity in H0. auto.  Qed.

  Lemma gjoin_join_list: forall w1 w2 w,
    gjoin w1 w2 w <-> join_list ([w1,w2]) w.
  Proof.
    split; intros.
    exists w2; splits; auto. apply join_list_single; auto.
    destruct H as [w2_[??]].
    apply join_list_single in H; subst w2_; auto.
  Qed.

Import Machine.Notation.
Open Scope wterm_scope.

  Fixpoint wterm_worlds b := match b with
  | w: t => w::nil
  | b;; t => wterm_worlds b
  | b1 || b2 => wterm_worlds b1 ++ wterm_worlds b2
  end%wterm.
  Lemma join_wterm_worlds: forall b w,
    join_wterms b w <-> join_list (wterm_worlds b) w.
  Proof.
    induction b; simpl; intros; eauto.
    split; intros.
    subst; exists gwemp; split; auto; apply gjoin_bot1.
    destruct H as [w'[??]]; subst.
    eapply gworld_unit_identity; apply gjoin_com; auto.
    split; [intros [w1[w2[?[??]]]] | intros ].
    eapply join_list_app; eauto; [apply IHb1|apply IHb2]; eauto. 
    edestruct join_list_app_inv as [w1[w2[?[??]]]]; eauto.
    do 2 eexists; splits; eauto.
    apply IHb1; eauto.
    apply IHb2; eauto.
  Qed.


Ltac cons_app l := match l with
  ?a :: ( ?b :: ?c ) => match cons_app (b::c) with
         ?d ++ ?e => constr:( [a] ++ d ++ e)
       |  ?d => constr:([a] ++ d)
    end
| ?a ++ ?b => match cons_app a with ?c => match cons_app b with ?d => constr:(c ++ d) end end
| ?a => a
end.
  Lemma join_list_group1': forall w ww wl1 wl2 wl3,
    join_list wl3 w ->
    join_list (wl1 ++ wl3 ++ wl2) ww ->
    join_list (wl1 ++ [w] ++ wl2) ww.
  Proof.
    intros.
    eapply join_list_group1 with (w1:=w) (wl1:=wl3) (wl2:=wl1++wl2); eauto.
    simpl.
    setoid_rewrite Permutation.Permutation_middle.
    reflexivity.
    eapply join_list_perm; eauto.
    permut3.
  Qed.
  Lemma join_list_group2': forall w ww wl1 wl2 wl3,
    join_list wl3 w ->
    join_list (wl1 ++ [w] ++ wl2) ww ->
    join_list (wl1 ++ wl3 ++ wl2) ww.
  Proof.
    intros.
    eapply join_list_group2 with (w1:=w) (wl1:=wl3) (wl2:=wl1++wl2); eauto.
    permut3.
    eapply join_list_perm; eauto.
    simpl.
    setoid_rewrite Permutation.Permutation_middle.
    reflexivity.
  Qed.
Ltac do_jrewrite1 H1 H2 :=
  let helper wl1 w1 wl2 w2 :=
    let wl2'' := cons_app wl2 in
    let wl2' := focus_env_rec [w1] wl2'' in
    match wl2' with
      ?wlA ++ [w1] ++ ?wlB => change wl2 with wl2' in H2;
           apply (join_list_group2' w1 w2 wlA wlB wl1 H1) in H2
    end in
  match type of H1 with
  | join_list ?wl1 ?w1 => match type of H2 with
      | join_list ?wl2 ?w2 => helper wl1 w1 wl2 w2
      end
  | join_wterms ?b1 ?w1 =>
      let H1':= fresh "H1" in
      generalize H1; intro H1';
      apply join_wterm_worlds in H1'; simpl in H1';
      do_jrewrite1 H1' H2;
      clear H1'
  end.
Tactic Notation "jrewrite<-" hyp(H1) "in" hyp(H2) := do_jrewrite1 H1 H2.
Ltac do_jrewrite2 H1 H2 :=
  (let helper wl1 w1 wl2 w2 :=
    let wl2'' := cons_app wl2 in
    let wl2' := focus_env_rec wl1 wl2'' in
    match wl2' with
      ?wlA ++ wl1 ++ ?wlB => let H:= fresh "H" in
        assert (H: Permutation.Permutation wl2 wl2') by solve [permut2 | permut3 | permut4];
        setoid_rewrite H in H2;
        apply (join_list_group1' w1 w2 wlA wlB wl1 H1) in H2;
        clear H
    end in
  match type of H1 with
  | join_list ?wl1 ?w1 => match type of H2 with
      | join_list ?wl2 ?w2 => helper wl1 w1 wl2 w2
      end
  | join_wterms ?b1 ?w1 =>
      let H1':= fresh "H1" in
      generalize H1; intro H1';
      apply join_wterm_worlds in H1'; simpl in H1';
      do_jrewrite2 H1' H2;
      clear H1'
  end).
Tactic Notation "jrewrite" hyp(H1) "in" hyp(H2) := do_jrewrite2 H1 H2.
Tactic Notation "jrewrite->" hyp(H1) "in" hyp(H2) := do_jrewrite2 H1 H2.



End JoinList.







Section Definitions.

(*
  Record ObserverFilter := MakeObserverFilter
  { O_filter: state->context->Prop;
    O_preservation: forall s C b s' C' b', O_filter s C -> cstep s C b s' C' b' -> O_filter s' C'
  }.

  Coercion O_filter : ObserverFilter >-> Funclass.
*)
  Import Machine.Notation.
  Import ContextSemantics.Notation.
  Import Permutation.
  Import JoinList.
  Open Scope wterm_scope.


  Fixpoint context_world C wC := match C with
  | cx_hole => wC=gwemp
  | cx_seq C t => context_world C wC
  | cx_par1 C1 b2 => exists wC1 w2, context_world C1 wC1 /\ join_wterms b2 w2 /\ gjoin wC1 w2 wC
  | cx_par2 b1 C2 => exists w1 wC2, join_wterms b1 w1 /\ context_world C2 wC2 /\ gjoin w1 wC2 wC
  end.


  Lemma context_world_join_wterms: forall C wC,
    context_world C wC <-> join_wterms (set_context C (gwemp:tm_skip)) wC.
  Proof.
    split; intros.

    revert wC H.
    induction C; simpl; intros; auto.
    destruct H as [wC1[w2[?[??]]]].
    exists wC1 w2; splits; auto.
    destruct H as [w1[wC2[?[??]]]].
    exists w1 wC2; splits; auto.

    revert wC H.
    induction C; simpl; intros; auto.
    destruct H as [w1[w2[?[??]]]].
    exists w1 w2; splits; auto.
    destruct H as [w1[w2[?[??]]]].
    exists w1 w2; splits; auto.
  Qed.

(*
  (* wo joins with every world in b, but the worlds of b might not join with themselves *)
  Fixpoint joins_wterm wo b := match b with
  | wtm_term w t => gjoins wo w
  | wtm_seq b t => joins_wterm wo b
  | wtm_par b1 b2 => joins_wterm wo b1 /\ joins_wterm wo b2
  end.

  Definition joins_with_list wo wl :=
    List.Forall (fun w' => gjoins wo w') wl.

  Lemma joins_with_list_nil: forall w,
    joins_with_list w nil.
  Proof. intros; red; auto. Qed.
  Lemma joins_with_list_cons: forall w wO wlO,
    gjoins w wO -> joins_with_list w wlO ->
    joins_with_list w (wO::wlO).
  Proof. intros; red; constructor; auto. Qed.
  Lemma joins_with_list_app: forall w wlO1 wlO2,
    joins_with_list w wlO1 -> joins_with_list w wlO2 ->
    joins_with_list w (wlO1++wlO2).
  Proof.
    intros; red. red in H; red in H0.
    rewrite List.Forall_forall. intros.
    edestruct in_app_or; eauto.
    rewrite List.Forall_forall in H; auto.
    rewrite List.Forall_forall in H0; auto.
  Qed.
*)



End Definitions.



Section Definitions2.
  Import Machine.Notation.
  Import ContextSemantics.Notation.
  Import Permutation.
  Open Scope wterm_scope.
  Import PermuteTactics.
  Import JoinList.

  Definition flatten_list {A:Type} (l: list (list A)) :=
    List.fold_right (fun (l1:list A) (l2: list A) => l1++l2) nil l.

  Definition ch_state_worlds s :=
    flatten_list (List.map (fun x => snd (fst (snd x))) (FMapPositive.PositiveMap.elements (s_c s))).

  Definition join_state s wS := 
    join_list (ch_state_worlds s) wS.

  Definition valid_machine_state wO s b :=
    exists wS wb ww,
      join_state s wS /\ join_wterms b wb /\ join_list ([wS,wb,wO]) ww.


  Lemma w_cons_prod_incr: forall w c,
    w_cons (w_prod_incr c w) = w_cons w.
  Proof. intros; apply w_cons_wupd_prod. Qed.
  Lemma w_env_prod_incr: forall w c,
    w_env (w_prod_incr c w) = w_env w.
  Proof. intros; apply w_env_wupd_prod. Qed.
  Lemma w_heap_prod_incr: forall w c,
    w_heap (w_prod_incr c w) = w_heap w.
  Proof. intros; apply w_heap_wupd_prod. Qed.
  Lemma gjoin_w_prod_incr: forall w1 w2 w c,
    can_send c w1 ->
    gjoin w1 w2 w ->
    gjoin (w_prod_incr c w1) w2 (w_prod_incr c w).
  Proof.
    intros.
    split; split; unfold fst,snd.
    destruct H0 as [[? _] _].
    unfold fst,snd in H0.
    intros c'.
    destruct w1 as [w1_p w1_c w1_m w1_h].
    destruct w2 as [w2_p w2_c w2_m w2_h].
    destruct w as [w_p w_c w_m w_h].
    apply join_channels_incr1.
    apply H.
    apply H0.
    do 2 rewrite w_cons_prod_incr; apply H0.
    do 2 rewrite w_env_prod_incr; apply H0.
    do 2 rewrite w_heap_prod_incr; apply H0.
  Qed.
  Lemma gjoin_w_prod_incr': forall w1 w2 w c,
    gjoin (w_prod_incr c w1) w2 (w_prod_incr c w) ->
    gjoin w1 w2 w.
  Proof.
    intros.
    destruct H as [[??][??]]; split; split; unfold fst,snd in *.
    intros c'.
    destruct w1 as [w1_p w1_c w1_m w1_h].
    destruct w2 as [w2_p w2_c w2_m w2_h].
    destruct w as [w_p w_c w_m w_h].
    eapply join_channels_decr1; apply H.
    do 2 rewrite w_cons_prod_incr in H0; apply H0.
    do 2 rewrite w_env_prod_incr in H1; apply H1.
    do 2 rewrite w_heap_prod_incr in H2; apply H2.
  Qed.
  Lemma w_prod_cons_incr: forall w c,
    w_prod (w_cons_incr c w) = w_prod w.
  Proof. intros; apply w_prod_wupd_cons. Qed.
  Lemma w_env_cons_incr: forall w c,
    w_env (w_cons_incr c w) = w_env w.
  Proof. intros; apply w_env_wupd_cons. Qed.
  Lemma w_heap_cons_incr: forall w c,
    w_heap (w_cons_incr c w) = w_heap w.
  Proof. intros; apply w_heap_wupd_cons. Qed.
  Lemma gjoin_w_cons_incr: forall w1 w2 w c,
    can_recv c w1 ->
    gjoin w1 w2 w ->
    gjoin (w_cons_incr c w1) w2 (w_cons_incr c w).
  Proof.
    intros.
    split; split; unfold fst,snd.
    do 2 rewrite w_prod_cons_incr; apply H0.
    destruct H0 as [[_ ?] _].
    unfold fst,snd in H0.
    intros c'.
    destruct w1 as [w1_p w1_c w1_m w1_h].
    destruct w2 as [w2_p w2_c w2_m w2_h].
    destruct w as [w_p w_c w_m w_h].
    apply join_channels_incr1.
    apply H.
    apply H0.
    do 2 rewrite w_env_cons_incr; apply H0.
    do 2 rewrite w_heap_cons_incr; apply H0.
  Qed.

  Lemma s_get_ch_supd_ch: forall s ch c,
    s_get_ch (supd_ch s ch) c = FMapPositive.PositiveMap.find c ch.
  Proof. intros; destruct s; reflexivity. Qed.
  Lemma s_ch_get_push_same': forall R w wl v vl s c,
    s_get_ch (s_push_ch s c w v) c = Some (R, wl, vl) ->
    exists wl' vl',
      s_get_ch s c = Some (R, wl', vl') /\ wl=wl'++[w] /\ vl=vl'++[v].
  Proof.
    intros.
    unfold s_push_ch,s_ch_apply in H.
    rewrite s_get_ch_supd_ch in H.
    destruct (s_get_ch s c) as [[[R' wl'] vl']|].
    rewrite POrdProps.P.F.add_eq_o in H; auto.
    inverts H.
    exists wl' vl'; splits; auto.
    rewrite POrdProps.P.F.remove_eq_o in H; auto.
    discriminate.
  Qed.

Import Permutation.
  Lemma ch_state_worlds_s_push_ch: forall s c w v,
    Permutation (ch_state_worlds (s_push_ch s c w v)) (w::ch_state_worlds s).
  Proof.
Admitted.
  Lemma ch_state_worlds_s_pop_ch: forall s c w v R wl vl,
    s_get_ch s c = Some (R,w::wl,v::vl) ->
    Permutation (w::ch_state_worlds (s_pop_ch s c)) (ch_state_worlds s).
  Proof.
Admitted.

  Lemma join_state_s_push_ch: forall s wS wS' wc c v,
    join_state s wS ->
    gjoin wc wS wS' ->
    join_state (s_push_ch s c wc v) wS'.
  Proof.
    intros.
    red.
    setoid_rewrite ch_state_worlds_s_push_ch.
    eapply join_list_cons; eauto.
  Qed.

(*
  Lemma joins_state_s_push_ch: forall w wS wS' s c v,
    join_state s wS ->
    gjoin w wS wS' ->
    join_state (s_push_ch s c w v) wS'.
  Proof.
    intros.
    red.
    setoid_rewrite ch_state_worlds_s_push_ch.
    eapply join_list_cons; eauto.
  Qed.
*)

(*
  Lemma step_frame: forall s w1 w2 w b s' b',
    step s b s' b' ->
    join_wterms b w1 ->
    joins_state w2 s ->
    gjoin w1 w2 w ->
    exists w1' w',
      gjoin w1' w2 w' /\ join_wterms b' w1' /\ joins_state w2 s'.
  Proof.
    intros.
Admitted.

  Lemma step_star_frame: forall s w1 w2 w b s' b',
    step_star s b s' b' ->
    join_wterms b w1 ->
    joins_state w2 s ->
    gjoin w1 w2 w ->
    exists w1' w',
      gjoin w1' w2 w' /\ join_wterms b' w1' /\ joins_state w2 s'.
  Proof.
    intros.
    revert w1 w2 w H0 H1 H2.
    induction H; intros.
    exists w1 w; splits; auto; apply step_nil.
    edestruct step_frame as [w1'[w'[?[??]]]]; eauto.
    edestruct IHstep_star as [w1'[w'[?[??]]]]; eauto.
    apply-> joins_state_s_iter; eauto.
  Qed.
*)

(*
  Lemma joins_state_w_cons_incr1: forall s w c,
    can_recv c w -> joins_state w s -> joins_state (w_cons_incr c w) s.
  Proof.
    repeat intro.
    apply H0 in H1.
    clear R vl c0.
    revert w s H H0 H1.
    induction wl; intros.
    apply joins_with_list_nil.
    inverts H1.
    apply joins_with_list_cons; eauto.
    destruct H4 as [w2 ?].
    exists (w_cons_incr c w2); eapply gjoin_w_cons_incr; eauto.
  Qed.
*)

  Lemma identity_channel_sa_simpl: forall t n e,
    @identity _ channel_sa.sa_ce (exist (fun a : Share.t * nat => identity (fst a) -> snd a = 0) 
         (t, n) e) <-> identity t.
  Proof.
    split; intros.
    repeat intro.
    unfold identity in H.
    pose (fun a => (forall a0 b : Share.t, join (fst a) a0 b -> a0 = b) -> snd a = 0) as P.
    assert (P (a,0)).
    subst P; simpl; intros; auto.
    assert (P (b,n)).
    subst P; simpl; intros; auto.
    apply split_identity with (c:=b) in H0.
    apply e; auto.
    intro; intros.
    apply H2; auto.
    spec H (exist P (a,0) H1).
    spec H (exist P (b,n) H2).
    spec H.
    simpl.
    split; simpl. apply H0.
    auto.
    inverts H.
    auto.
    intro; intros.
    destruct a; destruct b.
    apply exist_ext.
    destruct x; destruct x0.
    destruct H0. simpl in *.
    apply H in H0. subst.
    rewrite e; auto.
  Qed.
  Lemma join_channels_decr3: forall c ch1 ch2 ch3',
    has_access c ch1 ->
    join (channels_incr c ch1) ch2 ch3' ->
    exists ch3, ch3' = channels_incr c ch3.
  Proof.
Admitted.
  Lemma gjoin_w_prod_incr_inv: forall c w1 w2 w,
    can_send c w1 ->
    gjoin (w_prod_incr c w1) w2 w ->
    exists w', w = w_prod_incr c w'.
  Proof.
    intros.
    destruct H0 as [[? _]_].
    unfold fst,snd,w_prod_incr in *.
    rewrite w_prod_wupd_prod in H0.
    edestruct join_channels_decr3 as [w2' ?]; eauto.
    exists (wupd_prod w w2').
    destruct w.
    simpl; f_equal; auto.
  Qed.

(*
  Lemma joins_state_w_prod_incr2: forall s w c,
    can_send c w -> joins_state (w_prod_incr c w) s -> joins_state w s.
  Proof.
    repeat intro.
    apply H0 in H1.
    clear R vl c0.
    revert w s H H0 H1.
    induction wl; intros.
    apply joins_with_list_nil.
    inverts H1.
    apply joins_with_list_cons; eauto.
    destruct H4 as [w2 ?].
    edestruct gjoin_w_prod_incr_inv as [w2' ?]; eauto.
    subst.
    apply gjoin_w_prod_incr' in H1.
    exists w2'; auto.
  Qed.
*)



  Lemma can_send_w_prod_incr: forall w c,
    can_send c w <-> can_send c (w_prod_incr c w).
  Proof.
    destruct w.
    split; intros; intro; apply H; simpl in *.
    apply channels_incr_None in H0; congruence.
    apply channels_incr_None; congruence.
  Qed.

  Lemma can_recv_w_cons_incr: forall w c,
    can_recv c w <-> can_recv c (w_cons_incr c w).
  Proof.
    destruct w.
    split; intros; intro; apply H; simpl in *.
    apply channels_incr_None in H0; congruence.
    apply channels_incr_None; congruence.
  Qed.

  Lemma can_recv_w_env_set: forall w c x v,
    can_recv c w <-> can_recv c (w_env_set w x v).
  Proof.
    destruct w.
    split; intros; intro; apply H; simpl in *.
    rewrite w_cons_env_set in H0; auto.
    rewrite w_cons_env_set; auto.
  Qed.

  Lemma w_env_set_inv: forall w1 w2 w x v,
    gjoin w1 w2 (w_env_set w x v) ->
    w_env_share w1 x = Share.top ->
    exists w1',
      w1 = w_env_set w1' x v.
  Proof.
Admitted.

  Lemma join_state_s_iter: forall s wS s',
    s_iter s s' ->
    join_state s wS ->
    join_state s' wS.
  Proof.
    intros.
    destruct s; destruct s'.
    apply s_iter_respects_s_c in H. simpl in H.
    subst s_c0.
    apply H0.
  Qed.


  Lemma valid_machine_state_s_push_ch: forall wO c wb b wb' b' wc s v,
    can_send c wb ->
    gjoin wb' wc wb ->
    join_wterms b wb ->
    join_wterms b' wb' ->
    valid_machine_state wO s b ->
    valid_machine_state wO (s_push_ch s c wc v) b'.
  Proof.
    intros.
    destruct H3 as [wS[wb_[ww[?[??]]]]].
    replace wb_ with wb in *.
    eapply join_list_group2 with (w1:=wb) (wl1:=[wb',wc]) (wl2:=[wS,wO]) in H5; eauto.
    edestruct join_list_app_inv with (wl1:=[wS,wc]) (wl2:=[wO,wb']) as [wSc[wOb'[?[??]]]]; eauto.
    eapply join_list_perm; eauto.
    permut4.
    exists wSc wb' ww; splits; auto.
    eapply join_state_s_push_ch; eauto.
    apply gjoin_com.
    apply gjoin_join_list; auto.
    eapply join_list_group1 with (w1:=wSc) (wl2:=[wb',wO]) (wl1:=[wS,wc]); eauto.
    eapply join_list_perm; eauto.
    permut4.
    apply gjoin_join_list; auto.
    permut4.
    eapply join_wterms_eq; eauto.
  Qed.

  Lemma join_list_cons_inv: forall w w' w1 wl,
    gjoin w1 w' w ->
    join_list (w1::wl) w ->
    join_list wl w'.
  Proof.
    intros.
    destruct H0 as [w'_[??]].
    replace w'_ with w' in *; auto.
    eapply gjoin_canc; eapply gjoin_com; eauto.
  Qed.

  Lemma join_state_s_pop_ch: forall s wS wc c R wl v vl,
    join_state s wS ->
    s_get_ch s c = Some (R, wc::wl, v::vl) ->
    exists wS',
      gjoin wc wS' wS /\ join_state (s_pop_ch s c) wS'.
  Proof.
Admitted.

  Lemma valid_machine_state_s_pop_ch: forall wO c wb b wb' b' wc s R wl v vl,
    can_recv c wb ->
    gjoin wb wc wb' ->
    s_get_ch s c = Some (R, wc::wl, v::vl) ->
    join_wterms b wb ->
    join_wterms b' wb' ->
    valid_machine_state wO s b ->
    valid_machine_state wO (s_pop_ch s c) b'.
  Proof.
    intros.
    destruct H4 as [wS[wb_[ww[?[??]]]]].
    replace wb_ with wb in *.
    edestruct join_state_s_pop_ch as [wS'[??]]; eauto.
    exists wS' wb' ww; splits; auto.
    eapply join_list_group1 with (w1:=wb') (wl2:=[wS',wO]) (wl1:=[wb,wc]); eauto.
    eapply gjoin_join_list; eauto.
    permut4.
    eapply join_list_group2 with (w1:=wS) (wl2:=[wb,wO]) (wl1:=[wc,wS']); eauto.
    eapply gjoin_join_list; eauto.
    permut4.
    eapply join_wterms_eq; eauto.
  Qed.

  Lemma valid_machine_state_w_prod_incr1: forall wO s c w t,
    can_send c w ->
    valid_machine_state wO s (w_prod_incr c w: t) ->
    valid_machine_state wO s (w: t).
  Proof.
    introv Hcan ?.
    destruct H as [wS[wb[ww[?[??]]]]].
    simpl in H0; subst wb.
    change ([wS,w_prod_incr c w,wO]) with ([wS,w_prod_incr c w]++[wO]) in H1.
    edestruct join_list_app_inv as [wSw'[wO_[?[??]]]]; eauto.
    apply join_list_single in H2; subst wO_.
    apply gjoin_join_list in H0.
    apply gjoin_com in H0.
    edestruct gjoin_w_prod_incr_inv as [w' ?]; eauto.
    subst wSw'.
    edestruct gjoin_w_prod_incr_inv with (w1:=w') as [ww' ?]; eauto.
    eapply can_send_sub; eauto.
    exists wS.
    eapply gjoin_w_prod_incr'; eauto.
    subst ww.
    exists wS w ww'; splits; auto.
    reflexivity.
    eapply join_list_perm with (wl1:=[wO,w,wS]); eauto.
    exists w'; split.
    apply gjoin_join_list; auto.
    eapply gjoin_w_prod_incr'; eauto.
    apply gjoin_com.
    eapply gjoin_w_prod_incr'; eauto.
    permut4.
  Qed.

  Lemma valid_machine_state_w_prod_incr2: forall wO s c w t,
    can_send c w ->
    valid_machine_state wO s (w: t) ->
    valid_machine_state wO s (w_prod_incr c w: t).
  Proof.
    introv Hcan ?.
    destruct H as [wS[wb[ww[?[??]]]]].
    simpl in H0; subst wb.
    change ([wS,w,wO]) with ([wS,w]++[wO]) in H1.
    edestruct join_list_app_inv as [wSw[wO_[?[??]]]]; eauto.
    apply join_list_single in H2; subst wO_.
    apply gjoin_join_list in H0.
    apply gjoin_com in H0.
    exists wS (w_prod_incr c w) (w_prod_incr c ww); splits; auto.
    reflexivity.
    eapply join_list_perm with (wl1:=[wO,w_prod_incr c w,wS]); eauto.
    exists (w_prod_incr c wSw); split.
    apply gjoin_join_list; auto.
    eapply gjoin_w_prod_incr; eauto.
    apply gjoin_com.
    eapply gjoin_w_prod_incr; eauto.
    eapply can_send_sub; eauto.
    exists wS; auto.
    permut4.
  Qed.

  Lemma valid_machine_state_w_cons_incr2: forall wO s c w t1 t2,
    can_recv c w ->
    valid_machine_state wO s (w: t1) ->
    valid_machine_state wO s (w_cons_incr c w: t2).
  Proof.
    introv Hcan ?.
    destruct H as [wS[wb[ww[?[??]]]]].
    simpl in H0; subst wb.
    change ([wS,w,wO]) with ([wS,w]++[wO]) in H1.
    edestruct join_list_app_inv as [wSw[wO_[?[??]]]]; eauto.
    apply join_list_single in H2; subst wO_.
    apply gjoin_join_list in H0.
    apply gjoin_com in H0.
    exists wS (w_cons_incr c w) (w_cons_incr c ww); splits; auto.
    reflexivity.
    eapply join_list_perm with (wl1:=[wO,w_cons_incr c w,wS]); eauto.
    exists (w_cons_incr c wSw); split.
    apply gjoin_join_list; auto.
    eapply gjoin_w_cons_incr; eauto.
    apply gjoin_com.
    eapply gjoin_w_cons_incr; eauto.
    eapply can_recv_sub; eauto.
    exists wS; auto.
    permut4.
  Qed.

  Lemma w_env_share_sub: forall w1 w2 w x,
    gjoin w1 w2 w ->
    join_sub (w_env_share w x) (w_env_share w1 x).
  Proof.
Admitted.
  Lemma w_env_share_sub_top: forall  w1 w2 w x,
    w_env_share w1 x = Share.top ->
    gjoin w1 w2 w ->
    w_env_share w x = Share.top.
Admitted.

  Lemma valid_machine_state_w_env_set2: forall wO s x v w t,
    valid_machine_state wO s (w: t) ->
    valid_machine_state wO s (w_env_set w x v: t).
  Proof.
    intros.
    destruct H as [wS[wb[ww[?[??]]]]].
    simpl in H0; subst wb.
    destruct (Share.eq_dec (w_env_share w x) Share.top).
    exists wS (w_env_set w x v) (w_env_set ww x v); splits; auto.
    reflexivity.
    destruct H1 as [wwO[??]].
    apply gjoin_join_list in H0.
    exists (w_env_set wwO x v); splits; auto.
    apply gjoin_join_list.
    apply gjoin_w_env_set'; auto.
    apply gjoin_com.
    apply gjoin_w_env_set'.
    eapply w_env_share_sub_top; eauto.
    apply gjoin_com; auto.
    rewrite w_env_set_not_top; auto.
    exists wS w ww; splits; auto.
    reflexivity.
  Qed.

  Lemma gjoin_w_heap_set': forall w1 w2 w a v,
    MemMap.map_share a (w_heap w1) = Share.top ->
    gjoin w1 w2 w -> gjoin (w_heap_set w1 a v) w2 (w_heap_set w a v).
  Proof.
Admitted.

  Lemma w_heap_share_sub_top: forall w1 w2 w a,
    w_heap_share w1 a = Share.top ->
    gjoin w1 w2 w -> w_heap_share w a = Share.top.
Admitted.

  Lemma w_heap_set_not_top: forall w a v,
    w_heap_share w a <> Share.top -> w_heap_set w a v = w.
Admitted.

  Lemma valid_machine_state_w_heap_set2: forall wO s a v w t,
    valid_machine_state wO s (w: t) ->
    valid_machine_state wO s (w_heap_set w a v: t).
  Proof.
    intros.
    destruct H as [wS[wb[ww[?[??]]]]].
    simpl in H0; subst wb.
    destruct (Share.eq_dec (w_heap_share w a) Share.top).
    exists wS (w_heap_set w a v) (w_heap_set ww a v); splits; auto.
    reflexivity.
    destruct H1 as [wwO[??]].
    apply gjoin_join_list in H0.
    exists (w_heap_set wwO a v); splits; auto.
    apply gjoin_join_list.
    apply gjoin_w_heap_set'; auto.
    apply gjoin_com.
    apply gjoin_w_heap_set'.
    eapply w_heap_share_sub_top; eauto.
    apply gjoin_com; auto.
    rewrite w_heap_set_not_top; auto.
    exists wS w ww; splits; auto.
    reflexivity.
  Qed.

  Lemma ch_state_worlds_s_new_ch: forall s fresh_ch R,
    ch_state_worlds (s_new_ch s fresh_ch R) = ch_state_worlds s.
  Proof.
Admitted.

  Lemma join_state_s_new_ch: forall s wS fresh_ch R,
    join_state s wS ->
    join_state (s_new_ch s fresh_ch R) wS.
  Proof.
    intros.
    unfold join_state in *.
    rewrite ch_state_worlds_s_new_ch; auto.
  Qed.

  Lemma valid_machine_state_s_new_ch: forall wO b s fresh_ch R,
    valid_machine_state wO s b ->
    valid_machine_state wO (s_new_ch s fresh_ch R) b.
  Proof.
    intros.
    destruct H as [wS[wb[ww[?[??]]]]].
    exists wS wb ww; splits; auto.
    apply join_state_s_new_ch; auto.
  Qed.

  Lemma valid_machine_state_join_sub: forall wO s b wb b' wb',
    join_wterms b wb ->
    join_wterms b' wb' ->
    gjoin_sub wb' wb ->
    valid_machine_state wO s b ->
    valid_machine_state wO s b'.
  Proof.
    intros.
    destruct H2 as [wS[wb_[ww[?[??]]]]].
    replace wb_ with wb in *.
    destruct H1 as [w ?].
    eapply join_list_group2 with (w1:=wb) (wl1:=[wb',w]) (wl2:=[wS,wO]) in H4; eauto.
    eapply join_list_perm with (wl2:=[w,wb',wS,wO]) in H4.
    destruct H4 as [ww'[??]].
    exists wS wb' ww'; splits; auto.
    eapply join_list_perm; eauto.
    permut4.
    permut2.
    apply gjoin_join_list; auto.
    permut4.
    eapply join_wterms_eq; eauto.
  Qed.

  (* w joins with every world in wl *)
  Definition joins_with_list w wl :=
    List.Forall (fun w' => gjoins w w') wl.

  (* wo joins with every world in s, but the worlds of s might not join with themselves *)
  Definition joins_with_state wo s :=
    forall c R wl vl, s_get_ch s c = Some (R, wl, vl) -> joins_with_list wo wl.

  Lemma fresh_channel_local_props': forall s w fresh_ch,
    fresh_ch = fresh_channel s w ->
    let c := FreshChannels.fc_c fresh_ch in
    let w' := FreshChannels.fc_w fresh_ch in
    w_prod w c = None /\ w_cons w c = None /\ gjoin w (new_channel c) w' /\ joins_with_state (new_channel c) s.
Admitted.

  Lemma join_state_joins_with_state: forall s w wS,
    join_state s wS ->
    joins_with_state w s ->
    exists wS',
      gjoin w wS wS'.
Admitted.

  Lemma valid_machine_state_new_channel: forall wO s wb b fresh_ch,
    valid_machine_state wO s b ->
    join_wterms b wb ->
    fresh_ch = fresh_channel s wb ->
    valid_machine_state wO s (FreshChannels.fc_w fresh_ch: tm_skip).
Proof.
(* not QUITE true: need to modify fresh_channel so that it won't return a
channel name that will conflict with wO. *)
Admitted.

  Lemma ch_state_worlds_s_rm_ch: forall s c R,
    s_get_ch s c = Some (R, [], []) ->
    ch_state_worlds (s_rm_ch s c) = ch_state_worlds s.
Admitted.

  Lemma join_state_s_rm_ch: forall s wS c R,
    join_state s wS ->
    s_get_ch s c = Some (R, [], []) ->
    join_state (s_rm_ch s c) wS.
  Proof.
    unfold join_state; intros.
    erewrite ch_state_worlds_s_rm_ch; eauto.
  Qed.

  Lemma valid_machine_state_s_rm_ch: forall wO b s c R,
    valid_machine_state wO s b ->
    s_get_ch s c = Some (R, [], []) ->
    valid_machine_state wO (s_rm_ch s c) b.
  Proof.
    intros.
    destruct H as [wS[wb[ww[?[??]]]]].
    exists wS wb ww; splits; auto.
    eapply join_state_s_rm_ch; eauto.
  Qed.

  Lemma valid_machine_state_new_local: forall wO s w w' b b' x v,
    valid_machine_state wO s b ->
    gjoin w (new_local x v) w' ->
    join_wterms b w ->
    join_wterms b' w' ->
    valid_machine_state wO s b'.
Proof.
(* NOT TRUE, (but morally true): need to modify st_local so that any fresh variable chosen
does not conflict with either wO or s *)
Admitted.

  Lemma valid_machine_state_wterm: forall wO s b1 b2 wb,
    valid_machine_state wO s b1 ->
    join_wterms b1 wb ->
    join_wterms b2 wb ->
    valid_machine_state wO s b2.
  Proof.
    introv [wS[wb_[ww[?[??]]]]] ? ?.
    replace wb_ with wb in *.
    exists wS wb ww; splits; auto.
    eapply join_wterms_eq with (b:=b1); eauto.
  Qed.

  Lemma valid_machine_state_s_iter: forall wO s b s',
    s_iter s s' ->
    valid_machine_state wO s b ->
    valid_machine_state wO s' b.
  Proof.
    intros.
    destruct H0 as [wS[wb[ww[?[??]]]]].
    exists wS wb ww; splits; auto.
    eapply join_state_s_iter; eauto.
  Qed.



  Lemma valid_machine_state_par_observer: forall wO1 wO2 wO bO s b,
    join_wterms bO wO2 ->
    gjoin wO1 wO2 wO ->
    valid_machine_state wO1 s (b||bO) -> valid_machine_state wO s b.
  Proof.
    intros.
    destruct H1 as [wS[wbbO[ww[?[??]]]]].
    destruct H2 as [wb[wO2'[?[??]]]].
    replace wO2' with wO2 in *.
    exists wS wb ww; splits; auto.
    eapply join_list_group1 with (w1:=wO) (wl2:=[wS,wb]); eauto.
    apply gjoin_join_list in H0; eauto.
    permut4.
    eapply join_list_group2 with (wl1:=[wb,wO2]) (w1:=wbbO) (wl2:=[wO1,wS]); eauto.
    apply gjoin_join_list; auto.
    permut4.
    eapply join_list_perm; eauto.
    permut4.
    eapply join_wterms_eq; eauto.
  Qed.

  Lemma valid_machine_state_observer_par: forall wO1 wO2 wO bO s b,
    join_wterms bO wO2 ->
    gjoin wO1 wO2 wO ->
    valid_machine_state wO s b ->
    valid_machine_state wO1 s (b||bO).
  Proof.
    intros.
    destruct H1 as [wS[wb[ww[?[??]]]]].
    eapply join_list_group2 with (w1:=wO) (wl1:=[wO1,wO2]) (wl2:=[wS,wb]) in H3; eauto.
    apply join_list_perm with (wl2:=[wb,wO2]++[wO1,wS]) in H3.
    edestruct join_list_app_inv as [wbO2[wSO1[?[??]]]]; eauto.
    exists wS wbO2 ww; splits; auto.
    exists wb wO2; splits; auto.
    apply gjoin_join_list; auto.
    eapply join_list_group1 with (w1:=wbO2) (wl1:=[wb,wO2]) (wl2:=[wS,wO1]); eauto.
    permut4.
    eapply join_list_app; eauto.
    eapply join_list_perm; eauto. permut4.
    permut4.
    apply gjoin_join_list; auto.
    permut4.
  Qed.

  Lemma valid_machine_state_par_observer': forall wO1 bO s b,
    valid_machine_state wO1 s (b||bO) ->
    exists wO2 wO,
      join_wterms bO wO2 /\ gjoin wO1 wO2 wO /\ valid_machine_state wO s b.
  Proof.
    intros.
    destruct H as [wS[wbbO[ww[?[??]]]]].
    destruct H0 as [wb[wO2[?[??]]]].
    eapply join_list_group2 with (w1:=wbbO) (wl1:=[wb,wO2]) (wl2:=[wS,wO1]) in H1; eauto.
    eapply join_list_perm with (wl2:=[wO1,wO2]++[wb,wS]) in H1.
    edestruct join_list_app_inv as [wO[wbwS[?[??]]]]; eauto.
    exists wO2 wO; splits; auto.
    apply gjoin_join_list; auto.
    exists wS wb ww; splits; auto.
    eapply join_list_group1 with (w1:=wO) (wl1:=[wO1,wO2]) (wl2:=[wS,wb]); eauto.
    permut4.
    eapply join_list_perm; eauto.
    permut4.
    permut4.
    apply gjoin_join_list; auto.
    permut4.
  Qed.

  Lemma valid_machine_state_par_swap: forall wO s b1 b2,
    valid_machine_state wO s (b1||b2) ->
    valid_machine_state wO s (b2||b1).
  Proof.
    introv [wS[w12[ww[?[??]]]]].
    destruct H0 as [w1[w2[?[??]]]].
    exists wS w12 ww; splits; auto.
    exists w2 w1; splits; auto.
    apply gjoin_com; auto.
  Qed.

  Lemma valid_machine_state_step: forall wO s b s' b',
    step s b s' b' ->
    valid_machine_state wO s b ->
    valid_machine_state wO s' b'.
  Proof.
    intros.
    revert wO H0.
    step_cases (induction H) Case HH; intros; eauto.
    - Case "st_send".
    simpl in *; subst.
    eapply valid_machine_state_s_push_ch with (wb:=w_prod_incr c w_w') (b:=w_prod_incr c w_w':tm_send ec e); eauto.
    apply-> can_send_w_prod_incr; auto.
    reflexivity. reflexivity.
    eapply valid_machine_state_w_prod_incr2; eauto.
    - Case "st_recv".
    simpl in *; subst.
    eapply valid_machine_state_s_pop_ch with (wb:=w_cons_incr c (w_env_set w x v)) (b:=w_cons_incr c (w_env_set w x v):tm_recv x ec) (wb':=w_cons_incr c (w_env_set w_w' x v)); eauto.
    apply-> can_recv_w_cons_incr.
    apply-> can_recv_w_env_set; auto.
    eapply gjoin_w_cons_incr; eauto.
    apply-> can_recv_w_env_set; auto.
    apply gjoin_w_env_set'; eauto.
    reflexivity. reflexivity.
    eapply valid_machine_state_w_cons_incr2.
    apply-> can_recv_w_env_set; auto.
    eapply valid_machine_state_w_env_set2; eauto.
    - Case "st_newch".
    simpl in *; subst.
    remember (fresh_channel s w) as fresh_ch.
    apply valid_machine_state_s_new_ch.
    eapply valid_machine_state_w_env_set2; eauto.
    eapply valid_machine_state_new_channel; eauto.
    reflexivity.
    - Case "st_endch".
    simpl in *; subst.
    eapply valid_machine_state_s_rm_ch; eauto.
    eapply valid_machine_state_join_sub with (b:=w_w':tm_endch ec) (b':=w:tm_skip); eauto.
    reflexivity. reflexivity.
    exists w'; auto.
    - Case "st_assign".
    simpl in *; subst.
    eapply valid_machine_state_w_env_set2; eauto.
    - Case "st_load".
    simpl in *; subst.
    eapply valid_machine_state_w_env_set2; eauto.
    - Case "st_store".
    simpl in *; subst.
    eapply valid_machine_state_w_heap_set2; eauto.
    - Case "st_local".
    simpl in *; subst.
    eapply valid_machine_state_new_local; eauto.
    reflexivity. reflexivity.
    - Case "st_freelocal".
    simpl in *; subst.
    eapply valid_machine_state_join_sub with (b:=w: tm_freelocal x); eauto.
    reflexivity. reflexivity. exists (new_local x v); eauto.
    - Case "st_par".
    eapply valid_machine_state_wterm; eauto.
    reflexivity. exists w1 w2; splits; auto; reflexivity.
    - Case "st_par1".
    edestruct valid_machine_state_par_observer' as [wb2[wOb2[?[??]]]]; eauto.
    eapply valid_machine_state_observer_par; eauto.
    - Case "st_par2".
    apply valid_machine_state_par_swap.
    apply valid_machine_state_par_swap in H0.
    edestruct valid_machine_state_par_observer' as [wb1[wOb1[?[??]]]]; eauto.
    eapply valid_machine_state_observer_par; eauto.
    - Case "st_par_join".
    eapply valid_machine_state_wterm; eauto.
    exists w1 w2; splits; eauto; reflexivity.
    reflexivity.
  Qed.

  Lemma valid_machine_state_stepO: forall wO s b s' bO bO',
    step s bO s' bO' ->
    join_wterms bO wO ->
    valid_machine_state wO s b ->
    exists wO',
      join_wterms bO' wO' /\ valid_machine_state wO' s' b.
  Proof.
    intros.
    apply valid_machine_state_observer_par with (wO1:=gwemp) (wO2:=wO) (bO:=bO) in H1; auto.
    edestruct valid_machine_state_step as [wS'[wbO'[ww'[?[??]]]]].
    apply st_par2; eauto.
    eauto.
    destruct H3 as [wb[wO'[?[??]]]].
    exists wO'; split; auto.
    apply valid_machine_state_par_observer with (wO1:=gwemp) (wO2:=wO') (bO:=bO'); auto.
    apply gjoin_bot2.
    exists wS' wbO' ww'; splits; auto.
    exists wb wO'; splits; auto.
    apply gjoin_bot2.
  Qed.

  Lemma context_world_eq: forall C wO1 wO2,
    context_world C wO1 -> context_world C wO2 ->
    wO1 = wO2.
  Proof.
    setoid_rewrite context_world_join_wterms.
    intros; eapply join_wterms_eq; eauto.
  Qed.

  Definition dummyContext wO := cx_par1 cx_hole (wO:tm_skip).
  Lemma cstep_dummyContext: forall wO s b s' C' b',
    cstep s (dummyContext wO) b s' C' b' ->
    step s b s' b' /\ C'=dummyContext wO.
  Proof.
    intros.
    remember (dummyContext wO) as C.
    revert HeqC.
    induction H; intro; subst; try solve [discriminate].
    inverts HeqC.
    inverts H; auto.
    inverts HeqC.
    false; inverts H.
    split; auto.
  Qed.

  Lemma cstep_star_dummyContext: forall wO s b s' C' b',
    cstep_star s (dummyContext wO) b s' C' b' ->
    step_star s b s' b' /\ C'=dummyContext wO.
  Proof.
    intros.
    remember (dummyContext wO) as C.
    revert HeqC.
    induction H; intro; subst.
    split; auto; apply step_nil.
    edestruct cstep_dummyContext; eauto; subst C'.
    destruct IHcstep_star; auto; subst C''.
    split; auto; eapply step_cons; eauto.
    destruct IHcstep_star; auto; subst C''.
    split; auto; eapply st_iteration; eauto.
  Qed.

  Lemma context_world_dummyContext: forall wO, context_world (dummyContext wO) wO.
  Proof. intros. exists gwemp wO; splits; try reflexivity. apply gjoin_bot2. Qed.

  Lemma valid_machine_state_step_star: forall wO s b s' b',
    step_star s b s' b' ->
    valid_machine_state wO s b ->
    valid_machine_state wO s' b'.
  Proof.
    intros.
    induction H; auto.
    apply IHstep_star; eapply valid_machine_state_step; eauto.
    apply IHstep_star; eapply valid_machine_state_s_iter; eauto.
  Qed.

  Lemma valid_machine_state_step_starO: forall wO s b s' bO bO',
    step_star s bO s' bO' ->
    join_wterms bO wO ->
    valid_machine_state wO s b ->
    exists wO',
      join_wterms bO' wO' /\ valid_machine_state wO' s' b.
  Proof.
    intros.
    revert wO H0 H1.
    induction H; intros; eauto.
    edestruct valid_machine_state_stepO as [wO'[??]]; eauto.
    eapply IHstep_star; eauto; eapply valid_machine_state_s_iter; eauto.
  Qed.

  Lemma valid_machine_state_cstep: forall wO s C b s' C' b',
    cstep s C b s' C' b' ->
    context_world C wO ->
    valid_machine_state wO s b ->
    exists wO',
      context_world C' wO' /\ valid_machine_state wO' s' b'.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst.
    exists wO; split; auto.
    eapply valid_machine_state_step; eauto.
    subst b'. clear H.
    edestruct valid_machine_state_stepO as [wO'[??]]; eauto.
    apply step_simulates_cstep; eauto.
    apply context_world_join_wterms; eauto.
    exists wO'; split; auto.
    apply context_world_join_wterms; eauto.
  Qed.

  Lemma valid_machine_state_cstep_star: forall wO s C b s' C' b',
    cstep_star s C b s' C' b' ->
    context_world C wO ->
    valid_machine_state wO s b ->
    exists wO',
      context_world C' wO' /\ valid_machine_state wO' s' b'.
  Proof.
    intros.
    revert wO H0 H1.
    induction H; intros; eauto.
    edestruct valid_machine_state_cstep as [wO'[??]]; eauto.
    eapply IHcstep_star; eauto; eapply valid_machine_state_s_iter; eauto.
  Qed.

  Lemma valid_machine_state_cstep_star': forall wO wO' s C b s' C' b',
    cstep_star s C b s' C' b' ->
    context_world C wO ->
    context_world C' wO' ->
    valid_machine_state wO s b ->
    valid_machine_state wO' s' b'.
  Proof.
    intros.
    edestruct valid_machine_state_cstep_star as [wO'_[??]]; eauto.
    replace wO'_ with wO' in *; auto.
    eapply context_world_eq; eauto.
  Qed.

  Definition valid_machine_state2 wO sp bp sq bq :=
    valid_machine_state wO sp bp /\ valid_machine_state wO sq bq.

  Lemma valid_machine_state_stepO': forall wO wO' s b s' bO bO',
    step s bO s' bO' ->
    join_wterms bO wO ->
    join_wterms bO' wO' ->
    valid_machine_state wO s b ->
    valid_machine_state wO' s' b.
  Proof.
    intros.
    edestruct valid_machine_state_stepO as [wO'_[??]]; eauto.
    replace wO'_ with wO' in *; auto.
    eapply join_wterms_eq; eauto.
  Qed.

  Lemma valid_machine_state_step_starO': forall wO wO' s b s' bO bO',
    step_star s bO s' bO' ->
    join_wterms bO wO ->
    join_wterms bO' wO' ->
    valid_machine_state wO s b ->
    valid_machine_state wO' s' b.
  Proof.
    intros.
    edestruct valid_machine_state_step_starO as [wO'_[??]]; eauto.
    replace wO'_ with wO' in *; auto.
    eapply join_wterms_eq; eauto.
  Qed.



  Record pbisim_props R (wO:world) (s:state) (bp:wterm) (bq:wterm) : Prop := PBisim_props
  { pbp_halted1: halted bp -> step_star s bq s bp /\ R wO s bp bp;
    pbp_halted2: halted bq -> step_star s bp s bq /\ R wO s bq bq;
    pbp_s_iter: forall s',
      s_iter s s' ->
      R wO s' bp bq;
    pbp_context: forall bO s' wO' bO',
      step s bO s' bO' ->
      join_wterms bO wO ->
      join_wterms bO' wO' ->
      R wO' s' bp bq;
    pbp_step1: forall s' bp',
      step s bp s' bp' ->
      exists bq',
        step_star s bq s' bq' /\
        R wO s' bp' bq';
    pbp_cstep_star2: forall C s' wO' C' bq',
      cstep_star s C bq s' C' bq' ->
      context_world C wO ->
      context_world C' wO' ->
      (* either matches the behavior *)
      (exists bp',
        cstep_star s C bp s' C' bp' /\
        R wO' s' bp' bq')
      \/ (* or both will converge *)
      (exists s'' b'',
        cstep_star s C bp s'' C' b'' /\
        step_star_sans_s_iter s' bq' s'' b'' /\
        R wO' s'' b'' b'')
(*
    pbp_step_star: forall s' bq',
      step_star s bq s' bq' ->
      exists s'' bp'' bq'',
        step_star s bp s'' bp'' /\
        step_star s' bq' s'' bq'' /\
        R s'' bp'' bq'';
    pbp_cstep_star: forall C s' C' bq',
      valid_observerC C s bp ->
      valid_observerC C s bq ->
      cstep_star s C bq s' C' bq' ->
      exists s'' bp'' bq'',
        cstep_star s C bp s'' C' bp'' /\
        step_star s' bq' s'' bq'' /\
        R s'' bp'' bq''
*)
  }.

  Definition pbisimulation (R: world->state->wterm->wterm->Prop) : Prop :=
    forall wO s bp bq,
    R wO s bp bq ->
      pbisim_props R wO s bp bq.

  Definition pbisimilar wO s b1 b2 :=
    exists R, pbisimulation R /\ R wO s b1 b2.


(*
  Record dpbisim_props R (wO:world) (s:state) (bp:wterm) (bq:wterm) : Prop := PBisim_props
  { dpbp_halted1: halted bp -> step_star_sans_s_iter s bq s bp /\ R wO s bp bp;
    dpbp_halted2: halted bq -> step_star_sans_s_iter s bp s bq /\ R wO s bq bq;
    dpbp_s_iter: forall s',
      s_iter s s' ->
      R wO s' bp bq;
    dpbp_context: forall bO s' wO' bO',
      step s bO s' bO' ->
      join_wterms bO wO ->
      join_wterms bO' wO' ->
      R wO' s' bp bq;
    
    dpbp_step1: forall s' bp',
      step s bp s' bp' ->
      exists bq',
        step_star_sans_s_iter s bq s' bq' /\
        R wO s' bp' bq';
    dpbp_step2: forall s' bq',
      step s bq s' bq' ->
      (* either matches the behavior *)
      (exists bp',
        step_star_sans_s_iter s bp s' bp' /\
        R wO s' bp' bq')
      \/ (* or both will converge *)
      (exists s'' b'',
        step_star_sans_s_iter s bp s'' b'' /\
        step_star_sans_s_iter s' bq' s'' b'' /\
        R wO )
  }.

  Definition dpbisimulation (R: world->state->wterm->wterm->Prop) : Prop :=
    forall wO s bp bq,
    R wO s bp bq ->
      dpbisim_props R wO s bp bq.

  Definition dpbisimilar wO s b1 b2 :=
    exists R, dpbisimulation R /\ R wO s b1 b2.
*)
(*
  Inductive observer_lbl: world->wterm->world->wterm->Type :=
  | obs_lbl: forall wO bO wO' bO',
      join_wterms bO wO -> join_wterms bO' wO' -> observer_lbl wO bO wO' bO'.

  Inductive observer : world->Type :=
  | obs_nil: forall wO,
      observer wO
  | obs_cons: forall {wO bO wO' bO'} (l:observer_lbl wO bO wO' bO'),
      observer wO'->observer wO.

  Inductive ostep_star wO: observer wO->state->wterm->state->wterm->Prop :=
  | ostep_nil: forall s b,
      ostep_star wO (obs_nil wO) s b s b
  | ostep_label: forall bO bO' wO' (l:observer_lbl wO bO wO' bO') O s b s' b',
      step s bO s' bO' ->
      ostep_star wO' O s b s' b' ->
      ostep_star wO (obs_cons l O) s b s' b'
  | ostep_tao: forall O s b s' b' s'' b'',
      step s b s' b' ->
      ostep_star wO O s' b' s'' b'' ->
      ostep_star wO O s b s'' b''
  | ostep_iter: forall O s b s' s'' b'',
      s_iter s s' ->
      ostep_star wO O s' b s'' b'' ->
      ostep_star wO O s b s'' b''.



  Definition decoupled_bisimulation (R: forall wO, observer wO->state->wterm->state->wterm->Prop) :=
    forall wO (O: observer wO) sp bp sq bq,
    R wO O sp bp sq bq ->
    (halted bp -> ostep_star O sq bq sp bp /\ R wO (obs_nil wO) sp bp sp bp)
    /\
    s_i sp = s_i sq (* iteration states always match *)
    /\
    (forall sp',
      s_iter sp sp' ->
      exists sq',
        s_iter sq sq' /\ R wO O sp' bp sq' bq)
    /\
    (forall bO sp' wO' bO' (l:observer_lbl wO bO wO' bO'),
      step sp bO sp' bO' ->
      R wO (l::O) sp bp sq bq
      exists sq' bq' sq'',
        step_star_sans_s_iter sq bq sq' bq' /\
        step sq' bO sq'' bO' /\ R wO' sp' bp sq'' bq')
    /\
    (exists sq' bq',
      step_star_sans_s_iter sq bq sq' bq' /\ R wO sq' bq' sp bp)
    /\
    (forall sp' bp',
      step sp bp sp' bp' ->
      R wO sp' bp' sq bq)
  .
*)


  Definition decoupled_bisimulation (R: world->state->wterm->state->wterm->Prop) :=
    forall wO sp bp sq bq,
    R wO sp bp sq bq ->
    (halted bp -> step_star sq bq sp bp /\ R wO sp bp sp bp)
    /\
(*    (halted bq -> step_star sp bp sq bq /\ R sq bq sq bq)
    /\ *)
    s_i sp = s_i sq (* iteration state always matches *)
    /\
    (forall sp',
      s_iter sp sp' ->
      exists sq',
        s_iter sq sq' /\ R wO sp' bp sq' bq)
(*    /\
    (forall sq', s_iter sq sq' -> exists sp', s_iter sp sp' /\ R sp' bp sq' bq) *)
    /\
    (forall bO sp' wO' bO',
      step sp bO sp' bO' ->
      join_wterms bO wO ->
      join_wterms bO' wO' ->
      exists sq' bq' sq'',
        step_star_sans_s_iter sq bq sq' bq' /\
        step sq' bO sq'' bO' /\ R wO' sp' bp sq'' bq')
    /\
    (exists sq' bq',
      step_star_sans_s_iter sq bq sq' bq' /\ R wO sq' bq' sp bp)
    /\
    (forall sp' bp',
      step sp bp sp' bp' ->
      R wO sp' bp' sq bq)
  .


  Definition dbisimilar wO sp bp sq bq :=
    exists R, decoupled_bisimulation R /\ R wO sp bp sq bq /\ R wO sq bq sp bp.



  Definition dbisimilar1 wO sp bp sq bq :=
    exists R, decoupled_bisimulation R /\ R wO sp bp sq bq.


  Definition cbisimulation (R: world->state->wterm->state->wterm->Prop) :=
    forall wO sp bp sq bq,
    R wO sp bp sq bq ->
    (halted bp -> step_star sq bq sp bp /\ R wO sp bp sp bp)
    /\
(*    (halted bq -> step_star sp bp sq bq /\ R sq bq sq bq)
    /\ *)
    (forall sp', s_iter sp sp' -> exists sq', s_iter sq sq' /\ R wO sp' bp sq' bq)
(*    /\
    (forall sq', s_iter sq sq' -> exists sp', s_iter sp sp' /\ R sp' bp sq' bq) *)
    /\
    s_i sp = s_i sq (* iteration state always matches *)
    /\
    (forall C sp' wO' C' bp',
      cstep_star sp C bp sp' C' bp' ->
      context_world C wO ->
      context_world C' wO' ->
      exists sq' bq',
        cstep_star sq C bq sq' C' bq' /\
        R wO' sq' bq' sp' bp')
  .




(*  Definition cbisimilar sp bp sq bq :=
    exists R, cbisimulation R /\ R sp bp sq bq /\ R sq bq sp bp.
*)
  Definition cbisimilar wO sp bp sq bq :=
    exists R, cbisimulation R /\ R wO sp bp sq bq /\ R wO sq bq sp bp.


  Definition abisimilar wO sp bp sq bq :=
    exists R, cbisimulation R /\ R wO sp bp sq bq.

End Definitions2.









(*
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
*)

Section BisimulationRelations.
  Import Machine.Notation.
  Open Scope wterm_scope.

  Lemma pbisim_refl: forall wO s b,
    pbisimilar wO s b b.
  Proof. 
    intros.
    exists (fun (wO:world) (s:state) (bp:wterm) bq => bp=bq).
    split; auto.
    clear wO s b.
    intros wO s b_ b ?; subst b_.
    constructor; auto; try solve [split; auto; constructor].
    - intros; eexists; split; eauto. apply step_single; auto.
    - intros; eauto.
  Qed.

  Lemma step_pbisim1: forall wO s ba bb s' ba',
    step s ba s' ba' ->
    pbisimilar wO s ba bb ->
    exists bb',
      step_star s bb s' bb' /\ pbisimilar wO s' ba' bb'.
  Proof.
    introv H0 H; destruct H as [R[??]].
    edestruct pbp_step1 as [bb'[??]]; eauto.
    exists bb'; split; auto.
    exists R; split; auto.
  Qed.

  Lemma step_pbisim1': forall wO s ba bb s' ba',
    step s ba s' ba' ->
    pbisimilar wO s ba bb ->
    exists bb',
      step_star_sans_s_iter s bb s' bb' /\ pbisimilar wO s' ba' bb'.
  Proof.
    intros.
    edestruct step_pbisim1 as [bb'[??]]; eauto.
    exists bb'; split; auto.
    apply step_star_s_iter_step_star; auto.
    apply step_respecting_s_i in H; eauto.
  Qed.



  Lemma step_star_pbisim2: forall wO s ba bb s' bb',
    step_star s bb s' bb' ->
    pbisimilar wO s ba bb ->
    (exists ba', step_star s ba s' ba' /\ pbisimilar wO s' ba' bb') \/
    (exists s'' b'',
      step_star s ba s'' b'' /\ step_star_sans_s_iter s' bb' s'' b'').
  Proof.
    introv ? [R[??]].
    edestruct pbp_cstep_star2 with (C:=dummyContext wO) as [ [ba'[??]] | [s'''[b'''[?[??]]]] ]; eauto.
    apply step_star_add_context_cstep_star; eauto.
    apply context_world_dummyContext.
    apply context_world_dummyContext.
    edestruct cstep_star_dummyContext; eauto.
    left; exists ba'; splits; auto; exists R; auto.
    edestruct cstep_star_dummyContext; eauto.
  Qed.

  Lemma cstep_star_pbisim2: forall wO wO' s C ba bb s' C' bb',
    cstep_star s C bb s' C' bb' ->
    context_world C wO ->
    context_world C' wO' ->
    pbisimilar wO s ba bb ->
    (exists ba', cstep_star s C ba s' C' ba' /\ pbisimilar wO' s' ba' bb') \/
    (exists s'' b'',
      cstep_star s C ba s'' C' b'' /\ step_star_sans_s_iter s' bb' s'' b'').
  Proof.
    introv ? ? [R[??]].
    edestruct pbp_cstep_star2 as [ [ba'[??]] | [s'''[b'''[?[??]]]] ]; eauto.
    left; exists ba'; splits; auto; exists R; auto.
  Qed.

  Lemma step_pbisim2: forall wO s ba bb s' bb',
    step s bb s' bb' ->
    pbisimilar wO s ba bb ->
    (exists ba', step_star s ba s' ba' /\ pbisimilar wO s' ba' bb') \/
    (exists s'' b'',
      step_star s ba s'' b'' /\ step_star_sans_s_iter s' bb' s'' b'').
  Proof.
    intros.
    edestruct step_star_pbisim2 as [ [ba'[??]] | [s''[b''[??]]] ]; eauto.
    apply step_single; auto.
  Qed.

  Lemma step_pbisim2': forall wO s ba bb s' bb',
    step s bb s' bb' ->
    pbisimilar wO s ba bb ->
    (exists ba', step_star_sans_s_iter s ba s' ba' /\ pbisimilar wO s' ba' bb') \/
    (exists s'' b'',
      step_star_sans_s_iter s' bb' s'' b'' /\ step_star_sans_s_iter s ba s'' b'').
  Proof.
    intros.
    edestruct step_pbisim2 as [ [ba'[??]] | [s''[b''[??]]] ]; eauto.
    left.
    exists ba'; split; auto.
    apply step_star_s_iter_step_star; auto.
    eapply step_respecting_s_i; eauto.
    right.
    exists s'' b''; splits; auto.
    apply step_star_s_iter_step_star; auto.
    apply step_respecting_s_i in H; eauto.
    apply step_star_sans_s_iter_respects_s_i in H2.
    congruence.
  Qed.

  Lemma step_star_pbisim2'': forall wO s ba bb s' bb',
    step_star_sans_s_iter s bb s' bb' ->
    pbisimilar wO s ba bb ->
    (exists ba', step_star_sans_s_iter s ba s' ba' /\ pbisimilar wO s' ba' bb') \/
    (exists s'' b'',
      step_star_sans_s_iter s' bb' s'' b'' /\ step_star_sans_s_iter s ba s'' b'').
  Proof.
    intros.
    edestruct step_star_pbisim2 as [ [ba'[??]] | [s''[b''[??]]] ]; eauto.
    eapply step_star_step_star_sans_s_iter; eauto.
    left; exists ba'; splits; auto.
    apply step_star_s_iter_step_star; auto.
    eapply step_star_sans_s_iter_respects_s_i; eauto.
    right; exists s'' b''; splits; auto.
    apply step_star_s_iter_step_star; auto.
    eapply step_star_sans_s_iter_respects_s_i in H.
    eapply step_star_sans_s_iter_respects_s_i in H2.
    congruence.
  Qed.

  Lemma step_star_pbisim2': forall wO s ba bb s' bb',
    step_star s bb s' bb' ->
    pbisimilar wO s ba bb ->
    exists s'' ba'' bb'',
      step_star_sans_s_iter s' bb' s'' bb'' /\ step_star s ba s'' ba'' /\ pbisimilar wO s'' ba'' bb''.
  Proof.
    intros.
    edestruct step_star_pbisim2 as [ [ba'[??]] | [s''[b''[??]]] ]; eauto.
    exists s' ba' bb'; splits; auto. apply step_nil_.
    exists s'' b'' b''; splits; auto. apply pbisim_refl.
  Qed.


  Lemma cstep_pbisim2: forall wO wO' s C ba bb s' C' bb',
    cstep s C bb s' C' bb' ->
    context_world C wO ->
    context_world C' wO' ->
    pbisimilar wO s ba bb ->
    (exists ba', cstep_star s C ba s' C' ba' /\ pbisimilar wO' s' ba' bb') \/
    (exists s'' b'',
      cstep_star s C ba s'' C' b'' /\ step_star_sans_s_iter s' bb' s'' b'').
  Proof.
    intros.
    edestruct cstep_star_pbisim2
      with (ba:=ba) (bb:=bb) (C:=C) (C':=C')
      as [ [ba'[??]] | [s''[b''[??]]] ]; eauto.
    apply cstep_single_cstep_star; eauto.
  Qed.

  Lemma pbisim_s_iter: forall wO s s' ba bb,
    s_iter s s' ->
    pbisimilar wO s ba bb ->
    pbisimilar wO s' ba bb.
  Proof.
    introv ? [R[??]].
    exists R; split; auto.
    eapply pbp_s_iter; eauto.
  Qed.

  Lemma step_star_sans_s_iter_pbisim1: forall wO s ba bb s' ba',
    step_star_sans_s_iter s ba s' ba' ->
    pbisimilar wO s ba bb ->
    exists bb',
      step_star_sans_s_iter s bb s' bb' /\ pbisimilar wO s' ba' bb'.
  Proof.
    intros.
    revert bb H0.
    induction H; intros.
    exists bb; split; auto; apply step_nil_.
    rename b into ba. rename b' into ba'. rename b'' into ba''.
    edestruct step_pbisim1' as [bb'[??]]; eauto.
    edestruct IHstep_star_sans_s_iter as [bb''[??]]; clear IHstep_star_sans_s_iter; eauto.
    exists bb''; split; auto.
      eapply step_star_sans_s_iter_app; eauto.
  Qed.


  Lemma step_star_pbisim1: forall wO s ba bb s' ba',
    step_star s ba s' ba' ->
    pbisimilar wO s ba bb ->
    exists bb',
      step_star s bb s' bb' /\ pbisimilar wO s' ba' bb'.
  Proof.
    intros.
    revert bb H0.
    induction H; intros.
    exists bb; split; auto; apply step_nil.
    rename b into ba. rename b' into ba'. rename b'' into ba''.
    edestruct step_pbisim1 as [bb'[??]]; eauto.
    edestruct IHstep_star as [bb''[??]]; clear IHstep_star; eauto.
    exists bb''; split; auto.
      eapply step_star_app; eauto.
    rename b into ba. rename b'' into ba''.
    edestruct IHstep_star as [bb''[??]]; clear IHstep_star; eauto.
    eapply pbisim_s_iter; eauto.
    exists bb''; split; auto.
      eapply st_iteration; eauto.
  Qed.

  Lemma pbisim_context: forall wO wO' s bp bq bO s' bO',
    step s bO s' bO' ->
    join_wterms bO wO ->
    join_wterms bO' wO' ->
    pbisimilar wO s bp bq ->
    pbisimilar wO' s' bp bq.
  Proof.
    introv H0 Hvo [R[??]].
    exists R; split; auto.
    eapply pbp_context; eauto.
  Qed.

  Lemma cstep_pbisim1: forall wO wO' s C ba bb s' C' ba',
    cstep s C ba s' C' ba' ->
    context_world C wO ->
    context_world C' wO' ->
    pbisimilar wO s ba bb ->
    exists bb',
      cstep_star s C bb s' C' bb' /\ pbisimilar wO' s' ba' bb'.
  Proof.
    introv H0 HwO HwO' H.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    replace wO' with wO in *.
    edestruct step_pbisim1 as [bb'[??]]; eauto.
    exists bb'; splits; auto.
    apply step_star_add_context_cstep_star; auto.
    eapply context_world_eq; eauto.
    subst ba'. clear H0.
    exists bb; splits; auto.
    eapply cstep_context_cstep_star; eauto. apply cstep_nil.
    eapply pbisim_context; eauto.
    eapply step_simulates_cstep; eauto.
    apply context_world_join_wterms; auto.
    apply context_world_join_wterms; auto.
  Qed.

  Lemma cstep_star_pbisim1: forall wO s C ba bb s' C' ba',
    cstep_star s C ba s' C' ba' ->
    context_world C wO ->
    pbisimilar wO s ba bb ->
    valid_machine_state2 wO s ba s bb ->
    exists wO' bb',
      context_world C' wO' /\ cstep_star s C bb s' C' bb' /\ pbisimilar wO' s' ba' bb'.
  Proof.
    introv H HwO H0 Hvms.
    revert wO bb HwO H0 Hvms.
    induction H; intros.
    exists wO bb; splits; auto; apply cstep_nil.
    renames b to ba, b' to ba', b'' to ba''.
    edestruct valid_machine_state_cstep as [wO'[??]]; eauto.
    apply Hvms.
    edestruct cstep_pbisim1 as [bb'[??]]; eauto.
    edestruct IHcstep_star as [wO''[bb''[?[??]]]]; clear IHcstep_star; eauto.
    split; auto.
    eapply valid_machine_state_cstep_star'; eauto.
    apply Hvms.
    exists wO'' bb''; splits; auto.
      eapply cstep_star_app; eauto.
    renames b to ba, b'' to ba''.
    edestruct IHcstep_star as [wO''[bb''[?[??]]]]; clear IHcstep_star; eauto.
    eapply pbisim_s_iter; eauto.
    split; eapply valid_machine_state_s_iter; eauto; apply Hvms.
    exists wO'' bb''; splits; auto.
      eapply cstep_iteration; eauto.
  Qed.

  Lemma step_star_observer_context': forall s b bo s' b' bo',
    step_star s (b || bo) s' (b' || bo') <->
    cstep_star s (cx_par1 cx_hole bo) b s' (cx_par1 cx_hole bo') b'.
  Proof.
    split; intros.
    apply cstep_star_par_cx_par1.
    apply step_star_simulates_cstep_star_empty_context; auto.
    apply cstep_star_cx_par1_par in H.
    apply step_star_simulates_cstep_star in H; auto.
  Qed.

  Lemma valid_machine_state2_stepO: forall wO s bp bq s' bO bO',
    step s bO s' bO' ->
    join_wterms bO wO ->
    valid_machine_state2 wO s bp s bq ->
    exists wO', join_wterms bO' wO' /\ valid_machine_state2 wO' s' bp s' bq.
  Proof.
    intros.
    edestruct valid_machine_state_stepO with (b:=bp) as [wO'[??]]; eauto.
    apply H1.
    edestruct valid_machine_state_stepO with (b:=bq) as [wO'_[??]]; eauto.
    apply H1.
    replace wO'_ with wO' in *.
    exists wO'; split; auto.
    split; auto.
    eapply join_wterms_eq; eauto.
  Qed.

  Lemma bisim_pbisim: forall wO s bp bq,
    bisimilar s bp bq ->
    pbisimilar wO s bp bq.
  Proof.
    intros.
    exists (fun (wO:world) s bp bq => bisimilar s bp bq).
    split; auto.
    clear wO s bp bq H.
    intros wO s bp bq ?.
    constructor; intros.
    - split. eapply bisim_halted1; eauto. apply bisim_refl.
    - split. eapply bisim_halted2; eauto. apply bisim_refl.
    - eapply bisim_s_iter; eauto.
    - eapply bisim_context; eauto.
    - edestruct step_bisim1 as [bq'[??]]; eauto.
    - left.
      edestruct cstep_star_bisim2 as [bp'[??]]; eauto.
  Qed.

  Lemma pbisim_halted1: forall wO s ba bb,
    halted ba ->
    pbisimilar wO s ba bb ->
    step_star s bb s ba.
  Proof.
    introv ? [R[??]].
    edestruct pbp_halted1; eauto.
  Qed.

  Lemma pbisim_halted2: forall wO s ba bb,
    halted bb ->
    pbisimilar wO s ba bb ->
    step_star s ba s bb.
  Proof.
    introv ? [R[??]].
    edestruct pbp_halted2; eauto.
  Qed.

  Lemma valid_machine_state_step_star_sans_s_iter: forall wO s b s' b',
    valid_machine_state wO s b ->
    step_star_sans_s_iter s b s' b' ->
    valid_machine_state wO s' b'.
  Proof.
    intros.
    eapply valid_machine_state_step_star; eauto.
    apply step_star_step_star_sans_s_iter; auto.
  Qed.

  Lemma bisim_dbisim: forall wO s bp bq,
    bisimilar s bp bq ->
    dbisimilar wO s bp s bq.
  Proof.
    intros.
    exists (fun (wO:world) s bp s0 bq0 => exists bq, step_star_sans_s_iter s0 bq0 s bq /\ bisimilar s bp bq).
    split; eauto.
    * clear -wO; clear wO.
    intros wO s bp s0 bq0 [bq[??]].
    splits; intros.
    - split.
    eapply step_star_app.
    apply step_star_step_star_sans_s_iter; eauto.
    eapply bisim_halted1; eauto.
    exists bp; split. apply step_nil_. apply bisim_refl.
    - apply step_star_sans_s_iter_respects_s_i in H; congruence.
    - edestruct step_star_sans_s_iter_commute_s_iter1 as [s'[??]]; eauto.
    exists s'; split; auto.
    exists bq; split; auto.
    eapply bisim_s_iter; eauto.
    - exists s bq sp'; splits; auto.
    exists bq; split.
    apply step_nil_.
    eapply bisim_context; eauto.
    - exists s bq; split; auto.
    exists bp; split.
    apply step_nil_.
    apply bisim_symm; auto.
    - edestruct step_bisim1 as [bq'[??]]; eauto.
    exists  bq'; split; auto.
    eapply step_star_sans_s_iter_app; eauto.
    apply step_star_s_iter_step_star; eauto.
    eapply step_respecting_s_i; eauto.
    * split.
    exists bq; split; auto. apply step_nil_.
    exists bp; split; auto. apply step_nil_. apply bisim_symm; auto.
  Qed.

  Lemma step_star_sans_s_iter_skip_resolve: forall s w s' b',
    step_star_sans_s_iter s (w: tm_skip) s' b' -> s'=s /\ b' = w: tm_skip.
  Proof.
    intros.
    remember (w:tm_skip) as b.
    revert Heqb.
    induction H; intros; subst; auto.
    inverts H.
  Qed.


(*
  Lemma bisim_assign_send_recv: forall s w x e ec c R wO ww,
    eval ec (cname_value c) w ->
    total_send_recv c w ->
    s_get_ch s c = Some(R,nil,nil) ->
    gjoin wO w ww ->
    cbisimilar wO s (w: tm_assign x e) s (w: tm_send ec e;; tm_recv x ec).
  Proof.
Admitted.
*)

(*
  Lemma pbisim_dbisim: forall wO s bp bq,
    pbisimilar wO s bp bq ->
    dbisimilar wO s bp s bq.
  Proof.
    introv H.
    exists (fun wO s bp sq bq =>
      (exists bq', step_star_sans_s_iter sq bq s bq' /\ pbisimilar wO s bp bq') \/
(*       (exists bq', step_star_sans_s_iter sq bq s bq' /\ pbisimilar wO s bq' bp) \/
      (exists s'' b'', step_star_sans_s_iter sq bq s'' b'' /\ step_star_sans_s_iter s bp s'' b'') \/ *)
      (sq = s /\ pbisimilar wO sq bq bp) \/
      (exists s0 bp0, sq=s0 /\ step_star_sans_s_iter s0 bp0 s bp /\ pbisimilar wO s0 bq bp0)).
    split.
    clear wO s bp bq H.
    intros wO s bp sq bq [ [bq'[?[R[??]]]] | [ [??] | [s0[bp0[?[??]]]] ]].
(*    intros wO s bp sq bq [ [bq'[?[R[??]]]] | [ [??] | [??] ] ].*)


    - splits; intros.
    + edestruct pbp_halted1; eauto.
    split.
    eapply step_star_app; eauto.
    eapply step_star_step_star_sans_s_iter; eauto.
    left; exists bp; splits; auto.
    apply step_nil_.
    exists R; auto.
    + apply step_star_sans_s_iter_respects_s_i in H; congruence.
    + edestruct step_star_sans_s_iter_commute_s_iter1 as [sq'0[??]]; eauto.
    exists sq'0; splits; auto.
    left; exists bq'; split; auto.
    eapply pbisim_s_iter; eauto.
    exists R; auto.
    + exists s bq' sp'; splits; auto.
    left; exists bq'; split; auto.
    apply step_nil_.
    eapply pbisim_context; eauto.
    exists R; auto.
    + exists s bq'; split; auto.
    right; left; split; auto.
    exists R; auto.
    + edestruct step_pbisim1' as [bq''[??]]; eauto.
    exists R; eauto.
    left; exists bq''; split; auto.
    eapply step_star_sans_s_iter_app; eauto.

    - splits; intros; subst sq.
    + split.
    eapply pbisim_halted2; eauto.
    right; left; split; auto. apply pbisim_refl.
    + reflexivity.
    + exists sp'; split; auto.
    right; left; split; auto.
    eapply pbisim_s_iter; eauto.
    + exists s bq sp'; splits; auto.
    apply step_nil_.
    right; left; split; auto; eapply pbisim_context; eauto.
    + exists s bq; split; auto.
    apply step_nil_.
    left; exists bp; split; auto.
    apply step_nil_.
    + edestruct step_pbisim2' as [ [bq'[??]] | [s''[b''[??]]] ]; eauto.
    right; right.
    exists s bp; splits; auto.
    eapply step_cons_; eauto. apply step_nil_.
    right; right.
    exists s bp; splits; auto.
    eapply step_cons_; eauto. apply step_nil_.

    - splits; intros; subst sq.
    + admit.
    + admit.
    + admit.
    + Check step_star_pbisim2.
edestruct step_star_pbisim2'' as [ [bq'[??]] | [s''[b''[??]]] ]; eauto.
exists s bq' sp'; splits; auto.
right; left; split; auto.
eapply pbisim_context; eauto.




exists s b'' sp'; splits; auto.
exists bq'; split.



edestruct cstep_star_pbisim2 as [ [bq'[??]] | [s''[b''[??]]] ]; eauto.
eapply cstep_star_snoc.
apply step_star_add_context_cstep_star.
apply step_star_step_star_sans_s_iter; eauto.
apply cstep_skip_subst with (C:=cx_par1 cx_hole bO) (w:=gwemp).
apply cst_par1_step2; eauto.
exists gwemp wO; splits; auto. reflexivity. apply gjoin_bot2.
exists gwemp wO'; splits; auto. reflexivity. apply gjoin_bot2.

apply context_world_join_wterms; simpl; auto.
exists 


exists s sp'

    left; exists bq'; split; auto.

    - splits; intros.
    + split.
    eapply step_star_app.
    apply step_star_step_star_sans_s_iter; eauto.
    eapply pbisim_halted2; eauto.
    exists R; eauto.
    right; right; right; split; auto. apply pbisim_refl.
    + apply step_star_sans_s_iter_respects_s_i in H; congruence.
    + edestruct step_star_sans_s_iter_commute_s_iter1 as [sq'0[??]]; eauto.
    exists sq'0; splits; auto.
    right; left; exists bp'; split; auto.
    eapply pbisim_s_iter; eauto.
    exists R; auto.
    + exists s bp' sp'; splits; auto.
    right; right; right; split; auto.
    eapply pbisim_context; eauto.
    exists R; auto.
    + exists s bp'; split; auto.
    left; exists bp; split; auto.
    apply step_nil_.
    exists R; auto.
    + edestruct step_pbisim2' as [ [bq'[??]] | [s''[b''[??]]] ]; eauto.
    exists R; eauto.
    right; left; exists bq'; split; auto.
    eapply step_star_sans_s_iter_app; eauto.
    do 2 right; left. exists s'' b''; split; auto.
    eapply step_star_sans_s_iter_app; eauto.

    


    left; exists b


    - splits; intros.
    + split.
    eapply pbisim_halted2; eauto.
    right; left; split; auto. apply pbisim_refl.
    + reflexivity.
    + exists sp'; split; auto.
    right; left; split; auto.
    eapply pbisim_s_iter; eauto.
    + exists s bq sp'; splits; auto.
    apply step_nil_.
    right; left; split; auto; eapply pbisim_context; eauto.
    + exists s bq; split; auto.
    apply step_nil_.
    + edestruct step_pbisim2' as [ [bq'[??]] | [s''[b''[??]]] ]; eauto.
    left; exists bq'; split; auto.
    

    right; exists bq''; split; auto.
    eapply step_star_sans_s_iter_app; eauto.
    left; exists 
    exists R; auto.

edestruct step_pbisim2' as [bq'[??]]; eauto.
    

edestruct pbisim_halted2; eauto. pbp_halted2; eauto.
    split.
    eapply step_star_app; eauto.
    eapply step_star_step_star_sans_s_iter; eauto.
    left; exists bp; splits; auto.
    apply step_nil_.
    exists R; auto.
    + apply step_star_sans_s_iter_respects_s_i in H; congruence.
    + edestruct step_star_sans_s_iter_commute_s_iter1 as [sq'0[??]]; eauto.
    exists sq'0; splits; auto.
    right; exists bq'; split; auto.
    eapply pbisim_s_iter; eauto.
    exists R; auto.
    + exists s bq' sp'; splits; auto.
    right; exists bq'; split; auto.
    apply step_nil_.
    eapply pbisim_context; eauto.
    exists R; auto.
    + exists s bq'; split; auto.
    left; exists bp; split.
    apply step_nil_.
    exists R; auto.
    + edestruct step_pbisim2' as [ [bq''[??]] | [s''[b''[??]]] ]; eauto.
    exists R; eauto.
    right; exists bq''; split; auto.
    eapply step_star_sans_s_iter_app; eauto.
    left; exists 
    exists R; auto.

\- 
    - subst sq.
    splits; intros; try rename sp' into s'.
    + edestruct pbp_halted2; eauto. split; auto; left; splits; auto. exists R; auto.
    + reflexivity.
    + exists s'; split; auto; right; left; splits; auto.
    eapply pbisim_s_iter; eauto; exists R; auto.
    + exists s'; splits; auto.
    right; left; split; auto.
    eapply pbisim_context; eauto.
    exists R; eauto.
    + exists s bq; split.
    apply step_nil_.
    left; splits; auto.
    exists R; auto.
    + edestruct step_pbisim2' as [ [bq'[??]] | [s''[b''[??]]] ]; eauto.
    exists R; eauto.
    right; right.
    exists bq'; splits; auto.
    split.
    eapply valid_machine_state_step; eauto; apply Hvms.
    eapply valid_machine_state_step_star_sans_s_iter; eauto; apply Hvms.


    + edestruct cstep_star_pbisim1 with (bb:=bq) as [wO'_[bq'[?[??]]]]; eauto.
    exists R; auto.
    replace wO'_ with wO' in *.
    exists sp' bq'; split; auto.
    right; left; splits; auto.
    split; eapply valid_machine_state_cstep_star'; eauto; apply Hvms.
    eapply context_world_eq; eauto.

    - subst sq.
    splits; intros.
    + edestruct pbp_halted2; eauto. split; auto; left; splits; auto. exists R; auto.
    split; apply Hvms.
    + exists sp'; split; auto; right; left; splits; auto.
    eapply pbisim_s_iter; eauto; exists R; auto.
    split; eapply valid_machine_state_s_iter; eauto; apply Hvms.
    + reflexivity.
    + edestruct cstep_star_pbisim2 as [ [bq'[??]] | [s''[b''[??]]] ]; eauto.
    * exists R; eauto.
    * exists sp' bq'; split; eauto.
    left; splits; auto.
    split; eapply valid_machine_state_cstep_star'; eauto; apply Hvms.
    * exists s'' b''; splits; auto.
    do 2 right; exists b''; splits; auto.
    apply pbisim_refl.
    split; eapply valid_machine_state_cstep_star'; eauto; apply Hvms.

    - splits; intros.
    + edestruct pbp_halted1; eauto.
    split. eapply step_star_app; eauto.
    apply step_star_step_star_sans_s_iter; auto.
    left; splits; auto.
    apply pbisim_refl.
    split.
    apply Hvms.
    eapply valid_machine_state_step_star; eauto; apply Hvms.
    + edestruct step_star_sans_s_iter_commute_s_iter1 as [s0[??]]; eauto.
    exists s0; split; auto.
    do 2 right.
    exists bq'; splits; auto.
    eapply pbisim_s_iter; eauto.
    exists R; split; auto.
    split; eapply valid_machine_state_s_iter; eauto; apply Hvms.
    + symmetry; eapply step_star_sans_s_iter_respects_s_i; eauto.
    + edestruct cstep_star_pbisim1 with (ba:=bp) (bb:=bq') as [wO'_[bq''[?[??]]]]; eauto.
    exists R; auto.
    replace wO'_ with wO' in *.
    exists sp' bq''; splits; auto.
    eapply cstep_star_app; eauto.
    apply step_star_add_context_cstep_star; auto.
    apply step_star_step_star_sans_s_iter; auto.
    right; left; splits; auto.
    split; eapply valid_machine_state_cstep_star'; eauto; apply Hvms.
    eapply context_world_eq; eauto.

    - split; jauto. right; left; splits; auto.
    split; apply Hvms.
  Qed.

*)

  Lemma pbisim_cbisim: forall wO s bp bq,
    pbisimilar wO s bp bq ->
    valid_machine_state2 wO s bp s bq ->
    cbisimilar wO s bp s bq.
  Proof.
    introv H Hvms.
    exists (fun wO s bp sq bq =>
      (sq=s /\ pbisimilar wO s bp bq /\ valid_machine_state2 wO s bp s bq) \/
      (sq=s /\ pbisimilar wO s bq bp /\ valid_machine_state2 wO s bp s bq) \/
      (exists bq', step_star_sans_s_iter sq bq s bq' /\ pbisimilar wO s bp bq' /\ valid_machine_state2 wO s bp s bq')).
    split.
    clear wO s bp bq H Hvms.
    intros wO s bp sq bq [ [?[[R[??]]Hvms]] |[ [?[[R[??]]Hvms]] | [bq'[?[[R[??]]Hvms]]] ]].

    - subst sq.
    splits; intros.
    + edestruct pbp_halted1; eauto. split; auto; left; splits; auto. exists R; auto.
    split; apply Hvms.
    + exists sp'; split; auto; left; splits; auto.
    eapply pbisim_s_iter; eauto; exists R; auto.
    split; eapply valid_machine_state_s_iter; eauto; apply Hvms.
    + reflexivity.
    + edestruct cstep_star_pbisim1 with (bb:=bq) as [wO'_[bq'[?[??]]]]; eauto.
    exists R; auto.
    replace wO'_ with wO' in *.
    exists sp' bq'; split; auto.
    right; left; splits; auto.
    split; eapply valid_machine_state_cstep_star'; eauto; apply Hvms.
    eapply context_world_eq; eauto.

    - subst sq.
    splits; intros.
    + edestruct pbp_halted2; eauto. split; auto; left; splits; auto. exists R; auto.
    split; apply Hvms.
    + exists sp'; split; auto; right; left; splits; auto.
    eapply pbisim_s_iter; eauto; exists R; auto.
    split; eapply valid_machine_state_s_iter; eauto; apply Hvms.
    + reflexivity.
    + edestruct cstep_star_pbisim2 as [ [bq'[??]] | [s''[b''[??]]] ]; eauto.
    * exists R; eauto.
    * exists sp' bq'; split; eauto.
    left; splits; auto.
    split; eapply valid_machine_state_cstep_star'; eauto; apply Hvms.
    * exists s'' b''; splits; auto.
    do 2 right; exists b''; splits; auto.
    apply pbisim_refl.
    split; eapply valid_machine_state_cstep_star'; eauto; apply Hvms.

    - splits; intros.
    + edestruct pbp_halted1; eauto.
    split. eapply step_star_app; eauto.
    apply step_star_step_star_sans_s_iter; auto.
    left; splits; auto.
    apply pbisim_refl.
    split.
    apply Hvms.
    eapply valid_machine_state_step_star; eauto; apply Hvms.
    + edestruct step_star_sans_s_iter_commute_s_iter1 as [s0[??]]; eauto.
    exists s0; split; auto.
    do 2 right.
    exists bq'; splits; auto.
    eapply pbisim_s_iter; eauto.
    exists R; split; auto.
    split; eapply valid_machine_state_s_iter; eauto; apply Hvms.
    + symmetry; eapply step_star_sans_s_iter_respects_s_i; eauto.
    + edestruct cstep_star_pbisim1 with (ba:=bp) (bb:=bq') as [wO'_[bq''[?[??]]]]; eauto.
    exists R; auto.
    replace wO'_ with wO' in *.
    exists sp' bq''; splits; auto.
    eapply cstep_star_app; eauto.
    apply step_star_add_context_cstep_star; auto.
    apply step_star_step_star_sans_s_iter; auto.
    right; left; splits; auto.
    split; eapply valid_machine_state_cstep_star'; eauto; apply Hvms.
    eapply context_world_eq; eauto.

    - split; jauto. right; left; splits; auto.
    split; apply Hvms.
  Qed.

  Lemma bisim_cbisim: forall wO s bp bq,
    bisimilar s bp bq ->
    cbisimilar wO s bp s bq.
  Proof.
    introv H.
    exists (fun (wO:world) s bp sq bq => sq=s /\ bisimilar s bp bq).
    split.
    clear wO s bp bq H.
    intros wO s bp sq bq [??]; subst sq.

    - splits; intros.
    + splits; auto. eapply bisim_halted1; eauto. apply bisim_refl.
    + exists sp'; splits; auto. eapply bisim_s_iter; eauto.
    + reflexivity.
    + edestruct cstep_star_bisim1 with (bb:=bq) as [bq'[??]]; eauto.
    exists sp' bq'; splits; auto.
    apply bisim_symm; auto.

    - splits; auto. apply bisim_symm; auto.
  Qed.

  Lemma cbisim_abisim1: forall wO sp bp sq bq,
    cbisimilar wO sp bp sq bq -> abisimilar wO sp bp sq bq.
  Proof. introv [R[?[??]]]; eexists; eauto. Qed.

  Lemma cbisim_abisim2: forall wO sp bp sq bq,
    cbisimilar wO sp bp sq bq -> abisimilar wO sq bq sp bp.
  Proof. introv [R[?[??]]]; eexists; eauto. Qed.

  Lemma abisim_cbisim: forall wO sp bp sq bq,
    abisimilar wO sp bp sq bq -> abisimilar wO sq bq sp bp -> cbisimilar wO sp bp sq bq.
  Proof.
    introv [R1[??]] [R2[??]].
    exists (fun wO sp bp sq bq => R1 wO sp bp sq bq \/ R2 wO sp bp sq bq).
    splits; auto.
    clear wO sp bp sq bq H0 H2.
    intros wO sp bp sq bq [?|?].

    - edestruct H as [?[?[??]]]; eauto.
    splits; intros; eauto; try solve [intuition].
    + edestruct H3 as [sq'[??]]; eauto.
    + edestruct H5 as [sq''[bq''[??]]]; eauto.

    - edestruct H1 as [?[?[??]]]; eauto.
    splits; intros; eauto; try solve [intuition].
    + edestruct H3 as [sq'[??]]; eauto.
    + edestruct H5 as [sq''[bq''[??]]]; eauto.
  Qed.


  Lemma bisim_abisim1: forall wO s bp bq,
    bisimilar s bp bq ->
    abisimilar wO s bp s bq.
  Proof.
    intros.
    eapply cbisim_abisim1.
    apply bisim_cbisim; auto.
  Qed.

  Lemma bisim_abisim2: forall wO s bp bq,
    bisimilar s bp bq -> abisimilar wO s bq s bp.
  Proof. intros; apply bisim_abisim1. apply bisim_symm; auto. Qed.

  Lemma bisim_Cs_cbisim: forall wO s bp bq,
    bisimilar_Cs bp bq -> cbisimilar wO s bp s bq.
  Proof.
    intros; apply bisim_cbisim; auto; apply bisim_Cs_bisim; auto.
  Qed.


  Lemma cbisim_refl: forall wO s b,
    cbisimilar wO s b s b.
  Proof. 
    intros.
    apply bisim_cbisim. apply bisim_refl.
  Qed.

  Lemma cbisim_symm: forall wO sp bp sq bq,
    cbisimilar wO sp bp sq bq ->
    cbisimilar wO sq bq sp bp.
  Proof.
    intros.
    destruct H as [R[?[??]]].
    exists R; auto.
  Qed.

  Lemma context_world_dummyContext2: forall wO1 wO2,
    context_world (dummyContext wO1) wO2 -> wO1=wO2.
  Proof.
    introv [w1[w2[?[??]]]].
    simpl in *; subst w1 w2.
    apply gworld_unit_identity in H1; auto.
  Qed.

  Lemma step_star_abisim: forall wO sp bp sq bq sp' bp',
    step_star sp bp sp' bp' ->
    abisimilar wO sp bp sq bq ->
    exists sq' bq',
      step_star sq bq sq' bq' /\ abisimilar wO sq' bq' sp' bp'.
  Proof.
    introv H0 [R[H H1]].
    apply H in H1. destruct H1 as [_[_[_ ?]]]; eauto.
    edestruct H1 with (C:=dummyContext wO) as [sq'[bq'[??]]]; eauto.
    apply step_star_add_context_cstep_star; eauto.
    apply context_world_dummyContext.
    apply context_world_dummyContext.
    edestruct cstep_star_dummyContext; eauto.
    exists sq' bq'; splits; auto.
    exists R; split; auto.
  Qed.

  Lemma abisim_s_i_eq: forall wO sp bp sq bq,
    abisimilar wO sp bp sq bq -> s_i sp = s_i sq.
  Proof.
    introv [R[??]].
    edestruct H as [_[_[? _]]]; eauto.
  Qed.

  Lemma step_star_sans_s_iter_abisim: forall wO sp bp sq bq sp' bp',
    step_star_sans_s_iter sp bp sp' bp' ->
    abisimilar wO sp bp sq bq ->
    exists sq' bq',
      step_star_sans_s_iter sq bq sq' bq' /\ abisimilar wO sq' bq' sp' bp'.
  Proof.
    intros.
    edestruct step_star_abisim as [sq'[bq'[??]]]; eauto.
    apply step_star_step_star_sans_s_iter; eauto.
    exists sq' bq'; splits; auto.
    apply step_star_sans_s_iter_respects_s_i in H.
    apply abisim_s_i_eq in H0.
    apply abisim_s_i_eq in H2.
    apply step_star_s_iter_step_star; auto.
    congruence.
  Qed.

  Lemma step_abisim: forall wO sp bp sq bq sp' bp',
    step sp bp sp' bp' ->
    abisimilar wO sp bp sq bq ->
    exists sq' bq',
      step_star sq bq sq' bq' /\ abisimilar wO sq' bq' sp' bp'.
  Proof.
    intros.
    edestruct step_star_abisim as [sq'[bq'[??]]]; eauto.
    apply step_single; auto.
  Qed.

  Lemma step_abisim': forall wO sp bp sq bq sp' bp',
    step sp bp sp' bp' ->
    abisimilar wO sp bp sq bq ->
    exists sq' bq',
      step_star_sans_s_iter sq bq sq' bq' /\ abisimilar wO sq' bq' sp' bp'.
  Proof.
    intros.
    edestruct step_star_sans_s_iter_abisim as [sq'[bq'[??]]]; eauto.
    eapply step_cons_; eauto; apply step_nil_.
  Qed.

  Lemma cstep_star_abisim: forall wO wO' sp C bp sq bq sp' C' bp',
    cstep_star sp C bp sp' C' bp' ->
    context_world C wO ->
    context_world C' wO' ->
    abisimilar wO sp bp sq bq ->
    exists sq' bq',
      cstep_star sq C bq sq' C' bq' /\ abisimilar wO' sq' bq' sp' bp'.
  Proof.
    introv H0 HwO HwO' [R[H H1]].
    apply H in H1. destruct H1 as [_[_[_ ?]]]; eauto.
    edestruct H1 as [sq'[bq'[??]]]; eauto.
    exists sq' bq'; splits; auto.
    exists R; split; auto.
  Qed.

  Lemma cstep_abisim: forall wO wO' sp C bp sq bq sp' C' bp',
    cstep sp C bp sp' C' bp' ->
    context_world C wO ->
    context_world C' wO' ->
    abisimilar wO sp bp sq bq ->
    exists sq' bq',
      cstep_star sq C bq sq' C' bq' /\ abisimilar wO' sq' bq' sp' bp'.
  Proof.
    introv H0 HwO HwO' H.
    edestruct cstep_star_abisim with (C:=C) as [sq'[bq'[??]]]; eauto.
    eapply cstep_cons; eauto; apply cstep_nil.
  Qed.

  Lemma abisim_context: forall wO wO' sp bp sq bq bO sp' bO',
    step sp bO sp' bO' ->
    join_wterms bO wO ->
    join_wterms bO' wO' ->
    abisimilar wO sp bp sq bq ->
    exists sq' bq',
      cstep_star sq (cx_par1 cx_hole bO) bq sq' (cx_par1 cx_hole bO') bq' /\
      abisimilar wO' sq' bq' sp' bp.
  Proof.
    introv H0 HwO HwO' H.
    edestruct cstep_abisim as [sq'[bq'[??]]]; eauto.
    apply cst_par1_step2 with (C:=cx_hole); eauto.
    simpl; exists gwemp wO; splits; auto. apply gjoin_bot2.
    simpl; exists gwemp wO'; splits; auto. apply gjoin_bot2.
  Qed.

  Lemma abisim_s_iter: forall wO sp bp sq bq sp',
    s_iter sp sp' ->
    abisimilar wO sp bp sq bq ->
    exists sq',
      s_iter sq sq' /\ abisimilar wO sp' bp sq' bq.
  Proof.
    introv H [R[??]].
    apply H0 in H1; destruct H1 as [_[? _]].
    edestruct H1 as [sq'[??]]; eauto.
    exists sq'; split; auto.
    exists R; auto.
  Qed.

  Lemma abisim_s_iter': forall wO s bp bq s',
    s_iter s s' ->
    abisimilar wO s bp s bq ->
    abisimilar wO s' bp s' bq.
  Proof.
    intros.
    edestruct abisim_s_iter as [sq'[??]]; eauto.
    generalize (abisim_s_i_eq wO s' bp sq' bq H2); intro H3.
    inverts H; inverts H1.
    apply s_iter_get_eq_s_i with (i:=i0) in H3.
    destruct (Pos.eq_dec i0 i). subst i0.
    auto.
    false.
    rewrite s_iter_get_set_neq in H3; auto.
    rewrite s_iter_get_set_eq in H3.
    congruence.
  Qed.

  Lemma abisim_flip: forall wO sp bp sq bq,
    abisimilar wO sp bp sq bq ->
    exists sq' bq',
      step_star_sans_s_iter sq bq sq' bq' /\ abisimilar wO sq' bq' sp bp.
  Proof.
    intros.
    edestruct step_star_sans_s_iter_abisim as [sq'[bq'[??]]]; eauto.
    apply step_nil_.
  Qed.

  Lemma abisim_halted1: forall wO sp bp sq bq,
    halted bp ->
    abisimilar wO sp bp sq bq ->
    step_star sq bq sp bp.
  Proof.
    introv ? [R[H0 H1]].
    apply H0 in H1. destruct H1 as [[??] _]; eauto.
  Qed.

  Lemma abisim_halted1': forall wO sp bp sq bq,
    halted bp ->
    abisimilar wO sp bp sq bq ->
    step_star_sans_s_iter sq bq sp bp.
  Proof.
    intros.
    apply step_star_s_iter_step_star.
    eapply abisim_halted1; eauto.
    symmetry; eapply abisim_s_i_eq; eauto.
  Qed.

  Lemma abisim_s_iter_star: forall wO sp bp sq bq sp',
    s_iter_star sp sp' ->
    abisimilar wO sp bp sq bq ->
    exists sq',
      s_iter_star sq sq' /\ abisimilar wO sp' bp sq' bq.
  Proof.
    intros.
    revert sq bq H0.
    induction H; intros.
    exists sq; split; auto. apply s_iter_nil.
    edestruct abisim_s_iter as [sq'[??]]; eauto.
    edestruct IHs_iter_star with (sq:=sq') as [sq''[??]]; eauto.
    exists sq''; split; auto.
    eapply s_iter_cons; eauto.
  Qed.

  Lemma abisim_halted2: forall wO sp bp sq bq,
    halted bq ->
    abisimilar wO sp bp sq bq ->
    step_star sp bp sq bq.
  Proof.
    intros.
    edestruct abisim_flip as [sq'[bq'[??]]]; eauto.
    destruct H as [wq ?]; subst bq.
    inverts H1.
    eapply abisim_halted1; eauto.
    eexists; eauto.
    inverts H.
  Qed.

  Lemma abisim_halted2': forall wO sp bp sq bq,
    halted bq ->
    abisimilar wO sp bp sq bq ->
    step_star_sans_s_iter sp bp sq bq.
  Proof.
    intros.
    edestruct abisim_flip as [sq'[bq'[??]]]; eauto.
    destruct H as [wq ?]; subst bq.
    inverts H1.
    eapply abisim_halted1'; eauto.
    eexists; eauto.
    inverts H.
  Qed.

  Lemma cbisim_halted1: forall wO sp bp sq bq,
    halted bp ->
    cbisimilar wO sp bp sq bq ->
    step_star sq bq sp bp.
  Proof.
    intros; eapply abisim_halted1; auto; apply cbisim_abisim1; eauto.
  Qed.

  Lemma cbisim_halted2: forall wO sp bp sq bq,
    halted bq ->
    cbisimilar wO sp bp sq bq ->
    step_star sp bp sq bq.
  Proof.
    intros; eapply abisim_halted1; auto; apply cbisim_abisim2; eauto.
  Qed.

  Lemma cbisim_halted1': forall wO sp bp sq bq,
    halted bp ->
    cbisimilar wO sp bp sq bq ->
    step_star_sans_s_iter sq bq sp bp.
  Proof.
    intros; eapply abisim_halted1'; auto; apply cbisim_abisim1; eauto.
  Qed.

  Lemma cbisim_halted2': forall wO sp bp sq bq,
    halted bq ->
    cbisimilar wO sp bp sq bq ->
    step_star_sans_s_iter sp bp sq bq.
  Proof.
    intros; eapply abisim_halted1'; auto; apply cbisim_abisim2; eauto.
  Qed.


  Lemma abisim_halted_converge: forall wO sp bp sq bq sp' bp',
    step_star sp bp sp' bp' ->
    abisimilar wO sp bp sq bq ->
    halted bp' ->
    step_star sq bq sp' bp'.
  Proof.
    intros.
    edestruct step_star_abisim as [sq'[bq'[??]]]; eauto.
    eapply step_star_app; eauto.
    eapply abisim_halted2; eauto.
  Qed.

  Lemma abisim_halted_converge': forall wO sp bp sq bq sp' bp',
    step_star_sans_s_iter sp bp sp' bp' ->
    abisimilar wO sp bp sq bq ->
    halted bp' ->
    step_star_sans_s_iter sq bq sp' bp'.
  Proof.
    intros.
    edestruct step_star_sans_s_iter_abisim as [sq'[bq'[??]]]; eauto.
    eapply step_star_sans_s_iter_app; eauto.
    eapply abisim_halted2'; eauto.
  Qed.


  Lemma cbisim_halted_converge1: forall wO sp bp sq bq sp' bp',
    step_star sp bp sp' bp' ->
    cbisimilar wO sp bp sq bq ->
    halted bp' ->
    step_star sq bq sp' bp'.
  Proof.
    introv H0 H H1.
    eapply abisim_halted_converge; eauto.
    apply cbisim_abisim1; eauto.
  Qed.

  Lemma cbisim_halted_converge2: forall wO sp bp sq bq sq' bq',
    step_star sq bq sq' bq' ->
    cbisimilar wO sp bp sq bq ->
    halted bq' ->
    step_star sp bp sq' bq'.
  Proof.
    introv H0 H H1.
    eapply abisim_halted_converge; eauto.
    apply cbisim_abisim2; eauto.
  Qed.


  Lemma abisim_refl: forall wO s b,
    abisimilar wO s b s b.
  Proof.
    intros.
    apply bisim_abisim1.
    apply bisim_refl.
  Qed.

  Lemma abisim_past: forall wO sp bp sq bq sq0 bq0,
    step_star_sans_s_iter sq0 bq0 sq bq ->
    abisimilar wO sp bp sq bq ->
    abisimilar wO sp bp sq0 bq0.
  Proof.
    introv H H0.
    exists (fun wO sp bp sq0 bq0 => exists sq bq, step_star_sans_s_iter sq0 bq0 sq bq /\ abisimilar wO sp bp sq bq).
    split; eauto.
    clear wO sp bp sq bq sq0 bq0 H H0.
    intros wO sp bp sq0 bq0 [sq[bq[??]]].
    splits; intros.
    - split.
      eapply step_star_app.
      apply step_star_step_star_sans_s_iter; eauto.
      eapply abisim_halted1; eauto.
      exists sp bp; splits. apply step_nil_. apply abisim_refl.
    - edestruct abisim_s_iter as [sq'[??]]; eauto.
      edestruct step_star_sans_s_iter_commute_s_iter1 as [s'0[??]]; eauto.
      exists s'0; split; eauto.
    - apply abisim_s_i_eq in H0.
      apply step_star_sans_s_iter_respects_s_i in H.
      congruence.
    - edestruct cstep_star_abisim with (bp:=bp) (bq:=bq) as [sq'[bq'[??]]]; eauto.
      exists sq' bq'; split; auto.
      eapply cstep_star_app; eauto.
      apply step_star_add_context_cstep_star; auto.
      apply step_star_step_star_sans_s_iter; auto.
      exists sp' bp'; splits; auto. apply step_nil_.
  Qed.


  Lemma abisim_future: forall wO sp bp sq bq sp' bp',
    abisimilar wO sp bp sq bq ->
    step_star_sans_s_iter sp bp sp' bp' ->
    abisimilar wO sp' bp' sq bq.
  Proof.
    intros.
    exists (fun wO sp' bp' sq bq => exists sp bp, step_star_sans_s_iter sp bp sp' bp' /\ abisimilar wO sp bp sq bq).
    split; eauto.
    clear wO sp bp sq bq sp' bp' H H0.
    intros wO sp' bp' sq bq [sp[bp[??]]].
    splits; intros.
    - split.
    eapply abisim_halted_converge; eauto.
    apply step_star_step_star_sans_s_iter; auto.
    exists sp' bp'; split.
    apply step_nil_.
    apply abisim_refl.
    - edestruct step_star_sans_s_iter_commute_s_iter1 as [sp'1[??]]; eauto.
    clear H H1.
    edestruct abisim_s_iter as [sq'[??]]; eauto.
    exists sq'; split; auto.
    exists sp'1 bp; split; auto.
    - apply abisim_s_i_eq in H0.
    apply step_star_sans_s_iter_respects_s_i in H.
    congruence.
    - edestruct cstep_star_abisim with (C:=C) (C':=C') (bp:=bp) (bq:=bq) as [sq'[bq'[??]]]; eauto.
    eapply cstep_star_app; eauto.
    apply step_star_add_context_cstep_star; auto.
    apply step_star_step_star_sans_s_iter; auto.
    exists sq' bq'; splits; auto.
    exists sq' bq'; splits; auto. apply step_nil_.
  Qed.



  Lemma abisim_trans: forall wO sp bp sr br sq bq,
    abisimilar wO sp bp sr br ->
    abisimilar wO sr br sq bq ->
    abisimilar wO sp bp sq bq.
  Proof.
    introv H H0.
    exists (fun wO sp bp sq bq => exists sr br, abisimilar wO sp bp sr br /\ abisimilar wO sr br sq bq).
    split; eauto.
    clear wO sp bp sr br sq bq H H0.
    intros wO sp bp sq bq [sr[br[??]]].
    splits.

    - intro; split.
    eapply abisim_halted_converge; eauto.
    eapply abisim_halted1; eauto.
    exists sp bp; splits; try apply abisim_refl.
    - intros.
    edestruct abisim_s_iter as [sr'[??]]; eauto.
    edestruct abisim_s_iter with (sp:=sr) as [sq'[??]]; eauto.
    exists sq'; split; auto.
    exists sr' br; splits; auto.
    - apply abisim_s_i_eq in H.
    apply abisim_s_i_eq in H0.
    congruence.
    - intros.
    edestruct cstep_star_abisim with (bq:=br) as [sr'[br'[??]]]; eauto.
    edestruct cstep_star_abisim with (bq:=bq) as [sq'[bq'[??]]]; eauto.
    exists sq' bq'; splits; auto.
    exists sr' br'; splits; auto.
  Qed.


  Lemma cbisim_trans: forall wO sp bp sr br sq bq,
    cbisimilar wO sp bp sr br ->
    cbisimilar wO sr br sq bq ->
    cbisimilar wO sp bp sq bq.
  Proof.
    intros.
    apply abisim_cbisim.
    eapply abisim_trans; apply cbisim_abisim1; eauto.
    eapply abisim_trans; apply cbisim_abisim2; eauto.
  Qed.

  Lemma bisim_abisim: forall wO s bp bq,
    bisimilar s bp bq ->
    abisimilar wO s bp s bq.
  Proof.
    intros.
    apply cbisim_abisim1; auto.
    apply bisim_cbisim; auto.
  Qed.

  Lemma bisim_Cs_abisim: forall wO s bp bq,
    bisimilar_Cs bp bq ->
    abisimilar wO s bp s bq.
  Proof.
    intros.
    apply bisim_abisim; auto.
    apply bisim_Cs_bisim; auto.
  Qed.

End BisimulationRelations.










(*
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
*)

Module Morphisms.
  Export Bisimulation.Morphisms.


  Add Parametric Morphism: cbisimilar
  with signature eq ==> eq ==> equiv ==> eq ==> equiv ==> equiv
  as bisim_Cs_cbisim_m.
    intros; split; intros.
    eapply cbisim_trans.
    apply bisim_Cs_cbisim.
    apply bisim_Cs_symm; apply H.
    eapply cbisim_trans. eauto.
    apply bisim_Cs_cbisim. auto.
    eapply cbisim_trans.
    apply bisim_Cs_cbisim. eauto.
    eapply cbisim_trans. eauto.
    apply cbisim_symm.
    apply bisim_Cs_cbisim. auto.
  Qed.

  Add Parametric Morphism: abisimilar
  with signature eq ==> eq ==> equiv ==> eq ==> equiv ==> equiv
  as bisim_Cs_abisim_m.
    intros; split; intros.
    eapply abisim_trans.
    apply bisim_Cs_abisim.
    apply bisim_Cs_symm; apply H.
    eapply abisim_trans. eauto.
    apply bisim_Cs_abisim. auto.
    eapply abisim_trans.
    apply bisim_Cs_abisim. eauto.
    eapply abisim_trans. eauto.
    apply bisim_Cs_abisim. apply bisim_Cs_symm. auto.
  Qed.
  

End Morphisms.






(*
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
*)

Section Congruence.
  Import Machine.Notation.
  Import Notation.
  Import Notation.U.
  Import ContextSemantics.Notation.
  Import world_model.Notation.
  Import world_model.Notation.U.
  Import Morphisms.
  Open Scope forces_pred.
  Open Scope context_scope.
  Open Scope bsim_scope.
  Open Scope wterm_scope.

  Lemma abisim_halted_converge_cstep_star: forall wO wO' C sp bp sq bq s' C' b',
    cstep_star sp C bp s' C' b' ->
    abisimilar wO sp bp sq bq ->
    context_world C wO ->
    context_world C' wO' ->
    halted b' ->
    cstep_star sq C bq s' C' b'.
  Proof.
    intros.
    edestruct cstep_star_abisim as [sq'[bq'[??]]]; eauto.
    eapply cstep_star_app; eauto.
    apply step_star_add_context_cstep_star.
    eapply abisim_halted2; eauto.
  Qed.

  Lemma step_frame: forall s w b s' b' wF,
    step s b s' b' ->
    join_wterms b w ->
    gjoins w wF ->
    joins_with_state wF s ->
    exists w' w'wF,
      join_wterms b' w' /\
      gjoin w' wF w'wF /\
      join_wterms (extend_world b' wF) w'wF /\
      step s (extend_world b wF) s' (extend_world b' wF).
Admitted.

  Lemma joins_with_state_valid_machine_state: forall wO s b,
    valid_machine_state wO s b ->
    exists wb,
      join_wterms b wb /\ joins_with_state wb s.
Admitted.

  Lemma joins_with_list_sub: forall wl w w',
    joins_with_list w wl ->
    gjoin_sub w' w ->
    joins_with_list w' wl.
  Proof.
    intros; red; red in H.
    rewrite List.Forall_forall in *.
    intros.
    edestruct H; eauto.
    destruct H0 as [w'' ?].
    edestruct gjoin_assoc with (b:=w') (a:=w'') as [?[??]]; eauto.
    apply gjoin_com; auto.
    exists x1; auto.
  Qed.

  Lemma joins_with_state_sub: forall s w w',
    joins_with_state w s ->
    gjoin_sub w' w ->
    joins_with_state w' s.
  Proof.
    repeat intro.
    eapply joins_with_list_sub; eauto.
  Qed.

  Lemma flatten_step: forall wO s w b s' b',
    step s b s' b' ->
    join_wterms b w ->
    valid_machine_state wO s b ->
    exists t w',
      step s (w:t) s' (w':tm_skip) /\ join_wterms b' w'.
  Proof.
    intros.
    revert wO w H0 H1.
    step_cases (induction H) Case HH; intros;
      try solve [exists (tm_assert TT) w; split; simpl in *; subst; auto; apply st_assert; apply TT_gstar; reflexivity].
    - Case "st_send".
    simpl in *; subst. do 2 eexists; splits.
    eapply HH; eauto.
    reflexivity.
    - Case "st_recv".
    exists (tm_recv x ec) (w_cons_incr c (w_env_set w_w' x v)); split.
    eapply HH; simpl in *; subst; eauto.
    reflexivity.
    - Case "st_newch".
    exists (tm_newch x R) w'; split.
    eapply HH with (s':=s'); simpl in *; subst; try reflexivity.
    subst w'; reflexivity.
    - Case "st_endch".
    exists (tm_endch ec) w; split.
    eapply HH with (s':=s'); simpl in *; subst; try reflexivity; eauto.
    reflexivity.
    - Case "st_assign".
    exists (tm_assign x e) w'; split.
    eapply HH; simpl in *; subst; try reflexivity; eauto.
    reflexivity.
    - Case "st_load".
    exists (tm_load x e) w'; split.
    eapply HH; simpl in *; subst; try reflexivity; eauto.
    reflexivity.
    - Case "st_store".
    exists (tm_store e_dest e_value) w'; split.
    eapply HH; simpl in *; subst; try reflexivity; eauto.
    reflexivity.
    - Case "st_local".
    exists (tm_local e (fun _=>tm_skip)) w'; split.
    simpl in *; subst w0.
    change tm_skip with ((fun _ : var => tm_skip) x) at 2.
    eapply HH; try reflexivity; eauto.
    reflexivity.
    - Case "st_freelocal".
    exists (tm_freelocal x) w'; split.
    eapply HH; simpl in *; subst; try reflexivity; eauto.
    reflexivity.
    - Case "st_par".
    exists (tm_assert TT) w; split.
    simpl in *; subst; auto.
    apply st_assert; apply TT_gstar; reflexivity.
    exists w1 w2; splits; simpl; auto.
    - Case "st_seq1".
    eauto.
    - Case "st_par1".
    destruct H0 as [w1[w2[?[??]]]].
    edestruct valid_machine_state_par_observer' as [wb2[wOb2[?[??]]]]; eauto.
    edestruct IHstep as [t[w'[??]]]; eauto.
    edestruct step_frame with (b:=w1:t) as [w1'[w12'[?[?[??]]]]]; eauto.
    reflexivity.
    exists w; eauto.
    edestruct joins_with_state_valid_machine_state with (b:=b1||b2) as [w12_[??]]; eauto.
    destruct H9 as [w1_[w2_[?[??]]]].
    replace w1_ with w1 in *.
    replace w2_ with w2 in *.
    replace w12_ with w in *.
    eapply joins_with_state_sub with (w:=w); eauto.
    exists w1; auto.
    eapply gjoin_eq; eauto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
    erewrite extend_world_simpl in H12; eauto.
    simpl in H9; subst w1'.
    erewrite extend_world_simpl in H11; eauto.
    exists t w12'; split; auto.
    erewrite extend_world_simpl in H12; eauto.
    exists w' w2; splits; auto.
    - Case "st_par2".
    destruct H0 as [w1[w2[?[??]]]].
    apply valid_machine_state_par_swap in H1.
    edestruct valid_machine_state_par_observer' as [wb1[wOb1[?[??]]]]; eauto.
    edestruct IHstep as [t[w2'[??]]]; eauto.
    edestruct step_frame with (b:=w2:t) as [w2'_[w12'[?[?[??]]]]]; eauto.
    reflexivity.
    exists w; eauto.
    edestruct joins_with_state_valid_machine_state with (b:=b2||b1) as [w12_[??]]; eauto.
    destruct H9 as [w2_[w1_[?[??]]]].
    replace w1_ with w1 in *.
    replace w2_ with w2 in *.
    replace w12_ with w in *.
    eapply joins_with_state_sub with (w:=w); eauto.
    exists w2; auto.
    eapply gjoin_eq; eauto. apply gjoin_com; auto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
    simpl in H9; subst w2'_.
    erewrite extend_world_simpl in H12; eauto.
    2: apply gjoin_com; eauto.
    erewrite extend_world_simpl in H11; eauto.
    exists t w12'; split; auto.
    erewrite extend_world_simpl in H12; eauto.
    exists w1 w2'; splits; auto.
    apply gjoin_com; auto.
    - Case "st_par_join".
    exists (tm_assert TT) w; split.
    simpl in *; subst; auto.
    destruct H0 as [w1_[w2_[?[??]]]]; subst w1_ w2_.
    replace w0 with w in *.
    apply st_assert; apply TT_gstar; reflexivity.
    eapply gjoin_eq; eauto.
    reflexivity.
  Qed.

  Import PermuteTactics.
Import JoinList.

  Lemma valid_machine_state_observer_swap: forall wO s b bO,
    valid_machine_state wO s b ->
    join_wterms bO wO ->
    exists wb,
      join_wterms b wb /\ valid_machine_state wb s bO.
  Proof.
    intros.
    destruct H as [wS[wb[ww[?[??]]]]].
    exists wb; split; auto.
    exists wS wO ww; splits; auto.
    eapply join_list_perm; eauto.
    permut4.
  Qed.

  Lemma flatten_cstep: forall wO s C b s' C' b',
    cstep s C b s' C' b' ->
    context_world C wO ->
    valid_machine_state wO s b ->
    exists tO wO',
      context_world C' wO' /\
      cstep_star s (cx_par1 cx_hole (wO:tO)) b s' (cx_par1 cx_hole (wO':tm_skip)) b'.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    exists tm_skip wO; split; auto.
    apply step_single_cstep_star; auto.
    subst b'.
    clear H.
    apply step_simulates_cstep in H2.
    apply context_world_join_wterms in H0.
    edestruct valid_machine_state_observer_swap as [wb[??]]; eauto.
    edestruct flatten_step with (b:=C[gwemp:tm_skip]) as [tO[wO'[??]]]; eauto.
    exists tO wO'; split; auto.
    apply context_world_join_wterms; auto.
    apply cstep_single_cstep_star.
    apply cst_par1_step2; auto.
  Qed.


  Lemma flatten_cstep_star: forall wO s C b s' C' b',
    cstep_star s C b s' C' b' ->
    context_world C wO ->
    valid_machine_state wO s b ->
    exists tO wO',
      context_world C' wO' /\
      cstep_star s (cx_par1 cx_hole (wO:tO)) b s' (cx_par1 cx_hole (wO':tm_skip)) b'.
  Proof.
    intros.
    revert wO H0 H1.
    induction H; intros.
    exists tm_skip wO; split; auto; apply cstep_nil.
    edestruct flatten_cstep as [t[wO'[??]]]; eauto.
    edestruct valid_machine_state_cstep_star as [wO'_[??]]; eauto.
    exists gwemp wO; splits; auto; try reflexivity. apply gjoin_bot2.
    replace wO'_ with wO' in *.
    clear wO'_.
    edestruct IHcstep_star as [t'[wO''[??]]]; eauto.
    exists (t;;; t') wO''; split; auto.
    apply cstep_star_cx_par1_par in H4.
    apply cstep_star_cx_par1_par in H8.
    apply cstep_star_par_cx_par1.
    apply step_star_add_context_cstep_star.
    apply step_star_simulates_cstep_star in H4.
    apply step_star_simulates_cstep_star in H8.
    eapply step_cons.
    apply st_par2; apply st_seq.
    eapply step_star_app; eauto.
    eapply step_star_snoc.
    eapply step_star_par2_inner_seq_append; eauto.
    apply st_par2; apply st_seq_skip.
    destruct H5 as [wa[wb[?[??]]]].
    simpl in H5; subst wa.
    eapply gworld_unit_identity in H8.
    simpl in *; subst; auto.
    (**)
    edestruct IHcstep_star as [t'[wO''[??]]]; eauto.
    eapply valid_machine_state_s_iter; eauto.
    exists t' wO''; split; auto.
    eapply cstep_iteration; eauto.
  Qed.

  Lemma compose_context_world_cstep_star: forall s wO C1 b s' wO' C1' C2 b' s'' C2' b'',
    cstep_star s C1 b s' C1' b' ->
    cstep_star s' C2 b' s'' C2' b'' ->
    context_world C1 wO ->
    context_world C1' wO' ->
    context_world C2 wO' ->
    valid_machine_state wO s b ->
    exists wO'' C12 C12',
      context_world C12 wO /\
      context_world C12' wO'' /\
      context_world C2' wO'' /\
      cstep_star s C12 b s'' C12' b''.
  Proof.
    intros.
    edestruct flatten_cstep_star with (C:=C1) as [tO1[wO'_[??]]]; eauto.
    replace wO'_ with wO' in *. clear wO'_.
    edestruct flatten_cstep_star with (C:=C2) as [tO2[wO''[??]]]; eauto.
    eapply valid_machine_state_cstep_star' with (C:=C1); eauto.
    exists wO'' (cx_par1 cx_hole (wO: tO1;; tO2)) (cx_par1 cx_hole (wO'': tm_skip)); splits; auto.
    exists gwemp wO; splits; simpl; auto. apply gjoin_bot2.
    exists gwemp wO''; splits; simpl; auto. apply gjoin_bot2.
    apply cstep_star_par_cx_par1.
    eapply cstep_star_app.
    apply cstep_star_par2_inner_seq_append.
    apply cstep_star_cx_par1_par; eauto.
    eapply step_cstep_star.
    apply st_par2; apply st_seq_skip.
    apply cstep_star_cx_par1_par; eauto.
    eapply context_world_eq with (C:=C1'); eauto.    
  Qed.

  Lemma abisim_seq: forall wO sp bp sq bq tp tq,
    valid_machine_state2 wO sp bp sq bq ->
    abisimilar wO sp bp sq bq ->
    (forall C s' wO' C' w',
      cstep_star sp C bp s' C' (w':tm_skip) ->
      context_world C wO ->
      context_world C' wO' ->
      cbisimilar wO' s' (w':tp) s' (w':tq)) ->
    abisimilar wO sp (bp;; tp) sq (bq;; tq).
  Proof.
    intros.
    exists (fun wO sp bpt sq bqt =>
      (exists bp tp bq tq, bpt=bp;;tp /\ bqt=bq;;tq /\
        (forall C s' wO' C' w',
          cstep_star sp C bp s' C' (w':tm_skip) ->
          context_world C wO ->
          context_world C' wO' ->
          cbisimilar wO' s' (w':tp) s' (w':tq))
       /\ abisimilar wO sp bp sq bq /\ valid_machine_state2 wO sp bp sq bq)
      \/ abisimilar wO sp bpt sq bqt).
    split.
    2: left; jauto.
    clear wO sp bp sq bq tp tq H H0 H1.
    intros wO sp bpt sq bqt [[bp[tp[bq[tq[?[?[?[? Hvms]]]]]]]] | ?].

    * subst bpt bqt.
    splits.
    - intros [?]; discriminate.
    - intros.
    edestruct abisim_s_iter as [sq'[??]]; eauto.
    exists sq'; split; auto.
    left; exists bp tp bq tq; splits; auto.
    intros; eapply H1; eauto.
    eapply cstep_iteration; eauto.
    split; eapply valid_machine_state_s_iter; eauto; apply Hvms.
    - eapply abisim_s_i_eq; eauto.
    - intros C sp'' wO'' C'' bpt'' H; intros.
    edestruct cstep_star_seq_inversion as [ [s'[C'[w'[??]]]] | [bp''[??]] ]; eauto.
      + clear H.
      edestruct valid_machine_state_cstep_star with (wO:=wO) (C:=C) as [wO'[??]]; eauto.
      apply Hvms.
      edestruct cstep_star_abisim with (bp:=w':tp) as [sq''[bqt''[??]]]; eauto.
      apply cbisim_abisim1; eauto.
      exists sq'' bqt''; splits; auto.
      eapply cstep_star_seq_join; eauto.
      eapply abisim_halted_converge_cstep_star; eauto.
      eexists; eauto.
      + subst bpt''. clear H.
      edestruct cstep_star_abisim as [sq''[bq''[??]]]; eauto.
      exists sq'' (bq'';; tq); splits; auto.
      apply cstep_star_seq_append; auto.
      left; exists bq'' tq bp'' tp; splits; auto.
      intros C2 s''' wO''' C2' w'''; intros.
      edestruct compose_context_world_cstep_star
        with (C1:=C) (b:=bp) (C2:=C2) as [wO'''_[C12[C12'[?[?[??]]]]]]; eauto.
      eapply abisim_halted_converge_cstep_star; eauto.
      eexists; eauto.
      apply Hvms.
      replace wO'''_ with wO''' in *. clear wO'''_.
      apply cbisim_symm.
      eapply H1 with (C:=C12) (C':=C12') (wO':=wO'''); eauto.
      eapply context_world_eq; eauto.
      split; eapply valid_machine_state_cstep_star'; eauto; apply Hvms.

    * splits.
    - intros; split.
    eapply abisim_halted1; eauto.
    right; apply abisim_refl.
    - intros; edestruct abisim_s_iter as [sq'[??]]; eauto.
    - eapply abisim_s_i_eq; eauto.
    - intros.
    edestruct cstep_star_abisim as [sq'[bq'[??]]]; eauto.
  Qed.

  Lemma cbisim_seq: forall wO sp bp sq bq tp tq,
    valid_machine_state2 wO sp bp sq bq ->
    cbisimilar wO sp bp sq bq ->
    (forall C s' wO' C' w',
      cstep_star sp C bp s' C' (w':tm_skip) ->
      context_world C wO ->
      context_world C' wO' ->
      cbisimilar wO' s' (w':tp) s' (w':tq)) ->
    cbisimilar wO sp (bp;; tp) sq (bq;; tq).
  Proof.
    intros.
    apply abisim_cbisim; eapply abisim_seq; eauto.
    apply cbisim_abisim1; auto.
    split; apply H.
    apply cbisim_abisim2; auto.
    intros; apply cbisim_symm.
    eapply H1; eauto.
    eapply abisim_halted_converge_cstep_star; eauto.
    apply cbisim_abisim2; auto.
    eexists; eauto.
  Qed.


  Lemma abisim_seq1: forall wO sp bp sq bq t,
    valid_machine_state2 wO sp bp sq bq ->
    abisimilar wO sp bp sq bq ->
    abisimilar wO sp (bp;; t) sq (bq;; t).
  Proof.
    intros; eapply abisim_seq; auto; intros; apply cbisim_refl.
  Qed.

  Lemma abisim_seq2: forall wO s b tp tq,
    valid_machine_state wO s b ->
    (forall C s' wO' C' w',
      cstep_star s C b s' C' (w':tm_skip) ->
      context_world C wO ->
      context_world C' wO' ->
      cbisimilar wO' s' (w':tp) s' (w':tq)) ->
    abisimilar wO s (b;; tp) s (b;; tq).
  Proof.
    intros; eapply abisim_seq; auto; intros; auto.
    split; auto.
    apply abisim_refl.
  Qed.

  Lemma cbisim_seq1: forall wO sp bp sq bq t,
    valid_machine_state2 wO sp bp sq bq ->
    cbisimilar wO sp bp sq bq ->
    cbisimilar wO sp (bp;; t) sq (bq;; t).
  Proof.
    intros; eapply cbisim_seq; auto; intros; apply cbisim_refl.
  Qed.

  Lemma cbisim_seq2: forall wO s b tp tq,
    valid_machine_state wO s b ->
    (forall C s' wO' C' w',
      cstep_star s C b s' C' (w':tm_skip) ->
      context_world C wO ->
      context_world C' wO' ->
      cbisimilar wO' s' (w':tp) s' (w':tq)) ->
    cbisimilar wO s (b;; tp) s (b;; tq).
  Proof.
    intros; eapply cbisim_seq; auto; intros.
    split; auto.
    apply cbisim_refl; auto.
  Qed.

  Lemma cstep_par_compose_contexts1: forall s C b1 b2 s' C' b1' b2',
    cstep s C (b1 || b2) s' C' (b1' || b2') ->
    cstep s (compose_contexts C (cx_par1 cx_hole b2)) b1 s' (compose_contexts C' (cx_par1 cx_hole b2')) b1'.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto; clear H.
    subst C'.
    inverts H0.
    apply cstep_simulates_step; apply step_add_context; auto.
    apply cstep_composed_context_step2 with gwemp.
    apply cst_par1_step2; auto.
    inverts H1.
    apply cstep_composed_context_step1 with gwemp; auto.
  Qed.


  Lemma abisim_par1: forall wO wO2 sp bp sq bq w2 b2,
    abisimilar wO2 sp bp sq bq ->
    valid_machine_state2 wO2 sp bp sq bq ->
    join_wterms b2 w2 ->
    gjoin wO w2 wO2 ->
    abisimilar wO sp (bp || b2) sq (bq || b2).
  Proof.
    intros.
    exists (fun wO sp bp2 sq bq2 =>
      (exists bp bq wO2 w2 b2, bp2=bp||b2 /\ bq2=bq||b2 /\ join_wterms b2 w2 /\ gjoin wO w2 wO2 /\ valid_machine_state2 wO2 sp bp sq bq /\ abisimilar wO2 sp bp sq bq)
      \/ (sp=sq /\ bp2=bq2 /\ valid_machine_state wO sp bp2)).
    split.
    2: left; jauto.
    clear wO sp bp sq bq wO2 w2 b2 H H0 H1 H2.
    intros wO sp bp2 sq bq2 [[bp[bq[wO2[w2[b2[?[?[?[?[Hvms ?]]]]]]]]]] | [?[??]]].

    * subst bp2 bq2.
    splits.
    - introv [?]; discriminate.
    - intros; edestruct abisim_s_iter as [sq'[??]]; eauto.
    exists sq'; split; auto.
    left; do 5 eexists; splits; eauto.
    split; eapply valid_machine_state_s_iter; eauto; apply Hvms.
    - eapply abisim_s_i_eq; eauto.
    - intros C sp' wO' C' bp2'; intros.
    edestruct cstep_star_par_resolve
      as [ [bp'[b2' ?]] | [w1'[w2'[w'[?[??]]]]] ]; eauto.
    + subst bp2'.
    apply cstep_star_par_cx_par1 in H.
    edestruct valid_machine_state_cstep_star as [wO2'[??]]; eauto.
    exists wO w2; splits; eauto.
    apply Hvms.
    edestruct cstep_star_abisim as [sq'[bq'[??]]]; eauto.
    exists wO w2; splits; eauto.
    destruct H5 as [wO'_[w2'[?[??]]]].
    replace wO'_ with wO' in *.
    exists sq' (bq'||b2'); splits; auto.
    apply cstep_star_cx_par1_par; eauto.
    left; do 5 eexists; splits; eauto.
    split; eauto.
    eapply valid_machine_state_cstep_star'; eauto.
    exists wO w2; splits; eauto.
    exists wO' w2'; splits; eauto.
    apply Hvms.
    eapply context_world_eq; eauto.
    + subst bp2'. clear H.
    edestruct valid_machine_state_cstep_star as [wO2'[? ?]].
    apply cstep_star_par_cx_par1; eauto.
    exists wO w2; splits; eauto.
    apply Hvms.
    exists sp' (w': tm_skip); splits; auto.
    eapply cstep_star_snoc.
    apply cstep_star_cx_par1_par.
    eapply abisim_halted_converge_cstep_star; eauto.
    apply cstep_star_par_cx_par1; eauto.
    exists wO w2; splits; auto.
    eexists; eauto.
    apply step_cstep.
    apply st_par_join; auto.
    right; splits; auto.
    destruct H as [wO'_[w2'_[?[??]]]].
    simpl in H8; subst w2'_.
    replace wO'_ with wO' in *.
    eapply valid_machine_state_wterm. 
    eapply valid_machine_state_observer_par with (bO:=w2':tm_skip); eauto.
    reflexivity.
    exists w1' w2'; splits; eauto; reflexivity.
    reflexivity.
    eapply context_world_eq; eauto.

    * subst; splits.
    - intros; split; jauto.
    apply step_nil.
    - intros. exists sp'; split; auto.
    right; splits; auto.
    eapply valid_machine_state_s_iter; eauto.
    - reflexivity.
    - intros.
    exists sp' bp'; split; auto.
    right; splits; auto.
    eapply valid_machine_state_cstep_star'; eauto.
  Qed.

  Import Morphisms.

  Lemma abisim_par_swap: forall wO s b1 b2,
    abisimilar wO s (b1||b2) s (b2||b1).
  Proof.
    intros.
    setoid_rewrite bisim_Cs_par_swap at 1; auto.
    apply abisim_refl.
  Qed.

  Lemma cbisim_par_swap: forall wO s b1 b2,
    cbisimilar wO s (b1||b2) s (b2||b1).
  Proof.
    intros.
    setoid_rewrite bisim_Cs_par_swap at 1; auto.
    apply cbisim_refl.
  Qed.


  Lemma abisim_par2: forall wO wO1 w1 b1 sp bp sq bq,
    abisimilar wO1 sp bp sq bq ->
    join_wterms b1 w1 ->
    gjoin wO w1 wO1 ->
    valid_machine_state2 wO1 sp bp sq bq ->
    abisimilar wO sp (b1 || bp) sq (b1 || bq).
  Proof.
    intros.
    setoid_rewrite bisim_Cs_par_swap.
    eapply abisim_par1; eauto.
  Qed.

  Lemma abisim_par: forall wO wO1 wO2 w1 w2 sa sb sc bp1 bp2 bq1 bq2,
    abisimilar wO2 sa bp1 sb bq1 ->
    abisimilar wO1 sb bp2 sc bq2 ->
    join_wterms bq1 w1 ->
    join_wterms bp2 w2 ->
    gjoin wO w1 wO1 ->
    gjoin wO w2 wO2 ->
    valid_machine_state2 wO2 sa bp1 sb bq1 ->
    valid_machine_state2 wO1 sb bp2 sc bq2 ->
    abisimilar wO sa (bp1 || bp2) sc (bq1 || bq2).
  Proof.
    intros. eapply abisim_trans.
    eapply abisim_par1; eauto.
    eapply abisim_par2; eauto.
  Qed.

  Lemma cbisim_par1: forall wO wO2 sp bp sq bq w2 b2,
    cbisimilar wO2 sp bp sq bq ->
    join_wterms b2 w2 ->
    gjoin wO w2 wO2 ->
    valid_machine_state2 wO2 sp bp sq bq ->
    cbisimilar wO sp (bp || b2) sq (bq || b2).
  Proof.
    intros.
    apply abisim_cbisim.
    eapply abisim_par1; eauto.
    apply cbisim_abisim1; auto.
    eapply abisim_par1; eauto.
    apply cbisim_abisim2; auto.
    split; apply H2.
  Qed.

  Lemma cbisim_par2: forall wO wO1 w1 b1 sp bp sq bq,
    cbisimilar wO1 sp bp sq bq ->
    join_wterms b1 w1 ->
    gjoin wO w1 wO1 ->
    valid_machine_state2 wO1 sp bp sq bq ->
    cbisimilar wO sp (b1 || bp) sq (b1 || bq).
  Proof.
    intros.
    apply abisim_cbisim.
    eapply abisim_par2; eauto.
    apply cbisim_abisim1; auto.
    eapply abisim_par2; eauto.
    apply cbisim_abisim2; auto.
    split; apply H2.
  Qed.


  Lemma cbisim_par: forall wO wO1 wO2 w1 w2 sa sb sc bp1 bp2 bq1 bq2,
    cbisimilar wO2 sa bp1 sb bq1 ->
    cbisimilar wO1 sb bp2 sc bq2 ->
    join_wterms bq1 w1 ->
    join_wterms bp2 w2 ->
    gjoin wO w1 wO1 ->
    gjoin wO w2 wO2 ->
    valid_machine_state2 wO2 sa bp1 sb bq1 ->
    valid_machine_state2 wO1 sb bp2 sc bq2 ->
    cbisimilar wO sa (bp1 || bp2) sc (bq1 || bq2).
  Proof.
    intros.
    eapply cbisim_trans.
    eapply cbisim_par1; eauto.
    eapply cbisim_par2; eauto.
  Qed.

  Lemma abisim_if: forall wO s tp1 tp2 tq1 tq2 ep eq w,
    (expr_bool_equiv ep eq) ->
    (evalTrue ep w -> abisimilar wO s (w: tp1) s (w: tq1)) ->
    (evalFalse ep w -> abisimilar wO s (w: tp2) s (w: tq2)) ->
    abisimilar wO s (w: tm_if ep tp1 tp2) s (w: tm_if eq tq1 tq2).
  Proof.
    intros.
    setoid_rewrite<- H.
    destruct (evalTrueFalse_dec ep w) as [[?|?]|[??]].

    setoid_rewrite bisim_Cs_if_true; auto.
    setoid_rewrite bisim_Cs_if_false; auto.
    apply bisim_Cs_abisim.
    apply bisim_Cs_if_noeval; auto.
  Qed.

(*
  Lemma bisim_while: forall s w e z t1 t2 P,
    w |= P ->
    (forall s' C' w', w'|= P -> evalTrue e w' -> halted_assert s' C' (w':t1) P) ->
    (forall s' C' w', w'|= P -> evalTrue e w' -> halted_assert s' C' (w':t2) P) ->
    (forall C s' C' w' tt bb, cstep_star s C (w: tm_while_max e z t1) s' C' (w': tt;; bb) -> w'|= P -> evalTrue e w' -> bisimilar s' (w': t1) (w': t2)) ->
    bisimilar s (w: tm_while_max e z t1) (w: tm_while_max e z t2).
  Proof.
    intros.
    rename s into s0.
    rename w into w0.
    rename z into z0.
    exists (fun s ba bb =>
      (
        (exists z w, ba=w:tm_while_max e z t1 /\ bb=w:tm_while_max e z t2 /\ w|=P) \/
        (exists z bba bbb, ba=bba;; tm_while_max e z t1 /\ bb=bbb;; tm_while_max e z t2 /\ bisimilar s bba bbb /\ (forall C, halted_assert s C bba P)) \/
        ba=bb
      ) /\
      exists C0 C, cstep_star s0 C0 (w0: tm_while_max e z0 t1) s C ba).
    split.
    2: split; eauto; [ left; eauto | exists cx_hole cx_hole; apply cstep_nil].
    rename H into HPw0.
    intros s ba bb [[[z[w[?[??]]]]| [ [z[bba[bbb[?[?[??]]]]]] | ?] ] ?]; subst.
    splits.

    intros [?]; discriminate.
    intros [?]; discriminate.

    intros; split.
    left; exists z w; splits; auto.
    edestruct H5 as [C0[C ?]].
    exists C0 C.
    eapply cstep_star_iter_snoc; eauto.

    intros; split.
    left; exists z w; splits; auto.
    edestruct H5 as [C0[C0' ?]].
    exists (cx_par1 C0 b) (cx_par1 C0' b').
      eapply cstep_star_snoc.
      eapply cstep_star_par_cx_par1; eauto.
      apply cstep_star_par1_step; eauto.
      eapply cst_par1_step2; eauto.

    intros s' ba' ?.
    destruct H5 as [C0[C0' H7]].
    inverts H.
    exists (w: t2;; tm_while_max e z' t2); splits; auto.
      apply step_single; apply st_while_true; auto.
      right; left; exists z' (w: t1) (w:t2); splits; auto.
      eapply H2 with (C:=C0); eauto.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_true; eauto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_true; auto.
    exists (w: tm_skip); splits; auto.
      apply step_single; apply st_while_false; auto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_false; auto.
    exists (w: tm_while_max e z1 t2); splits; auto.
      apply step_single; apply st_while_go; auto.
      left; exists z1 w; splits; auto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_go; auto.
    exists (w: tm_skip); splits; auto.
      apply step_single; apply st_while_limit; auto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_limit; auto.

    intros s' bb' ?.
    destruct H5 as [C0[C0' H7]].
    inverts H.
    exists (w: t1;; tm_while_max e z' t1); splits; auto.
      apply step_single; apply st_while_true; auto.
      right; left; exists z' (w: t1) (w:t2); splits; auto.
      eapply H2 with (C:=C0); eauto.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_true; eauto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_true; auto.
    exists (w: tm_skip); splits; auto.
      apply step_single; apply st_while_false; auto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_false; eauto.
    exists (w: tm_while_max e z1 t1); splits; auto.
      apply step_single; apply st_while_go; auto.
      left; exists z1 w; splits; auto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_go; eauto.
    exists (w: tm_skip); splits; auto.
      apply step_single; apply st_while_limit; auto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_single_cstep_star.
      apply st_while_limit; eauto.

    splits.
    intros [?]; discriminate.
    intros [?]; discriminate.

    intros; split.
    right; left; exists z bba bbb; splits; auto.
    eapply bisim_s_iter; eauto.
    intro C; intro; intros; eapply H5.
    eapply cstep_iteration; eauto.
    edestruct H6 as [C0[C ?]].
    exists C0 C.
    eapply cstep_star_iter_snoc; eauto.

    intros; split.
    right; left; exists z bba bbb; splits; auto.
    eapply bisim_context; eauto.
    intro C; intro; intros; eapply H5 with (C:=cx_par1 C b).
    eapply cstep_cons.
    apply cst_par1_step2; eauto.
    apply cstep_star_par_cx_par1; apply cstep_star_par1_step; eauto.
    edestruct H6 as [C0[C0' ?]].
    exists (cx_par1 C0 b) (cx_par1 C0' b').
      eapply cstep_star_snoc.
      eapply cstep_star_par_cx_par1; eauto.
      apply cstep_star_par1_step; eauto.
      eapply cst_par1_step2; eauto.

    intros s' ba' ?.
    destruct H6 as [C0[C0' H7]].
    inverts H.
    edestruct step_bisim1 with (ba:=bba) as [bbb'[??]]; eauto.
    exists (bbb';; tm_while_max e z t2); splits; auto.
      apply step_star_seq_app; auto.
      right; left; exists z b1' bbb'; splits; auto.
      intros C1; intro; intros; eapply H5.
      eapply step_cstep_star; eauto.
      exists C0 C0'.
      eapply cstep_star_snoc; eauto.
      apply cstep_simulates_step; apply step_add_context; apply st_seq1; auto.
    exists (w: tm_while_max e z t2); splits; auto.
      eapply step_star_seq_join; eauto.
      eapply bisim_halted1; eauto. eexists; eauto.
      apply step_nil.
      left; exists z w; splits; auto.
      eapply H5; eauto. apply cstep_nil with (C:=cx_hole).
      exists C0 C0'.
      eapply cstep_star_app; eauto. apply step_single_cstep_star; apply st_seq_skip; auto.

    intros s' bb' ?.
    destruct H6 as [C0[C0' H7]].
    inverts H.
    edestruct step_bisim2 with (bb:=bbb) as [bba'[??]]; eauto.
    exists (bba';; tm_while_max e z t1); splits; auto.
      apply step_star_seq_app; auto.
      right; left; exists z bba' b1'; splits; auto.
      intros C1; intro; intros; eapply H5.
      eapply cstep_star_app; eauto.
      apply step_star_add_context_cstep_star; auto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      apply step_star_add_context_cstep_star; apply step_star_seq_app; auto.
    exists (w: tm_while_max e z t1); splits; auto.
      eapply step_star_seq_join; eauto.
      eapply bisim_halted2; eauto. eexists; eauto.
      apply step_nil.
      left; exists z w; splits; auto.
      eapply H5; eauto.
      apply step_star_add_context_cstep_star with (C:=cx_hole).
      eapply bisim_halted2; eauto. eexists; eauto.
      exists C0 C0'.
      eapply cstep_star_app; eauto.
      eapply cstep_star_seq_join; eauto.
      apply step_star_add_context_cstep_star.
      eapply bisim_halted2; eauto. eexists; eauto.
      apply cstep_nil.

    destruct H3 as [C0[C0' ?]].
    splits; auto.
    intros.
    split; auto.
    apply step_nil.
    intros.
    split; eauto.
    intro; split.
    apply step_nil.
    intros; splits; eauto.
    intros; splits; eauto.
    exists C0 C0'; eapply cstep_star_iter_snoc; eauto.
    intros; splits; eauto.
    exists (cx_par1 C0 b) (cx_par1 C0' b').
      eapply cstep_star_snoc.
      eapply cstep_star_par_cx_par1; eauto.
      apply cstep_star_par1_step; eauto.
      eapply cst_par1_step2; eauto.
    intros; exists b1'; splits; eauto.
    apply step_single; auto.
    exists C0 C0'.
    eapply cstep_star_snoc; eauto. apply cstep_simulates_step; apply step_add_context; auto.
    intros; exists b2'; splits; eauto.
    apply step_single; auto.
    exists C0 C0'.
    eapply cstep_star_snoc; eauto. apply cstep_simulates_step; apply step_add_context; auto.
  Qed.


  Lemma bisim_step: forall s ba bb,
    (forall s : state, halted ba -> step_star s bb s ba) ->
    (forall s : state, halted bb -> step_star s ba s bb) ->
    (forall b s'0 b'0 s' ba', step_star s b s'0 b'0 -> step s'0 ba s' ba' -> ba'<>ba -> exists bb', step_star s'0 bb s' bb' /\ bisimilar s' ba' bb') ->
    (forall b s'0 b'0 s' bb', step_star s b s'0 b'0 -> step s'0 bb s' bb' -> bb'<>bb -> exists ba', step_star s'0 ba s' ba' /\ bisimilar s' ba' bb') ->
    bisimilar s ba bb.
  Proof.
    intros.
    rename s into s0.
    exists (fun s ba bb => (exists b0 b, step_star s0 b0 s b) /\
      (((forall s : state, halted ba -> step_star s bb s ba) /\ (forall s : state, halted bb -> step_star s ba s bb) /\
      (forall b s'0 b'0 s' ba', step_star s b s'0 b'0 -> step s'0 ba s' ba' -> ba'<>ba -> exists bb', step_star s'0 bb s' bb' /\ bisimilar s' ba' bb') /\
      (forall b s'0 b'0 s' bb', step_star s b s'0 b'0 -> step s'0 bb s' bb' -> bb'<>bb -> exists ba', step_star s'0 ba s' ba' /\ bisimilar s' bb' ba')
      ) \/ bisimilar s ba bb)).
    split.
    eapply bisimulation_symm_equiv.
    clear ba bb H H0 H1 H2.
    intros s ba bb [[b0[b?]][[?[?[??]]]|?]]; subst.
    splits; eauto.

    intro; splits.
    apply H0; auto.
    eauto.
    right; apply bisim_refl.

    intros; split.
    exists b0 b.
      eapply step_star_iter_snoc; eauto.
    left; splits; intros; auto.
    edestruct H2 as [bb'[??]]; eauto.
    eapply st_iteration; eauto.
    edestruct H3 as [ba'[??]]; eauto.
    eapply st_iteration; eauto.

    intros; split.
    exists (b0||b1) (b||b').
      eapply step_star_snoc.
      apply step_star_par1_step; eauto.
      apply st_par2; auto.
    left; splits; intros; auto.
    edestruct H2 as [bb'[??]]; eauto.
    eapply step_cons.
    apply st_par2; eauto.
    apply step_star_par1_step; eauto.
    edestruct H3 as [ba'[??]]; eauto.
    eapply step_cons.
    apply st_par2; eauto.
    apply step_star_par1_step; eauto.

    intros s' ba' ?.
    destruct (wterm_eq_dec ba' ba).
    subst ba'.
    assert (s'=s); subst; eauto. eapply step_spin_no_change; eauto.
    exists bb; splits; auto. apply step_nil.
      exists b0 b; auto.
    edestruct H2 as [bb'[??]]; clear H2; eauto.
      apply step_nil with (b:=ba).
    exists bb'; splits; auto.
      exists (b0||bb) (b||bb'); auto.
      eapply step_star_app.
      apply step_star_par1_step; eauto.
      apply step_star_par2_step; auto.

    splits; intros; auto.
    splits; intros; eauto.
    apply bisim_halted1; auto.
    right; apply bisim_refl.
    splits; intros; eauto.
    exists b0 b. eapply step_star_iter_snoc; eauto.
    right; eapply bisim_s_iter; eauto.
    splits; intros; eauto.
    exists (b0||b1) (b||b'); auto.
      eapply step_star_snoc.
      apply step_star_par1_step; eauto.
      apply st_par2; auto.
    right; eapply bisim_context; eauto.
    edestruct step_bisim1 as [bb'[??]]; eauto.
    exists bb'; splits; auto.
    exists (b0||ba) (b||b1'); auto.
      eapply step_star_snoc.
      apply step_star_par1_step; eauto.
      apply st_par2; auto.
    eauto.
    right; apply bisim_symm; auto.
    splits; intros; eauto.
    exists (gwemp:tm_skip) (gwemp:tm_skip); apply step_nil.
    left; splits; auto.
    intros.
    edestruct H2 as [ba'[??]]; eauto.
    exists ba'; split; auto.
    apply bisim_symm; auto.
  Qed.

  Lemma bisim_tm_par: forall s w P1 ta1 tb1 P2 ta2 tb2,
    (forall b s' b' w1, step_star s b s' b' -> w1|=gsep_con P1 TT -> bisimilar s' (w1: ta1) (w1:tb1)) ->
    (forall b s' b' w2, step_star s b s' b' -> w2|=gsep_con P2 TT -> bisimilar s' (w2: ta2) (w2:tb2)) ->
    bisimilar s (w: tm_par P1 ta1 P2 ta2) (w: tm_par P1 tb1 P2 tb2).
  Proof.
    intros.
    apply bisim_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H2.
    exists (w1: tb1 || w2: tb2); split.
      apply step_single.
      apply st_par; auto.
      apply bisim_par.
      eapply H; eauto.
      eapply H0; eauto.
    intros.
    inverts H2.
    exists (w1: ta1 || w2: ta2); split.
      apply step_single.
      apply st_par; auto.
      apply bisim_par.
      eapply H; eauto.
      eapply H0; eauto.
  Qed.

  Lemma bisim_s_iter_star: forall s s' ba bb,
    s_iter_star s s' ->
    bisimilar s ba bb ->
    bisimilar s' ba bb.
  Proof.
    intros.
    induction H; auto.
    apply IHs_iter_star.
    eapply bisim_s_iter; eauto.
  Qed.

  Lemma bisim_step_star_context: forall s b s' b' ba bb,
    step_star s b s' b' ->
    bisimilar s ba bb ->
    bisimilar s' ba bb.
  Proof.
    intros.
    induction H; auto.
    apply IHstep_star; eapply bisim_context; eauto.
    apply IHstep_star; eapply bisim_s_iter; eauto.
  Qed.

  Lemma bisim_assert_intro: forall w P s t1 t2,
    (w|= gsep_con P TT -> bisimilar s (w: t1) (w: t2)) ->
    bisimilar s (w: tm_assert P;; t1) (w: tm_assert P;; t2).
  Proof.
    intros.
    apply bisim_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H1. inverts H7.
    exists (w: t2)%wterm; split.
      eapply step_cons.
      apply st_seq1; apply st_assert; eauto.
      apply step_single.
      apply st_seq_skip.
      setoid_rewrite bisim_Cs_seq_skip1'.
      eapply bisim_step_star_context; eauto.
    intros.
    inverts H1. inverts H7.
    exists (w: t1)%wterm; split.
      eapply step_cons.
      apply st_seq1; apply st_assert; eauto.
      apply step_single.
      apply st_seq_skip.
      setoid_rewrite bisim_Cs_seq_skip1'.
      eapply bisim_step_star_context; eauto.
  Qed.

  Lemma bisim_seq1_rewrite: forall s b1 b1' b2 b2' t3 t4,
    bisimilar s b1 b1' ->
    bisimilar s b2 b2' ->
    bisimilar s (b1';; t3) (b2';; t4) ->
    bisimilar s (b1;; t3) (b2;; t4).
  Proof.
    intros.
    eapply bisim_trans.
    apply bisim_seq1; eauto.
    eapply bisim_trans; eauto.
    apply bisim_symm.
    apply bisim_seq1; auto.
  Qed.

  Lemma bisim_seq1_rewrite': forall s b1 b2 t3 t4,
    bisimilar s b1 b2 ->
    bisimilar s (b2;; t3) (b2;; t4) ->
    bisimilar s (b1;; t3) (b1;; t4).
  Proof.
    intros.
    eapply bisim_seq1_rewrite; eauto.
  Qed.


  Lemma halted_assert_bisim: forall s C ba bb P,
    halted_assert s C ba P ->
    bisimilar s ba bb ->
    halted_assert s C bb P.
  Proof.
    intros.
    intro; intros.
    eapply H.
    eapply bisim_halted_converge1; eauto.
    apply bisim_symm; auto.
    eexists; eauto.
  Qed.

*)

End Congruence.


















