Require Import Machine.
Require Import ContextSemantics.
Require Import Bisimulation.
Require Import HoareLogic.
Require Import WorldLogic.
(*Require Import Termination.*)
Require Import world_model.
Require Import msl.msl_standard.
Require Import LibTactics.
Require Import SfLib.
Require Import StepStarCs.
Require Import StepStar.
Require Import IterationBounds.

Section Definitions.
  Import world_model.Notation.
  Import world_model.Notation.U.
  Import Machine.Notation.
  Import ContextSemantics.Notation.

  Open Scope context_scope.
  Open Scope wterm_scope.

End Definitions.


Section Lemmas.
  Import world_model.Notation.
  Import world_model.Notation.U.
  Import Machine.Notation.
  Import ContextSemantics.Notation.
  Import world_model.Morphisms.
  Import WorldLogic.Notation WorldLogic.Notation.U WorldLogic.Tactics.

  Open Scope join_scope.
  Open Scope context_scope.
  Open Scope wterm_scope.
  Open Scope forces_pred.

(*
  Definition term_impl s b1 b2 := forall C s' C' w1' b2',
    cstep_star s C (b1 || b2) s' C' (w1':tm_skip || b2') ->
    terminates s' C' b2' w1'.
*)

  Definition term_ensures_pred s b P := forall C,
    halted_assert s C b P.

(*
  Definition safe s b:= forall C s' C' b',
    cstep_star s C b s' C' b' ->
    (exists s'' b'', step s' b' s'' b'') \/ halted b'.
*)

  Definition must_terminate s b := forall C s' C' b',
    cstep_star s C b s' C' b' ->
    exists s'' w'', step_star s' b' s'' (w'':tm_skip).

  Definition extended_world b w2 b':= exists w1 w,
    join_wterms b w1 /\ gjoin w1 w2 w /\ b' = extend_world b w2.

  Definition respects_global_state (b0: wterm):= forall s b s' b',
    step_star_Cs b0 b -> step s b s' b' -> s' = s.

(*
  Definition frameable_R wF ba bb := exists b, ba=extend_world b wF /\ bb=b||wF:tm_skip.

  Definition frameable_bisim_Cs (b:wterm) := forall w wF,
    join_wterms b w ->
    gjoins w wF ->
    exists R, bisimulation_Cs R /\ R (extend_world b wF)  (b || wF:tm_skip) /\ (forall ba bb, R ba bb -> frameable_R wF ba bb).
*)

  Definition frameable_bisim_Cs (b:wterm) := forall w wF,
    join_wterms b w ->
    gjoins w wF ->
    bisimilar_Cs (extend_world b wF)  (b || wF:tm_skip).

(*
  Definition semantic_equiv s (ba bb: wterm) := forall C s' C' w',
    cstep_star s C ba s' C' (w':tm_skip) <-> cstep_star s C bb s' C' (w':tm_skip).
*)

  Definition par_will_join b1 b2:= forall w1' w2',
    step_star_Cs (b1 || b2) (w1':tm_skip || w2':tm_skip) ->
    exists w', gjoin w1' w2' w'.

  Definition empty_state:= makeState empty_ch_state empty_iters_state.


  Lemma bisim_Cs_frameable_extend_world: forall b1 w1 w2,
    frameable_bisim_Cs b1 ->
    join_wterms b1 w1 -> gjoins w1 w2 ->
    bisimilar_Cs (extend_world b1 w2) (b1||w2:tm_skip).
  Proof.
    intros.
    spec H w1 w2 H0 H1.
    auto.
  Qed.

(*
  Lemma term_impl_preservation: forall s C b1 b2 s' C' b1' b2',
    term_impl s b1 b2 ->
    cstep_star s C (b1 || b2) s' C' (b1' || b2') ->
    term_impl s' b1' b2'.
  Proof.
    unfold term_impl; intros.
    repeat intro.
    eapply H with (C:=compose_contexts C C0) (C':= compose_contexts C' C'0) (s':=s'0).
    eapply cstep_star_app.
    eapply cstep_star_composed_context_step1'; eauto.
    eapply cstep_star_composed_context_step2'; eauto.
    eapply cstep_star_composed_context_step2'; eauto.
  Qed.

  Lemma term_impl_s_iter: forall s b1 b2 s',
    term_impl s b1 b2 ->
    s_iter s s' ->
    term_impl s' b1 b2.
  Proof.
    intros.
    apply term_impl_preservation with s cx_hole b1 b2 cx_hole; auto.
    eapply cstep_iteration; eauto.
    apply cstep_nil.
  Qed.

  Lemma term_impl_step_context: forall s b1 b2 bC s' bC',
    term_impl s b1 b2 ->
    step s bC s' bC' ->
    term_impl s' b1 b2.
  Proof.
    intros.
    apply term_impl_preservation with s (cx_par1 cx_hole bC) b1 b2 (cx_par1 cx_hole bC'); auto.
    apply cstep_star_par_cx_par1.
    apply cstep_star_par2_step.
    eapply step_single_cstep_star; eauto.
  Qed.

  Lemma term_impl_step1: forall s b1 b2 s' b1',
    term_impl s b1 b2 ->
    step s b1 s' b1' ->
    term_impl s' b1' b2.
  Proof.
    intros.
    apply term_impl_preservation with s cx_hole b1 b2 cx_hole; auto.
    eapply step_single_cstep_star; apply st_par1; auto.
  Qed.

  Lemma term_impl_step2: forall s b1 b2 s' b2',
    term_impl s b1 b2 ->
    step s b2 s' b2' ->
    term_impl s' b1 b2'.
  Proof.
    intros.
    apply term_impl_preservation with s cx_hole b1 b2 cx_hole; auto.
    eapply step_single_cstep_star; apply st_par2; auto.
  Qed.
*)
  Lemma term_ensures_pred_preservation: forall s C b P s' C' b',
    term_ensures_pred s b P ->
    cstep_star s C b s' C' b' ->
    term_ensures_pred s' b' P.
  Proof.
    intros.
    repeat intro.
    eapply H with (C:=compose_contexts C C0) (C':= compose_contexts C' C'0) (s':=s'0).
    eapply cstep_star_app.
    eapply cstep_star_composed_context_step1'; eauto.
    eapply cstep_star_composed_context_step2'; eauto.
  Qed.

  Lemma term_ensures_pred_s_iter: forall s b P s',
    term_ensures_pred s b P ->
    s_iter s s' ->
    term_ensures_pred s' b P.
  Proof.
    intros.
    apply term_ensures_pred_preservation with s cx_hole b cx_hole; auto.
    eapply cstep_iteration; eauto.
    apply cstep_nil.
  Qed.

  Lemma term_ensures_pred_step_context: forall s b P bC s' bC',
    term_ensures_pred s b P ->
    step s bC s' bC' ->
    term_ensures_pred s' b P.
  Proof.
    intros.
    apply term_ensures_pred_preservation with s (cx_par1 cx_hole bC) b (cx_par1 cx_hole bC'); auto.
    apply cstep_star_par_cx_par1.
    apply cstep_star_par2_step.
    eapply step_single_cstep_star; eauto.
  Qed.

  Lemma term_ensures_pred_step: forall s b P s' b',
    term_ensures_pred s b P ->
    step s b s' b' ->
    term_ensures_pred s' b' P.
  Proof.
    intros.
    apply term_ensures_pred_preservation with s cx_hole b cx_hole; auto.
    eapply step_single_cstep_star; eauto.
  Qed.

(*
  Lemma wno_channels_step_preservation: forall s b s' b',
    wno_channels b -> step s b s' b' -> wno_channels b'.
Admitted.

  Lemma wno_heap_step_preservation: forall s b s' b',
    wno_heap b -> step s b s' b' -> wno_heap b'.
Admitted.
*)

  Lemma bisim_Cs_det_step b : forall s s' b',
    deterministic_Cs_wterm b ->
    step s b s' b' -> bisimilar_Cs b b'.
  Proof.
    intros.
    eapply H.
    eapply step_star_Cs_cons.
    apply step_single_cstep_star with (C:=cx_hole); eauto.
    apply step_star_Cs_nil.
  Qed.

  Lemma bisim_Cs_det_step_star b : forall s s' b',
    deterministic_Cs_wterm b ->
    step_star s b s' b' -> bisimilar_Cs b b'.
  Proof.
    intros.
    eapply H.
    eapply step_star_Cs_cons.
    apply step_star_add_context_cstep_star with (C:=cx_hole); eauto.
    apply step_star_Cs_nil.
  Qed.

  Lemma bisim_Cs_det_cstep_star b : forall s C s' C' b',
    deterministic_Cs_wterm b ->
    cstep_star s C b s' C' b' -> bisimilar_Cs b b'.
  Proof.
    intros.
    eapply H.
    eapply step_star_Cs_cons; eauto.
    apply step_star_Cs_nil.
  Qed.

  Lemma deterministic_wterm_preservation_step: forall s b s' b',
    deterministic_Cs_wterm b ->
    step s b s' b' ->
    deterministic_Cs_wterm b'.
  Proof.
    intros.
    repeat intro.
    eapply bisim_Cs_trans with b.
    symmetry.
    eapply bisim_Cs_det_step; eauto.
    eapply H.
    eapply step_star_Cs_cons; eauto.
    apply step_single_cstep_star with (C:=cx_hole); eauto.
  Qed.

  Lemma deterministic_wterm_preservation_cstep_star: forall s C b s' C' b',
    deterministic_Cs_wterm b ->
    cstep_star s C b s' C' b' ->
    deterministic_Cs_wterm b'.
  Proof.
    intros.
    repeat intro.
    eapply bisim_Cs_trans with b.
    symmetry.
    eapply bisim_Cs_det_cstep_star; eauto.
    eapply H.
    eapply step_star_Cs_cons; eauto.
  Qed.

  Lemma par_will_join_preservation_step1: forall s b1 b2 s' b1',
    par_will_join b1 b2 ->
    step s b1 s' b1' ->
    par_will_join b1' b2.
  Proof.
    intros.
    repeat intro.
    eapply H.
    eapply step_star_Cs_cons; eauto.
    apply step_single_cstep_star with (C:=cx_hole).
    apply st_par1; eauto.
  Qed.

  Lemma par_will_join_preservation_step2: forall s b1 b2 s' b2',
    par_will_join b1 b2 ->
    step s b2 s' b2' ->
    par_will_join b1 b2'.
  Proof.
    intros.
    repeat intro.
    eapply H.
    eapply step_star_Cs_cons; eauto.
    apply step_single_cstep_star with (C:=cx_hole).
    apply st_par2; eauto.
  Qed.

  Lemma deterministic_wterm_par: forall b1 b2,
    deterministic_Cs_wterm b1 ->
    deterministic_Cs_wterm b2 ->
    par_will_join b1 b2 ->
    deterministic_Cs_wterm (b1 || b2).
  Proof.
    intros.
    intros b' ?.
    remember (b1||b2) as b.
    revert b1 b2 Heqb H H0 H1.
    induction H2; intros; subst.
    reflexivity.

    eapply bisim_Cs_trans.
    clear IHstep_star_Cs H2.
    remember (b1||b2) as b.
    revert b1 b2 Heqb H0 H1 H3.
    induction H; intros; subst.
    reflexivity.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    clear H.
    inverts H4.
    setoid_rewrite bisim_Cs_det_step with (b:=b1); eauto.
    eapply IHcstep_star; eauto.
    eapply deterministic_wterm_preservation_step with (b:=b1); eauto.
    eapply par_will_join_preservation_step1; eauto.
    setoid_rewrite bisim_Cs_det_step with (b:=b2); eauto.
    eapply IHcstep_star; eauto.
    eapply deterministic_wterm_preservation_step with (b:=b2); eauto.
    eapply par_will_join_preservation_step2; eauto.
    eapply transitivity.
    apply bisim_Cs_par_skip; eauto.
    apply cstep_star_skip_resolve in H0. subst b''.
    reflexivity.
    subst b'.
    eapply IHcstep_star; eauto.
    eapply IHcstep_star; eauto.
    edestruct cstep_star_par_resolve as [[b1'[b2' ?]]|[w1'[w2'[w'[?[??]]]]]]; eauto.
    subst.
    eapply IHstep_star_Cs; eauto.
    apply cstep_star_par_cx_par1 in H.
    eapply deterministic_wterm_preservation_cstep_star with (b:=b1); eauto.
    apply cstep_star_par_cx_par2 in H.
    eapply deterministic_wterm_preservation_cstep_star with (b:=b2); eauto.
    repeat intro.
    eapply H3; eauto.
    eapply step_star_Cs_cons; eauto.
    subst ba'.
    clear IHstep_star_Cs.
    apply step_star_Cs_skip_resolve in H2.
    subst ba''.
    reflexivity.
  Qed.

(*
  Inductive deterministic_wterm : wterm->Prop :=
  | det_halted: forall b, halted b -> deterministic_wterm b
  | det_cons: forall b b',
      (forall s, step s b s b') ->
      (forall s' s'' b'', step s' b s'' b'' -> b''=b' /\ s''=s') ->
      deterministic_wterm b' ->
      deterministic_wterm b.
*)

  Lemma step_halted_false: forall s b s' b',
    halted b ->
    ~step s b s' b'.
  Proof.
    introv [w ?]; subst.
    intro. inverts H.
  Qed.


(*
  Lemma safe_preservation_context: forall s b bC s' bC',
    safe s b ->
    step s bC s' bC' ->
    safe s' b.
  Proof.
    intros.
    repeat intro; eapply H.
    eapply cstep_cons with (C:=cx_par1 C bC).
    apply cst_par1_step2; eauto.
    apply cstep_star_par_cx_par1.
    apply cstep_star_par1_step; eauto.
  Qed.

  Lemma safe_preservation_step: forall s b s' b',
    safe s b ->
    step s b s' b' ->
    safe s' b'.
  Proof.
    intros.
    repeat intro; eapply H. eapply step_cstep_star; eauto.
  Qed.

  Lemma safe_preservation_s_iter: forall s b s',
    safe s b ->
    s_iter s s' ->
    safe s' b.
  Proof.
    intros.
    repeat intro; eapply H.
    eapply cstep_iteration; eauto.
  Qed.



  Lemma must_term_det_halt: forall s b,
    safe s b ->
    must_terminate s b ->
    exists s' w,
      step_star s b s' (w: tm_skip).
  Proof.
    intros.
    edestruct H0 as [s'[w' ?]].
    apply cstep_nil with (C:=cx_hole).
    exists s' w'; auto.
  Qed.

  Lemma bisim_must_term_det_halt: forall s b,
    safe s b ->
    must_terminate s b ->
    deterministic_wterm b ->
    exists w,
      bisimilar_Cs b (w: tm_skip).
  Proof.
    intros.
    edestruct must_term_det_halt as [s'[w ?]]; eauto.
    exists w.
    eapply bisim_Cs_det_step_star; eauto.
  Qed.

  Lemma bisim_semantic_equivalence: forall s ba bb,
    safe s ba -> safe s bb ->
    semantic_equiv s ba bb ->
    must_terminate s ba ->
    must_terminate s bb ->
    deterministic_wterm ba ->
    deterministic_wterm bb ->
    bisimilar s ba bb.
  Proof.
    introv Hsafe_ba Hsafe_bb.
    introv Hequiv Hterm_ba Hterm_bb Hdet_ba Hdet_bb.
    edestruct bisim_must_term_det_halt with (b:=ba) as [wa H4]; eauto.
    setoid_rewrite H4.
    edestruct bisim_must_term_det_halt with (b:=bb) as [wb H5]; eauto.
    assert (H6: step_star s ba s (wa:tm_skip)).
      eapply bisim_Cs_halted2; eauto.
      eexists; eauto.
    apply step_star_add_context_cstep_star with (C:=cx_hole) in H6.
    apply Hequiv in H6.
    edestruct cstep_star_bisim_Cs1 as [bb'[H7 H8]]; eauto.
    apply cstep_star_skip_resolve in H7.
    subst bb'.
    setoid_rewrite H8.
    apply bisim_symm.
    apply bisim_Cs_bisim; auto.
  Qed.
 *)

(*
  Lemma frameable_bisim_Cs_preservation: forall s C b s' C' b',
    frameable_bisim_Cs b ->
    cstep_star s C b s' C' b' ->
    frameable_bisim_Cs b'.
  Proof.
    unfold frameable_bisim_Cs.
    intros.
    rename w into w2.
    spec H w2.
    destruct H as [b_[??]].
    destruct H as [w1[w[?[??]]]].
    subst b_.
    edestruct cstep_star_bisim_Cs1 as [bb'[??]].
    2: eauto.
    eapply cstep_star_extend_world; eauto.
    edestruct cstep_star_par_skip_resolve as [b1'[?[?| [w1'[w'[?[??]]]] ]]]; eauto.
    subst bb'.
    clear H3.
    exists (extend_world b' w2); split.
    *)

  Lemma must_terminate_preservation_step: forall s b s' b',
    must_terminate s b ->
    step s b s' b' ->
    must_terminate s' b'.
  Proof.
    intros.
    repeat intro; eapply H.
    eapply step_cstep_star; eauto.
  Qed.

  Lemma must_terminate_preservation_s_iter: forall s b s',
    must_terminate s b ->
    s_iter s s' ->
    must_terminate s' b.
  Proof.
    intros.
    repeat intro; eapply H.
    eapply cstep_iteration; eauto.
  Qed.

  Lemma must_terminate_preservation_context: forall s bC b s' bC',
    must_terminate s b ->
    step s bC s' bC' ->
    must_terminate s' b.
  Proof.
    intros.
    repeat intro; eapply H with (C:=cx_par1 C bC) (C':=cx_par1 C' bC').
    eapply cstep_context_cstep_star; eauto.
    apply cst_par1_step2; eauto.
    apply cstep_star_par_cx_par1.
    apply cstep_star_par1_step; auto.
  Qed.

  Lemma must_terminate_preservation_context_step_star: forall s bC b s' bC',
    must_terminate s b ->
    step_star s bC s' bC' ->
    must_terminate s' b.
  Proof.
    intros.
    induction H0; auto.
    apply IHstep_star.
    eapply must_terminate_preservation_context; eauto.
    apply IHstep_star.
    eapply must_terminate_preservation_s_iter; eauto.
  Qed.

  Lemma must_terminate_par: forall s b1 b2,
    must_terminate s b1 ->
    must_terminate s b2 ->
    par_will_join b1 b2 ->
    must_terminate s (b1 || b2).
  Proof.
    intros.
    repeat intro.
    remember (b1||b2) as b.
    revert b1 b2 Heqb H H0 H1.
    induction H2; intros; subst.

    edestruct H as [s'[w1' ?]].
    apply cstep_nil with (C:=cx_hole).
    edestruct H0 with (C:=cx_par1 cx_hole b1) (C':=cx_par1 cx_hole (w1':tm_skip)) as [s''[w2' ?]].
    apply cstep_star_par_cx_par1.
    apply cstep_star_par2_step; auto.
    apply step_star_add_context_cstep_star; eauto.
    edestruct H1 as [w'' ?].
    eapply step_star_Cs_cons.
    apply step_star_add_context_cstep_star with (C:=cx_hole); eauto.
    eapply step_star_app.
    apply step_star_par1_step; eauto.
    apply step_star_par2_step; eauto.
    apply step_star_Cs_nil.
    exists s'' w''.
    eapply step_star_app.
    eapply step_star_app.
    apply step_star_par1_step; eauto.
    apply step_star_par2_step; eauto.
    apply step_single; apply st_par_join; auto.

    edestruct cstep_inversion as [[?]|[?]]; eauto; clear H.
    subst C'.
    inverts H4.
    edestruct IHcstep_star as [s'''[w''' ?]]; clear IHcstep_star; eauto.
    eapply must_terminate_preservation_step with (b:=b1); eauto.
    eapply must_terminate_preservation_context with (b:=b2); eauto.
    eapply par_will_join_preservation_step1; eauto.
    edestruct IHcstep_star as [s'''[w''' ?]]; clear IHcstep_star; eauto.
    eapply must_terminate_preservation_context with (b:=b1); eauto.
    eapply must_terminate_preservation_step with (b:=b2); eauto.
    eapply par_will_join_preservation_step2; eauto.
    clear IHcstep_star.
    eapply cstep_star_skip_resolve in H2. subst b''.
    exists s'' w. apply step_nil.
    subst b'.
    apply step_simulates_cstep in H4.
    edestruct IHcstep_star as [s'''[w''' ?]]; clear IHcstep_star; eauto.
    eapply must_terminate_preservation_context with (b:=b1); eauto.
    eapply must_terminate_preservation_context with (b:=b2); eauto.

    edestruct IHcstep_star as [s'''[w''' ?]]; clear IHcstep_star; eauto.
    eapply must_terminate_preservation_s_iter with (b:=b1); eauto.
    eapply must_terminate_preservation_s_iter with (b:=b2); eauto.
  Qed.

(*
  Lemma safe_par: forall s b1 b2,
    safe s b1 -> safe s b2 ->
    par_will_join b1 b2 ->
    safe s (b1 || b2).
  Proof.
    intros.
    repeat intro.
    remember (b1||b2) as b.
    revert b1 b2 Heqb H H0 H1.
    induction H2; intros; subst; eauto.

    edestruct H as [[s'[b1' ?]] | [w1 ?]].
    apply cstep_nil with (C:=C).
    left; exists s' (b1'||b2). apply st_par1; auto.
    edestruct H0 as [[s'[b2' ?]] | [w2 ?]].
    apply cstep_nil with (C:=C).
    left; exists s' (b1||b2'). apply st_par2; auto.
    subst.
    edestruct H1 as [w ?].
    apply step_star_Cs_nil.
    left; exists s (w:tm_skip).
    apply st_par_join; auto.

    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    clear H.
    inverts H4.
    edestruct IHcstep_star as [[s'''[b''' ?]] | [w''' ?]]; eauto.
    eapply safe_preservation_step with (b:=b1); eauto.
    eapply safe_preservation_context with (b:=b2); eauto.
    eapply par_will_join_preservation_step1; eauto.
    subst b''.
    right; eexists; eauto.
    edestruct IHcstep_star as [[s'''[b''' ?]] | [w''' ?]]; eauto.
    eapply safe_preservation_context with (b:=b1); eauto.
    eapply safe_preservation_step with (b:=b2); eauto.
    eapply par_will_join_preservation_step2; eauto.
    subst b''.
    right; eexists; eauto.
    apply cstep_star_skip_resolve in H2.
    subst b''.
    right; eexists; eauto.
    subst b'.
    apply step_simulates_cstep in H4.
    eapply IHcstep_star; eauto.
    eapply safe_preservation_context with (b:=b1); eauto.
    eapply safe_preservation_context with (b:=b2); eauto.

    eapply IHcstep_star; eauto.
    eapply safe_preservation_s_iter with (b:=b1); eauto.
    eapply safe_preservation_s_iter with (b:=b2); eauto.
  Qed.

  Lemma sem_eq_simpl_par: forall s w1 t1 w2 t2,
    semantic_equiv s (w1 : t1 || w2 : t2) ((w1 : t1 || w2 : tm_skip); t2).
Admitted.

*)


  Lemma must_terminate_preservation_cstep_star: forall s C b s' C' b',
    must_terminate s b ->
    cstep_star s C b s' C' b' ->
    must_terminate s' b'.
  Proof.
    intros.
    revert H.
    induction H0; intros; auto.
    apply IHcstep_star; clear IHcstep_star.
    edestruct cstep_inversion as [[?]|[?]]; eauto; clear H.
    subst C'.
    eapply must_terminate_preservation_step; eauto.
    subst b'.
    apply step_simulates_cstep in H2.
    eapply must_terminate_preservation_context; eauto.
    apply IHcstep_star; clear IHcstep_star.
    eapply must_terminate_preservation_s_iter; eauto.
  Qed.

  Lemma must_terminate_seq: forall s b1 t2,
    must_terminate s b1 ->
    (forall C s' C' w', cstep_star s C b1 s' C' (w':tm_skip) -> must_terminate s' (w':t2)) ->
    must_terminate s (b1;; t2).
  Proof.
    intros.
    repeat intro.
    edestruct cstep_star_seq_inversion as [ [s'0[C'0[w'[??]]]] | [b1''[??]] ]; eauto; clear H1.
    apply H0 in H2.
    eapply H2; eauto.
    subst b'.
    edestruct H as [s''[w1' ?]]; eauto.
    edestruct H0 as [s'''[w2'' ?]].
      eapply cstep_star_app; eauto.
      apply step_star_add_context_cstep_star; eauto.
      apply cstep_nil with (C:=cx_hole).
    exists s''' w2''.
      eapply step_star_seq_join; eauto.
  Qed.

  Lemma deterministic_wterm_seq: forall b1 t2,
    deterministic_Cs_wterm b1 ->
    (forall w', step_star_Cs b1 (w':tm_skip) -> deterministic_Cs_wterm (w':t2)) ->
    deterministic_Cs_wterm (b1;; t2).
  Proof.
    intros.
    intros b' ?.
    edestruct step_star_Cs_seq_inversion as [ [w'[??]] | [b1''[??]] ]; eauto; clear H1.

    spec H0 w' H2.
    apply H in H2.
    setoid_rewrite H2.
    setoid_rewrite bisim_Cs_seq_skip1'.
    apply H0; auto.

    subst b'.
    apply H in H2.
    setoid_rewrite H2.
    reflexivity.
  Qed.

(*
  Lemma safe_seq: forall s b1 t2,
    safe s b1 ->
    (forall C s' C' w', cstep_star s C b1 s' C' (w':tm_skip) -> safe s' (w':t2)) ->
    safe s (b1; t2).
  Proof.
    intros.
    intros C s' C' b' ?.
    edestruct cstep_star_seq_inversion as [ [s'0[C'0[w'[??]]]] | [b1''[??]] ]; eauto; clear H1.
    apply H0 in H2.
    eapply H2; eauto.
    subst b'.
    apply H in H2.
    left.
    edestruct H2 as [ [s''[b'' ?]] | ?]; clear H2.
    exists s'' (b''; t2); apply st_seq1; auto.
    destruct H1 as [w2 ?].
    subst b1''.
    exists s' (w2: t2).
    apply st_seq_skip.
  Qed.
*)


  Lemma must_terminate_frame: forall s b w1 w2 w,
    must_terminate s b ->
    frameable_bisim_Cs b ->
    join_wterms b w1 -> gjoin w1 w2 w ->
    must_terminate s (extend_world b w2).
Admitted.


  Lemma deterministic_wterm_frame: forall wb b wF,
    deterministic_Cs_wterm b ->
    frameable_bisim_Cs b ->
    join_wterms b wb ->
    gjoins wb wF ->
    deterministic_Cs_wterm (extend_world b wF).
  Proof.
    unfold deterministic_Cs_wterm.
    intros.
    generalize H0; intro H0'.
    spec H0 wb wF H1 H2.
    edestruct step_star_Cs_bisim_Cs1 as [b'e[??]]; eauto.
    edestruct step_star_Cs_par_skip_resolve as [b'0[?[ ? | [w'[ww[?[??]]]] ]]]; eauto.
    subst b'e.
    apply H in H6.
    setoid_rewrite bisim_Cs_frameable_extend_world; eauto.
    setoid_rewrite H5.
    apply bisim_Cs_par1.
    auto.
    (**)
    subst.
    apply H in H6.
    setoid_rewrite bisim_Cs_frameable_extend_world; eauto.
    setoid_rewrite H5.
    setoid_rewrite H6.
    apply bisim_Cs_par_skip; auto.
  Qed.

  Lemma respects_global_state_preservation_ste: forall s b s' b',
    respects_global_state b ->
    step s b s' b' ->
    respects_global_state b'.
  Proof.
    unfold respects_global_state; intros.
    eapply H; eauto.
    eapply step_star_Cs_cons; eauto. apply step_single_cstep_star with (C:=cx_hole); eauto.
  Qed.

  Lemma respects_global_state_preservation_cstep_star: forall s C b s' C' b',
    respects_global_state b ->
    cstep_star s C b s' C' b' ->
    respects_global_state b'.
  Proof.
    unfold respects_global_state; intros.
    eapply H; eauto.
    eapply step_star_Cs_cons; eauto.
  Qed.

  Lemma respects_global_state_preservation_step_star_Cs: forall b b',
    respects_global_state b ->
    step_star_Cs b b' ->
    respects_global_state b'.
  Proof.
    unfold respects_global_state; intros.
    eapply H. 2: eauto.
    eapply step_star_Cs_app; eauto.
  Qed.

  Lemma respects_global_state_step_star: forall b b'0 s s' b',
    respects_global_state b ->
    step_star_Cs b b'0 ->
    step_star s b'0 s' b' ->
    s_iter_star s s'.
  Proof.
    intros.
    revert b H H0.
    induction H1; intros; auto.
    apply s_iter_nil.
    replace s' with s in IHstep_star.
    eapply IHstep_star; eauto.
    eapply step_star_Cs_app; eauto.
    eapply step_star_Cs_cons; eauto.
    eapply step_single_cstep_star with (C:=cx_hole); eauto.
    eapply step_star_Cs_nil.
    symmetry.
    eapply H0; eauto.
    eapply s_iter_cons; eauto.
  Qed.

  Lemma respects_global_state_frame: forall wb b wF,
    respects_global_state b ->
    frameable_bisim_Cs b ->
    join_wterms b wb ->
    gjoins wb wF ->
    respects_global_state (extend_world b wF).
  Proof.
    intros.
    unfold respects_global_state.
    intros s b'0 s' b' ? ?.
    spec H0 wb wF H1 H2.
    edestruct step_star_Cs_bisim_Cs1 as [b'e[??]]; eauto.
    edestruct step_star_Cs_par_skip_resolve as [b'1[?[ ? | [w'[ww[?[??]]]] ]]]; eauto.
    subst b'e.
    edestruct step_bisim_Cs1 with (ba:=b'0) as [b''[??]]; eauto.
    edestruct step_star_par_skip_resolve as [b'2[? [?| [w1'[w'[?[??]]]] ]]]; eauto.
    subst.
    eapply s_iter_star_step_eq; eauto.
    eapply respects_global_state_step_star; eauto.
    subst.
    eapply s_iter_star_step_eq; eauto.
    eapply respects_global_state_step_star; eauto.
    subst.
    eapply skip_step_bisim_Cs_respects_state; eauto.
  Qed.

(*
  Lemma safe_skip: forall s w, safe s (w: tm_skip).
  Proof.
    intros. intros C s' C' b' ?.
    apply cstep_star_skip_resolve in H.
    subst b'.
    right; eexists; eauto.
  Qed.

  Lemma safe_preservation_context_cstep_star: forall s b C bC s' C' bC',
    safe s b -> cstep_star s C bC s' C' bC' -> safe s' b.
  Proof.
    intros.
    induction H0; auto.
    apply IHcstep_star.
    apply step_simulates_cstep in H0.
    eapply safe_preservation_context; eauto.
    apply IHcstep_star.
    eapply safe_preservation_s_iter; eauto.
  Qed.
*)

  Lemma gjoin_w_env_set1: forall w1 w2 w2' w3 w3' x v,
    gjoin w1 w2 w3 ->
    w2' = w_env_set w2 x v ->
    w3' = w_env_set w3 x v ->
    gjoin w1 w2' w3'.
Admitted.

  Lemma gjoin_w_heap_set1: forall w1 w2 w2' w3 w3' x v,
    gjoin w1 w2 w3 ->
    w2' = w_heap_set w2 x v ->
    w3' = w_heap_set w3 x v ->
    gjoin w1 w2' w3'.
Admitted.

  Lemma w_env_share_sub_top: forall x w w',
    gjoin_sub w w' ->
    w_env_share w x = Share.top ->
    w_env_share w' x = Share.top.
  Proof.
    unfold w_env_share. intros.
    destruct H as [w2 ?]. 
    destruct w; destruct w2; destruct w'. simpl in *.
    destruct H as [_[? _]]. unfold fst,snd in *.
    rename w_env into m1; rename w_env0 into m2; rename w_env1 into m3.
    eapply LocalsMap.map_join_char with (a:=x) in H.
    destruct H as [? _].
    apply join_com in H. destruct H as [_ ?].
    rewrite H0 in H.
    rewrite Share.lub_top in H.
    auto.
  Qed.

  Lemma w_heap_share_sub_top: forall x w w',
    gjoin_sub w w' ->
    w_heap_share w x = Share.top ->
    w_heap_share w' x = Share.top.
  Proof.
    unfold w_heap_share. intros.
    destruct H as [w2 ?]. 
    destruct w; destruct w2; destruct w'. simpl in *.
    destruct H as [_[_ ?]]. unfold fst,snd in *.
    rename w_heap into m1; rename w_heap0 into m2; rename w_heap1 into m3.
    eapply MemMap.map_join_char with (a:=x) in H.
    destruct H as [? _].
    apply join_com in H. destruct H as [_ ?].
    rewrite H0 in H.
    rewrite Share.lub_top in H.
    auto.
  Qed.

  Lemma w_heap_get_sub: forall a v w1 w3,
    gjoin_sub w1 w3 ->
    w_heap_get w1 a = Some v ->
    w_heap_get w3 a = Some v.
  Proof.
    unfold w_heap_get; intros.
    destruct H as [w2 ?].
    destruct w1; destruct w2; destruct w3; simpl in *.
    destruct H as [_[_ ?]]. unfold fst,snd in *.
    eapply MemMap.map_join_char with (a:=a) in H.
    intuition.
    rewrite<- H3. rewrite<- H2. auto.
    false.
    rewrite<- H. auto.
  Qed.

  Lemma gjoin_w_env_set2: forall w1 w2 w2' w3' x v,
    gjoin w1 w2' w3' ->
    w2' = w_env_set w2 x v ->
    exists w3 v0,
      gjoin w1 w2 w3 /\ w_env_get w2 x = Some v0 /\ w3 = w_env_set w3' x v0.
Admitted.

  Lemma gjoin_w_heap_set2: forall w1 w2 w2' w3' x v,
    gjoin w1 w2' w3' ->
    w2' = w_heap_set w2 x v ->
    exists w3 v0,
      gjoin w1 w2 w3 /\ w_heap_get w2 x = Some v0 /\ w3 = w_heap_set w3' x v0.
Admitted.


  Lemma extend_world_step: forall w2 s b s' b',
    step s b s' b' ->
    step s (extend_world b w2) s' (extend_world b' w2).
  Proof.
    intros.
    step_cases (induction H) Case HH;
      try (simpl; destruct gjoin_dec as [[ww]| ?]; eapply HH; eauto);
      match goal with
      | [|- evalTrue _ _] => eapply pIsTrue_gframe; eauto; apply gjoin_com; eauto
      | [|- evalFalse _ _] => eapply pIsFalse_gframe; eauto; apply gjoin_com; eauto
      | _ => auto
      end.

    - Case "st_send".
admit.
    - Case "st_recv".
admit.
    - Case "st_recv_blocked".
admit.
    - Case "st_recv_blocked".
admit.
    - Case "st_recv_blocked".
eapply can_recv_sub; eauto. exists w2; auto.

    - Case "st_newch".
(* proof: we have full permission. so no resources will enter or leave this thread *)
admit.

    - Case "st_endch".
admit.

    - Case "st_assert".
    setoid_rewrite<- TT_gsplit.
    setoid_rewrite gstar_assoc.
    apply gjoin_com in g.
    exists w w2; splits; auto. reflexivity.

    - Case "st_assign".
    simpl; destruct gjoin_dec as [[ww]| ?].
    generalize g; intro H2.
    eapply gjoin_w_env_set1 with (x:=x) (v:=v) (w2':=w') in g; auto.
    match_gjoin_left.
    eapply HH; eauto.
    apply gjoin_com in H2.
    eapply eval_gframe; eauto.
    eapply w_env_share_sub_top; eauto. eexists; eauto.
    (**)
    destruct gjoin_dec as [[w3' ?]|].
    false.
    edestruct gjoin_w_env_set2 as [w3[v0[?[??]]]]; eauto.
    eapply n; eauto.
    eapply HH; eauto.

    - Case "st_load".
    simpl; destruct gjoin_dec as [[ww]| ?].
    generalize g; intro H2.
    eapply gjoin_w_env_set1 with (x:=x) (v:=v) (w2':=w') in g; auto.
    match_gjoin_left.
    eapply HH; eauto.
    apply gjoin_com in H2.
    eapply eval_gframe; eauto.
    eapply w_heap_get_sub; eauto. eexists; eauto.
    (**)
    destruct gjoin_dec as [[w3' ?]|].
    false.
    edestruct gjoin_w_env_set2 as [w3[v0[?[??]]]]; eauto.
    eapply n; eauto.
    eapply HH; eauto.

    - Case "st_store".
    simpl; destruct gjoin_dec as [[ww]| ?].
    generalize g; intro H3.
    eapply gjoin_w_heap_set1 with (x:=dest_addr) (v:=v) (w2':=w') in g; auto.
    match_gjoin_left.
    apply gjoin_com in H3.
    eapply HH; eauto.
    eapply eval_gframe; eauto.
    eapply eval_gframe; eauto.
    eapply w_heap_share_sub_top; eauto. eexists; eauto.
    (**)
    destruct gjoin_dec as [[w3' ?]|].
    false.
    edestruct gjoin_w_heap_set2 as [w3[v0[?[??]]]]; eauto.
    eapply n; eauto.
    eapply HH; eauto.

    - Case "st_local".
admit.

    - Case "st_freelocal".
admit.

    - Case "st_par".
    simpl.
    destruct gjoin_dec as [[ww]|?].
    apply gjoin_com in g.
    apply gjoin_com in H1.
    edestruct gjoin_assoc with (a:=w0) (b:=w1) as [w12 [??]]; eauto.
    match_gjoin_left.
    eapply HH; eauto.
    setoid_rewrite<- TT_gsplit.
    setoid_rewrite gstar_assoc.
    exists w1 w2; splits; eauto. reflexivity.
    apply gjoin_com; auto.
    (**)
admit.

    - Case "st_seq1".
    simpl. apply st_seq1; eauto.

    - Case "st_par1".
    simpl. apply st_par1; eauto.

    - Case "st_par2".
    simpl. apply st_par2; eauto.

    - Case "st_par_join".
admit.
  Qed.

(*
  Lemma safe_frame: forall s b w1 w2 w,
    safe s b ->
    frameable_bisim_Cs b ->
    join_wterms b w1 -> gjoin w1 w2 w ->
    safe s (extend_world b w2).
  Proof.
    intros.
clear H0.
    intros C s' C' b' ?.
    remember (extend_world b w2) as bb.
    revert b Heqbb H H1 H2.
    induction H0; intros; subst.
    edestruct H as [ [s''[b'' ?]] | ?].
    apply cstep_nil with (C:=C).
    left; exists s'' (extend_world b'' w2).
    apply extend_world_step; auto.
    right.
    destruct H0 as [w0 ?].
    subst b0; simpl.
    destruct (gjoin_dec w2 w0).
    destruct s0 as [w20 ?].
    eexists; eauto.
    eexists; eauto.
Admitted.

  Lemma bisim_simple_par: forall s w1 t1 w2 t2,
    safe s (w1:t1) -> safe s (w2:t2) ->
    gjoins w1 w2 ->
    par_will_join (w1:t1) (w2:t2) ->
    par_will_join (w1:t1) (w2:tm_skip) ->
    must_terminate s (w1: t1) ->
    must_terminate s (w2: t2) ->
    deterministic_wterm (w1: t1) ->
    deterministic_wterm (w2: t2) ->
    frameable_bisim_Cs (w1: t1) ->
    frameable_bisim_Cs (w2: t2) ->
    bisimilar s (w1: t1 || w2: t2) (extend_world (w1: t1) w2; t2).
  Proof.
    intros.
    setoid_rewrite bisim_Cs_frameable_extend_world; eauto.
    eapply bisim_semantic_equivalence; eauto.

    eapply safe_par; eauto.

    apply safe_seq.
    eapply safe_par; eauto.
    apply safe_skip.
    intros.
    edestruct cstep_star_par_resolve as [ [b1'[b2' ?]] | [w1'[w2'[w''[?[??]]]]] ]; eauto; clear H8.
    inverts H11.
    inverts H13.
    apply cstep_star_par_skip_resolve in H11.
    destruct H11 as [b1'[?[?|[w1''[w'''[?[??]]]]]]].
    2: discriminate.
    inverts H11.
    eapply safe_preservation_context_cstep_star; eauto.
    apply join_com in H12.
    eapply safe_frame with (w2:=w1') in H0; eauto.
    erewrite extend_world_simpl in H0; eauto.
    reflexivity.

    apply sem_eq_simpl_par.

    apply must_terminate_par; eauto.

    apply must_terminate_seq; eauto.
    apply must_terminate_par; eauto.
    repeat intro. apply cstep_star_skip_resolve in H10. subst b'; exists s' w2. apply step_nil.
    intros.
    edestruct cstep_star_par_resolve as [ [b1'[b2' ?]] | [w1'[w2'[w''[?[??]]]]] ]; eauto; clear H10.
    inverts H11.
    inverts H13.
    apply cstep_star_par_skip_resolve in H11.
    destruct H11 as [b1'[?[?|[w1''[w'''[?[??]]]]]]].
    2: discriminate.
    inverts H11.
    eapply must_terminate_preservation_context_step_star in H5; eauto.
    2: apply step_star_simulates_cstep_star; eauto.
    apply join_com in H12.
    eapply must_terminate_frame in H5; eauto.
    erewrite extend_world_simpl in H5; eauto.
    reflexivity.

    apply deterministic_wterm_par; eauto.
    eapply deterministic_wterm_seq; eauto.
    eapply deterministic_wterm_par; eauto.
    repeat intro.
    apply step_star_Cs_skip_resolve in H10. subst b'. reflexivity.
    intros.
    edestruct step_star_Cs_par_skip_resolve as [b1'[?[?| [w1'[w''[?[??]]]] ]]]; eauto; clear H.
    inverts H12.
    inverts H14.
    subst b1'.
    clear H10.
    apply join_com in H13.
    rewrite<- extend_world_simpl with (w1:=w2) (w2:=w1'); auto.
    eapply deterministic_wterm_frame; eauto.
    reflexivity.
    eexists; eauto.
    reflexivity.
  Qed.

  Lemma safe_preservation_cstep: forall s C b s' C' b',
    safe s b -> cstep s C b s' C' b' -> safe s' b'.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    eapply safe_preservation_step; eauto.
    subst b'.
    eapply safe_preservation_context; eauto.
    apply step_simulates_cstep; eauto.
  Qed.
  Lemma safe_preservation_cstep_star: forall s C b s' C' b',
    safe s b -> cstep_star s C b s' C' b' -> safe s' b'.
  Proof.
    intros.
    induction H0; auto.
    apply IHcstep_star; eapply safe_preservation_cstep; eauto.
    apply IHcstep_star; eapply safe_preservation_s_iter; eauto.
  Qed.

  Lemma safe_tm_par: forall s w P1 t1 P2 t2,
    (forall w1, w1|=gsep_con P1 TT -> safe s (w1:t1)) ->
    (forall w2, w2|=gsep_con P2 TT -> safe s (w2:t2)) ->
    (forall w1 w2, w1|=gsep_con P1 TT -> w2|=gsep_con P2 TT -> par_will_join (w1:t1) (w2:t2)) ->
    w |= gsep_con (gsep_con P1 P2) TT ->
    safe s (w: tm_par P1 t1 P2 t2).
  Proof.
    intros.
    intros C s' C' b' ?.
    edestruct cstep_star_tm_par_resolve as [ [??] | [ [w1[w2[b1'[b2'[?[?[?[??]]]]]]]] | [w1[w2[w1'[w2'[w'[?[?[?[?[??]]]]]]]]]] ]]; eauto.
    subst b'.
    setoid_rewrite<- gstar_assoc in H2.
    destruct H2 as [w1[w2T[?[??]]]].
    left; exists s' (w1:t1 || w2T:t2).
      apply st_par; auto. apply TT_gstar. auto.
    subst b'.
    eapply safe_preservation_cstep_star with (b:=w1:t1||w2:t2); eauto.
    apply safe_par; auto.
    apply cstep_nil with (C:=C).
    subst b'.
    right; eexists; eauto.
  Qed.
*)


  Lemma must_terminate_tm_par: forall s P1 t1 P2 t2 w,
    (forall w1, w1|=gsep_con P1 TT -> must_terminate s (w1: t1)) ->
    (forall w2, w2|=gsep_con P2 TT -> must_terminate s (w2: t2)) ->
    (forall w1 w2, w1|=gsep_con P1 TT -> w2|=gsep_con P2 TT -> par_will_join (w1:t1) (w2:t2)) ->
    w |= gsep_con (gsep_con P1 P2) TT ->
    must_terminate s (w: tm_par P1 t1 P2 t2).
  Proof.
    intros.
    setoid_rewrite<- gstar_assoc in H2.
    intros C s' C' b' ?.
    edestruct cstep_star_tm_par_resolve as [ [??] | [ [w1[w2[b1'[b2'[?[?[?[??]]]]]]]] | [w1[w2[w1'[w2'[w'[?[?[?[?[??]]]]]]]]]] ]]; eauto.
    subst b'.
    destruct H2 as [w1[w2T[?[??]]]].
    spec H1 w1 w2T.
    spec H1. apply TT_gstar; auto.
    spec H1; auto.
    spec H w1. spec H. apply TT_gstar; auto.
    spec H0 w2T. spec H0; auto.
    edestruct H as [s'0[w1' ?]].
       eapply cstep_star_skip_subst; eauto.
    eapply must_terminate_preservation_context_step_star in H0.
    Focus 2. apply step_star_simulates_cstep_star with (C:=C); eauto.
    eapply must_terminate_preservation_context_step_star in H0; eauto.
    edestruct H0 as [s'1[w2' ?]].
      apply cstep_nil with (C:=C).
    edestruct H1 as [w' ?].
      eapply step_star_Cs_cons.
      apply step_star_add_context_cstep_star with (C:=C); apply step_star_par1_step; eauto.
      eapply step_star_Cs_cons.
      apply step_star_add_context_cstep_star with (C:=C); apply step_star_par2_step; eauto.
      eapply step_star_Cs_nil.
    exists s'1 w'.
      eapply step_cons.
      apply st_par; eauto. apply TT_gstar; auto.
      eapply step_star_snoc.
      eapply step_star_app.
      apply step_star_par1_step; eauto.
      apply step_star_par2_step; eauto.
      apply st_par_join; auto.
    subst b'.
    edestruct must_terminate_preservation_cstep_star as [s'0[w1' ?]].
    eapply H; eauto.
    eapply cstep_star_par_cx_par1; eauto.
    eapply cstep_nil with (C:=C).
    edestruct H0 as [s'1[w2' ?]].
      eauto.
      eapply cstep_star_par_cx_par2. 
      eapply cstep_star_app.
      eauto.
      apply cstep_star_par1_step.
      apply step_star_add_context_cstep_star; eauto.
    edestruct H1 with (w1:=w1) (w2:=w2)as [w' ?]; eauto.
      eapply step_star_Cs_cons; eauto.
      eapply step_star_Cs_cons.
      eapply cstep_star_app.
      eapply cstep_star_par1_step.
      apply step_star_add_context_cstep_star with (C:=C); eauto.
      eapply cstep_star_par2_step.
      apply step_star_add_context_cstep_star; eauto.
      apply step_star_Cs_nil.
    exists s'1 w'.
      eapply step_star_app.
      eapply step_star_par1_step; eauto.
      eapply step_star_snoc.
      eapply step_star_par2_step; eauto.
      apply st_par_join; auto.
    subst b'.
    exists s' w'; apply step_nil.
  Qed.



  Lemma join_wterms_extend_world2: forall b wF w,
    join_wterms (extend_world b wF) w ->
    exists wb, join_wterms b wb /\ (gjoin wb wF w \/ extend_world b wF=b).
  Proof.
    induction b; intros; simpl in *.
    destruct gjoin_dec as [[ww]|?].
    simpl in H; subst w0.
    exists w; split; auto.
    left.
    eapply @join_com; eauto.
    simpl in H; subst w0.
    exists w; split; auto.
    (**)
    edestruct IHb as [wb[??]]; eauto.
    exists wb; split; auto.
    destruct H1; auto.
    right; f_equal. auto.
    (**)
    destruct H as [w1[w2[?[??]]]].
    clear IHb2.
    edestruct IHb1 as [wb[?[?|?]]]; eauto.
    edestruct @join_assoc with (H:=gworld_sa) (a:=wF) (b:=wb) as [wb2[??]]; eauto.
    exists wb2; split; eauto.
    left; eapply @join_com; eauto.
    rewrite H3 in H.
    assert (w1=wb); try subst w1.
      eapply join_wterms_eq; eauto.
    exists w; split; eauto.
    right; f_equal; auto.
  Qed.


  Lemma frameable_bisim_Cs_frame: forall wF1 wb t,
    frameable_bisim_Cs (wb:t) ->
    frameable_bisim_Cs (extend_world (wb:t) wF1).
  Proof.
    intros.
    intros wbF1 wF2; intros.
    destruct H1 as [w ?].
    edestruct join_wterms_extend_world2 as [wb'[?[?|?]]]; eauto.
    simpl in H2; subst wb'.
    edestruct @join_assoc with (H:=gworld_sa) (a:=wb) (b:=wF1) as [wF[??]]; eauto.
    rewrite extend_world_compose with (wF:=wF) (wb:=wb) (wbF:=w); eauto.
    setoid_rewrite bisim_Cs_frameable_extend_world; eauto;
        [ | reflexivity | eexists; eauto | reflexivity | eexists; eauto]. 
    setoid_rewrite<- bisim_Cs_par_assoc.
    setoid_rewrite bisim_Cs_par_skip; eauto.
    reflexivity.
    reflexivity.
    (**)
    rewrite H3 in *.
    simpl in H2; subst wb'.
    simpl in H0; subst wbF1.
    eapply H; eauto.
    reflexivity.
    eapply join_joins; eauto.
  Qed.

(*
  Tactic Notation "match_gjoin_left_destruct_exactly" constr(a) constr(w1) constr(w2) constr(w) :=
      match a with
      | gjoin_dec w1 w2 =>
          let w' := fresh "w" in let H':= fresh "H" in
            destruct a as [ [w' H'] | H' ];
            [ assert (w'=w); [eapply (@join_eq world gworld_sa w1 w2); eauto | subst w']
            | false; eapply H'; eauto]
      | gjoin_dec w2 w1 =>
          let w' := fresh "w" in let H':= fresh "H" in
            destruct a as [ [w' H'] | H' ];
            [ assert (w'=w); [eapply (@join_eq world gworld_sa w1 w2); eauto | subst w']
            | false; eapply H'; eapply (@join_com world gworld_sa); eauto]
      end.

  Tactic Notation "match_gjoin_left_exactly" constr(w1) constr(w2) constr(w) :=
    match goal with
    | [ |- context [ match ?a with inleft _ => _ | inright _ => _ end ] ] =>
      match_gjoin_left_destruct_exactly a w1 w2 w
    | [ HH: context [ match ?a with inleft _ => _ | inright _ => _ end ] |- _ ] => 
      match_gjoin_left_destruct_exactly a w1 w2 w
    end.

  Tactic Notation "match_gjoin_left":=
    match reverse goal with
    | H: @join world gworld_sa ?w1 ?w2 ?w |- _  => match_gjoin_left_exactly w1 w2 w
    | H: gjoin ?w1 ?w2 ?w |- _  => match_gjoin_left_exactly w1 w2 w
    end.
*)

  Lemma frameable_bisim_Cs_par: forall w P1 P2 t1 t2,
    gprecise P1 -> gprecise P2 ->
    w|= gsep_con P1 P2 ->
    (forall w1, w1|=P1 -> frameable_bisim_Cs (w1: t1)) ->
    (forall w2, w2|=P2 -> frameable_bisim_Cs (w2: t2)) ->
    frameable_bisim_Cs (w: tm_par P1 t1 P2 t2).
  Proof.
    introv HprecP1 HprecP2 Hw12.
    intros.
    unfold frameable_bisim_Cs.
    intros.
    simpl in H1. subst w0.
    destruct H2 as [ww ?].
    apply bisim_Cs_step.
    simpl; match_gjoin_left; introv [?]; discriminate.
    simpl; match_gjoin_left; introv [?]; discriminate.
    intros.
    clear H3.
    simpl in H2.
    match_gjoin_left.
    inverts H2.
    rename w1 into w1F; rename w2 into w2F.
    destruct Hw12 as [w1[w2[?[??]]]].
    edestruct gprecise_TT_sub with (P1:=P1) (P2:=P2) (w1:=w1) (w3:=w1F) (w2:=w2) (w4:=w2F); eauto.
    destruct H6 as [wF1 ?].
    destruct H7 as [wF2 ?].
    exists (w1: t1 || w2: t2 || wF:tm_skip); split.
    apply step_single.    
    apply st_par1; apply st_par; auto; apply TT_gstar; auto.
    rewrite<- (extend_world_simpl w1 wF1 w1F t1); auto.
    rewrite<- (extend_world_simpl w2 wF2 w2F t2); auto.
    setoid_rewrite bisim_Cs_frameable_extend_world; eauto;
        [ | reflexivity | eexists; eauto | reflexivity | eexists; eauto]. 
    setoid_rewrite bisim_Cs_par_assoc.
    setoid_rewrite<- bisim_Cs_par_assoc at 2.
    setoid_rewrite bisim_Cs_par_swap at 3.
    setoid_rewrite bisim_Cs_par_assoc.
    setoid_rewrite<- bisim_Cs_par_assoc at 1.
    setoid_rewrite bisim_Cs_par_skip.
    reflexivity.
    eapply gjoin_sub_framing with (P1:=P1) (P2:=P2); eauto.
    (**)
    intros.
    clear H3.
    inverts H2.
    inverts H7.
    edestruct @join_assoc with (H:=gworld_sa) (a:=w1) (b:=w2) as [w20[??]]; eauto.
    exists (w1: t1 || w20: t2); split.
    simpl; match_gjoin_left.
    apply step_single.    
    apply st_par; auto.
    setoid_rewrite<- TT_gsplit.
    setoid_rewrite gstar_assoc.
    exists w2 wF; splits; auto.
    reflexivity.
    rewrite<- (extend_world_simpl w2 wF w20 t2); auto.
    setoid_rewrite bisim_Cs_frameable_extend_world; eauto;
        [ |  | eexists; eauto | eexists; eauto]. 
    setoid_rewrite bisim_Cs_par_assoc.
    reflexivity.
    rename w1 into w1F; rename w2 into w2F.
    destruct H11 as [w2[wF2[?[??]]]].
    rewrite<- (extend_world_simpl w2 wF2 w2F t2); auto.
    apply frameable_bisim_Cs_frame; auto.
    inverts H7.
  Qed.


  Lemma triple_join_exists_gworld: @triple_join_exists world gworld_sa.
  Proof.
Admitted. (* this is true because gworld_sa is based off of share algebras; it's probably true for any reasonable sa *)

  Lemma gjoins_derives_step_frame: forall s w b w' b' wF,
    join_wterms b w ->
    join_wterms b' w' ->
    step s b s b' ->
    gjoins w' wF ->
    gjoins w wF.
  Proof.
    intros.
    remember s as s' in H1 at 2.
    revert w w' wF Heqs' H H0 H2.
    step_cases (induction H1) Case; intros; simpl in *; try subst; try solve [simpl in *; subst; eauto].

    Case "st_send".
    false.
    clear -Heqs' H1.
    generalize H1; intro H1'.
    apply s_ch_get_push_same with (w:=w') (v:=v) in H1'.
    rewrite<- Heqs' in H1.
    rewrite H1 in H1'.
    inverts H1'.
    rewrite app_nil_end in H2 at 1.
    apply app_inv_head in H2.
    inverts H2.

    Case "st_recv".
    false. clear -Heqs' H5. rename H5 into H.
    generalize H; intro H'.
    apply s_ch_get_pop_same in H'.
    rewrite<- Heqs' in H.
    rewrite H in H'.
    inverts H'.
    assert (length (v::q) = length q) by (f_equal; auto).
    simpl in H0. false; omega.

    Case "st_newch".
    false.
    destruct (fresh_channel_global_props s w0 (fresh_channel s w0)) as [_ ?]; auto.
    generalize (s_ch_get_newch_same R s w0 (FreshChannels.fc_c (fresh_channel s w0)) (fresh_channel s w0)); intro H2.
    spec H2; auto.
    spec H2; auto.
    rewrite H1 in H at 1.
    rewrite H in H2.
    discriminate.

    Case "st_endch".
    false. clear -Heqs' H2.
    generalize (s_ch_get_rm_same s c); intro.
    rewrite Heqs' in H.
    rewrite H2 in H.
    discriminate.

    Case "st_assign".
    destruct H4 as [ww' ?].
    apply gjoin_com in H1.
    edestruct gjoin_w_env_set2 as [ww[v0[?[??]]]]; eauto.
    eexists; apply gjoin_com; eauto.

    Case "st_load".
    destruct H4 as [ww' ?].
    apply gjoin_com in H1.
    edestruct gjoin_w_env_set2 as [ww[v0[?[??]]]]; eauto.
    eexists; apply gjoin_com; eauto.

    Case "st_store".
    destruct H5 as [ww' ?].
    apply gjoin_com in H2.
    edestruct gjoin_w_heap_set2 as [ww[v0[?[??]]]]; eauto.
    eexists; apply gjoin_com; eauto.

    Case "st_local".
    destruct H3 as [ww' ?].
    apply gjoin_com in H1.
admit.

    Case "st_freelocal".
admit.

    Case "st_par".
    destruct H3 as [w3[w4[?[??]]]]; subst w3 w4.
    assert (w'=w0); subst; auto.
    eapply @join_eq with (H:=gworld_sa); eauto.

    Case "st_par1".
    destruct H as [w1[w2[?[??]]]].
    destruct H0 as [w1'[w2'[?[??]]]].
    assert (w2'=w2); try subst w2'.
      eapply join_wterms_eq; eauto.
    destruct H2 as [wbF' ?].
    edestruct @join_assoc with (H:=gworld_sa) (a:=w2) (b:=w1') as [w1F'[??]]; eauto.
    spec IHstep w1 w1' wF (refl_equal s) H H0 (@join_joins _ _ _ _ _ H7).
    destruct IHstep as [w1F ?].
    clear H H3 H0 H1 H5.
    edestruct @join_assoc with (H:=gworld_sa) (a:=w1') (b:=w2) as [w2F[??]]; eauto.
    clear -H4  H9 H.
    edestruct (triple_join_exists_gworld w1 w2 wF) as [wbF ?]; eauto.
    eapply join_joins; eauto.

    Case "st_par2".
    destruct H as [w1[w2[?[??]]]].
    destruct H0 as [w1'[w2'[?[??]]]].
    assert (w1'=w1); try subst w1'.
      eapply join_wterms_eq; eauto.
    destruct H2 as [wbF' ?].
    edestruct @join_assoc with (H:=gworld_sa) (a:=w1) (b:=w2') as [w2F'[??]]; eauto.
    spec IHstep w2 w2' wF (refl_equal s) H3 H5 (@join_joins _ _ _ _ _ H7).
    destruct IHstep as [w2F ?].
    edestruct @join_assoc with (H:=gworld_sa) (a:=w2') (b:=w1) as [w1F[??]]; eauto.
    edestruct (triple_join_exists_gworld w1 w2 wF) as [wbF ?]; eauto.
    eapply join_joins; eauto.

    Case "st_par_join".
    destruct H0 as [w1'[w2'[?[??]]]]; subst w1' w2'.
    assert (w0 = w'); try subst w0; auto.
      eapply @join_eq with (H:=gworld_sa); eauto.
  Qed.

  Lemma extend_world_skip_inversion: forall w b wF,
    extend_world b wF = w: tm_skip ->
    (b = w: tm_skip /\ ~gjoins w wF) \/ (exists wb, b=wb:tm_skip /\ gjoin wb wF w).
  Proof.
    intros.
    destruct b.
    simpl in H.
    destruct gjoin_dec as [[wbF]| ?].
    eapply @join_com in g.
    inverts H. eauto.
    inverts H.
    left; split; auto.
    intro H.
    destruct H.
    eapply n; eauto.
    eapply @join_com; eauto.
    simpl in H. inverts H.
    simpl in H. inverts H.
  Qed.

(*  Definition frameable_R wF ba bb:= exists wb b w, join_wterms b wb /\ gjoin wb wF w /\ ba=extend_world b wF /\ (bb=b||wF:tm_skip \/ bb=ba).
  Lemma asfdasdf: forall wF,
    bisimulation_Cs (frameable_R wF).
  Proof.
    intros wF ba bb [wb[b[w[?[?[?[?|?]]]]]]]; subst.
    splits.
    intros.
    destruct H1 as [w' ?].
    edestruct extend_world_skip_inversion as [[??]|[wb'[??]]]; eauto.
    subst b.
    simpl in H. subst w'.
    false; eapply H3; eexists; eauto.
    subst b.
    simpl in H; subst wb'.
    assert (w'=w); try subst w'.
      eapply @join_eq with (H:=gworld_sa); eauto.
    rewrite H1.
    split.
    intros.
    apply step_single. apply st_par_join; auto.
    exists wb (wb: tm_skip) w; splits; auto. reflexivity.
    introv [?]; discriminate.
    intros.
    *)
    
(*
  Lemma frameable_bisim_Cs_step: forall s w b b',
    join_wterms b w ->
    frameable_bisim_Cs b ->
    step s b s b' ->
    frameable_bisim_Cs b'.
  Proof.
    intros.
    intros w' wF ? ?.
    spec H0 w wF H.
    spec H0.
    eapply gjoins_derives_step_frame; eauto.
    eapply extend_world_step with (w2:=wF) in H1.
    edestruct step_bisim_Cs1 as [bb'[??]]; eauto.


      apply st_par1; eauto.


      rewrite<- Heqs' in H0.
SearchAbout s_get_ch.
    unfold s_get_ch,s_push_ch
      unfold s_push_ch in Heqs'.
      destruct s.
      inverts Heqs'.
      destruct s_get_ch.
      destruct c0; destruct p.
    SearchAbout FMapPositive.PositiveMap.add.
*)



  Lemma bisim_Cs_par_frame: forall F1 F2 w w1 w2 w12 wF t1 t2,
    respects_global_state (w1: t1) ->
    respects_global_state (w2: t2) ->
    frameable_bisim_Cs (w1 : t1) ->
    frameable_bisim_Cs (w2 : t2) ->
    deterministic_Cs_wterm (w1 : t1) ->
    deterministic_Cs_wterm (w2 : t2) ->
    gjoin w1 w2 w12 ->
    gjoin w12 wF w ->
    wF|=gsep_con F1 F2 ->
    bisimilar_Cs (w : tm_par (gsep_con (exactly w1) F1) t1 (gsep_con (exactly w2) F2) t2) (w1 : t1 || w2 : t2 || wF : tm_skip).
  Proof.
    introv Hresp1 Hresp2.
    introv Hframe1 Hframe2.
    introv Hdet1 Hdet2.
    intros.
    apply bisim_Cs_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros. clear H3.
    inverts H2.
    setoid_rewrite<- gstar_assoc in H10.
    setoid_rewrite<- gstar_assoc in H11.
    rename w0 into w1F1; rename w3 into w2F2.
    destruct H10 as [w1'[wF1[?[??]]]]. simpl in H3; subst w1'.
    destruct H11 as [w2'[wF2[?[??]]]]. simpl in H5; subst w2'.
    eexists; split.
      apply step_nil.
      rewrite<- (extend_world_simpl w1 wF1 w1F1 t1); auto.
      rewrite<- (extend_world_simpl w2 wF2 w2F2 t2); auto.
      setoid_rewrite bisim_Cs_frameable_extend_world; eauto;
        [ | reflexivity | eexists; eauto | reflexivity | eexists; eauto]. 
      setoid_rewrite bisim_Cs_par_assoc.
      setoid_rewrite<- bisim_Cs_par_assoc at 2.
      setoid_rewrite bisim_Cs_par_swap at 3.
      setoid_rewrite bisim_Cs_par_assoc.
      setoid_rewrite<- bisim_Cs_par_assoc at 1.
      edestruct @join_assoc with (H:=gworld_sa) (a:=w1) (b:=wF1) as [w2F[??]]; eauto.
      edestruct @join_assoc with (H:=gworld_sa) (a:=w1) (b:=w2) as [w2F'[??]]; eauto.
      edestruct @join_assoc with (H:=gworld_sa) (a:=w2) (b:=wF2) as [wF'[??]]; eauto.
      assert (w2F' = w2F); try subst w2F'.
        eapply @join_canc with (H:=gworld_sa); eauto.
      assert (wF' = wF); try subst wF'.
        eapply @join_canc with (H:=gworld_sa); eauto.
      eapply @join_com in H10.
      setoid_rewrite bisim_Cs_par_skip; eauto.
      reflexivity.
    (**)
    intros. clear H3.
    destruct H1 as [wF1[wF2[?[??]]]].
    edestruct @join_assoc with (H:=gworld_sa) (a:=w1) (b:=w2) as [w2F[??]]; eauto.
    edestruct @join_assoc with (H:=gworld_sa) (a:=wF1) (b:=wF2) as [w2F2[??]]; eauto.
    edestruct @join_assoc with (H:=gworld_sa) (a:=w2F2) (b:=wF1) as [w1F1[??]]; eauto.
    eapply @join_com in H9.
    eapply @join_com in H7.
    eapply @join_com in H10.
    inverts H2.
    inverts H15.
    assert (s'=s); try subst s'.
      eapply Hresp1; eauto. apply step_star_Cs_nil.
    exists (w1F1: t1 || w2F2: t2); split.
      eapply step_single.
      apply st_par; eauto.
      apply TT_gstar; auto. exists w1 wF1; splits; eauto. reflexivity.
      apply TT_gstar; auto. exists w2 wF2; splits; eauto. reflexivity.
      rewrite<- (extend_world_simpl w1 wF1 w1F1 t1); auto.
      setoid_rewrite bisim_Cs_frameable_extend_world; eauto.
      setoid_rewrite (Hdet1 b1'0).
      setoid_rewrite<- bisim_Cs_par_assoc.
      apply bisim_Cs_par2.
      rewrite<- (extend_world_simpl w2 wF2 w2F2 t2); auto.
      setoid_rewrite bisim_Cs_frameable_extend_world; eauto.
      setoid_rewrite bisim_Cs_par_assoc.
      setoid_rewrite bisim_Cs_par_swap at 2.
      setoid_rewrite<- bisim_Cs_par_assoc at 1.
      setoid_rewrite bisim_Cs_par_skip; eauto.
      reflexivity.
      reflexivity.
      eexists; eauto.
      eapply step_star_Cs_cons. apply step_single_cstep_star with (C:=cx_hole); eauto. apply step_star_Cs_nil.
      reflexivity.
      eexists; eauto.
    assert (s'=s); try subst s'.
      eapply Hresp2; eauto. apply step_star_Cs_nil.
    exists (w1F1: t1 || w2F2: t2); split.
      eapply step_single.
      apply st_par; eauto.
      apply TT_gstar; auto. exists w1 wF1; splits; eauto. reflexivity.
      apply TT_gstar; auto. exists w2 wF2; splits; eauto. reflexivity.
      rewrite<- (extend_world_simpl w1 wF1 w1F1 t1); auto.
      rewrite<- (extend_world_simpl w2 wF2 w2F2 t2); auto.
      setoid_rewrite bisim_Cs_frameable_extend_world; eauto.
      setoid_rewrite (Hdet2 b2').
      setoid_rewrite<- bisim_Cs_par_assoc.
      apply bisim_Cs_par2.
      setoid_rewrite bisim_Cs_par_assoc.
      setoid_rewrite bisim_Cs_par_swap at 2.
      setoid_rewrite<- bisim_Cs_par_assoc at 1.
      setoid_rewrite bisim_Cs_par_skip; eauto.
      reflexivity.
      eapply step_star_Cs_cons. apply step_single_cstep_star with (C:=cx_hole); eauto. apply step_star_Cs_nil.
      reflexivity.
      eexists; eauto.
      reflexivity.
      eexists; eauto.
    assert (w4=w12); try subst w4.
      eapply @join_eq with (H:=gworld_sa); eauto.
    exists (w : tm_skip); split.
      eapply step_cons.
      apply st_par; eauto.
      apply TT_gstar; auto. exists w1 wF1; splits; eauto. reflexivity.
      apply TT_gstar; auto. exists w2 wF2; splits; eauto. reflexivity.
      apply step_single.
      apply st_par_join; auto.
      setoid_rewrite bisim_Cs_par_skip; eauto.
      reflexivity.
    inverts H15.
  Qed.
     

  Lemma bisim_simple_par_dist: forall s w1 t1 w2 t2 P3 t3 P4 t4,
    gprecise P3 -> gprecise P4 ->
    term_ensures_pred s (w1:t1) (gsep_con P3 TT) ->
    term_ensures_pred s (w2:t2) (gsep_con P4 TT) ->
    par_will_join (w1:t1) (w2:t2) ->
    (forall w3, w3|=P3 -> frameable_bisim_Cs (w3: t3)) ->
    (forall w4, w4|=P4 -> frameable_bisim_Cs (w4: t4)) ->
    (forall w3, w3|=P3 -> deterministic_Cs_wterm (w3: t3)) ->
    (forall w4, w4|=P4 -> deterministic_Cs_wterm (w4: t4)) ->
    (forall w3, w3|=P3 -> respects_global_state (w3: t3)) ->
    (forall w4, w4|=P4 -> respects_global_state (w4: t4)) ->
    bisimilar s ((w1: t1 || w2: t2);; tm_par P3 t3 P4 t4) ((w1: t1;; t3) || (w2: t2;; t4)).
  Proof.
    introv HprecP3 HprecP4.
    introv HtermP3 HtermP4.
    introv Hparjoins12.
    introv Hframe3 Hframe4.
    introv Hdet3 Hdet4.
    introv Hresp3 Hresp4.
    rename s into s0.
    exists (fun s ba bb =>
      (exists b1 b2 C0 C,
        cstep_star s0 C0 (w1: t1 || w2: t2) s C (b1||b2) /\
        ba = (b1 || b2);; tm_par P3 t3 P4 t4 /\ bb = (b1;; t3) || (b2;;t4))
      \/
      (exists w3 b3 b2 C0 s1 C1 b2'0 C,
        cstep_star s0 C0 (w1: t1 || w2: t2) s1 C1 (w3:tm_skip||b2'0) /\
        cstep_star s1 C1 (w3: t3||b2'0) s C (b3||b2) /\
        ba = (w3: tm_skip || b2);; tm_par P3 t3 P4 t4 /\ bb = b3 || (b2;; t4))
      \/
      (exists b1 w4 b4 C0 s1 C1 b1'0 C,
        cstep_star s0 C0 (w1: t1 || w2: t2) s1 C1 (b1'0||w4:tm_skip) /\
        cstep_star s1 C1 (b1'0 || w4: t4) s C (b1 || b4) /\
        ba = (b1 || w4:tm_skip);; tm_par P3 t3 P4 t4 /\ bb = (b1;; t3) || b4)
      \/
      bisimilar s ba bb).
    split.
    2: left; do 4 eexists; splits; eauto; apply cstep_nil with (C:=cx_hole).
    intros s ba bb [ [b1[b2[C0[C[?[??]]]]]] | [ [w3[b3[b2[C0[s1[C1[b2'0[C[?[?[??]]]]]]]]]]] | [ [b1[w4[b4[C0[s1[C1[b1'0[C[?[?[??]]]]]]]]]]] | ? ]]]; subst.

    splits.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    left; exists b1 b2 C0 C; splits; auto.
      eapply cstep_star_iter_snoc; eauto.
    intros.
    left; exists b1 b2 (cx_par1 C0 b) (cx_par1 C b'); splits; auto.
      eapply cstep_star_snoc.
      apply cstep_star_par_cx_par1.
      apply cstep_star_par1_step; eauto.
      apply cst_par1_step2; auto.
    intros s' b' ?.
    inverts H0.
    inverts H5.
    exists ((b1'0;;t3) || (b2;; t4)); split.
      apply step_single; apply st_par1; apply st_seq1; auto.
      left; exists b1'0 b2 C0 C; splits; auto.
      eapply cstep_star_snoc; eauto.
      apply cstep_simulates_step; apply step_add_context; apply st_par1; auto.
    exists ((b1;;t3) || (b2';; t4)); split.
      apply step_single; apply st_par2; apply st_seq1; auto.
      left; exists b1 b2' C0 C; splits; auto.
      eapply cstep_star_snoc; eauto.
      apply cstep_simulates_step; apply step_add_context; apply st_par2; auto.
    rename w3 into w4F2; rename w0 into w3F1.
    exists (w3F1: t3 || w4F2: t4); split.
      eapply step_cons.
      apply st_par1; apply st_seq_skip.
      eapply step_single.
      apply st_par2; apply st_seq_skip.
      right. right. right.
      setoid_rewrite bisim_Cs_seq_skip1'.
      apply bisim_Cs_bisim.
      assert (w3F1|= gsep_con P3 TT).
        eapply HtermP3. apply cstep_star_par_cx_par1; eauto.
      assert (w4F2|= gsep_con P4 TT).
        eapply HtermP4. apply cstep_star_par_cx_par2; eauto.
      destruct H0 as [w3[wF1[?[??]]]].
      destruct H1 as [w4[wF2[?[??]]]].
      rewrite<- (extend_world_simpl w3 wF1 w3F1 t3); auto.
      rewrite<- (extend_world_simpl w4 wF2 w4F2 t4); auto.
      setoid_rewrite bisim_Cs_frameable_extend_world; eauto;
        [ | reflexivity | eexists; eauto | reflexivity | eexists; eauto]. 
      setoid_rewrite bisim_Cs_par_assoc.
      setoid_rewrite<- bisim_Cs_par_assoc at 2.
      setoid_rewrite bisim_Cs_par_swap at 3.
      setoid_rewrite bisim_Cs_par_assoc.
      setoid_rewrite<- bisim_Cs_par_assoc at 1.
      edestruct @join_assoc with (H:=gworld_sa) (a:=wF1) (b:=w3) as [w34F2[??]]; eauto.
      edestruct @join_assoc with (H:=gworld_sa) (a:=wF2) (b:=w4) as [w34[??]]; eauto.
      edestruct @join_assoc with (H:=gworld_sa) (a:=w34) (b:=wF2) as [wF[??]]; eauto.
      eapply @join_com in H9.
      eapply @join_com in H11.
      setoid_rewrite bisim_Cs_tm_par_par_precise_frame with (w1:=w3) (w2:=w4); eauto.
      setoid_rewrite bisim_Cs_par_frame; eauto.
      setoid_rewrite bisim_Cs_par_skip; eauto.
      reflexivity.
      eapply TT_gstar. reflexivity.
    (**)
    intros s' b' ?.
    inverts H0.
    inverts H5.
    rename b1'0 into b1'.
    exists ((b1' || b2);; tm_par P3 t3 P4 t4); split.
      apply step_single; apply st_seq1; apply st_par1; auto.
      left; exists b1' b2 C0 C; splits; auto.
      eapply cstep_star_snoc; eauto.
      apply cstep_simulates_step; apply step_add_context; apply st_par1; auto.
    eexists; split.
      apply step_nil.
      right; left. exists w (w:t3) b2 C0 s' C; exists b2 C; splits; auto. apply cstep_nil.
    inverts H5. rename b1' into b2'.
    exists ((b1 || b2');; tm_par P3 t3 P4 t4); split.
      apply step_single; apply st_seq1; apply st_par2; auto.
      left; exists b1 b2' C0 C; splits; auto.
      eapply cstep_star_snoc; eauto.
      apply cstep_simulates_step; apply step_add_context; apply st_par2; auto.
    eexists; split.
      apply step_nil.
      right; right; left; exists b1 w (w:t4) C0 s' C; exists b1 C; splits; auto. apply cstep_nil.

    splits.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    right; left; do 8 eexists; splits; eauto.
      eapply cstep_star_iter_snoc; eauto.
    intros.
    right; left; do 5 eexists; exists (cx_par1 C1 b); do 2 eexists; splits; eauto.
      apply cstep_star_par_cx_par1.
      apply cstep_star_par1_step; eauto.
      eapply cstep_star_snoc; eauto.
      apply cstep_star_par_cx_par1.
      apply cstep_star_par1_step; eauto.
      apply cst_par1_step2; eauto.
    intros s' b' ?.
    inverts H1.
    inverts H6.
    false; inverts H5.
    exists (b3 || (b2';; t4)); split.
      apply step_single; apply st_par2; apply st_seq1; auto.
      right; left; do 8 eexists; splits; eauto.
      eapply cstep_star_snoc; eauto.
      apply cstep_simulates_step; apply step_add_context; apply st_par2; auto.
    rename w3 into w3F1; rename w4 into w4F2.
    assert (w3F1|= gsep_con P3 TT).
      eapply HtermP3. apply cstep_star_par_cx_par1; eauto.
    assert (w4F2|= gsep_con P4 TT).
      eapply HtermP4.
      apply cstep_star_par_cx_par2 with (b1:=w1: t1;; t3).
      eapply cstep_star_app; eauto.
      eapply cstep_star_snoc.
      apply cstep_star_par1_inner_seq_append; eauto.
      apply cstep_simulates_step; apply step_add_context; apply st_par1; apply st_seq_skip.
    destruct H1 as [w3[wF1[?[??]]]].
    destruct H2 as [w4[wF2[?[??]]]].
    edestruct @join_assoc with (H:=gworld_sa) (a:=wF1) (b:=w3) as [w34F2[??]]; eauto.
    edestruct @join_assoc with (H:=gworld_sa) (a:=wF2) (b:=w4) as [w34[??]]; eauto.
    edestruct @join_assoc with (H:=gworld_sa) (a:=w34) (b:=wF2) as [wF[??]]; eauto.
    exists (b3 || w4F2:t4); split.
      apply step_single; apply st_par2; apply st_seq_skip.
      right; right; right.
      setoid_rewrite bisim_Cs_seq_skip1'.
      apply bisim_trans with (w3F1:t3 || w4F2:t4).
      apply bisim_Cs_bisim.
      rewrite<- (extend_world_simpl w3 wF1 w3F1 t3); auto.
      rewrite<- (extend_world_simpl w4 wF2 w4F2 t4); auto.
      setoid_rewrite bisim_Cs_frameable_extend_world; eauto;
        [ | reflexivity | eexists; eauto | reflexivity | eexists; eauto]. 
      setoid_rewrite bisim_Cs_par_assoc.
      setoid_rewrite<- bisim_Cs_par_assoc at 2.
      setoid_rewrite bisim_Cs_par_swap at 3.
      setoid_rewrite bisim_Cs_par_assoc.
      setoid_rewrite<- bisim_Cs_par_assoc at 1.
      eapply @join_com in H10.
      eapply @join_com in H12.
      setoid_rewrite bisim_Cs_tm_par_par_precise_frame with (w1:=w3) (w2:=w4); eauto.
      setoid_rewrite bisim_Cs_par_frame; eauto.
      setoid_rewrite bisim_Cs_par_skip; eauto.
      reflexivity.
      eapply TT_gstar. reflexivity.
      apply bisim_par1.
      apply bisim_Cs_bisim.
      rewrite<- (extend_world_simpl w3 wF1 w3F1 t3); auto.
      eapply (deterministic_wterm_frame w3 (w3:t3) wF1); eauto.
      reflexivity. eexists; eauto.
      rewrite (extend_world_simpl w3 wF1 w3F1 t3); auto.
        eapply step_star_Cs_cons.
        apply cstep_star_par_cx_par1; eauto.
        apply step_star_Cs_nil.
    intros s' b' ?.
    inverts H1.
    rename w3 into w3F1.
    assert (w3F1|= gsep_con P3 TT).
      eapply HtermP3. apply cstep_star_par_cx_par1; eauto.
    destruct H1 as [w3[wF1[?[??]]]].
    assert (s'=s).
      eapply (respects_global_state_frame w3 (w3:t3) wF1); eauto.
      reflexivity. eapply join_joins; eauto.
      rewrite (extend_world_simpl w3 wF1 w3F1 t3); auto.
      eapply step_star_Cs_cons.
      apply cstep_star_par_cx_par1; eauto.
      apply step_star_Cs_nil.
    subst s'.
    eexists; split.
      apply step_nil.
      right; left; eexists; exists b1'; do 6 eexists; splits; eauto.
        eapply cstep_star_snoc; eauto.
        apply cstep_simulates_step; apply step_add_context; apply st_par1; auto.
    inverts H6. rename b1' into b2'.
    exists ((w3 : tm_skip || b2');; tm_par P3 t3 P4 t4); split.
      apply step_single; apply st_seq1; apply st_par2; auto.
      right; left; do 8 eexists; splits; eauto.
      eapply cstep_star_snoc; eauto.
      apply cstep_simulates_step; apply step_add_context; apply st_par2; auto.
    rename w3 into w3F1; rename w into w4F2.
    edestruct Hparjoins12 as [w ?].
    eapply step_star_Cs_cons; eauto.
    eapply step_star_Cs_cons.
    apply cstep_star_par2_step.
    apply cstep_star_par_cx_par2; eauto.
    apply step_star_Cs_nil.
    assert (w3F1|= gsep_con P3 TT).
      eapply HtermP3. apply cstep_star_par_cx_par1; eauto.
    assert (w4F2|= gsep_con P4 TT).
      eapply HtermP4.
      apply cstep_star_par_cx_par2 with (b1:=w1: t1;; t3).
      eapply cstep_star_app; eauto.
      eapply cstep_star_snoc.
      apply cstep_star_par1_inner_seq_append; eauto.
      apply cstep_simulates_step; apply step_add_context; apply st_par1; apply st_seq_skip.
    destruct H2 as [w3[wF1[?[??]]]].
    destruct H3 as [w4[wF2[?[??]]]].
    edestruct @join_assoc with (H:=gworld_sa) (a:=wF1) (b:=w3) as [w34F2[??]]; eauto.
    edestruct @join_assoc with (H:=gworld_sa) (a:=wF2) (b:=w4) as [w34[??]]; eauto.
    edestruct @join_assoc with (H:=gworld_sa) (a:=w34) (b:=wF2) as [wF[??]]; eauto.
    apply @join_com in H10.
    apply @join_com in H12.
    exists (w: tm_par P3 t3 P4 t4); split.
      eapply step_cons.
      apply st_seq1.
      apply st_par_join; eauto.
      apply step_single.
      apply st_seq_skip.
      right; right; right.
      apply bisim_trans with (w3F1:t3 || w4F2:t4).
      apply bisim_Cs_bisim.
      setoid_rewrite bisim_Cs_tm_par_par_precise_frame with (w1:=w3) (w2:=w4); eauto.
      setoid_rewrite bisim_Cs_par_frame; eauto.
      rewrite<- (extend_world_simpl w3 wF1 w3F1 t3); auto.
      rewrite<- (extend_world_simpl w4 wF2 w4F2 t4); auto.
      setoid_rewrite bisim_Cs_frameable_extend_world; eauto;
        [ | reflexivity | eexists; eauto | reflexivity | eexists; eauto]. 
      setoid_rewrite bisim_Cs_par_assoc.
      setoid_rewrite<- bisim_Cs_par_assoc at 3.
      setoid_rewrite bisim_Cs_par_swap at 5.
      setoid_rewrite bisim_Cs_par_assoc.
      setoid_rewrite<- bisim_Cs_par_assoc at 2.
      setoid_rewrite bisim_Cs_par_skip; eauto.
      reflexivity.
      apply TT_gstar; reflexivity.
      apply bisim_Cs_bisim.
      apply bisim_Cs_par1.
      rewrite<- (extend_world_simpl w3 wF1 w3F1 t3); auto.
      eapply (deterministic_wterm_frame w3 (w3:t3) wF1); eauto.
      reflexivity. eexists; eauto.
      rewrite (extend_world_simpl w3 wF1 w3F1 t3); auto.
        eapply step_star_Cs_cons.
        apply cstep_star_par_cx_par1; eauto.
        apply step_star_Cs_nil.

    splits.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    right; right; left; do 8 eexists; splits; eauto.
      eapply cstep_star_iter_snoc; eauto.
    intros.
    right; right; left; do 5 eexists; exists (cx_par1 C1 b); do 2 eexists; splits; eauto.
      apply cstep_star_par_cx_par1.
      apply cstep_star_par1_step; eauto.
      eapply cstep_star_snoc; eauto.
      apply cstep_star_par_cx_par1.
      apply cstep_star_par1_step; eauto.
      apply cst_par1_step2; eauto.
    intros s' b' ?.
    inverts H1.
    inverts H6. rename b1'1 into b1'.
    exists ((b1';;t3) || b4); split.
      apply step_single; apply st_par1; apply st_seq1; auto.
      right; right; left; do 8 eexists; splits; eauto.
      eapply cstep_star_snoc; eauto.
      apply cstep_simulates_step; apply step_add_context; apply st_par1; auto.
    false; inverts H5.
    rename w0 into w3F1; rename w4 into w4F2.
    assert (w3F1|= gsep_con P3 TT).
      eapply HtermP3.
      apply cstep_star_par_cx_par1 with (b2:=w2: t2;; t4); eauto.
      eapply cstep_star_app; eauto.
      eapply cstep_star_snoc.
      apply cstep_star_par2_inner_seq_append; eauto.
      apply cstep_simulates_step; apply step_add_context; apply st_par2; apply st_seq_skip.
    assert (w4F2|= gsep_con P4 TT).
      eapply HtermP4. apply cstep_star_par_cx_par2; eauto.
    destruct H1 as [w3[wF1[?[??]]]].
    destruct H2 as [w4[wF2[?[??]]]].
    edestruct @join_assoc with (H:=gworld_sa) (a:=wF1) (b:=w3) as [w34F2[??]]; eauto.
    edestruct @join_assoc with (H:=gworld_sa) (a:=wF2) (b:=w4) as [w34[??]]; eauto.
    edestruct @join_assoc with (H:=gworld_sa) (a:=w34) (b:=wF2) as [wF[??]]; eauto.
    apply @join_com in H10.
    apply @join_com in H12.
    exists (w3F1:t3 || b4); split.
      apply step_single; apply st_par1; apply st_seq_skip.
      right; right; right.
      setoid_rewrite bisim_Cs_seq_skip1'.
      apply bisim_Cs_bisim.
      apply bisim_Cs_trans with (w3F1:t3 || w4F2:t4).
      setoid_rewrite bisim_Cs_tm_par_par_precise_frame with (w1:=w3) (w2:=w4); eauto.
      setoid_rewrite bisim_Cs_par_frame; eauto.
      rewrite<- (extend_world_simpl w3 wF1 w3F1 t3); auto.
      rewrite<- (extend_world_simpl w4 wF2 w4F2 t4); auto.
      setoid_rewrite bisim_Cs_frameable_extend_world; eauto;
        [ | reflexivity | eexists; eauto | reflexivity | eexists; eauto]. 
      setoid_rewrite bisim_Cs_par_assoc.
      setoid_rewrite<- bisim_Cs_par_assoc at 3.
      setoid_rewrite bisim_Cs_par_swap at 5.
      setoid_rewrite bisim_Cs_par_assoc.
      setoid_rewrite<- bisim_Cs_par_assoc at 2.
      setoid_rewrite bisim_Cs_par_skip; eauto.
      reflexivity.
      apply TT_gstar; reflexivity.
      apply bisim_Cs_par2.
      rewrite<- (extend_world_simpl w4 wF2 w4F2 t4); auto.
      eapply (deterministic_wterm_frame w4 (w4:t4) wF2); eauto.
      reflexivity. eexists; eauto.
      rewrite (extend_world_simpl w4 wF2 w4F2 t4); auto.
        eapply step_star_Cs_cons.
        apply cstep_star_par_cx_par2; eauto.
        apply step_star_Cs_nil.
    intros s' b' ?.
    inverts H1.
    inverts H6. rename b1'1 into b1'.
    exists ((b1' || w4:tm_skip);; tm_par P3 t3 P4 t4); split.
      apply step_single; apply st_seq1; apply st_par1; auto.
      right; right; left; do 8 eexists; splits; eauto.
      eapply cstep_star_snoc; eauto.
      apply cstep_simulates_step; apply step_add_context; apply st_par1; auto.
    rename w into w3F1. rename w4 into w4F2.
    edestruct Hparjoins12 as [w ?].
    eapply step_star_Cs_cons; eauto.
    eapply step_star_Cs_cons.
    apply cstep_star_par1_step.
    apply cstep_star_par_cx_par1; eauto.
    apply step_star_Cs_nil.
    assert (w3F1|= gsep_con P3 TT).
      eapply HtermP3.
      apply cstep_star_par_cx_par1 with (b2:=w2: t2;; t4); eauto.
      eapply cstep_star_app; eauto.
      eapply cstep_star_snoc.
      apply cstep_star_par2_inner_seq_append; eauto.
      apply cstep_simulates_step; apply step_add_context; apply st_par2; apply st_seq_skip.
    assert (w4F2|= gsep_con P4 TT).
      eapply HtermP4. apply cstep_star_par_cx_par2; eauto.
    destruct H2 as [w3[wF1[?[??]]]].
    destruct H3 as [w4[wF2[?[??]]]].
    edestruct @join_assoc with (H:=gworld_sa) (a:=wF1) (b:=w3) as [w34F2[??]]; eauto.
    edestruct @join_assoc with (H:=gworld_sa) (a:=wF2) (b:=w4) as [w34[??]]; eauto.
    edestruct @join_assoc with (H:=gworld_sa) (a:=w34) (b:=wF2) as [wF[??]]; eauto.
    apply @join_com in H10.
    apply @join_com in H12.
    exists (w: tm_par P3 t3 P4 t4); split.
      eapply step_cons.
      apply st_seq1.
      apply st_par_join; eauto.
      apply step_single.
      apply st_seq_skip.
      right; right; right.
      apply bisim_Cs_bisim.
      apply bisim_Cs_trans with (w3F1:t3 || w4F2:t4).
      setoid_rewrite bisim_Cs_tm_par_par_precise_frame with (w1:=w3) (w2:=w4); eauto.
      setoid_rewrite bisim_Cs_par_frame; eauto.
      rewrite<- (extend_world_simpl w3 wF1 w3F1 t3); auto.
      rewrite<- (extend_world_simpl w4 wF2 w4F2 t4); auto.
      setoid_rewrite bisim_Cs_frameable_extend_world; eauto;
        [ | reflexivity | eexists; eauto | reflexivity | eexists; eauto]. 
      setoid_rewrite bisim_Cs_par_assoc.
      setoid_rewrite<- bisim_Cs_par_assoc at 3.
      setoid_rewrite bisim_Cs_par_swap at 5.
      setoid_rewrite bisim_Cs_par_assoc.
      setoid_rewrite<- bisim_Cs_par_assoc at 2.
      setoid_rewrite bisim_Cs_par_skip; eauto.
      reflexivity.
      apply TT_gstar; reflexivity.
      apply bisim_Cs_par2.
      rewrite<- (extend_world_simpl w4 wF2 w4F2 t4); auto.
      eapply (deterministic_wterm_frame w4 (w4:t4) wF2); eauto.
      reflexivity. eexists; eauto.
      rewrite (extend_world_simpl w4 wF2 w4F2 t4); auto.
        eapply step_star_Cs_cons.
        apply cstep_star_par_cx_par2; eauto.
        apply step_star_Cs_nil.
    rename w4 into w4F2.
    assert (w4F2|= gsep_con P4 TT).
      eapply HtermP4. apply cstep_star_par_cx_par2; eauto.
    destruct H1 as [w4[wF2[?[??]]]].
    assert (s'=s).
      eapply (respects_global_state_frame w4 (w4:t4) wF2); eauto.
      reflexivity. eapply join_joins; eauto.
      rewrite (extend_world_simpl w4 wF2 w4F2 t4); auto.
      eapply step_star_Cs_cons.
      apply cstep_star_par_cx_par2; eauto.
      apply step_star_Cs_nil.
    subst s'.
    eexists; split.
      apply step_nil.
      right; right; left; do 2 eexists; exists b2'; do 5 eexists; splits; eauto.
        eapply cstep_star_snoc; eauto.
        apply cstep_simulates_step; apply step_add_context; apply st_par2; auto.

    splits; intros.
    split.
    apply bisim_halted1; auto.
    right; right; right; apply bisim_refl.
    split.
    apply bisim_halted2; auto.
    right; right; right; apply bisim_refl.
    right; right; right; eapply bisim_s_iter; eauto.
    right; right; right; eapply bisim_context; eauto.
    edestruct step_bisim1 as [bb'[??]]; eauto.
    exists bb'; split; auto.
    edestruct step_bisim2 as [ba'[??]]; eauto.
    exists ba'; split; auto.
  Qed.

  Lemma term_ensures_pred_skip: forall s w P,
    w|= P -> term_ensures_pred s (w: tm_skip) P.
  Proof.
    repeat intro.
    apply cstep_star_skip_resolve in H0.
    inverts H0; auto.
  Qed.


End Lemmas.
