Require Import Machine.
Require Import world_model.
Require Import msl.msl_standard.
Require Import SfLib.
Require Import LibTactics.
Require Import IterationBounds.
(*Import Machine.PMapFacts.*)

(*Require Import Omega.*)

Section Definitions.
  Import Machine.Notation.

End Definitions.

  Ltac step_cons n := match n with
   | 0 => apply step_nil
   | S ?n => eapply step_cons; [ | step_cons n]
  end.
  Ltac step_cons' n := match n with
   | 0 => eauto
   | S ?n => eapply step_cons; [| step_cons' n]
  end.



Section Lemmas.
  Import Notation.
  Import Machine.Notation.
  Import world_model.Notation.
  Open Scope forces_pred.
  Open Scope wterm_scope.

  Lemma step_star_seq_inversion: forall s b1 t2 s'' b'',
    step_star s (b1;; t2) s'' b'' ->
    (exists s', exists w',
    step_star s b1 s' (w':tm_skip) /\ step_star s' (w':t2) s'' b'')
    \/ (exists b1'', step_star s b1 s'' b1'' /\ b'' = b1'';; t2).
  Proof.
    intros.
    remember (b1;; t2) as b.
    revert b1 t2 Heqb.
    induction H; intros; subst.
    right; exists b1; splits; auto; apply step_nil.
    inverts H.
 
    edestruct IHstep_star as [[s'1[w'[??]]] | [b1''[??]]]; eauto.
    left; exists s'1 w'; split; auto; eapply step_cons; eauto.
    subst; right; exists b1''; split; auto; eapply step_cons; eauto.
 
    clear IHstep_star; left; exists s' w; split; auto; apply step_nil. 
 
    edestruct IHstep_star as [[s'1[w'[??]]] | [b1''[??]]]; eauto.
    left; exists s'1 w'; split; auto; eapply st_iteration; eauto.
    subst. right; exists b1''; split; auto; eapply st_iteration; eauto.
  Qed.

  Lemma step_single: forall s b s' b',
    step s b s' b' -> step_star s b s' b'.
  Proof. intros; eapply step_cons; eauto; apply step_nil. Qed.
    
  Lemma step_star_app: forall s b s' b' s'' b'',
    step_star s b s' b' ->
    step_star s' b' s'' b'' ->
    step_star s b s'' b''.
  Proof. intros. induction H; auto. eapply step_cons; eauto. eapply st_iteration; eauto. Qed.

  Lemma step_star_seq_app: forall s b t s' b',
    step_star s b s' b' ->
    step_star s (b;;t) s' (b';;t).
  Proof. intros. induction H. apply step_nil. eapply step_cons; eauto. apply st_seq1; auto. eapply st_iteration; eauto. Qed.

  Lemma step_star_snoc: forall s b s' b' s'' b'',
    step_star s b s' b' -> step s' b' s'' b'' ->
    step_star s b s'' b''.
  Proof.
    intros.
    revert s'' b'' H0; induction H; intros.
    apply step_single; auto.
    eapply step_cons; eauto.
    eapply st_iteration; eauto.
  Qed.

  Lemma step_star_iter_snoc: forall s b s' b' s'',
    step_star s b s' b' -> s_iter s' s'' ->
    step_star s b s'' b'.
  Proof.
    intros.
    revert s'' H0; induction H; intros.
    eapply st_iteration; eauto; apply step_nil.
    eapply step_cons; eauto.
    eapply st_iteration; eauto.
  Qed.


  Lemma step_star_seq_join: forall s b1 s' w2 t2 s'' b2'',
    step_star s b1 s' (w2:tm_skip) ->
    step_star s' (w2:t2) s'' b2'' ->
    step_star s (b1;;t2) s'' b2''.
  Proof. 
    intros.
    eapply step_star_app; eauto.
    eapply step_star_snoc.
    apply step_star_seq_app; eauto.
    apply st_seq_skip.
  Qed.

  Lemma step_star_seq_join': forall s b1 s' w2 t2,
    step_star s b1 s' (w2:tm_skip) ->
    step_star s (b1;;t2) s' (w2:t2).
  Proof. 
    intros.
    eapply step_star_seq_join; eauto.
    apply step_nil.
  Qed.

  Lemma step_star_par1_step: forall s b1 b2 s' b1',
    step_star s b1 s' b1' ->
    step_star s (b1 || b2) s' (b1' || b2).
  Proof. intros; induction H. apply step_nil. eapply step_cons; eauto. apply st_par1; auto. eapply st_iteration; eauto. Qed.

  Lemma step_star_par2_step: forall s b1 b2 s' b2',
    step_star s b2 s' b2' ->
    step_star s (b1 || b2) s' (b1 || b2').
  Proof. intros; induction H. apply step_nil. eapply step_cons; eauto. apply st_par2; auto. eapply st_iteration; eauto. Qed.

  Lemma step_star_skip_s_iter: forall s w s' w',
    step_star s (w:tm_skip) s' (w':tm_skip) -> s_iter_star s s'.
  Proof.
    intros.
    remember (w:tm_skip)%wterm as b.
    remember (w':tm_skip)%wterm as b'.
    revert w w' Heqb Heqb'.
    induction H; intros; subst.
    apply s_iter_nil.
    inverts H.
    eapply s_iter_cons; eauto.
  Qed.

  Lemma step_star_skip_resolve: forall s w s' b',
    step_star s (w : tm_skip) s' b' -> s_iter_star s s' /\ b' = (w : tm_skip)%wterm.
  Proof.
    intros. remember (w:tm_skip)%wterm as b.
    induction H; intros; subst.
    split; auto; constructor.
    inverts H.
    destruct IHstep_star; subst; auto.
    split; auto.
    eapply s_iter_cons; eauto.
  Qed.

  Lemma step_star_par_swap: forall s b1 b2 s' b',
    step_star s (b1 || b2) s' b' ->
    exists b'0,
      step_star s (b2 || b1) s' b'0 /\
      ((exists b1' b2', b'=b1'||b2' /\ b'0=b2'||b1') \/ exists w, b'=w:tm_skip /\ b'0=w:tm_skip).
  Proof.
    intros.
    remember (b1||b2) as b.
    revert b1 b2 Heqb.
    induction H; intros; subst.

    eexists; split. apply step_nil. eauto.

    inverts H.
    edestruct IHstep_star as [b'0[? [ [b1''[b2''[??]]] | [w[??]] ] ]]; clear IHstep_star; eauto.
    subst.
    exists (b2'' || b1''); split; eauto.
      eapply step_cons; eauto. apply st_par2; auto.
    subst.
    exists (w: tm_skip); split; eauto.
      eapply step_cons. apply st_par2; eauto. eauto.
    edestruct IHstep_star as [b'0[? [ [b1''[b2''[??]]] | [w[??]] ] ]]; clear IHstep_star; eauto.
    subst.
    exists (b2'' || b1''); split; eauto.
      eapply step_cons; eauto. apply st_par1; auto.
    subst.
    exists (w: tm_skip); split; eauto.
      eapply step_cons. apply st_par1; eauto. eauto.
    clear IHstep_star.
    edestruct step_star_skip_resolve as [??]; eauto.
    subst b''.
    apply gjoin_com in H5.
    exists (w:tm_skip); split; eauto.
      eapply step_cons; eauto. apply st_par_join; eauto.

    edestruct IHstep_star as [b'0[? [ [b1''[b2''[??]]] | [w[??]] ] ]]; clear IHstep_star; eauto.
    subst.
    exists (b2'' || b1''); split; eauto.
      eapply st_iteration; eauto.
    subst.
    exists (w: tm_skip); split; eauto.
      eapply st_iteration; eauto.
  Qed.

  Lemma step_star_par1_seq_inversion : forall s b1 t2 b3 s'' b'',
    step_star s (b1;; t2 || b3) s'' b'' ->
    (exists b1' b3',
    step_star s (b1 || b3) s'' (b1' || b3') /\
    b'' = (b1';; t2 || b3')%wterm) \/
    (exists s' w2 b3',
      step_star s (b1 || b3) s' (w2 : tm_skip || b3') /\
      step_star s' (w2 : t2 || b3') s'' b'').
  Proof.
    intros.
    remember (b1;; t2 || b3) as b.
    revert b1 t2 b3 Heqb.
    induction H; intros; subst.

    left; do 2 eexists; split; auto; apply step_nil.

    inverts H.
    inverts H5. rename b1'0 into b1'.
    edestruct IHstep_star as [ [b1''[b2''[??]]] | [s'0[b1'0[w2'0[??]]]] ]; clear IHstep_star; eauto.
    subst.
    left; do 2 eexists; split; auto.
    eapply step_cons; eauto. apply st_par1; eauto.
    right; do 3 eexists; split; eauto.
    eapply step_cons; eauto. apply st_par1; eauto.
    clear IHstep_star.
    right; do 3 eexists; split; eauto. apply step_nil.
    rename b2' into b3'.
    edestruct IHstep_star as [ [b1''[b2''[??]]] | [s'0[b1'0[w2'0[??]]]] ]; clear IHstep_star; eauto.
    subst.
    left; do 2 eexists; split; auto.
    eapply step_cons; eauto. apply st_par2; eauto.
    right; do 3 eexists; split; eauto.
    eapply step_cons; eauto. apply st_par2; eauto.

    edestruct IHstep_star as [ [b1''[b2''[??]]] | [s'0[b1'0[w2'0[??]]]] ]; clear IHstep_star; eauto.
    subst.
    left; do 2 eexists; split; auto.
    eapply st_iteration; eauto.
    right; do 3 eexists; split; eauto.
    eapply st_iteration; eauto.
  Qed.

  Lemma step_star_par2_seq_inversion : forall s b1 b2 t3 s'' b'',
    step_star s (b1 || (b2;;t3)) s'' b'' ->
    (exists b1' b2',
      step_star s (b1 || b2) s'' (b1' || b2') /\
      b'' = (b1' || (b2';;t3))%wterm) \/
      (exists s' w3 b1',
        step_star s (b1 || b2) s' (b1' || w3:tm_skip) /\
        step_star s' (b1' || w3:t3) s'' b'').
  Proof.
    intros.
    remember (b1 || (b2;; t3)) as b.
    revert b1 b2 t3 Heqb.
    induction H; intros; subst.

    left; do 2 eexists; split; auto; apply step_nil.

    inverts H.
    edestruct IHstep_star as [ [b1''[b2''[??]]] | [s'0[b1'0[w2'0[??]]]] ]; clear IHstep_star; eauto.
    subst.
    left; do 2 eexists; split; auto.
    eapply step_cons; eauto. apply st_par1; eauto.
    right; do 3 eexists; split; eauto.
    eapply step_cons; eauto. apply st_par1; eauto.
    inverts H5.
    rename b1' into b2'.
    edestruct IHstep_star as [ [b1''[b2''[??]]] | [s'0[b1'0[w2'0[??]]]] ]; clear IHstep_star; eauto.
    subst.
    left; do 2 eexists; split; auto.
    eapply step_cons; eauto. apply st_par2; eauto.
    right; do 3 eexists; split; eauto.
    eapply step_cons; eauto. apply st_par2; eauto.
    clear IHstep_star.
    right; do 3 eexists; split; eauto. apply step_nil.


    edestruct IHstep_star as [ [b1''[b2''[??]]] | [s'0[b1'0[w2'0[??]]]] ]; clear IHstep_star; eauto.
    subst.
    left; do 2 eexists; split; auto.
    eapply st_iteration; eauto.
    right; do 3 eexists; split; eauto.
    eapply st_iteration; eauto.
  Qed.

  Lemma step_star_par1_inner_seq_append : forall s b1 b2 s' b1' b2' t3,
    step_star s (b1||b2) s' (b1'||b2') ->
    step_star s (b1;;t3 || b2) s' (b1';;t3 || b2').
  Proof.
    intros.
    remember (b1||b2) as b.
    remember (b1'||b2') as b'.
    revert b1 b2 b1' b2' Heqb Heqb'.
    induction H; intros; subst.

    inverts Heqb'.
    apply step_nil.

    inverts H.
    eapply step_cons; eauto; apply st_par1; apply st_seq1; auto.
    eapply step_cons; eauto; apply st_par2; auto.
    clear IHstep_star.
    false. apply step_star_skip_resolve in H0. destruct H0. discriminate.

    eapply st_iteration; eauto; apply st_par1; apply st_seq1; auto.
  Qed.

  Lemma step_star_par2_inner_seq_append : forall s b1 b2 s' b1' b2' t3,
    step_star s (b1||b2) s' (b1'||b2') ->
    step_star s (b1 || (b2;;t3)) s' (b1' || (b2';;t3)).
  Proof.
    intros.
    remember (b1||b2) as b.
    remember (b1'||b2') as b'.
    revert b1 b2 b1' b2' Heqb Heqb'.
    induction H; intros; subst.

    inverts Heqb'.
    apply step_nil.

    inverts H.
    eapply step_cons; eauto; apply st_par1; auto.
    eapply step_cons; eauto; apply st_par2; apply st_seq1; auto.
    clear IHstep_star.
    false. apply step_star_skip_resolve in H0. destruct H0. discriminate.

    eapply st_iteration; eauto; apply st_par1; apply st_seq1; auto.
  Qed.

  Lemma step_star_tm_seq: forall s w t1 t2 s' b',
    step_star s (w: tm_seq t1 t2) s' b' ->
    b'=w:tm_seq t1 t2 \/ step_star s (w: t1;; t2) s' b'.
  Proof.
    intros.
    remember (w: tm_seq t1 t2) as b.
    induction H; subst; auto.
    right; inverts H; auto.
    destruct IHstep_star; auto.
    right; eapply st_iteration; eauto.
  Qed.
   
  Lemma step_star_assert_resolve: forall s s' P w b',
    step_star s (w : tm_assert P) s' b' ->
    (b'=w:tm_assert P \/ b'=w:tm_skip /\ w|=gsep_con P TT)%wterm.
  Proof.
    intros.
    remember (w: tm_assert P)%wterm as b.
    revert Heqb.
    induction H; intros; subst; eauto.
    inverts H; auto.
    edestruct step_star_skip_resolve; eauto.
  Qed.

  Lemma step_star_assert_skip_resolve: forall s b P s' w',
    step_star s (b;; tm_assert P) s' (w':tm_skip) ->
    w' |= gsep_con P TT.
  Proof.
    intros.
    edestruct step_star_seq_inversion as [[s'0[w'0[??]]] | [bx'[??]]]; eauto.
    edestruct step_star_assert_resolve as [?|[??]]; eauto.
    discriminate.
    inversion H2; subst; auto.
    discriminate.
  Qed.

  Lemma step_star_go_preservation: forall s b s' b' i,
    s_iter_get s i = Some iter_go ->
    step_star s b s' b' ->
    s_iter_get s' i = Some iter_go.
  Proof.
    intros.
    induction H0; auto.
    apply IHstep_star.
    eapply step_go_preservation; eauto.
    apply IHstep_star.
    eapply s_iter_go_preservation; eauto.
  Qed.

  Lemma s_iter_step_star: forall s b s',
    s_iter_star s s' -> step_star s b s' b.
  Proof.
    intros. induction H.
    apply step_nil.
    eapply st_iteration; eauto.
  Qed.

  Lemma step_star_s_iter_preservation: forall i s b s'' b'',
    s_iter_get s i = Some iter_go ->
    step_star s b s'' b'' ->
    s_iter_get s'' i = Some iter_go.
  Proof.
    intros.
    induction H0; auto.
    apply IHstep_star.
    unfold s_iter_get in *.
    erewrite<- step_s_iter; eauto.
    apply IHstep_star.
    inverts H0.
    destruct (Pos.eq_dec i0 i); subst.
    apply s_iter_get_set_eq.
    rewrite s_iter_get_set_neq; auto.
  Qed.

  Lemma step_star_par_skip_resolve: forall s b1 w2 s' b',
    step_star s (b1 || w2 : tm_skip) s' b' ->
    exists b1',
      step_star s b1 s' b1' /\
      (b' = b1' || w2 : tm_skip \/
       (exists w1' w',
        b1' = w1' : tm_skip /\ gjoin w1' w2 w' /\ b' = w' : tm_skip)).
  Proof.
    intros.
    remember (b1||w2:tm_skip) as b.
    revert b1 w2 Heqb.
    induction H; intros; subst.

    eexists; split; eauto. apply step_nil.

    inverts H.
    edestruct IHstep_star as [b1''[? [? | [w1''[w''[?[??]]]]] ]]; clear IHstep_star; eauto.
    subst.
    eexists; split; eauto; eapply step_cons; eauto.
    subst.
    eexists; split; iauto; eapply step_cons; eauto.
    inverts H5.
    clear IHstep_star.
    edestruct step_star_skip_resolve; eauto.
    subst b''.
    eexists; split; iauto.
    apply s_iter_step_star; auto.
    
    edestruct IHstep_star as [b1''[? [? | [w1''[w''[?[??]]]]] ]]; clear IHstep_star; eauto.
    subst.
    eexists; split; eauto; eapply st_iteration; eauto.
    subst.
    eexists; split; iauto; eapply st_iteration; eauto.
  Qed.



End Lemmas.
