Require Import Machine.
Require Import ContextSemantics.
Require Import StepStar.
Require Import world_model.
Require Import msl.msl_standard.
Require Import SfLib.
Require Import LibTactics.

Import Machine.Notation.
Import world_model.Notation.
Import world_model.Notation.U.
Open Scope wterm_scope.
Open Scope term_scope.
Open Scope forces_pred.

  Inductive step_star_Cs: wterm->wterm->Prop :=
  | step_star_Cs_nil: forall ba,
      step_star_Cs ba ba
  | step_star_Cs_cons: forall s C ba s' C' ba' ba'',
      cstep_star s C ba s' C' ba' -> step_star_Cs ba' ba'' -> step_star_Cs ba ba''.

  Definition halted_assert_Cs (b:wterm) (P: pred world) := forall w',
    step_star_Cs b (w':tm_skip) -> P w'.




  Lemma step_single_Cs: forall s b s' b',
    step s b s' b' ->
    step_star_Cs b b'.
  Proof.
    intros.
    eapply step_star_Cs_cons.
    apply step_single_cstep_star with (C:=cx_hole); eauto.
    apply step_star_Cs_nil.
  Qed.

  Lemma step_star_Cs_skip_resolve: forall w b',
    step_star_Cs (w:tm_skip) b' -> b' = w:tm_skip.
  Proof.
    intros.
    remember (w:tm_skip) as b.
    revert Heqb.
    induction H; intros; subst; auto.
    apply cstep_star_skip_resolve in H. subst. auto.
  Qed.

  Lemma step_star_Cs_assert_resolve: forall P w b',
    step_star_Cs (w : tm_assert P) b' ->
    b'=w:tm_assert P \/ (b'=w:tm_skip /\ w|=gsep_con P TT).
  Proof.
    intros.
    remember (w: tm_assert P) as b.
    revert Heqb.
    induction H; intros; subst; auto.
    edestruct cstep_star_assert_resolve as [?|[??]]; eauto.
    subst; auto.
    subst.
    apply step_star_Cs_skip_resolve in H0; subst; eauto.
  Qed.

  Lemma step_star_Cs_while_inf_skip_resolve: forall w e t w',
    step_star_Cs (w: tm_while_max e i_inf t) (w':tm_skip) ->
    evalFalse e w'.
  Proof.
    intros.
    remember (w : tm_while_max e i_inf t) as b.
    remember (w' : tm_skip) as b'.
    assert ((exists w, b=w : tm_while_max e i_inf t) \/ (exists bt, b=bt;; tm_while_max e i_inf t)).
      eauto.
    clear w Heqb.
    revert w' H0 Heqb'.
    induction H; intros; subst.
    destruct H0 as [[?]|[?]]; discriminate.
    destruct H1 as [[w?]|[bt?]]; subst.
    edestruct cstep_star_while_resolve as [[??] | [[x''[??]] | [w'' ?]]]; eauto.
    eapply IHstep_star_Cs; eauto.
    destruct H1 as [? | [i[n[z1[?[??]]]]]].
    apply iter_decrs_inf1 in H1. subst. eauto.
    discriminate.
    subst.
    apply step_star_Cs_skip_resolve in H0. inverts H0.
    edestruct cstep_star_while_max_skip_resolve as [?|[[n[??]]|[i[n[n'[?[??]]]]]]]; eauto.
    false.
    false.
    edestruct cstep_star_seq_inversion as [[s'0[C'0[w'0[??]]]]|[b'0[??]]]; eauto.
    clear H.
    edestruct cstep_star_while_resolve as [[??] | [[z'[??]] | [w'' ?]]]; eauto.
    destruct H.
    apply iter_decrs_inf1 in H. subst z'; auto.
    destruct H as [i[n[z[?[??]]]]]; discriminate.
    subst.
    apply step_star_Cs_skip_resolve in H0. inverts H0.
    edestruct cstep_star_while_max_skip_resolve as [?|[[n[??]]|[i[n[n'[?[??]]]]]]]; eauto.
    false.
    false.
  Qed.

  Lemma step_star_Cs_seq_inversion: forall b1 t2 b'',
       step_star_Cs (b1;; t2) b'' ->
       (exists w',
        step_star_Cs b1 (w' : tm_skip) /\ step_star_Cs (w' : t2) b'') \/
       (exists b1'', step_star_Cs b1 b1'' /\ b'' = b1'';; t2).
  Proof.
    first [generalize apply step_star_Cs_seq_inversion; false | auto ].
    intros.
    remember (b1;; t2) as b.
    revert b1 t2 Heqb.
    induction H; intros; subst.
    right; exists b1; split; auto. apply step_star_Cs_nil.
    edestruct cstep_star_seq_inversion as [ [s'0[C'0[w'[??]]]] | [b1''[??]] ]; eauto; clear H.
    left; exists w'; split.
      eapply step_star_Cs_cons; eauto. apply step_star_Cs_nil.
      eapply step_star_Cs_cons; eauto.
    subst ba'.
    edestruct IHstep_star_Cs as [ [w'[??]] | [b1'''[??]] ]; clear IHstep_star_Cs; eauto.
    left; exists w'; split; auto.
      eapply step_star_Cs_cons; eauto.
    subst ba''.
    right; exists b1'''; split; auto.
      eapply step_star_Cs_cons; eauto.
  Qed.



  Lemma step_star_Cs_par_resolve: forall b1 b2 b',
    step_star_Cs (b1 || b2) b' ->
    (exists b1' b2', b' = b1' || b2') \/
    (exists w1' w2' w',
     step_star_Cs (b1 || b2) (w1' : tm_skip || w2' : tm_skip) /\
     gjoin w1' w2' w' /\ b' = w' : tm_skip).
  Proof.
    intros.
    remember (b1||b2) as b.
    revert b1 b2 Heqb.
    induction H; intros; subst; eauto.
    edestruct cstep_star_par_resolve as [ [b1'[b2' ?]] | [w1'[w2'[w''[?[??]]]]] ]; eauto.
    subst ba'.
    edestruct IHstep_star_Cs as [ [b1''[b2'' ?]] | [w1''[w2''[w''[?[??]]]]] ]; clear IHstep_star_Cs; eauto.
    subst ba''.
    right; exists w1'' w2'' w''; splits; auto.
      eapply step_star_Cs_cons; eauto.
    subst ba'.
    clear IHstep_star_Cs.
    apply step_star_Cs_skip_resolve in H0.
    subst ba''.
    right; exists w1' w2' w''; splits; auto.
      eapply step_star_Cs_cons. apply H1.
      apply step_star_Cs_nil.
  Qed.

  Lemma step_star_Cs_par_skip_resolve: forall b1 w2 b',
    step_star_Cs (b1 || w2 : tm_skip) b' ->
    exists b1',
      step_star_Cs b1 b1' /\
      (b' = b1' || w2 : tm_skip \/
      (exists w1' w', b1' = w1' : tm_skip /\ gjoin w1' w2 w' /\ b' = w' : tm_skip)).
  Proof.
    intros.
    remember (b1||w2:tm_skip) as b.
    revert b1 Heqb.
    induction H; intros; subst.
    exists b1; split; auto.
      apply step_star_Cs_nil.
    edestruct cstep_star_par_skip_resolve as [b1'[?[?| [w1'[w'[?[??]]]] ]]]; eauto; clear H.
    subst ba'.
    edestruct IHstep_star_Cs as [b1''[?[?| [w1''[w''[?[??]]]] ]]]; clear IHstep_star_Cs; eauto.
    subst ba''.
    exists b1''; split; auto.
      eapply step_star_Cs_cons; eauto.
    subst b1'' ba''.
    exists (w1'':tm_skip); split.
      eapply step_star_Cs_cons; eauto.
      right; exists w1'' w''; splits; auto.
    subst b1' ba'.
    apply step_star_Cs_skip_resolve in H0.
    subst ba''.
    clear IHstep_star_Cs.
    exists (w1':tm_skip); split.
      eapply step_star_Cs_cons; eauto. apply step_star_Cs_nil.
      right; exists w1' w'; splits; auto.
  Qed.

  Lemma step_star_Cs_app: forall b b' b'',
    step_star_Cs b b' -> step_star_Cs b' b'' -> step_star_Cs b b''.
  Proof.
    intros.
    induction H; auto.
    eapply step_star_Cs_cons; eauto.
  Qed.

  Lemma step_star_Cs_step_inversion: forall b b',
    step_star_Cs b b' ->
    (exists s s' b'0, step s b s' b'0 /\ step_star_Cs b'0 b' /\ b'0<>b) \/ b'=b.
  Proof.
    intros.
    induction H; auto.
    edestruct cstep_star_step_inversion as [ [s'0[C'0[s'1[b'1[?[?[??]]]]]]] | [??] ]; eauto.
    left; do 3 eexists; splits; eauto.
    eapply step_star_Cs_cons; eauto.
    subst ba'.
    destruct IHstep_star_Cs as [ [s0[s0'[b'0[??]]]] | ?]; auto.
    left; do 3 eexists; split; eauto.
  Qed.

  Lemma step_star_Cs_step_star: forall s b s' b',
    step_star s b s' b' ->
    step_star_Cs b b'.
  Proof.
    intros.
    eapply step_star_Cs_cons.
    apply step_star_simulates_cstep_star_empty_context; eauto.
    apply step_star_Cs_nil.
  Qed.

  Lemma step_star_Cs_seq_app: forall b b' t,
    step_star_Cs b b' ->
    step_star_Cs (b;; t) (b';; t).
  Proof.
    intros.
    induction H.
    apply step_star_Cs_nil.
    eapply step_star_Cs_cons; eauto.
    eapply cstep_star_seq_append; eauto.
  Qed.



Section IgnoringGlobalState.

  Definition ignores_global_state (b0: wterm):= forall s b s' b',
    step_star_Cs b0 b -> step s b s' b' -> s' = s /\ (forall s, step s b s b').

  Lemma step_star_ignores_global_state: forall s b s' b',
    ignores_global_state b ->
    step_star s b s' b' ->
    s_iter_star s s' /\ (forall s, step_star s b s b').
  Proof.
    intros.
    induction H0; intros.
    split. apply s_iter_nil. intros; apply step_nil.
    destruct IHstep_star.
    repeat intro.
    edestruct H; eauto.
    eapply step_star_Cs_app; eauto. eapply step_single_Cs; eauto.
    edestruct H; eauto.
    apply step_star_Cs_nil.
    subst s'.
    split; auto; intros.
    eapply step_cons; eauto.
    destruct IHstep_star.
    repeat intro.
    edestruct H; eauto.
    split; auto.
    eapply s_iter_cons; eauto.
  Qed.

  Lemma cstep_star_ignores_global_state: forall s C b s' C' b',
    ignores_global_state b ->
    cstep_star s C b s' C' b' ->
    cstep_star s C (gwemp:tm_skip) s' C' (gwemp:tm_skip) /\ (forall s, step_star s b s b').
  Proof.
    intros.
    induction H0; intros.
    split. apply cstep_nil. intros; apply step_nil.
    destruct IHcstep_star.
    repeat intro.
    edestruct H; eauto.
    eapply step_star_Cs_cons; eauto. eapply cstep_single_cstep_star; eauto.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'. 
    edestruct H; eauto.
    apply step_star_Cs_nil.
    subst s'.
    split; auto; intros.
    eapply step_cons; eauto.
    subst b'.
    split; auto.
    eapply cstep_cons; eauto.
    destruct IHcstep_star.
    repeat intro.
    edestruct H; eauto.
    split; auto.
    eapply cstep_iteration; eauto.
  Qed.

  Lemma step_star_Cs_ignores_global_state: forall s b b',
    ignores_global_state b ->
    step_star_Cs b b' ->
    step_star s b s b'.
  Proof.
    intros.
    revert s.
    induction H0; intros.
    apply step_nil.

    spec IHstep_star_Cs.
    repeat intro.
    eapply H; auto. eapply step_star_Cs_cons; eauto.
    edestruct cstep_star_ignores_global_state as [??]; eauto.
    eapply step_star_app; eauto.
  Qed.

  Lemma ignores_global_state_simpl: forall b,
    simple_step_wterm b ->
    state_independent_step_wterm b ->
    ignores_global_state b.
  Proof.
    repeat intro.
    edestruct step_star_Cs_step_inversion as [ [s0[s0'[b'0[?[??]]]]] | ?]; eauto.
    edestruct H as [? [? | [w' ?]]]; eauto; subst s0' b'0.
    false.
    apply step_star_Cs_skip_resolve in H4; subst.
    inverts H2.
    subst.
    edestruct H as [? [? | [w' ?]]]; eauto; subst s' b'.
    split; intros; eauto.
    split; intros; eauto.
  Qed.

  Lemma ignores_global_state_tm_assign: forall w x e,
    ignores_global_state (w: tm_assign x e).
  Proof.
    intros; apply ignores_global_state_simpl; repeat intro; inverts H; eauto.
    econstructor; eauto.
  Qed.

  Lemma ignores_global_state_tm_load: forall w x e,
    ignores_global_state (w: tm_load x e).
  Proof.
    intros; apply ignores_global_state_simpl; repeat intro; inverts H; eauto.
    econstructor; eauto.
  Qed.

  Lemma ignores_global_state_tm_store: forall w e1 e2,
    ignores_global_state (w: tm_store e1 e2).
  Proof.
    intros; apply ignores_global_state_simpl; repeat intro; inverts H; eauto.
    econstructor; eauto.
  Qed.

  Lemma ignores_global_state_seq: forall b1 t2,
    ignores_global_state b1 ->
    (forall w, ignores_global_state (w: t2)) ->
    ignores_global_state (b1;; t2).
  Proof.
    repeat intro.
    edestruct step_star_Cs_seq_inversion as [ [w'1[??]] | [?[??]] ]; eauto; clear H1.
    edestruct H0; eauto.
    subst.
    inverts H2.
    edestruct H; eauto.
    subst; split; intros; auto. apply st_seq1; eauto.
    split; intros; eauto.
    apply st_seq_skip.
  Qed.



End IgnoringGlobalState.