Require Import Machine.
Require Import ContextSemantics.
Require Import Bisimulation.
Require Import world_model.
Require Import WorldLogic.
Require Import HoareLogic.
Require Import msl.msl_standard.
Require Import LibTactics.
Require Import SfLib.

Section Definitions.
  Import world_model.Notation.
  Import world_model.Notation.U.
  Import Machine.Notation.
  Import WorldLogic.Notation.
  Import WorldLogic.Notation.U.
  Import WorldLogic.Tactics.
  Open Scope forces_pred.
  Open Scope entails_pred.
  Open Scope join_scope.

Require Import ParBisimulation.

  Definition terminates s wC b := forall C s' wC' C' b',
    context_world C wC ->
    context_world C' wC' ->
    cstep_star s C b s' C' b' ->
    exists s'' w'',
      step_star_sans_s_iter s' b' s'' (w'':tm_skip).

  Definition term_impl s wC b1 b2 := forall C s' wC' wC1' C' w1' b2',
    context_world C wC ->
    context_world C' wC' ->
    gjoin wC' w1' wC1' ->
    cstep_star s C (b1 || b2) s' C' (w1': tm_skip || b2') ->
    terminates s' wC1' b2'.

  Definition coterminate s wC b1 b2 :=
    term_impl s wC b1 b2 /\ term_impl s wC b2 b2.

  Definition termination_impl P1 t1 P2 t2:= forall s wC w1 w2,
    w1 |= P1 -> w2 |= P2 ->
    valid_machine_state wC s (w1: t1 || w2: t2) ->
    term_impl s wC (w1: t1) (w2: t2).




  Lemma has_total_endpoint_can_send: forall c w w' w_w',
    has_total_endpoint c (w_prod w') ->
    gjoin w w' w_w' ->
    can_send c w_w'.
  Proof.
    introv H H0.
    pose True as H1.
    intro H2.
    destruct H0 as [[? _]_].
    spec H0 c.
    simpl in H.
    red in H.
    simpl in H0.
    rewrite H2 in H0.
    inverts H0.
    rewrite H6 in H; auto.
    rewrite<- H5 in H; auto.
  Qed.

  Lemma has_total_endpoint_access_conflict: forall c ch1 ch2,
    has_total_endpoint c ch1 ->
    has_access c ch2 ->
    joins ch1 ch2 ->
    False.
  Proof.
    introv H H0 H1.
    destruct H1 as [ch ?].
    spec H1 c.
    red in H.
    simpl in H1.
    apply H0.
    inverts H1; auto.
    rewrite<- H3 in H; false.
    rewrite<- H2 in H.
    false.
    clear H2 H3 H4 H0.
    destruct x as [[[??]?]?].
    destruct y as [[[??]?]?].
    destruct z as [[[??]?]?].
    red in H5.
    rewrite identity_channel_sa_simpl in *.
    apply n2.
    apply join_com in H5.
    destruct H5 as [? _].
    simpl in *. subst.
    clear -H0.
    destruct H0 as [? _].
    rewrite Share.glb_top in H.
    subst; apply bot_identity.
  Qed.

  Lemma has_total_endpoint_send_conflict: forall c w1 w2,
    has_total_endpoint c (w_prod w1) ->
    can_send c w2 ->
    gjoins w1 w2 ->
    False.
  Proof.
    intros.
    eapply has_total_endpoint_access_conflict; eauto.
    destruct H1 as [w ?].
    eexists; apply H1.
  Qed.

  Lemma has_total_endpoint_recv_conflict: forall c w1 w2,
    has_total_endpoint c (w_cons w1) ->
    can_recv c w2 ->
    gjoins w1 w2 ->
    False.
  Proof.
    intros.
    eapply has_total_endpoint_access_conflict; eauto.
    destruct H1 as [w ?].
    eexists; apply H1.
  Qed.

  Lemma total_send_recv_can_send_recv: forall c w,
    total_send_recv c w ->
    can_send c w /\ can_recv c w.
  Proof.
    intros.
    destruct H; unfold has_total_endpoint in *.
    split; intro.
    rewrite H1 in H; auto.
    rewrite H1 in H0; auto.
  Qed.

  Lemma has_total_endpoint_sub: forall c ch1 ch,
    join_sub ch1 ch ->
    has_total_endpoint c ch1 ->
    has_total_endpoint c ch.
  Proof.
    unfold has_total_endpoint; intros.
    destruct H as [ch2 ?].
    spec H c.
    inverts H; auto.
    rewrite<- H2 in H0; false.
    false.
    rewrite<- H1 in H0.
    red in H4.
    destruct x as [[[??]?]?].
    destruct y as [[[??]?]?].
    destruct z as [[[??]?]?].
    apply join_com in H4.
    destruct H4.
    simpl in *.
    clear H1 H2 H3 e n0 e1 n4.
    rewrite identity_channel_sa_simpl in *.
    subst.
    destruct H.
    rewrite Share.glb_top in H. subst t0.
    apply n2.
    apply bot_identity.
  Qed.



  Lemma total_send_recv_sub: forall c w1 w,
    total_send_recv c w1 ->
    gjoin_sub w1 w ->
    total_send_recv c w.
  Proof.
    intros.
    destruct H0 as [w2 ?].
    destruct H as [??].
    split; eapply has_total_endpoint_sub.
    exists (w_prod w2); apply H0. auto.
    exists (w_cons w2); apply H0. auto.
  Qed.

  (* declares that channel c has resource invariant function R *)
  Definition s_has_ch_ri s c R:= exists wl vl,
    s_get_ch s c = Some (R, wl, vl).

  Lemma has_total_endpoint_super: forall c ch1 ch2 ch,
    has_total_endpoint c ch ->
    join ch1 ch2 ch ->
    ~has_access c ch2 ->
    has_total_endpoint c ch1.
  Proof.
    intros.
    red in H1. unfold has_access in H1.
    red in H. red.
    spec H0 c.
    case_eq (ch2 c); intros.
    false. apply H1. congruence.
    rewrite H2 in H0.
    inverts H0; auto.
    rewrite<- H6 in H; auto.
    rewrite H4.
    auto.
  Qed.

  Lemma can_send_super: forall c w1 w2 w,
    can_send c w ->
    gjoin w1 w2 w ->
    (can_send c w1 \/ can_send c w2).
Admitted.

  Lemma can_recv_super: forall c w1 w2 w,
    can_recv c w ->
    gjoin w1 w2 w ->
    (can_recv c w1 \/ can_recv c w2).
Admitted.


  Lemma terminates_now: forall wC s b,
    terminates s wC b ->
    exists s' w',
      step_star_sans_s_iter s b s' (w': tm_skip).
  Proof.
    intros.
    edestruct H with (C:=dummyContext wC) (C':=dummyContext wC) as [s'[w' ?]]; eauto.
    apply context_world_dummyContext.
    apply context_world_dummyContext.
    apply cstep_nil.
  Qed.

  Lemma terminates_s_iter: forall wC s b s',
    terminates s wC b ->
    s_iter s s' ->
    terminates s' wC b.
  Proof.
    unfold terminates; intros.
    edestruct H with (C:=C) as [s''[w'' ?]]; eauto.
    eapply cstep_iteration; eauto.
  Qed.





  Lemma terminates_step: forall wC s b s' b',
    terminates s wC b ->
    step s b s' b' ->
    terminates s' wC b'.
  Proof.
    unfold terminates; intros.
    edestruct H with (C:=C) as [s''[w'' ?]]; eauto.
    eapply step_cstep_star; eauto.
  Qed.

  Lemma terminates_step_star: forall wC s b s' b',
    terminates s wC b ->
    step_star s b s' b' ->
    terminates s' wC b'.
  Proof.
    unfold terminates; intros.
    edestruct H with (C:=C) as [s''[w'' ?]]; eauto.
    eapply cstep_star_app; eauto.
    eapply step_star_add_context_cstep_star; eauto.
  Qed.

  Lemma terminates_cstep: forall s wC C b s' C' b',
    terminates s wC b ->
    cstep s C b s' C' b' ->
    context_world C wC ->
    valid_machine_state wC s b ->
    exists wC',
      context_world C' wC' /\ terminates s' wC' b'.
  Proof.
    unfold terminates; intros.
    edestruct valid_machine_state_cstep as [wC'[??]]; eauto.
    exists wC'; split; auto; intros.
    edestruct compose_context_world_cstep_star
      with (C1:=C) (C1':=C') (C2:=C0) as [wC''[C12[C12'[?[?[??]]]]]]; eauto.
    apply cstep_single_cstep_star; auto.
  Qed.

  Lemma terminates_cstep': forall s wC C b s' wC' C' b',
    terminates s wC b ->
    cstep s C b s' C' b' ->
    context_world C wC ->
    context_world C' wC' ->
    valid_machine_state wC s b ->
    terminates s' wC' b'.
  Proof.
    intros.
    edestruct terminates_cstep as [wC'_[??]]; eauto.
    replace wC'_ with wC' in *; auto.
    eapply context_world_eq; eauto.
  Qed.

  Lemma terminates_cstep_star: forall s wC C b s' C' b',
    terminates s wC b ->
    cstep_star s C b s' C' b' ->
    context_world C wC ->
    valid_machine_state wC s b ->
    exists wC',
      context_world C' wC' /\ terminates s' wC' b'.
  Proof.
    unfold terminates; intros.
    edestruct valid_machine_state_cstep_star as [wC'[??]]; eauto.
    exists wC'; split; auto; intros.
    edestruct compose_context_world_cstep_star
      with (C1:=C) (C1':=C') (C2:=C0) as [wC''[C12[C12'[?[?[??]]]]]]; eauto.
  Qed.

  Lemma terminates_cstep_star': forall s wC C b s' wC' C' b',
    terminates s wC b ->
    cstep_star s C b s' C' b' ->
    context_world C wC ->
    context_world C' wC' ->
    valid_machine_state wC s b ->
    terminates s' wC' b'.
  Proof.
    intros.
    edestruct terminates_cstep_star as [wC'_[??]]; eauto.
    replace wC'_ with wC' in *; auto.
    eapply context_world_eq; eauto.
  Qed.

  Lemma terminates_step_starO: forall s wO bO b s' bO',
    terminates s wO b ->
    step_star s bO s' bO' ->
    join_wterms bO wO ->
    valid_machine_state wO s b ->
    exists wO',
      join_wterms bO' wO' /\ terminates s' wO' b.
  Proof.
    unfold terminates; intros.
    edestruct terminates_cstep_star
      with (C:=cx_par1 cx_hole bO) (C':=cx_par1 cx_hole bO') (b:=b) (b':=b)
      as [wO'[??]]; eauto.
    apply cstep_star_par_cx_par1.
    apply cstep_star_par2_step.
    apply step_star_add_context_cstep_star; eauto.
    exists gwemp wO; splits; auto.
    reflexivity. apply gjoin_bot2.
    destruct H3 as [gwemp_[wO'_[?[??]]]].
    simpl in H3; subst gwemp_.
    apply gworld_unit_identity in H6; subst wO'_; eauto.
  Qed.

Import StepStar.

  Lemma tm_simulates_wtm: forall (b:wterm) w,
    join_wterms b w ->
    forall s,
      step_star s (w:just_term b) s b.
  Proof.
    intros.
    revert w H.
    induction b; intros.
    inverts H.
    simpl; eapply step_nil.

    eapply step_cons. apply st_seq.
    apply step_star_seq_app; auto.

    destruct H as [w1[w2[?[??]]]].
    simpl.
    destruct (join_wterms_dec b1) as [[w1_ H2]|].
    replace w1_ with w1 in *.
    destruct (join_wterms_dec b2) as [[w2_ H3]|].
    replace w2_ with w2 in *.
    eapply step_cons.
    eapply st_par; eauto.
    apply TT_gstar; reflexivity.
    apply TT_gstar; reflexivity.
    eapply step_star_app.
    apply step_star_par1_step; eauto.
    apply step_star_par2_step; eauto.
    eapply join_wterms_eq; eauto.
    false. eapply n; eauto.
    eapply join_wterms_eq; eauto.
    false. eapply n; eauto.
  Qed.

  Lemma cstep_par_decouple_context: forall s C b s' C' b',
    cstep s C b s' C' b' ->
    cstep s (cx_par1 cx_hole (set_context C (gwemp:tm_skip))) b s' (cx_par1 cx_hole (set_context C' (gwemp:tm_skip))) b'.
  Proof.
    intros.
    edestruct cstep_inversion as [[?]|[?]]; eauto.
    subst C'.
    apply step_cstep; auto.
    subst b'.
    apply cst_par1_step2.
    apply step_simulates_cstep; auto.
  Qed.

  Lemma cstep_star_par_decouple_context: forall s C b s' C' b',
    cstep_star s C b s' C' b' ->
    cstep_star s (cx_par1 cx_hole (set_context C (gwemp:tm_skip))) b s' (cx_par1 cx_hole (set_context C' (gwemp:tm_skip))) b'.
  Proof.
    intros.
    induction H.
    apply cstep_nil.
    eapply cstep_cons; eauto.
    apply cstep_par_decouple_context; auto.
    eapply cstep_iteration; eauto.
  Qed.

  Lemma terminates_step_starO_skip: forall s wO bO b s' wO',
    terminates s wO b ->
    step_star s bO s' (wO': tm_skip) ->
    join_wterms bO wO ->
    terminates s' wO' b.
  Proof.
    unfold terminates; intros.
    remember (just_term (set_context C (gwemp:tm_skip))) as tC.
    eapply H with (C:=cx_par1 cx_hole (bO;; tC)) (C':=cx_par1 cx_hole (set_context C' (gwemp:tm_skip))); eauto.
    exists gwemp wO; splits; auto. reflexivity. apply gjoin_bot2.
    exists gwemp wC'; splits; auto. reflexivity.
    apply context_world_join_wterms; auto. apply gjoin_bot2.
    eapply cstep_star_app.
    apply step_star_observer_context'.
    apply step_star_par2_step; eauto.
    apply step_star_seq_app; eauto.
    eapply cstep_star_app.
    apply step_star_observer_context'.
    apply step_star_par2_step.
    eapply step_cons.
    apply st_seq_skip.
    subst tC.
    eapply tm_simulates_wtm.
    apply context_world_join_wterms; auto.
    apply cstep_star_par_decouple_context; auto.
  Qed.


  Lemma terminates_stepO: forall s wO bO b s' bO',
    terminates s wO b ->
    step s bO s' bO' ->
    join_wterms bO wO ->
    valid_machine_state wO s b ->
    exists wO',
      join_wterms bO' wO' /\ terminates s' wO' b.
  Proof.
    intros; eapply terminates_step_starO; eauto.
    apply step_single; auto.
  Qed.

  Lemma terminates_step_starO': forall s wO bO b s' wO' bO',
    terminates s wO b ->
    step_star s bO s' bO' ->
    join_wterms bO wO ->
    join_wterms bO' wO' ->
    valid_machine_state wO s b ->
    terminates s' wO' b.
  Proof.
    intros.
    edestruct terminates_step_starO as [wO'_[??]]; eauto.
    replace wO'_ with wO' in *; auto.
    eapply join_wterms_eq; eauto.
  Qed.

  Lemma terminates_stepO': forall s wO bO b s' wO' bO',
    terminates s wO b ->
    step s bO s' bO' ->
    join_wterms bO wO ->
    join_wterms bO' wO' ->
    valid_machine_state wO s b ->
    terminates s' wO' b.
  Proof.
    intros.
    eapply terminates_step_starO'; eauto.
    apply step_single; auto.
  Qed.

  Lemma valid_machine_state_wterm': forall wO s b1 b2,
    valid_machine_state wO s b1 ->
    (forall wb, join_wterms b1 wb -> join_wterms b2 wb) ->
    valid_machine_state wO s b2.
  Proof.
    intros.
    destruct H as [wS[wb[ww[?[??]]]]].
    exists wS wb ww; splits; eauto.
  Qed.


  Lemma valid_machine_state_join_wterms: forall b wC s,
    valid_machine_state wC s b ->
    exists wb, join_wterms b wb.
  Proof.
    introv [_[wb[_[_[? _]]]]]; eauto.
  Qed.

  Lemma term_impl_step: forall s wC b1 b2 s' b1' b2',
    term_impl s wC b1 b2 ->
    step s (b1 || b2) s' (b1' || b2') ->
    term_impl s' wC b1' b2'.
  Proof.
    unfold term_impl; intros.
    eapply H; eauto.
    eapply step_cstep_star; eauto.
  Qed.

  Lemma term_impl_s_iter: forall s wO b1 b2 s',
    term_impl s wO b1 b2 ->
    s_iter s s' ->
    term_impl s' wO b1 b2.
  Proof.
    unfold term_impl; intros.
    eapply H; eauto.
    eapply cstep_iteration; eauto.
  Qed.

  Lemma term_impl_cstep: forall s wC C b1 b2 s' C' b1' b2',
    term_impl s wC b1 b2 ->
    context_world C wC ->
    cstep s C (b1 || b2) s' C' (b1' || b2') ->
    valid_machine_state wC s (b1||b2) ->
    exists wC',
      context_world C' wC' /\ term_impl s' wC' b1' b2'.
  Proof.
    unfold term_impl; intros.
    edestruct valid_machine_state_cstep as [wC'[??]]; eauto.
    exists wC'; split; auto.
    intros.
    edestruct compose_context_world_cstep_star
      with (C1:=C) (C1':=C') (C2:=C0) as [wC''[C12[C12'[?[?[??]]]]]]; eauto.
    apply cstep_single_cstep_star; auto.
    replace wC'0 with wC'' in *.
    clear H11.
    eapply H; eauto.
    eapply context_world_eq; eauto.
  Qed.

  Lemma term_impl_step_star: forall s wC b1 b2 s' b1' b2',
    term_impl s wC b1 b2 ->
    step_star s (b1 || b2) s' (b1' || b2') ->
    term_impl s' wC b1' b2'.
  Proof.
    unfold term_impl; intros.
    eapply H; eauto.
    eapply cstep_star_app; eauto.
    eapply step_star_add_context_cstep_star; eauto.
  Qed.

  Lemma term_impl_cstep_star: forall s wC C b1 b2 s' C' b1' b2',
    term_impl s wC b1 b2 ->
    context_world C wC ->
    cstep_star s C (b1 || b2) s' C' (b1' || b2') ->
    valid_machine_state wC s (b1||b2) ->
    exists wC',
      context_world C' wC' /\ term_impl s' wC' b1' b2'.
  Proof.
    unfold term_impl; intros.
    edestruct valid_machine_state_cstep_star as [wC'[??]]; eauto.
    exists wC'; split; auto.
    intros.
    edestruct compose_context_world_cstep_star
      with (C1:=C) (C1':=C') (C2:=C0) as [wC''[C12[C12'[?[?[??]]]]]]; eauto.
    replace wC'0 with wC'' in *.
    clear H11.
    eapply H; eauto.
    eapply context_world_eq; eauto.
  Qed.

  Lemma term_impl_stepO: forall s wO bO b1 b2 s' bO',
    term_impl s wO b1 b2 ->
    join_wterms bO wO ->
    step s bO s' bO' ->
    valid_machine_state wO s (b1||b2) ->
    exists wO',
      join_wterms bO' wO' /\ term_impl s' wO' b1 b2.
  Proof.
    intros.
    edestruct term_impl_cstep with (C:=cx_par1 cx_hole bO) as [wO'[??]]; eauto.
    apply context_world_join_wterms; eauto.
    exists gwemp wO; splits; eauto.
    reflexivity. apply gjoin_bot2.
    apply cst_par1_step2; eauto.
    destruct H3 as [gwemp_[wO'_[?[??]]]].
    simpl in H3; subst.
    eapply gworld_unit_identity in H6.
    subst; eauto.
  Qed.

  Lemma term_impl_step_starO: forall s wO bO b1 b2 s' bO',
    term_impl s wO b1 b2 ->
    join_wterms bO wO ->
    step_star s bO s' bO' ->
    valid_machine_state wO s (b1||b2) ->
    exists wO',
      join_wterms bO' wO' /\ term_impl s' wO' b1 b2.
  Proof.
    intros.
    edestruct term_impl_cstep_star with (C:=cx_par1 cx_hole bO) as [wO'[??]]; eauto.
    apply context_world_join_wterms; eauto.
    exists gwemp wO; splits; eauto.
    reflexivity. apply gjoin_bot2.
    apply step_star_observer_context'.
    apply StepStar.step_star_par2_step; eauto.
    destruct H3 as [gwemp_[wO'_[?[??]]]].
    simpl in H3; subst.
    eapply gworld_unit_identity in H6.
    subst; eauto.
  Qed.

  Lemma term_impl_stepO': forall s wO bO b1 b2 s' wO' bO',
    term_impl s wO b1 b2 ->
    join_wterms bO wO ->
    join_wterms bO' wO' ->
    step s bO s' bO' ->
    valid_machine_state wO s (b1||b2) ->
    term_impl s' wO' b1 b2.
  Proof.
    intros.
    edestruct term_impl_stepO as [wO'_[??]]; eauto.
    replace wO'_ with wO' in *; auto.
    eapply join_wterms_eq; eauto.
  Qed.

  Lemma term_impl_cstep_star': forall s wC C b1 b2 s' wC' C' b1' b2',
    term_impl s wC b1 b2 ->
    context_world C wC ->
    context_world C' wC' ->
    cstep_star s C (b1 || b2) s' C' (b1' || b2') ->
    valid_machine_state wC s (b1||b2) ->
    term_impl s' wC' b1' b2'.
  Proof.
    intros.
    edestruct term_impl_cstep_star as [wC'_[??]]; eauto.
    replace wC'_ with wC' in *; auto.
    eapply context_world_eq; eauto.
  Qed.



  Lemma step_context_inv: forall s C b s' Cb',
    step s (set_context C b) s' Cb' ->
    (exists b',
      step s b s' b' /\ set_context C b' = Cb') \/
    (exists C',
      step s (set_context C (gwemp:tm_skip)) s' (set_context C' (gwemp:tm_skip)) /\
      Cb' = (set_context C' b)) \/
    (exists w,
      b = w:tm_skip)%wterm.
  Proof.
    intros.
    revert s b s' Cb' H.
    induction C; intros; simpl in *.
    - left; eauto.
    - inverts H.
    edestruct IHC as [ [b'[??]] | [ [C'[??]] | [w ?] ]]; clear IHC; eauto; subst b1'.
    left; exists b'; split; simpl; auto.
    right; left; exists (cx_seq C' t); split; simpl; auto. apply st_seq1; auto.
    destruct C; simpl in H1; try discriminate.
    subst b; eauto.
    - rename w into b2.
    inverts H.
    edestruct IHC as [ [b'[??]] | [ [C'[??]] | [w ?] ]]; clear IHC; eauto; subst b1'.
    left; exists b'; split; simpl; auto.
    right; left; exists (cx_par1 C' b2); split; simpl; auto. apply st_par1; auto.
    right; left; exists (cx_par1 C b2'); split; simpl; auto. apply st_par2; auto.
    destruct C; simpl in H0; try discriminate.
    subst b; eauto.
    - rename w into b1.
    inverts H.
    right; left; exists (cx_par2 b1' C); split; simpl; auto. apply st_par1; auto.
    edestruct IHC as [ [b'[??]] | [ [C'[??]] | [w ?] ]]; clear IHC; eauto; subst b2'.
    left; exists b'; split; simpl; auto.
    right; left; exists (cx_par2 b1 C'); split; simpl; auto. apply st_par2; auto.
    destruct C; simpl in H2; try discriminate.
    subst b; eauto.
  Qed.


(*
  Lemma terminates_tm_endch_inv: forall s wO w ec c C,
    terminates s wO (set_context C (w: tm_endch ec)) ->
    cannot_recv_channel s (set_context C (w: tm_endch ec)) c ->
    eval ec (cname_value c) w ->
    exists R,
      s_get_ch s c = Some(R,nil,nil) /\ total_send_recv c w.
  Proof.
Admitted.
*)
(*
    intros.
    edestruct terminates_now as [s'[w' ?]]; eauto.
    clear wO H.
    remember (set_context C (w: tm_endch ec)) as b.
    remember (w':tm_skip)%wterm as b'.
    revert w w' C Heqb Heqb' H0.
    induction H1; intros; subst.
    destruct C; simpl in Heqb'; discriminate.
    rename b' into bb'.
    edestruct step_context_inv as [ [b'[??]] | [ [C'[??]] | [w_ ?] ]]; eauto; try subst bb'.
    - inverts H2.
    assert (cname_value c0 = cname_value c).
      eapply eval_eq; eauto.
      eapply eval_gframe; eauto.
    inverts H2.
    exists R; split; auto.
    eapply total_send_recv_sub; eauto. exists w0; auto.
    - edestruct IHstep_star_sans_s_iter as [R[??]]; eauto.
    exists 

    inverts H; eauto.
    exists c R; splits; auto.
    eapply eval_gframe; eauto.
    eapply total_send_recv_sub; eauto.
    exists w0; auto.
  Qed.
  *)  

(*
  Lemma terminates_tm_endch_inv: forall s wC w ec,
    terminates s wC (w: tm_endch ec) ->
    exists c R,
      eval ec (cname_value c) w /\ s_get_ch s c = Some(R,nil,nil) /\ total_send_recv c w.
  Proof.
    intros.
    edestruct terminates_now as [s'[w' ?]]; eauto.
    clear wC H.
    remember (w: tm_endch ec)%wterm as b.
    remember (w':tm_skip)%wterm as b'.
    revert Heqb Heqb'.
    induction H0; intros; subst.
    discriminate.
    inverts H; eauto.
    exists c R; splits; auto.
    eapply eval_gframe; eauto.
    eapply total_send_recv_sub; eauto.
    exists w0; auto.
  Qed.
*)

  Lemma w_env_share_wupd_cons: forall w ch,
    w_env_share (wupd_cons w ch) = w_env_share w.
  Proof.
    intros.
    destruct w as [w_p w_c w_e w_h].
    reflexivity.
  Qed.

(*
  Lemma empty_channel_queue': forall s c w x R wl vl,
    s_get_ch s c = Some (R, wl, vl) ->
    can_recv c w ->
    w_env_share w x = Share.top ->
    exists t s' w' wl' vl',
      step_star s (w: t) s' (w': tm_skip) /\
      s_get_ch s' c = Some (R, wl', vl') /\
      (wl'=nil\/vl'=nil).
  Proof.
    intros.
    revert s w vl H H0 H1.
    induction wl; intros.
    exists tm_skip s w (@nil world) vl; split; eauto.
    apply step_nil.
    destruct vl.
    exists tm_skip s w (a::wl) (@nil value); split; eauto.
    apply step_nil.
    assert (gjoins w a).
admit.
    destruct H2 as [w_w' ?].
    destruct (IHwl (s_pop_ch s c) (w_cons_incr c (w_env_set w_w' x v)) vl)
      as [t[s''[w''[wl''[vl''[?[??]]]]]]]; auto.
    eapply s_ch_get_pop_same; eauto.
    apply-> can_recv_w_cons_incr.
    apply-> can_recv_upd_var.
    eapply can_recv_sub; eauto.
    exists a; auto.
    unfold w_cons_incr.
    rewrite w_env_share_wupd_cons.
    rewrite w_env_share_set.
    eapply w_env_share_sub_top; eauto.
    exists (tm_recv x (expr_value (cname_value c));;; t)%term.
    do 4 eexists; splits.
    eapply step_cons.
    apply st_seq.
    eapply step_cons.
    apply st_seq1. eapply st_recv; eauto.
    reflexivity.
admit.
    eapply step_cons; eauto.
    apply st_seq_skip.
    eauto.
    auto.
  Qed.
*)
  Lemma w_env_share_new_local_same: forall x v,
    w_env_share (new_local x v) x = Share.top.
  Proof. intros. unfold new_local, w_env_share. apply LocalsMap.singleton_spec. Qed.

(*
  Lemma empty_channel_queue: forall s c w R wl vl,
    s_get_ch s c = Some (R, wl, vl) ->
    can_recv c w ->
    exists t s' w' wl' vl',
      step_star s (w: t) s' (w': tm_skip) /\
      s_get_ch s' c = Some (R, wl', vl') /\
      (wl'=nil\/vl'=nil).
  Proof.
    intros.
    remember (fresh_var w (nat_value 0)) as fresh_x.
    destruct (fresh_var_local_props w (nat_value 0) fresh_x) as [??]; auto.
    edestruct empty_channel_queue' with (w:=FreshLocals.fl_w fresh_x) (x:=FreshLocals.fl_x fresh_x) as [t[s'[w'[wl'[vl'[?[??]]]]]]]; eauto.
    eapply can_recv_sub; eauto. eexists; eauto.
    eapply w_env_share_sub_top.
    2: apply gjoin_com; eauto.
    apply w_env_share_new_local_same.
    exists (tm_local (expr_value (nat_value 0)) (fun x => t)).
    exists s' w' wl' vl'; splits; auto.
    eapply step_cons.
    eapply st_local; eauto.
    reflexivity.
    auto.
  Qed.

  Lemma terminate_tm_recv_inv: forall wO s w x ec C c,
    terminates s wO (set_context C (w: tm_recv x ec)) ->
    cannot_recv_channel s (set_context C (w: tm_recv x ec)) c ->
    eval ec (cname_value c) w ->
    exists R wc wl v vl,
      s_get_ch s c = Some(R,wc::wl,v::vl).
  Proof.
Admitted.
*)
(*
    intros.
    edestruct terminates_now as [s'[w' ?]]; eauto.
    clear -H0.
    remember (w: tm_recv x ec)%wterm as b.
    remember (w':tm_skip)%wterm as b'.
    revert Heqb Heqb'.
    induction H0; intros; subst.
    discriminate.
    inverts H; eauto.
    exists c R w'0 wl v q; split; auto.
  Qed.

  Lemma terminate_tm_recv_inv: forall s w x ec wC,
    terminates s wC (w: tm_recv x ec) ->
    exists c R wc wl v vl,
      eval ec (cname_value c) w /\ s_get_ch s c = Some(R,wc::wl,v::vl).
  Proof.
    intros.
    edestruct terminates_now as [s'[w' ?]]; eauto.
    clear -H0.
    remember (w: tm_recv x ec)%wterm as b.
    remember (w':tm_skip)%wterm as b'.
    revert Heqb Heqb'.
    induction H0; intros; subst.
    discriminate.
    inverts H; eauto.
    exists c R w'0 wl v q; split; auto.
  Qed.
*)
(*
  Lemma terminates_can_recv_conflict: forall s wO C w1 x ec c,
    terminates s wO (set_context C (w1: tm_recv x ec))%wterm ->
    cannot_recv_channel s (set_context C (w1: tm_recv x ec)) c ->
    eval ec (cname_value c) w1 ->
    can_recv c wO ->
    False.
  Proof.
    intros.
    case_eq (s_get_ch s c); intros.
    destruct c0 as [[R wl]vl].
    edestruct empty_channel_queue as [t[s'[w'[wl'[vl'[?[??]]]]]]]; eauto.
    edestruct terminate_tm_recv_inv with (s:=s') as [R_[wc_[wl_[v_[vl_ ?]]]]]; eauto.
    eapply terminates_step_starO_skip; eauto. reflexivity.
    eapply cannot_recv_channel_step_starO; eauto.
    rewrite H5 in H7.
    inverts H7.
    destruct H6; discriminate.
    edestruct terminate_tm_recv_inv as [R_[wc_[wl_[v_[vl_ ?]]]]]; eauto.
    rewrite H3 in H4; discriminate.
  Qed.
*)

(*
  Lemma terminates_can_recv_conflict: forall s wC w1 x ec c,
    terminates s wC (w1: tm_recv x ec)%wterm ->
    eval ec (cname_value c) w1 ->
    can_recv c wC ->
    False.
  Proof.
    intros.
    case_eq (s_get_ch s c); intros.
    destruct c0 as [[R wl]vl].
    edestruct empty_channel_queue as [t[s'[w'[wl'[vl'[?[??]]]]]]]; eauto.
    eapply terminates_step_starO_skip in H3; eauto.
    edestruct terminate_tm_recv_inv as [c_[R_[wc_[wl_[v_[vl_[??]]]]]]]; eauto.
    assert (cname_value c_ = cname_value c).
      eapply eval_eq; eauto.
    inverts H8.
    rewrite H4 in H7.
    inverts H7.
    destruct H5; discriminate.
    reflexivity.
    edestruct terminate_tm_recv_inv as [c_[R_[wc_[wl_[v_[vl_[??]]]]]]]; eauto.
    assert (cname_value c_ = cname_value c).
      eapply eval_eq; eauto.
    inverts H5.
    rewrite H2 in H4; discriminate.
  Qed.
*)

(*
  Lemma terminates_sub: forall s wC wC' b,
    terminates s wC b ->
    gjoin_sub wC' wC ->
    valid_machine_state wC s b ->
    terminates s wC' b.
  Proof.
    unfold terminates; intros.
    destruct H0 as [w' ?].
    eapply H with (C:=cx_par1 C (w':tm_skip)) (C':=cx_par1 C' (w':tm_skip)); eauto.
    exists wC' w'; splits; auto.  reflexivity.
    exists wC'0 w'; splits; auto.  reflexivity.
    ea

  Qed.
*)

  Lemma step_star_sans_s_iter_seq_inversion: forall s b1 t2 s'' b'',
    step_star_sans_s_iter s (b1;; t2) s'' b'' ->
    (exists s' w',
      step_star_sans_s_iter s b1 s' (w': tm_skip) /\
      step_star_sans_s_iter s' (w': t2) s'' b'') \/
    (exists b1'', step_star_sans_s_iter s b1 s'' b1'' /\ b'' = (b1'';; t2)%wterm).
  Proof.
    intros.
    remember (b1;; t2)%wterm as b.
    revert b1 Heqb.
    induction H; intros; subst.
    right; exists b1; split; auto. apply step_nil_.
    inverts H.
    edestruct IHstep_star_sans_s_iter as [ [s'0[w'0[??]]] | [b1''[??]] ]; eauto.
    left; exists s'0 w'0; split; auto.
    eapply step_cons_; eauto.
    subst b''.
    right; exists b1''; split; auto.
    eapply step_cons_; eauto.
    left; exists s' w; split; auto.
    apply step_nil_.
  Qed.

  Lemma terminates_seq1: forall s wC b t,
    terminates s wC (b;; t) ->
    terminates s wC b.
  Proof.
    unfold terminates; intros.
    edestruct H with (b':=(b';;t)%wterm) as [s''[w'' ?]]; eauto.
    eapply cstep_star_seq_append; eauto.
    edestruct step_star_sans_s_iter_seq_inversion as [ [s'0[w'0[??]]] | [b''[??]] ]; eauto.
    discriminate.
  Qed.

  Lemma terminates_seq2: forall s wC C b s' wC' C' w' t,
    terminates s wC (b;; t) ->
    context_world C wC ->
    context_world C' wC' ->
    cstep_star s C b s' C' (w': tm_skip) ->
    valid_machine_state wC s (b;; t) ->
    terminates s' wC' (w':t).
  Proof.
    intros.
    eapply terminates_cstep_star'; eauto.
    eapply cstep_star_seq_join; eauto.
    apply cstep_nil.
  Qed.

  Lemma term_impl_seq_appL: forall s wC b1 t b2,
    term_impl s wC b1 b2 ->
    valid_machine_state wC s (b1||b2) ->
    term_impl s wC (b1;; t) b2.
  Proof.
    unfold term_impl; intros.
    edestruct cstep_star_par_seq_inversion as [ [b2'_[b1'[??]]] | [s'0[C'0[b2'0[w1'0[??]]]]] ]; eauto.
    apply cstep_star_par_swap; eauto.
    discriminate.
    clear H4.
    apply cstep_star_par_swap in H5.
    apply cstep_star_par_swap in H6.
    edestruct valid_machine_state_cstep_star with (s:=s) as [wC'0[??]]; eauto.
    edestruct valid_machine_state_par_observer' as [wC2'1[wC'1[?[??]]]].
    apply valid_machine_state_par_swap; eauto.
    simpl in H8; subst wC2'1.
    eapply H with (wC':=wC'0) in H5; eauto.
    clear H.
    red in H5.
    eapply terminates_cstep_star' with (C:=cx_par1 C'0 (w1'0:t)) (C':=cx_par1 C' (w1':tm_skip)); eauto.
    apply cstep_star_par_cx_par1.
    apply cstep_star_par_swap; eauto.
    exists wC'0 w1'0; splits; auto. reflexivity.
    exists wC' w1'; splits; auto. reflexivity.
  Qed.

  Lemma term_impl_seq_appR: forall s wC b1 t b2,
    term_impl s wC b1 (b2;; t) ->
    term_impl s wC b1 b2.
  Proof.
    unfold term_impl; intros.
    eapply terminates_seq1; eauto.
    eapply H; eauto.
    apply cstep_star_par2_inner_seq_append; eauto.
  Qed.






End Definitions.
(*
Module Notation.
  Delimit Scope term with term_scope.
  Delimit Scope pred with pred.
  Module U.
    Notation "A ⊢ 〖 P1 〗 t1 '⇊' 〖 P2 〗 t2" := (term_impl A%pred P1%pred t1%term P2%pred t2%term)
      (at level 55, no associativity, format  "A ⊢ 〖 P1 〗  t1  '⇊'  〖 P2 〗  t2") : term_scope.
(*    Notation "A ⊢ 〖 G1 ; P1 〗 t1 '⇈' 〖 G2 ; P2 〗 t2" := (diverge_impl A%pred G1 P1%pred t1%term G2 P2%pred t2%term)
      (at level 55, no associativity, format  "A ⊢ 〖 G1 ; P1 〗  t1  '⇈'  〖 G2 ; P2 〗  t2") : term_scope.*)
    Notation "A ⊢ 〖 P1 〗 t1 '↕↕' 〖 P2 〗 t2" := (coterminate A%pred P1%pred t1%term P2%pred t2%term)
      (at level 55, no associativity, format  "A ⊢ 〖 P1 〗  t1  '↕↕'  〖 P2 〗  t2") : term_scope.
    Notation " ⊢ 〖 P1 〗 t1 ⇊ 〖 P2 〗 t2" := (term_impl TT P1%pred t1%term P2%pred t2%term)
      (at level 55, no associativity, format  "⊢ 〖 P1 〗  t1  ⇊  〖 P2 〗  t2") : term_scope.
(*    Notation " ⊢ 〖 G1 ; P1 〗 t1 '⇈' 〖 G2 ; P2 〗 t2" := (diverge_impl TT G1 P1%pred t1%term G2 P2%pred t2%term)
      (at level 55, no associativity, format  "⊢ 〖 G1 ; P1 〗  t1  '⇈'  〖 G2 ; P2 〗  t2") : term_scope.*)
    Notation " ⊢ 〖 P1 〗 t1 '↕↕' 〖 P2 〗 t2" := (coterminate TT P1%pred t1%term P2%pred t2%term)
      (at level 55, no associativity, format  "⊢ 〖 P1 〗  t1  '↕↕'  〖 P2 〗  t2") : term_scope.
  End U.
End Notation.
*)

(*
Section Lemmas.
  Import world_model.Notation.
  Import world_model.Notation.U.
  Import Machine.Notation.
  Import HoareLogic.Notation.
  Import HoareLogic.Notation.U.
  Import Notation.
  Import Notation.U.
  Import ContextSemantics.Notation.
  Import WorldLogic.Notation.
  Import WorldLogic.Notation.U.
  Import WorldLogic.Tactics.
  Import world_model.Morphisms.

  Open Scope htriple_scope.
  Open Scope context_scope.
  Open Scope term_scope.
  Open Scope join_scope.
  Open Scope forces_pred.
  Open Scope entails_pred.

  Lemma term_impl_seq_appL: forall A P1 t1 P2 t2 t3,
    term_impl A P1 t1 P2 t3 ->  
    term_impl A P1 (tm_seq t1 t2) P2 t3.
  Proof.
    intros.
    unfold term_impl, terminates in *.
    introv Hw12 Hww HP1 HP2 H0 H1.

    apply cstep_star_par_swap in H0.
    edestruct cstep_star_term_par_seq_inversion as [?|[?|?]]; eauto.
    destruct H2 as [b1'_[b2'00[??]]]. inverts H3.
    destruct H2 as [s'1[C'1[b1'1[w2'[??]]]]].
    clear H0.
    eapply H with (w1':=w2'); eauto.
    apply cstep_star_par_swap. eauto.
    eapply cstep_star_par2_inner_tm_seq_append with (t3:=t0) in H3.
    eapply cstep_star_app; eauto.
    eapply step_cstep_star; eauto.
    apply st_par2. apply st_seq_skip.

    destruct H2 as [?[??]]; subst. inverts H4.
  Qed.


(*
Lemma cstep_star_compose_weaken_context': forall s C b s' C' b' C2,
  cstep_star s C b s' C' b' ->
  cstep_star s (compose_contexts C C2) b s' (compose_contexts C' C2) b'.
Admitted.


Lemma term_impl_equiv: forall A G1 P1 t1 G2 P2 t2,
  term_impl A G1 P1 t1 G2 P2 t2 -> term_impl_prepend_seq A G1 P1 t1 G2 P2 t2.
Proof.
  introv H.
  intros s C w1 w2 bb s' C' b1' s'' C'' w1' b2'' He Hww Henv1 Henv2 H1 H2 H3 H4.
  intros t3 s''' C''' b2''' b3''' H5.
  apply cstep_star_par_swap in H5.
  move H at bottom.
  unfold term_impl in H.
  assert (cstep_star s (compose_contexts C (cx_par1 cx_hole bb)) (w1:t1||w2:t2) s'' (compose_contexts C'' (cx_par1 cx_hole (w2:tm_skip))) (w1':tm_skip||b2''))%wterm.
    eapply cstep_star_app.
    apply cstep_star_par1_step.
    apply cstep_star_par_compose_contexts1; eauto.
    apply cstep_star_compose_weaken_context'; auto.
  apply H in H0; auto; clear H.
  unfold terminates in H0.
  destruct H0 with t3 s''' (compose_contexts C''' (cx_par1 cx_hole (w2 : tm_skip))) b2''' b3''' as [s''''[C''''[w2'''' ?]]].
  apply cstep_star_par_compose_contexts1; eauto.
  apply cstep_star_par1_step.
  apply cstep_star_par_swap.
  apply H5.
Lemma cstep_star_compose_contexts_skip2: forall s C w b s' C' b',
  cstep_star s (compose_contexts C (cx_par1 cx_hole (w:tm_skip))) b s' C' b' ->
  exists C'0,
    cstep_star s C b s' C'0 b'.
Proof.
  intros. remember (compose_contexts C (cx_par1 cx_hole (w : tm_skip))) as C_.
  revert C w HeqC_.
  induction H; intros; subst.
  exists C0; apply cstep_nil.
  edestruct cstep_composed_context_inversion as [?|[?|?]]; eauto.
  destruct H1 as [C'0[?[??]]]. subst.
  destruct IHcstep_star with C'0 w as [C''0 H2].
  reflexivity.
  exists C''0. eapply cstep_cons; eauto. apply cstep_skip_subst with gwemp; auto.
  destruct H1 as [C'0[?[??]]]. subst.
  elimtype False. clear -H1.
  inversion H1; subst.
  inversion H5; subst; simpl in H; inversion H.
  inversion H5.
  simpl in H. inversion H; subst; inversion H4.
  destruct H1 as [??]. subst.
  destruct IHcstep_star with C0 w as [C''0 H2].
  reflexivity.
  exists C''0. eapply step_cstep_star; eauto.
Qed.
  apply cstep_star_compose_contexts_skip2 in H.
  destruct H as [C''''0 ?].
  exists s'''' C''''0 w2''''; auto.
  destruct Hww as [ww Hww].
  exists ww.
  rewrite Hww. solvew.
Qed.
*)

(*
Lemma term_diverge_impl_equiv: forall A G1 P1 t1 G2 P2 t2,
  term_impl A G1 P1 t1 G2 P2 t2 <-> diverge_impl A G2 P2 t2 G1 P1 t1.
Proof.
  unfold term_impl, diverge_impl, terminates; split; intros.
  edestruct H as [s''[C''[w1''?]]]; eauto.
  apply cstep_star_par_swap; eauto.
  edestruct H as [s''[C''[w1''?]]]; eauto.
  apply cstep_star_par_swap; eauto.
Qed.


Lemma term_diverge_impl_prepend_seq_equiv: forall A G1 P1 t1 G2 P2 t2,
  term_impl_prepend_seq A G1 P1 t1 G2 P2 t2 <-> diverge_impl_prepend_seq A G2 P2 t2 G1 P1 t1.
Proof.
  unfold term_impl_prepend_seq, diverge_impl_prepend_seq, terminates; split; intros.
  edestruct H as [s'''[C'''[w1'''?]]]; eauto.
  apply cstep_star_par_swap; eauto.
  apply cstep_star_par_swap; eauto.
  edestruct H as [s'''[C'''[w1'''?]]]; eauto.
  apply cstep_star_par_swap; eauto.
  apply cstep_star_par_swap; eauto.
Qed.

Lemma cstep_star_context_step_inversion: forall s C b s''' C''' b''',
  cstep_star s C b s''' C''' b''' ->
  (exists s', exists C', exists s'', exists b'',
  cstep_star s C (gwemp:tm_skip) s' C' (gwemp:tm_skip) /\
  step s' b s'' b'' /\
  cstep_star s'' C' b'' s''' C''' b''')
  \/ (cstep_star s C (gwemp:tm_skip) s''' C''' (gwemp:tm_skip) /\ b'''=b).
Admitted.

Lemma recv_commutes_left_of_context: forall s C wC w x c n s'' C'' w'',
  join_wterms C[gwemp:tm_skip] wC ->
  gjoins wC w ->  
  has_endpoint c Share.top n (w_cons w) ->
  cstep_star s C (w:tm_recv x c) s'' C'' (w'':tm_skip) ->
  exists s',
    step s (w:tm_recv x c) s' (w'':tm_skip) /\
    cstep_star s' C (gwemp:tm_skip) s'' C'' (gwemp:tm_skip).
Admitted.

Lemma cstep_star_channel_noninterference: forall (s:state) (C:context) (b:wterm) s' C' b' ww (w':world) c,
  ww ||= s (x) C (x) b (x) w' ->
  has_total_endpoint c (w_cons w') ->
  cstep_star s C b s' C' b' ->
  (s_get_ch s c = None /\ s_get_ch s' c = None) \/
  (exists q, exists q', exists R, exists wcl, exists wcl',
    s_get_ch s c = Some (R,wcl,q) /\ s_get_ch s' c = Some (R,wcl++wcl',q++q')).
Admitted.
*)


(*
Lemma term_impl_recv_send: forall G1 G2 c p m e n x,
  m >= m ->
  term_impl G1 (pSend c p m) 0 (tm_send c e) G2 (pRecv c Share.top n) 0 (tm_recv x c).
Proof.
  introv H He Hww Henv1 Henv2 H0 H1 H2.
  intros  t3 s'' C'' b2'' w3' H3.
  apply cstep_star_par_swap in H3.
  edestruct cstep_star_context_step_inversion.
    apply H2. clear H2.
  destruct H4 as [s'0[C'0[s'1[b''[?[??]]]]]].
  inversion H4; subst. clear H4.
  inversion H10; subst.
  assert (w=w1').
    apply cstep_star_par_compose_contexts1 in H5.
    apply cstep_star_skip_resolve in H5. inversion H5; auto.
  subst w.
  destruct Hww as [ww Hww].
  move Hww at bottom.
  apply cstep_star_par_swap in H5.
  apply cstep_star_remove_par_skip in H5.
  edestruct cstep_star_context_step_inversion.
    apply H5.
  destruct H4 as [s'1[C'1[s'2[b2'2[?[??]]]]]].
assert (Hww'0: ww ⊫ s_push_ch s'0 c w' v ⨁ C'0 ⨁ (gwemp:tm_skip)%wterm ⨁ w2).
admit.
  edestruct cstep_star_channel_noninterference with (ww:=ww) (s:= s_push_ch s'0 c w' v) (C:=C'0) (w':=w2) (b:=(gwemp:tm_skip)%wterm).
  apply Hww'0.
  apply pRecv_total_endpoint with n.
  apply H1.
  apply H4.
  destruct H13.
  erewrite s_ch_get_push_same in H13; eauto. discriminate.
  destruct H13 as [q'[q''[R'[wcl[wcl'[??]]]]]].
  erewrite s_ch_get_push_same in H13; eauto.
  inversion H13; subst R' wcl q'. clear H13.
  inversion H6; subst.
  apply cstep_star_skip_resolve in H7. subst b2'.
  apply cstep_star_par_compose_contexts2 in H3.
  apply cstep_star_skip_resolve in H3. subst b2''.
  do 3 eexists; apply cstep_nil.
  elimtype False.
  eapply app_cons_not_nil.
  rewrite H22 in H14. inversion H14.
  symmetry in H19.
  eapply List.app_eq_nil in H19.
  destruct H19. subst q''.
  symmetry. apply H13. 
  destruct H4; subst b2'.
  
  SearchAbout List.app. 

Lemma cstep_star_channel_noninterference: forall (s:state) (C:context) (b:wterm) s' C' b' ww (w':gworld) c,
  ww ||= s (+) C (+) b (+) w' ->
  has_total_endpoint c (gw_cons w') ->
  cstep_star s C b s' C' b' ->
  (s_get_ch s c = None /\ s_get_ch s' c = None) \/
  (exists q, exists q', exists R, exists wcl, exists wcl',
    s_get_ch s c = Some (R,wcl,q) /\ s_get_ch s' c = Some (R,wcl++wcl',q++q')).
Admitted.


  patternw C in Hww.
  edestruct @Join_split1 with (ab:=ww) as [wC]; eauto.
  joinablew_ex C w2 Hww.
  destruct H6 as [wCw2 H6].
  rewrite<- H4 in H6.
Lemma pRecv_eval_expr: forall w c p e,
  w |= pRecv c p e ->
  exists n:nat, w |= pRecv c p n.
Proof.
  intros. destruct H as [n]. exists n. exists n. split. split. reflexivity. apply H. apply H.
Qed.
  apply pRecv_eval_expr in H1.
  destruct H1 as [n' H1].
  edestruct recv_commutes_left_of_context with (c:=c) as [s'00[??]].
  apply H4. reflexivity. exists wCw2.
  apply Join_join_equiv; apply H6.
  apply pRecv_has_endpoint.
  apply H1.
  apply H5.

Check world_model.pRecv_eval_expr.
Check recv_commutes_left_of_context.
  apply H1.



Qed.
*)

(*
Lemma diverge_impl_seq_appR: forall A G1 P1 t1 G2 P2 t2 t3,
  diverge_impl A G1 P1 t1 G2 P2 t2 ->  
  diverge_impl A G1 P1 t1 G2 P2 (tm_seq t2 t3).
Proof.
  intros.
  unfold diverge_impl, terminates in *.
  introv He Hww Henv1 Henv2 H0 H1 H2 H3.
  rename w3 into w4; rename t0 into t4.
  edestruct cstep_star_term_par_seq_inversion as [?|[?|?]]; eauto.

  destruct H4 as [?[?[??]]]; discriminate.

  destruct H4 as [s'1[C'1[b1'1[w3[??]]]]].
  edestruct H with (t2:=tm_seq t3 t4) as [s'''[C'''[w1'''?]]]; eauto.
  eapply cstep_star_app.
  apply cstep_star_par2_inner_tm_seq_append; eauto.
  eapply step_cstep_star.
  apply st_par2.
  apply st_seq_skip.
  eauto.

  destruct H4 as [?[??]]; discriminate.
Qed.

Lemma term_impl_seq_appL: forall A G1 G2 P1 t1 P2 t2 t3,
  term_impl A G1 P1 t1 G2 P2 t3 ->  
  term_impl A G1 P1 (tm_seq t1 t2) G2 P2 t3.
Proof.
  intros.
  apply term_diverge_impl_equiv.
  apply diverge_impl_seq_appR.
  apply term_diverge_impl_equiv.
  auto.
Qed.

Lemma diverge_impl_seqL1: forall A G1 P1 t1 G2 P2 t2 t3,
  diverge_impl A G1 P1 (tm_seq t1 t3) G2 P2 t2 ->  
  diverge_impl A G1 P1 t1 G2 P2 t2.
Proof.
  intros.
  unfold diverge_impl, terminates in *.
  introv He Hww Henv1 Henv2 H0 H1 H2 H3.
  rename w3 into w4; rename t0 into t4.
  edestruct H with (t2:=t4) as [s'''[C'''[w1'''?]]].
  eauto. eauto. auto. auto. auto. auto.
  apply cstep_star_par1_inner_tm_seq_append.
  eauto.
  apply cstep_star_par1_inner_seq_append.
  eauto.
  edestruct cstep_star_seq_inversion as [?|?]; eauto.

  destruct H5 as [s''0[C''0[w''0[??]]]].
  do 3 eexists; eauto.

  destruct H5 as [?[_[? _]]]; discriminate.
Qed.

Lemma term_impl_seqR1: forall A G1 P1 t1 G2 P2 t2 t3,
  term_impl A G1 P1 t1 G2 P2 (tm_seq t2 t3) ->  
  term_impl A G1 P1 t1 G2 P2 t2.
Proof.
  intros.
  apply term_diverge_impl_equiv.
  apply diverge_impl_seqL1 with t3.
  apply term_diverge_impl_equiv.
  auto.
Qed.

Lemma terminates_seq_appR: forall s C b1 w2 t2 s' C' b1' w2',
  terminates s C b1 w2 ->
  cstep_star s C (b1||w2:t2) s' C' (b1'||w2':tm_skip) ->
  terminates s' C' b1' w2'.
Proof.
  unfold terminates; intros.
  rename w2' into w3; rename t0 into t3.
  edestruct H with (t2:=tm_seq t2 t3) as [s''[C''[w1''?]]]; eauto.
  eapply cstep_star_app.
  apply cstep_star_par2_inner_tm_seq_append; eauto.
  eapply step_cstep_star; eauto.
  apply st_par2. apply st_seq_skip.
Qed.

Lemma terminates_seqL1: forall s C b1 t3 w2,
  terminates s C (b1; t3)%wterm w2 ->
  terminates s C b1%wterm w2.
Proof.
  unfold terminates; intros.
  edestruct H as [s''[C''[w3''?]]].
  apply cstep_star_par1_inner_seq_append; eauto.
  edestruct cstep_star_seq_inversion as [?|?]; eauto.

  destruct H2 as [s''0[C''0[w''0[??]]]].
  do 3 eexists; eauto.

  destruct H2 as [?[_[? _]]]; discriminate.
Qed.

Lemma terminates_tm_seqL1: forall s C w1 t1 t3 w2,
  terminates s C (w1: tm_seq t1 t3)%wterm w2 ->
  terminates s C (w1: t1)%wterm w2.
Proof.
  unfold terminates; intros.
  edestruct H as [s''[C''[w3''?]]].
  apply cstep_star_par1_inner_tm_seq_append; eauto.
  edestruct cstep_star_seq_inversion as [?|?]; eauto.

  destruct H2 as [s''0[C''0[w''0[??]]]].
  do 3 eexists; eauto.

  destruct H2 as [?[_[? _]]]; discriminate.
Qed.

Lemma terminates_seqL2: forall s C b1 t3 w2 w3 s' C',
  terminates s C (b1; t3)%wterm w2 ->
  cstep_star s C b1 s' C' (w3:tm_skip) ->
  terminates s' C' (w3: t3)%wterm w2.
Proof.
  intros.
  unfold terminates in *; intros.
  edestruct H as [s''[C''[w1''?]]].
  eapply cstep_star_app.
  apply cstep_star_par1_step.
  apply cstep_star_seq_append.
  apply H0.
  eapply step_cstep_star.
  apply st_par1.
  apply st_seq_skip.
  apply H1.
  do 3 eexists; eauto.
Qed.

Lemma terminates_tm_seqL2: forall s C w1 t1 t3 w2 w3 s' C',
  terminates s C (w1: tm_seq t1 t3)%wterm w2 ->
  cstep_star s C (w1:t1) s' C' (w3:tm_skip) ->
  terminates s' C' (w3: t3)%wterm w2.
Proof.
  intros.
  unfold terminates in *; intros.
  edestruct H as [s''[C''[w1''?]]].
  eapply cstep_star_app.
  apply cstep_star_par1_step.
  apply seq_tm_seq_equiv2'.
  apply cstep_star_seq_append.
  apply H0.
  eapply step_cstep_star.
  apply st_par1.
  apply st_seq_skip.
  apply H1.
  do 3 eexists; eauto.
Qed.

Lemma terminates_par1_while: forall s C w1 e t1 w2 t2 s' C' b1' b2',
  terminates s C (w1: tm_while e t1)%wterm w2 ->
  cstep_star s C (w1: tm_while e t1 || w2:t2) s' C' (b1'||b2') ->
  exists s'', exists C'', exists w1'',
    cstep_star s' C' (b1'||b2') s'' C'' (w1'':tm_skip || b2').
Admitted.

Lemma term_impl_seq: forall D A12 A34 G1 G2 P1 P2 P3 P4 t1 t2 t3 t4,
  (D,G1) |-h {{P1*P2&&A12}} (tm_par P1 t1 P2 t2) {{P3*P4&&A34}} ->
  (D,G1) |-h {{P1}} t1 {{P3}} ->
  (D,G2) |-h {{P2}} t2 {{P4}} ->
  term_impl A12 G1 P1 t1 G2 P2 t2 ->
  term_impl A34 G1 P3 t3 G2 P4 t4 ->
  term_impl A12 G1 P1 (tm_seq t1 t3) G2 P2 (tm_seq t2 t4).
Proof.
  simpl; introv Hh12 H H0 H1 H2.
unfold term_impl.

intros s C w1 w2 s'' C'' b24'' w3' He Hww Henv1 Henv2 H3 H4 H5.
edestruct cstep_star_term_par1_seq_inversion as [H6|[H6|H6]]; eauto.
destruct H6 as [b1'[b24'[??]]]; discriminate.
2: destruct H6 as [_[_ H6]]; discriminate.
destruct H6 as [s'[C'[w3[b24'[??]]]]].
edestruct cstep_star_term_par_seq_inversion as [H8|[H8|H8]].
  apply H6.
destruct H8 as [b1'[b2'[??]]].
inversion H9; subst ;clear H9.
edestruct cstep_star_par_seq_inversion as [H9|H9].
  apply H7.
(* Case: t1||t2 --> skip||b2' / t3||b2' --> skip||b2'' (b2'' = b24'') *)
  destruct H9 as [b1''[b2''[??]]].
  inversion H10; subst; clear H10.
  clear H6 H5 H7. intros t5 s''' C''' b24''' w5'''  H5.
  apply cstep_star_par_swap in H5.   
  edestruct cstep_star_par_seq_inversion as [H10|H10].
    apply H5.
  (* Case: t5||b2'' --> b5'''||b2''' *)
    destruct H10 as [w5'''_[b2'''[H10 H11]]].
    inversion H11; subst w5'''_ b24'''; clear H11.
    clear H5.
    apply H1 in H8; auto. unfold terminates in H8.
    specialize (H8 (tm_seq t3 t5) s''' C''' b2''' w5''').
    destruct H8 as [s_4[C_4[w4 H8]]].
    apply cstep_star_par_swap.
    eapply cstep_star_app.
    eapply cstep_star_par1_inner_tm_seq_append; apply H9.
    eapply step_cstep_star. apply st_par1. apply st_seq_skip. apply H10.
    move H2 at bottom.
    apply term_impl_equiv in H2. unfold term_impl_prepend_seq in H2.
Admitted.
(*
    specialize (H2 s' (compose_contexts C' (cx_par1 cx_hole (w3':t5)%wterm)) w3 w4 b2' s_4 (compose_contexts C_4 (cx_par1 cx_hole (w5''':tm_skip))) (w3':tm_skip)%wterm s_4 (compose_contexts C_4 (cx_par1 cx_hole (w5''':tm_skip))) w3' (w4:t4)%wterm).
    unfold terminates in H2.
    edestruct H2 as [s_5[C_5[w4_5 ?]]]; clear H2.
admit.
admit.
admit.
admit.
admit.
admit.
    eapply cstep_star_app.
    apply cstep_star_compose_weaken_context'. apply H9.
    apply cstep_star_par_swap.
    apply cstep_star_par1_step.
    apply cstep_star_par_compose_contexts1.
    apply cstep_star_par_swap.
    eapply cstep_star_app.
    apply H10. apply cstep_star_par2_step. apply H8.
    apply cstep_nil.
    apply cstep_nil.
    apply cstep_star_compose_contexts_skip2 in H5.
    destruct H5 as [C_5' ?].
    exists s_5 C_5' w4_5.
    eapply cstep_star_seq_join; split; eauto.
  (* Case: t5||b2'' --> b5'''1||skip *)
    destruct H10 as [s''1[C''1[b5''1[w4[H10 H11]]]]].
    clear H5. rename b24''' into b4'''.
    move H2 at bottom.
    apply term_impl_equiv in H2. unfold term_impl_prepend_seq in H2.
    specialize (H2 s' (compose_contexts C' (cx_par1 cx_hole (w3':t5)%wterm)) w3 w4 b2' s''1 (compose_contexts C''1 (cx_par1 cx_hole b5''1)) (w3':tm_skip)%wterm s''1 (compose_contexts C''1 (cx_par1 cx_hole b5''1)) w3' (w4:t4)%wterm).
    unfold terminates in H2.
    edestruct H2 as [s_4[C_4[w4_4 ?]]].
admit.
admit.
admit.
admit.
admit.
admit.
    eapply cstep_star_app.
    apply cstep_star_compose_weaken_context'. apply H9.
    apply cstep_star_par2_step.
    apply cstep_star_par_compose_contexts1.
    apply cstep_star_par_swap. auto.
    apply cstep_nil.
    apply cstep_star_par1_step.
    apply cstep_star_par_compose_contexts1.
    apply cstep_star_par_swap. eauto.
    apply cstep_star_compose_contexts_skip2 in H5.
    destruct H5 as [C_4' ?].
    exists s_4 C_4' w4_4. auto.
(* Case: t1||t2 --> skip||b2' / t3||b2' --> b3''||skip / b3''||t4 --> skip||b4''' (b4'''=b24'') *)
  destruct H9 as [s'0[C'0[b3'0[w4[H9 H10]]]]].
  clear H6 H5 H7.
  rename b24'' into b4''.
  intros t5 s''' C''' b4''' w5'''  H5.
  move H2 at bottom.
  apply term_impl_equiv in H2. unfold term_impl_prepend_seq in H2.
  specialize (H2 s' C' w3 w4 b2' s'0 C'0 b3'0 s'' C'' w3' b4'').
  edestruct H2 as [s_4[C_4[w4_4 ?]]]; eauto.
admit.
admit.
admit.
admit.
admit.
admit.
(* Case: t1||t2 --> b1'||skip / b1'||t4 --> skip||b4'' / t3||b4'' -->  skip||b4'' *)
  rename b24'' into b4''. rename b24' into b4'.
  destruct H8 as [s'0[C'0[b1'0[w4[H9 H10]]]]].
  clear H6 H5.
  intros t5 s''' C''' b4''' w5'''  H5.
  move H2 at bottom.
unfold term_impl in H2.
  specialize (H2 s'0 (compose_contexts C'0 (cx_par1 cx_hole b1'0)) w3 w4 s'' (compose_contexts C'' (cx_par1 cx_hole (w3:tm_skip)%wterm)) b4'').
  edestruct H2 as [s_4[C_4[w4_4 ?]]].
admit.
admit.
admit.
admit.
admit.
admit.
    eapply cstep_star_app.
    apply cstep_star_par2_step.
    apply cstep_star_par_compose_contexts1.
    apply cstep_star_par_swap. eauto.
    apply cstep_star_compose_weaken_context'. apply H7.
    apply cstep_star_compose_weaken_context'. apply H5.
    apply cstep_star_compose_contexts_skip2 in H6.
    destruct H6 as [C_4' ?].
    exists s_4 C_4' w4_4. auto.
(* Case t1=skip: t3||t2;t4 --> skip||b24'' *)
  destruct H8 as [?[??]]; subst.
  inversion H10; subst t1 w1 b24'; clear H10.
  clear H5 H6.
  edestruct cstep_star_term_par_seq_inversion as [H8|[H8|H8]].
    apply H7.
  (* Case t3||t2 --> skip||b2'' (b24'' = b2'';t4) *)
    destruct H8 as [b3''[b2''[??]]].
    inversion H6; subst b3'' b24''; clear H6.
    clear H7.
    unfold terminates.
    intros t5 s''' C''' b24''' w5''' H6.
    apply cstep_star_par_swap in H6.
    edestruct cstep_star_par_seq_inversion as [H7|H7].
      apply H6.
    destruct H7 as [b5'''[b2'''[??]]].
    inversion H8; subst b5''' b24'''; clear H8 H6.
    edestruct H1 as [s_4[C_4[w4 ?]]]. eauto. eauto.
admit.
admit.
admit.
admit.
    apply cstep_nil.
    apply cstep_star_par_swap.
    instantiate (5:= tm_seq t3 t5).
    eapply cstep_star_app.
    apply cstep_star_par1_inner_tm_seq_append. apply H5.
    eapply step_cstep_star. apply st_par1. apply st_seq_skip.
    apply H7.
    move H2 at bottom.
    apply term_impl_equiv in H2. unfold term_impl_prepend_seq in H2.
    specialize (H2 s (compose_contexts C (cx_par1 cx_hole (w3':t5)%wterm)) w3 w4 (w2:t2)%wterm s_4 (compose_contexts C_4 (cx_par1 cx_hole (w5''':tm_skip))) (w3':tm_skip)%wterm s_4 (compose_contexts C_4 (cx_par1 cx_hole (w5''':tm_skip))) w3' (w4:t4)%wterm).
    unfold terminates in H2.
    edestruct H2 as [s_5[C_5[w4_5 ?]]].
admit.
admit.
admit.
admit.
admit.
admit.
    eapply cstep_star_app.
    apply cstep_star_compose_weaken_context'. apply H5.
    apply cstep_star_par2_step.
    apply cstep_star_par_compose_contexts1.
    apply cstep_star_par_swap.
    eapply cstep_star_app.
    apply H7. apply cstep_star_par2_step. apply H6.
    apply cstep_nil.
    apply cstep_nil.
    apply cstep_star_compose_contexts_skip2 in H8.
    destruct H8 as [C_5' ?].
    exists s_5 C_5' w4_5.
    eapply cstep_star_seq_join; split; eauto.
    destruct H7 as [s''0[C''0[b5''0[w4[??]]]]].
    clear H6. rename b24''' into b4'''.
    move H2 at bottom.
    apply term_impl_equiv in H2. unfold term_impl_prepend_seq in H2.
    specialize (H2 s (compose_contexts C (cx_par1 cx_hole (w3':t5)%wterm)) w3 w4 (w2:t2)%wterm s''0 (compose_contexts C''0 (cx_par1 cx_hole b5''0)) (w3':tm_skip)%wterm s''' (compose_contexts C''' (cx_par1 cx_hole (w5''':tm_skip))) w3' b4''').
    unfold terminates in H2.
    edestruct H2 as [s_4[C_4[w4_4 ?]]]; auto.
admit.
admit.
admit.
admit.
admit.
    eapply cstep_star_app.
    apply cstep_star_compose_weaken_context'. apply H5.
    apply cstep_star_par2_step.
    apply cstep_star_par_compose_contexts1.
    apply cstep_star_par_swap. auto.
    apply cstep_star_par2_step.
    apply cstep_star_par_compose_contexts1.
    apply cstep_star_par_swap. auto.
    apply cstep_nil.
    apply cstep_star_compose_contexts_skip2 in H6.
    destruct H6 as [C_4' ?].
    exists s_4 C_4' w4_4. auto.
  (* Case t3||t2 --> skip||b2'' (b24'' = b2'';t4) *)
    destruct H8 as [s'[C'[b3'[w4[??]]]]].
    rename b24'' into b4''.
    intros t5 s''' C''' b4''' w5' H8.
    move H2 at bottom.
    apply term_impl_equiv in H2. unfold term_impl_prepend_seq in H2.
    specialize (H2 s C w3 w4 (w2:t2)%wterm s' C' b3' s'' C'' w3' b4'').
    edestruct H2 as [s_4[C_4[w4' ?]]]; auto.
admit.
admit.
admit.
admit.
admit.
    apply H8.
    eauto.    
  (* Case t3=skip *)
    destruct H8 as [?[??]]; inversion H8; subst s'' C'' w3 t3 b24''; clear H8.
    clear H7.
    intros t5 s' C' b24' w5' H6.
    apply cstep_star_par_swap in H6.
    edestruct cstep_star_term_par_seq_inversion as [H8|[H8|H8]].
      apply H6.
    destruct H8 as [b5'[b2'[??]]].
    inversion H7; subst b5' b24'; clear H7.
    clear H6.
    unfold term_impl in H1.
    edestruct H1 as [s''[C''[w4 ?]]]; auto.
    eauto. eauto. admit. admit. eauto. eauto.
    apply cstep_nil.
    apply cstep_star_par_swap.
    apply H5.
    clear H1; move H2 at bottom.
    apply term_impl_equiv in H2. unfold term_impl_prepend_seq in H2.
    specialize (H2 s (compose_contexts C (cx_par1 cx_hole (w3':t5)%wterm)) w3' w4 (w2:t2)%wterm s'' (compose_contexts C'' (cx_par1 cx_hole (w5':tm_skip))) (w3':tm_skip)%wterm s'' (compose_contexts C'' (cx_par1 cx_hole (w5':tm_skip))) w3' (w4:t4)%wterm).
    edestruct H2 as [s'''[C'''[w4' ?]]]; auto.
admit.
admit.
admit.
admit.
admit.
    apply cstep_star_par2_step.
    apply cstep_star_par_compose_contexts1.
    apply cstep_star_par_swap.
    eapply cstep_star_app; eauto.
    apply cstep_star_par2_step; auto.
    apply cstep_nil.
    apply cstep_nil.
    apply cstep_star_compose_contexts_skip2 in H1.
    destruct H1 as [C'''0 ?].
    exists s''' C'''0 w4'.
    eapply cstep_star_seq_join; split; eauto.

    destruct H8 as [s'0[C'0[b5'0[w4[??]]]]].
    clear H6. rename b24' into b4'.
    clear H1; move H2 at bottom.
    apply term_impl_equiv in H2. unfold term_impl_prepend_seq in H2.
    specialize (H2 s (compose_contexts C (cx_par1 cx_hole (w3':t5)%wterm)) w3' w4 (w2:t2)%wterm s'0 (compose_contexts C'0 (cx_par1 cx_hole b5'0)) (w3':tm_skip)%wterm s' (compose_contexts C' (cx_par1 cx_hole (w5':tm_skip))) w3' b4').
    edestruct H2 as [s''[C''[w4' ?]]]; auto.
admit.
admit.
admit.
admit.
admit.
    apply cstep_star_par2_step.
    apply cstep_star_par_compose_contexts1.
    apply cstep_star_par_swap. auto.
    apply cstep_star_par2_step.
    apply cstep_star_par_compose_contexts1.
    apply cstep_star_par_swap. auto.
    apply cstep_nil.
    apply cstep_star_compose_contexts_skip2 in H1.
    destruct H1 as [C''0 ?].
    exists s'' C''0 w4'. auto.

    destruct H8 as [?[??]].
    inversion H8; subst s' C' w5' t5 b24'; clear H8 H6.
    edestruct H1 with (s:=s) (C:=C) as [s'[C'[w4 ?]]]. eauto. eauto.
admit.
admit.
auto.
auto.
    apply cstep_nil. apply cstep_nil.
    apply term_impl_equiv in H2. unfold term_impl_prepend_seq in H2.
    specialize (H2 s C w3' w4 (w2:t2)%wterm s' C' (w3':tm_skip)%wterm s' C' w3' (w4:t4)%wterm).
    edestruct H2 as [s''[C''[w4' ?]]]; auto.
admit.
admit.
admit.
admit.
admit.
    apply cstep_star_par2_step. auto.
    apply cstep_nil.
    apply cstep_nil.
    exists s'' C'' w4'.
    eapply step_cstep_star.
    constructor.
    eapply cstep_star_seq_join; split; eauto.
Qed.
*)
*)


End Lemmas.

*)