Require Import Machine.
Require Import StepStar.
Require Import ContextSemantics.
Require Import Bisimulation.
Require Import HoareLogic.
Require Import WorldLogic.
Require Import msl.msl_standard.
Require Import world_model.
Require Import LibTactics.
Require Import SfLib.
Require Import IterationSubstitution.
Require Import LoopBisimulations.
Require Import SimpleParallelization.
Require Import Termination.
Require Import Parallelization.
Require Import ParBisimulation.

Import ParBisimulation.JoinList.
Import Machine.Notation.
Open Scope wterm_scope.
Require Import world_model.


(*
  Arguments Scope tm_par [ pred term_scope pred term_scope ].
  Arguments Scope tm_assert [ pred ].
  Arguments Scope tm_while_max [ _ _ term_scope ].
  Arguments Scope tm_if [ _ term_scope term_scope ].
  Arguments Scope tm_seq [ _ term_scope term_scope ].
*)

Section Definitions.
  Import Machine.Notation.
  Import ContextSemantics.Notation.
  Import Bisimulation.Notation.
  Import HoareLogic.Notation.
  Import HoareLogic.Notation.U.
  Import world_model.Notation.
  Import world_model.Notation.U.
  Open Scope forces_pred.
  Open Scope bsim_scope.

  Definition wno_channels (b: wterm) :=
    exists w, join_wterms b w /\ w_prod w = channels_unit /\ w_cons w = channels_unit.
  Definition wno_heap (b:wterm) :=
    exists w, join_wterms b w /\ w_heap w = heap_unit.

  Definition transformation_sound P ta tb :=
    forall F s w,
      w |= gsep_con P F->
      bisimilar s (w:ta) (w:tb).

  Definition channel_types:= cname->(value->pred world).
  Record valid_state (C:channel_types) s locals wT :=
  { vs_joins: join_list (locals++ch_state_worlds s) wT;
    vs_total_ch: forall c,
      (w_prod wT c <> None \/ w_cons wT c <> None) ->
      total_send_recv c wT;
    vs_ch_typ: forall c p1 p2 n m,
      has_endpoint c p1 n (w_prod wT) ->
      has_endpoint c p2 m (w_cons wT) ->
      n >= m /\
      (exists R wl vl,
        s_get_ch s c = Some(R,wl,vl) /\
        length wl = length vl /\
        length wl = n - m /\
        C c = R)
  }.

  Definition transformation_sound' C P ta tb :=
    forall wO F s w wT,
      w |= gsep_con P F->
      valid_state C s ([wO,w]%list) wT ->
      cbisimilar wO s (w:ta) s (w:tb).

  Notation "|-t {{ P }} ta === tb " := (transformation_sound P%pred ta%term tb%term)
    (no associativity, at level 70, P at level 99, ta at level 0, tb at next level)
    : transformation_scope.
  Notation "C |-t {{ P }} ta ==== tb " := (transformation_sound' C P%pred ta%term tb%term)
    (no associativity, at level 70, P at level 99, ta at level 0, tb at next level)
    : transformation_scope.
  Open Scope transformation_scope.

  Import IterationSubstitution.Notation.

  Definition logical_loop (P: pred world) (e:expr) (n:nat) (f:iter_schema) :=
    |-t {{P}} tm_assert (pIsFalse e);;; f i_inf === tm_assert (pIsFalse e) /\
    |-t  {{P}}(f i_inf;;; tm_assert (pIsFalse e)) === f i_inf /\
    (forall z u, |-t  {{P}}(f (iter_plus2 z (i_plus n u))) === (f z;;; f u)).


End Definitions.

Module Notation.
  Import Machine.Notation.
  Import world_model.Notation.

  Delimit Scope transformation_scope with transformation.
  Delimit Scope trans_scope_helper with trans_helper.

  Notation "|-t {{ P }} ta === tb " := (transformation_sound P%pred ta%term tb%term)
    (no associativity, at level 70, P at level 99, ta at level 0, tb at next level)
    : transformation_scope.
  Notation "C |-t {{ P }} ta ==== tb " := (transformation_sound' C P%pred ta%term tb%term)
    (no associativity, at level 70, P at level 99, ta at level 0, tb at next level)
    : transformation_scope.



  Module Type Tests.
    Import Machine.Notation.
    Open Scope transformation_scope.
    Parameter T_Reflexivity: forall P t, |-t {{P}} t === t.
    Definition abc:= nat. (* ':=' is not broken! *)
    Lemma test_list: [1,2,3,4] = 1::2::3::4::nil. reflexivity. Qed.
    Goal (forall (P: pred world) (t1 t2: term),
      |-t {{P}} t1 === t2). Abort.
    Goal (forall (P: pred world) (t1 t2: term),
      |-t {{P}} t1 === t2 /\ True). Abort.
    Goal (forall (P: pred world) (t1 t2: term),
      |-t {{P && P * P}} t1;;;t2 === t2;;;t1). Abort.
    Goal (forall (P: pred world) (e:expr) (x:ident),
      |-t {{P}} tm_while e (( x::= [e] )) === tm_skip). Abort.
    Lemma test_load1: forall (x:ident) (e:expr),
      |-t {{TT}} x::= [e] === tm_load x e.
    Proof. intros; apply T_Reflexivity. Qed.
    Lemma test_load2: forall (x:ident) (e:expr),
      |-t {{TT}} (tm_load x e) === x::= [e].
    Proof. intros; apply T_Reflexivity. Qed.
    Lemma test_assign1: forall (x:ident) (e:expr),
      |-t {{TT}} x::= e === tm_assign x e.
    Proof. intros; apply T_Reflexivity. Qed.
    Lemma test_assign2: forall (x:ident) (e:expr),
      |-t {{TT}} (tm_assign x e) === x::= e.
    Proof. intros; apply T_Reflexivity. Qed.
    Lemma test_store1: forall (e1 e2:expr),
      |-t {{TT}} e1<- e2 === tm_store e1 e2.
    Proof. intros; apply T_Reflexivity. Qed.
    Lemma test_store2: forall (e1 e2:expr),
      |-t {{TT}} (tm_store e1 e2) === e1<- e2.
    Proof. intros; apply T_Reflexivity. Qed.
    Lemma test_store3: forall (x:ident) (e:expr),
      |-t {{TT}} expr_plus x 1 <- e;;; x::= e === tm_seq (tm_store (expr_plus x 1) e) (tm_assign x e).
    Proof. intros; apply T_Reflexivity. Qed.
    Lemma test_load_assign1: forall (x:ident) (e:expr),
      |-t {{TT}} x::= [e];;; x::= e === tm_seq (tm_load x e) (tm_assign x e).
    Proof. intros; apply T_Reflexivity. Qed.
  End Tests.


(*
  Definition guarded_thread := prod (pred world) term.
  Definition thread_list := list guarded_thread.
  Fixpoint compile_thread_list_helper (head:guarded_thread)(tl: thread_list) :=
    match tl with
    | nil => head
    | (P, t)::tl => compile_thread_list_helper (((fst head)*P)%pred, tm_par (fst head) (snd head) P t) tl
    end.
  Definition compile_thread_list (tl: thread_list) :=
    match tl with
    | nil => tm_skip
    | pt::tl => snd (compile_thread_list_helper pt tl)
    end.
  Eval simpl in (compile_thread_list_helper (TT,tm_skip) ((TT,tm_skip)::(TT,tm_skip)::nil)).
  Eval simpl in (compile_thread_list ((TT,tm_skip)::(TT,tm_skip)::(TT,tm_skip)::nil)).


  Notation "[ ( PX ::: X ) ||| .. ||| ( PY ::: Y ) ]" := (compile_thread_list (@cons guarded_thread (PX,X) .. (@cons guarded_thread (PY,Y) nil) ..))
    (left associativity, PX at level 0, PY at level 0, at level 30) : term_scope.


  Delimit Scope guarded_thread_scope with guarded_thread.
  Notation " P ::: t " := (@pair (pred world) term P%pred t%term) (no associativity, t at level 0, at level 25) : guarded_thread_scope.
(*  Notation " {[ P ]} t " := (@pair (pred world) term P%pred t%term) (no associativity, t at level 0, at level 20) : guarded_thread_scope.*)
  Notation "[ PX ||| .. ||| PY ]" := (compile_thread_list (@cons guarded_thread PX%guarded_thread .. (@cons guarded_thread PY%guarded_thread nil) ..))
    (left associativity, PX at level 0, PY at level 0, at level 30) : term_scope.

 Eval simpl in (( [ TT*emp:: tm_skip ||| (TT:: tm_skip) ||| (TT:: tm_skip) ] )%term).
 Check ( |-t {{emp}} (tm_skip;;tm_skip);;tm_skip === tm_skip;;(tm_skip;;tm_skip) )%transformation. 
*)


(*
 Check ( |-t {{emp}} (tm_skip;;tm_skip);;tm_skip === tm_skip;;(tm_skip;;tm_skip) )%transformation. 
 Check ( |-t {{emp}} (tm_skip;;tm_skip);;tm_skip === tm_assert emp;; tm_skip )%transformation. 
*)
  Module U.
    Notation "|-t {{ P }} ta ≡ tb " := (transformation_sound P%pred ta%term tb%term)
      (no associativity, at level 70, P at level 99, ta at level 0, tb at next level)
      : transformation_scope.
  End U.


End Notation.

Module Morphisms.
  Require Import Setoid.
  Require Export Morphisms.
  Export SetoidTactics.
  Export ParBisimulation.Morphisms.
  Require Export SetoidClass.
  Import Notation.
  Open Scope transformation_scope.

  Add Parametric Morphism: transformation_sound
  with signature equiv ==> equiv ==> equiv ==> iff
  as transformation_m.
  Proof.
    intros P1 P2 ? t1 t2 ? t3 t4 ?.
    unfold transformation_sound in *.
    split; intros.
    setoid_rewrite<- H0.
    setoid_rewrite<- H1.
    apply H2 with F.
    setoid_rewrite H; auto.
    setoid_rewrite H0.
    setoid_rewrite H1.
    apply H2 with F.
    setoid_rewrite<- H; auto.
  Qed.

  Add Parametric Morphism: transformation_sound'
  with signature eq ==> equiv ==> equiv ==> equiv ==> iff
  as transformation'_m.
  Proof.
    intros C P1 P2 ? t1 t2 ? t3 t4 ?.
    unfold transformation_sound' in *.
    split; intros.
    setoid_rewrite<- H0.
    setoid_rewrite<- H1.
    apply H2 with F wT.
    setoid_rewrite H; auto.
    auto.
    setoid_rewrite H0.
    setoid_rewrite H1.
    apply H2 with F wT.
    setoid_rewrite<- H; auto.
    auto.
  Qed.

  Add Parametric Morphism: halted_assert
  with signature eq ==> eq ==> equiv ==> equiv ==> iff
  as halted_assert_m.
  Proof.
    intros s C ba bb ? Pa Pb ?.
    unfold halted_assert.
    split; intros.
    setoid_rewrite<- H0.
    eapply H1.
    eapply bisim_halted_converge2; eauto.
    apply bisim_Cs_bisim; auto.
    eexists; eauto.
    setoid_rewrite H0.
    eapply H1.
    eapply bisim_halted_converge1; eauto.
    apply bisim_Cs_bisim; auto.
    eexists; eauto.
  Qed.

  Add Parametric Morphism: frame_halted_assert
  with signature equiv ==> equiv ==> equiv ==> iff
  as frame_halted_assert_m.
  Proof.
    intros Pa Pb ? ta tb ? Qa Qb ?.
    unfold frame_halted_assert.
    split; intros.
    setoid_rewrite<- H1.
    setoid_rewrite<- H0.
    eapply H2; eauto.
    setoid_rewrite H.
    auto.
    setoid_rewrite H1.
    setoid_rewrite H0.
    eapply H2; eauto.
    setoid_rewrite<- H.
    auto.
  Qed.

  Add Parametric Morphism: htriple
  with signature equiv ==> equiv ==> equiv ==> iff
  as htriple_m.
  Proof.
    intros Pa Pb ? ta tb ? Qa Qb ?.
    unfold htriple, frame_halted_assert.
    split; intros.
    setoid_rewrite<- H1.
    setoid_rewrite<- H0.
    eapply H2; eauto.
    setoid_rewrite H.
    auto.
    setoid_rewrite H1.
    setoid_rewrite H0.
    eapply H2; eauto.
    setoid_rewrite<- H.
    auto.
  Qed.


End Morphisms.


Section SimpleTransformations.
  Import Machine.Notation.
  Import ContextSemantics.Notation.
  Import Bisimulation.Notation.
  Import world_model.Notation.
  Import world_model.Notation.U.
  Import HoareLogic.Notation.
  Import HoareLogic.Notation.U.
  Import WorldLogic.Notation.
  Import WorldLogic.Notation.U.
  Import WorldLogic.Tactics.
  Import world_model.Morphisms.
(*  Import HoareLogic.Morphisms.*)
  Import Notation.
  Import Notation.U.
  Import IterationSubstitution.Notation.


  Open Scope forces_pred.
  Open Scope bsim_scope.
  Open Scope transformation_scope.
  Open Scope join_scope.
  Open Scope entails_pred.

(*
  Theorem htriple_bisim1: forall D G P Q ta tb,
    htriple_soundness (D,G) P ta Q ->
    (forall s C w F, env_typing_correct w G -> w |= gsep_con P F -> C |-t (s,w:ta) ~~ (s,w:tb)) ->
    htriple_soundness (D,G) P tb Q.
  Proof.
    unfold htriple_soundness; intros.
    specialize (H0 s C w F H2).
    spec H0; auto.
    destruct H0 as [_ ?].
    destruct H0 as [R[??]]. apply H0 in H5.
    edestruct H5 as [sa''[ba''[sb''[bb''[?[?[??]]]]]]]; clear H5; eauto.
    eapply H; eauto.
    simpl in *.
    destruct H1 as [ww[?[?[??]]]].
    exists ww; split; eauto.
    splits; eauto.
    rewrite H5. solvew.
    apply step_star_skip_resolve in H6. destruct H6; subst.
    destruct H8; subst. eexists; eauto.
    eauto.
  Qed.

  Theorem htriple_bisim2: forall D G P Q ta tb,
    htriple_soundness (D,G) P tb Q ->
    (forall s C w F, env_typing_correct w G ->w |= gsep_con P F -> C |-t (s,w:ta) ~~ (s,w:tb)) ->
    htriple_soundness (D,G) P ta Q.
  Proof.
    intros.
    eapply htriple_bisim1; eauto.
    intros. apply tbisim_symm.
    eauto.
  Qed.
*)

  Open Scope htriple_scope.

  Lemma htriple_term_ensures_pred: forall s w P t Q,
    w|= P ->
    |-h {{P}} t {{Q}} ->
    term_ensures_pred s (w: t) Q.
  Proof.
    unfold frame_halted_assert, term_ensures_pred, halted_assert.
    intros.    
    setoid_rewrite<- gstar_unit2.
    eapply H0; eauto.
    setoid_rewrite gstar_unit2.
    auto.
  Qed.


  Lemma T_term_bisim_Cs: forall P t1 t2,
    term_bisimilar_Cs t1 t2 ->
    |-t {{ P }} t1 === t2.
  Proof.
    unfold transformation_sound; intros.
    apply bisim_Cs_bisim.
    apply H.
  Qed.

  Lemma T_bisim_Cs: forall P t1 t2,
    (forall w F, w|=gsep_con P F -> bisimilar_Cs (w:t1) (w:t2)) ->
    |-t {{ P }} t1 === t2.
  Proof.
    unfold transformation_sound; intros.
    apply bisim_Cs_bisim.
    eapply H; eauto.
  Qed.

  Lemma T_bisim: forall P t1 t2,
    (forall s w F, w|=gsep_con P F -> bisimilar s (w:t1) (w:t2)) ->
    |-t {{ P }} t1 === t2.
  Proof.
    unfold transformation_sound; intros.
    eapply H; eauto.
  Qed.

  Lemma T_Reflexivity: forall P t,
    |-t {{P}} t === t.
  Proof.
    intros.
    apply T_bisim_Cs; reflexivity.
  Qed.

  Lemma T_Symmetry: forall P t1 t2,
    |-t {{P}} t1 === t2 ->
    |-t {{P}} t2 === t1.
  Proof.
    intros.
    apply T_bisim; intros.
    apply bisim_symm; eauto.
  Qed.

  Lemma T_Transitivity: forall P t1 t2 t3,
    |-t {{P}} t1 === t2 ->
    |-t {{P}} t2 === t3 ->
    |-t {{P}} t1 === t3.
  Proof.
    intros.
    apply T_bisim; intros.
    eapply bisim_trans; eauto.
  Qed.

  Lemma T_Weaken: forall P1 P2 t1 t2,
    |-t {{P1}} t1 === t2 ->
    P2 |-- P1 ->
    |-t {{P2}} t1 === t2.
  Proof.
    unfold transformation_sound in *; intros.
    apply H with F.
    apply gstar_intro with (z:=P1) (w:=F) (x:=P2) (y:=F); auto.
  Qed.

  Lemma T_And1: forall P1 P2 t1 t2,
    |-t {{P1}} t1 === t2 ->
    |-t {{P1 && P2}} t1 === t2.
  Proof.
    intros.
    apply T_Weaken with P1; auto.
    eapply and_elim1; eauto.
  Qed.

  Lemma T_And2: forall P1 P2 t1 t2,
    |-t {{P2}} t1 === t2 ->
    |-t {{P1 && P2}} t1 === t2.
  Proof.
    intros.
    setoid_rewrite and_comm. apply T_And1; auto.
  Qed.

  Lemma T_Or: forall P1 P2 t1 t2,
    |-t {{P1}} t1 === t2 ->
    |-t {{P2}} t1 === t2 ->
    |-t {{P1 || P2}} t1 === t2.
  Proof.
    intros.
    apply T_bisim; intros.
    destruct H1 as [w1[w2[?[??]]]].
    destruct H2.
    apply H with F.
    exists w1 w2; auto.
    apply H0 with F.
    exists w1 w2; auto.
  Qed.


  Lemma T_Frame: forall F P t1 t2,
    |-t {{P}} t1 === t2 ->
    |-t {{gsep_con P F}} t1 === t2.
  Proof.
    unfold transformation_sound.
    introv H H0.
    eapply H; eauto.
    rewrite gstar_assoc.
    apply H0.
  Qed.

  Lemma T_Frame': forall F P t1 t2,
    |-t {{P}} t1 === t2 ->
    |-t {{P * F}} t1 === t2.
  Proof.
    intros.
    eapply T_Weaken with (gsep_con P (rm_vars F)).
    apply T_Frame; auto.
    apply star_gstar.
  Qed.

  Lemma T_Par: forall P1 P2 t1 t2 t3 t4,
    |-t {{P1}} t1 === t3 ->
    |-t {{P2}} t2 === t4 ->
    |-t {{gsep_con P1 P2}} tm_par P1 t1 P2 t2 === tm_par P1 t3 P2 t4.
  Proof.
    unfold transformation_sound in *.
    intros.
    apply bisim_tm_par; eauto.
  Qed.

(*
  Lemma cstep_star_extend_gworld: forall s C w1 w2 w b s' C' b' b2 b2',
    cstep_star s C b s' C' b' ->
    join_wterms b w1 ->
    gjoin w1 w2 w ->
    b2 = extend_world b w2 ->
    b2' = extend_world b' w2 ->
    cstep_star s C (b2) s' C' (b2').
  Admitted.
*)

  Lemma T_Seq: forall P1 P2 t1 t2 t3 t4,
    (forall s C w F, w|=gsep_con P1 F -> halted_assert s C (w: t1) (gsep_con P2 F)) ->
    |-t {{P1}} t1 === t3 ->
    |-t {{P2}} t2 === t4 ->
    |-t {{P1}} t1;;; t2 === t3;;; t4.
  Proof.
    unfold transformation_sound in *.
    intros.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_seq; eauto.
    intros.
    eapply H1 with (F:=F).
    eapply H; eauto.
  Qed.

  Lemma T_Seq2: forall P1 P2 t1 t2 t4,
    (forall s C w F, w|=gsep_con P1 F -> halted_assert s C (w: t1) (gsep_con P2 F)) ->
    |-t {{P2}} t2 === t4 ->
    |-t {{P1}} t1;;; t2 === t1;;; t4.
  Proof.
    intros.
    apply T_Seq with P2; intros; eauto.
    apply T_Reflexivity.
  Qed.


  Lemma T_SeqAssoc: forall P t1 t2 t3,
    |-t {{P}} t1;;; (t2;;; t3) === (t1;;; t2);;; t3.
  Proof.
    intros.
    setoid_rewrite bisim_Cs_tm_seq_assoc.
    apply T_Reflexivity.
  Qed.


  Lemma T_If': forall P e12 e34 t1 t2 t3 t4,
    expr_bool_equiv e12 e34 ->
    |-t {{P}} t1 === t3 ->
    |-t {{P}} t2 === t4 ->
    |-t {{P}} tm_if e12 t1 t2 === tm_if e34 t3 t4.
  Proof.
    unfold transformation_sound in *.
    intros.
    setoid_rewrite H.
    apply bisim_if; intros; eauto.
  Qed.

  Lemma T_IfT': forall P e1 e2 t1 t2,
    expr_bool_equiv e1 e2 ->
    |-t {{P}} t1 === t2 ->
    |-t {{P}} tm_ift e1 t1 === tm_ift e2 t2.
  Proof.
    intros; apply T_If'; auto.
    apply T_Reflexivity.
  Qed.

  Lemma T_If: forall P e12 e34 t1 t2 t3 t4,
    expr_bool_equiv e12 e34 ->
    |-t {{P && pIsTrue e12}} t1 === t3 ->
    |-t {{P && pIsFalse e12}} t2 === t4 ->
    |-t {{P && evaluable e12}} tm_if e12 t1 t2 === tm_if e34 t3 t4.
  Proof.
    unfold transformation_sound in *.
    intros.
    setoid_rewrite H.
    apply bisim_if; intros; eauto.
    eapply H0 with (F:=F).
    eapply pIsTrue_gstar_up; eauto.
    split; auto.
    apply H; auto.
    eapply H1 with (F:=F).
    eapply pIsFalse_gstar_up; eauto.
    split; auto.
    apply H; auto.
  Qed.

  Lemma T_If_Assert_True: forall P e t1 t2,
    |-t {{P}} tm_if e t1 t2 === tm_if e (tm_assert (pIsTrue e);;; t1) t2.
  Proof.
    unfold transformation_sound in *.
    intros.

    destruct (evalTrueFalse_dec e w) as [[?|?]|[??]].
    setoid_rewrite bisim_Cs_if_true; auto.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    setoid_rewrite<- bisim_Cs_seq_skip1' at 1.
    apply bisim_seq1.
    setoid_rewrite bisim_Cs_assert_skip; auto.
    apply bisim_refl.
    apply TT_gstar; auto.

    setoid_rewrite bisim_Cs_if_false; auto.
    apply bisim_refl.

    apply bisim_Cs_bisim.
    apply bisim_Cs_if_noeval; auto.
  Qed.

  Lemma T_If_Assert_False: forall P e t1 t2,
    |-t {{P}} tm_if e t1 t2 === tm_if e t1 (tm_assert (pIsFalse e);;; t2).
  Proof.
    unfold transformation_sound in *.
    intros.

    destruct (evalTrueFalse_dec e w) as [[?|?]|[??]].
    setoid_rewrite bisim_Cs_if_true; auto.
    apply bisim_refl.

    setoid_rewrite bisim_Cs_if_false; auto.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    setoid_rewrite<- bisim_Cs_seq_skip1' at 1.
    apply bisim_seq1.
    setoid_rewrite bisim_Cs_assert_skip; auto.
    apply bisim_refl.
    apply TT_gstar; auto.

    apply bisim_Cs_bisim.
    apply bisim_Cs_if_noeval; auto.
  Qed.

  Lemma T_IfT_Split: forall P e t1 t2,
    |-t {{ P }} tm_ift e (t1;;; tm_ift e t2) === tm_ift e t1;;; tm_ift e t2.
  Proof.
    intros.
    setoid_rewrite bisim_Cs_ift_split.
    apply T_Reflexivity.
  Qed.

  Lemma T_If_True: forall e t1 t2,
    |-t {{ pIsTrue e }} tm_if e t1 t2 === t1.
  Proof.
    intros.
    apply T_bisim_Cs; intros.
    apply bisim_Cs_if_true.
    eapply pIsTrue_gstar; eauto.
  Qed.

  Lemma T_If_False: forall e t1 t2,
    |-t {{ pIsFalse e }} tm_if e t1 t2 === t2.
  Proof.
    intros.
    apply T_bisim_Cs; intros.
    apply bisim_Cs_if_false.
    eapply pIsFalse_gstar; eauto.
  Qed.

  Lemma T_If_Fate: forall P e t,
    |-t {{ P && is_bool e }} tm_if e t t === t.
  Proof.
    intros.
    setoid_rewrite is_bool_dec.
    setoid_rewrite and_dist.
    eapply T_Weaken with (pIsTrue e || pIsFalse e)%pred.
    apply T_Or.
    apply T_If_True.
    apply T_If_False.
    apply or_elim.
    apply or_intro1; apply and_elim4; auto.
    apply or_intro2; apply and_elim4; auto.
  Qed.

  Lemma T_Assert_Skip: forall P1 P2 F,
    P1 |-- P2 ->
    |-t {{ gsep_overlap P1 F }} tm_assert P2 === tm_skip.
  Proof.
    intros.
    apply T_bisim_Cs.
    intros.
    apply bisim_Cs_assert_skip with (gsep_con TT TT); auto.
    revert w H0.
    setoid_rewrite gstar_assoc.
    apply gstar_intro.
    setoid_rewrite gsep_overlap_con.
    apply and_elim3.
    apply gstar_intro; auto.
    repeat intro; simpl; auto.
  Qed.

  Lemma T_Assert_Intro: forall P Q t1 t2,
    |-t {{gsep_overlap P Q}} t1 === t2 ->
    |-t {{P}} tm_assert Q;;; t1 === tm_assert Q;;; t2.
  Proof.
    intros.
    apply T_bisim; intros.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_seq2; intros.
    edestruct cstep_star_assert_resolve as [?|[??]]; eauto.
    discriminate.
    inverts H2.
    eapply H with (exactly gwemp); eauto.
    setoid_rewrite gstar_unit2.
    setoid_rewrite gsep_overlap_con.
    split; auto.
    clear H1 H3; revert w H0; apply gstar_intro; auto.
    apply TT_intro.
  Qed.

  Lemma T_While: forall P e1 e2 z t1 t2,
    expr_bool_equiv e1 e2 ->
    |-h {{P && evaluable e1}} t1 {{P && evaluable e1}} ->
    |-t {{P && pIsTrue e1}} t1 === t2 ->
    |-t {{P && evaluable e1}} tm_while_max e1 z t1 === tm_while_max e2 z t2.
  Proof.
    unfold transformation_sound in *.
    intros.
    setoid_rewrite H.
    apply bisim_while with (gsep_con (P&&evaluable e1) F)%pred; intros; eauto.
    repeat intro; eapply H0 with (w:=w'); eauto.
    eapply bisim_halted_converge2; eauto.
    apply H1 with F.
    eapply pIsTrue_gstar_up; eauto.
    split; auto.
    setoid_rewrite H; auto.
    eexists; eauto.
    apply H1 with F.
    eapply pIsTrue_gstar_up; eauto.
    split; auto.
    setoid_rewrite H; auto.
  Qed.

  Lemma T_While': forall P e1 e2 z t1 t2,
    expr_bool_equiv e1 e2 ->
    |-t {{pIsTrue e1}} t1 === t2 ->
    |-t {{P}} tm_while_max e1 z t1 === tm_while_max e2 z t2.
  Proof.
    intros.
    apply T_Weaken with TT.
    apply T_bisim. intros.
    setoid_rewrite<- H; auto.
    apply bisim_while with TT.
    simpl; auto.
    repeat intro; simpl; auto.
    repeat intro; simpl; auto.
    intros.
    eapply H0.
    apply TT_gstar; auto.
    repeat intro; simpl; auto.
  Qed.

  Lemma T_Assert_Redundant: forall P P1 P2,
    P1 |-- P2 ->
    |-t {{ P }} tm_assert P1;;; tm_assert P2 === tm_assert P1.
  Proof.
    intros.
    apply T_bisim_Cs.
    intros.
    apply bisim_Cs_assert_assert_shadow; auto.
  Qed.

  Lemma T_Assert_And: forall P P1 P2,
    |-t {{ P }} tm_assert P1;;; tm_assert P2 === tm_assert (gsep_overlap P1 P2)%pred.
  Proof.
    intros.
    apply T_bisim_Cs.
    intros.
    apply bisim_Cs_assert_assert_and.
  Qed.

  Lemma T_Skip1: forall P t,
    |-t {{ P }} tm_skip;;; t === t.
  Proof.
    intros.
    apply T_bisim_Cs; intros.
    apply bisim_Cs_tm_seq_skip1'.
  Qed.

  Lemma T_Skip2: forall P t,
    |-t {{ P }} t;;; tm_skip === t.
  Proof.
    intros.
    apply T_bisim_Cs; intros.
    apply bisim_Cs_tm_seq_skip2'.
  Qed.

  Lemma T_Par_Swap: forall P P1 t1 P2 t2,
    |-t {{ P }} tm_par P1 t1 P2 t2 === tm_par P2 t2 P1 t1.
  Proof.
    intros.
    apply T_bisim_Cs; intros.
    apply bisim_Cs_tm_par_swap.
  Qed.

  Lemma T_Par_Skip: forall P1 P2,
    |-t {{ gsep_con P1 P2 }} tm_par P1 tm_skip P2 tm_skip === tm_skip.
  Proof.
    intros.
    apply T_bisim_Cs; intros.
    setoid_rewrite<- gstar_assoc in H.
    edestruct H as [w1[w2f[?[??]]]].
    eapply bisim_Cs_tm_par_skip with (w1:=w1) (w2:=w2f) (F1:=TT); eauto.
    apply TT_gstar; auto.
  Qed.

  Lemma s_iter_det: forall s s' s'',
    s_iter s s' ->
    s_iter s s'' ->
    s_i s' = s_i s'' ->
    s' = s''.
  Proof.
    intros.
    inverts H.
    inverts H0.
    destruct (Pos.eq_dec i0 i); subst; auto.
    false.
    assert (s_iter_get (s_iter_set s i iter_go) i = s_iter_get (s_iter_set s i0 iter_go) i).
      apply s_iter_get_eq_s_i; auto.
    rewrite s_iter_get_set_eq in H0.
    rewrite s_iter_get_set_neq in H0; auto.
    congruence.
  Qed.

  Lemma cbisim_s_iter: forall wO s s' bp bq,
    s_iter s s' ->
    cbisimilar wO s bp s bq ->
    cbisimilar wO s' bp s' bq.
  Proof.
    intros.
    apply abisim_cbisim.
    edestruct abisim_s_iter as [s''[??]]; eauto.
    apply cbisim_abisim1; eauto.
    apply (s_iter_det s s' s'') in H1; auto. subst s''.
    apply H2.
    eapply abisim_s_i_eq; eauto.
    edestruct abisim_s_iter as [s''[??]]; eauto.
    apply cbisim_abisim2; eauto.
    apply (s_iter_det s s' s'') in H1; auto. subst s''.
    apply H2.
    eapply abisim_s_i_eq; eauto.
  Qed.

  Lemma abisim_step: forall wO s bp bq,
    ~halted bp -> ~halted bq ->
    valid_machine_state wO s bp ->
    valid_machine_state wO s bq ->
    (forall bO s' wO' bO' bp' s'', step_star s bO s' bO' -> join_wterms bO wO -> join_wterms bO' wO' -> step s' bp s'' bp' -> exists bq', step_star_sans_s_iter s' bq s'' bq' /\ cbisimilar wO' s'' bp' s'' bq') ->
    (forall bO s' wO' bO' bq' s'', step_star s bO s' bO' -> join_wterms bO wO -> join_wterms bO' wO' -> step s' bq s'' bq' -> exists bp', step_star_sans_s_iter s' bp s'' bp' /\ cbisimilar wO' s'' bq' s'' bp') ->
    abisimilar wO s bp s bq.
  Proof.
    intros.
    exists (fun wO sp bp sq bq =>
    ( sq=sp /\
      ~halted bp /\ ~halted bq /\
      valid_machine_state wO sp bp /\
      valid_machine_state wO sq bq /\
      (forall bO s' wO' bO' bp' s'', step_star sp bO s' bO' -> join_wterms bO wO -> join_wterms bO' wO' -> step s' bp s'' bp' -> exists bq', step_star_sans_s_iter s' bq s'' bq' /\ cbisimilar wO' s'' bp' s'' bq') /\
      (forall bO s' wO' bO' bq' s'', step_star sq bO s' bO' -> join_wterms bO wO -> join_wterms bO' wO' -> step s' bq s'' bq' -> exists bp', step_star_sans_s_iter s' bp s'' bp' /\ cbisimilar wO' s'' bq' s'' bp')
    ) \/ (
      abisimilar wO sp bp sq bq
    )).
    splits; iauto.
    clear wO s bp bq H H0 H1 H2 H3 H4.
    intros wO s bp sq bq [ [?[?[?[Hvms1[Hvms2[??]]]]]] | ? ]; try subst sq.
    * splits.
    - intros; false.
    - intros s' ?.
    exists s'; split; auto.
    left; splits; auto.
    eapply valid_machine_state_s_iter; eauto.
    eapply valid_machine_state_s_iter; eauto.
    intros.
    eapply H2; eauto.
    eapply st_iteration; eauto.
    intros.
    eapply H3; eauto.
    eapply st_iteration; eauto.
    - reflexivity.
    - intros. rename sp' into s'.
    edestruct cstep_star_step_inversion as [ [s'0[C'0[s'1[b'1[?[?[??]]]]]]] | [??]]; eauto.
    clear H.
    edestruct valid_machine_state_cstep_star with (C:=C) (b:=bp) as [wC'0[HwC'0 Hvms']]; eauto.
    eapply cstep_star_skip_subst; eauto.
    edestruct H2 as [bq'[??]]; eauto.
    eapply step_star_simulates_cstep_star; eauto.
    apply context_world_join_wterms; auto.
    apply context_world_join_wterms; eauto.
    edestruct cstep_star_abisim with (C:=C'0) (bp:=b'1) as [sq''[bq''[??]]]; eauto.
    apply cbisim_abisim1; eauto.
    exists sq'' bq''; split.
    eapply cstep_star_app.
    eapply cstep_star_skip_subst; eauto.
    eapply cstep_star_app; eauto.
    apply step_star_add_context_cstep_star.
    apply step_star_step_star_sans_s_iter; auto.
    right; auto.
    (**)
    subst.
    exists s' bq; split.
    eapply cstep_star_skip_subst; eauto.
    left; splits; auto.
    clear H.
    eapply valid_machine_state_cstep_star' with (C:=C) (C':=C'); eauto.
    eapply cstep_star_skip_subst; eauto.
    eapply valid_machine_state_cstep_star' with (C:=C) (C':=C'); eauto.
    intros. rename bp' into bq'.
    clear H.
    edestruct compose_context_world_cstep_star
      with (C1:=C) (C1':=C') (C2:=cx_par1 cx_hole bO) (C2':=cx_par1 cx_hole bO') as [wO'0_[C12[C12'[?[?[??]]]]]]; eauto.
    apply cstep_star_par_cx_par1.
    apply step_star_add_context_cstep_star.
    apply step_star_par2_step; eauto.
    exists gwemp wO'; splits; simpl; auto. apply gjoin_bot2.
    edestruct valid_machine_state_join_wterms with (b:=bp) as [wp Hw]; eauto.
    eapply valid_machine_state_join_sub with (b:=bp); eauto.
    reflexivity.
    exists wp. apply gjoin_bot2.
    replace wO'0_ with wO'0 in *. clear wO'0_.

    edestruct H3 with (bO:=set_context C12(gwemp:tm_skip)) (bO':=set_context C12'(gwemp:tm_skip)) as [bp'[??]]; eauto.
    apply step_star_simulates_cstep_star; eauto.
    apply context_world_join_wterms; auto.
    apply context_world_join_wterms; eauto.
    eapply context_world_eq; eauto.
    exists gwemp wO'0; splits; auto.
    reflexivity. apply gjoin_bot2.

    intros. rename bq' into bp'.
    clear H.
    edestruct compose_context_world_cstep_star
      with (C1:=C) (C1':=C') (C2:=cx_par1 cx_hole bO) (C2':=cx_par1 cx_hole bO') as [wO'0_[C12[C12'[?[?[??]]]]]]; eauto.
    apply cstep_star_par_cx_par1.
    apply step_star_add_context_cstep_star.
    apply step_star_par2_step; eauto.
    exists gwemp wO'; splits; simpl; auto. apply gjoin_bot2.
    edestruct valid_machine_state_join_wterms with (b:=bp) as [wp Hw]; eauto.
    eapply valid_machine_state_join_sub with (b:=bp); eauto.
    reflexivity.
    exists wp. apply gjoin_bot2.
    replace wO'0_ with wO'0 in *. clear wO'0_.

    edestruct H2 with (bO:=set_context C12(gwemp:tm_skip)) (bO':=set_context C12'(gwemp:tm_skip)) as [bq'[??]]; eauto.
    apply step_star_simulates_cstep_star; eauto.
    apply context_world_join_wterms; auto.
    apply context_world_join_wterms; eauto.

    eapply context_world_eq; eauto.
    exists gwemp wO'0; splits; auto.
    reflexivity. apply gjoin_bot2.

    * splits.
    intros.
    split.
    eapply abisim_halted1; eauto.
    right; apply abisim_refl.
    intros.
    edestruct abisim_s_iter as [sq'[??]]; eauto.
    eapply abisim_s_i_eq; eauto.
    intros.
    edestruct cstep_star_abisim as [sq'[bq'[??]]]; eauto.
  Qed.

  Lemma cbisim_step: forall wO s bp bq,
    ~halted bp -> ~halted bq ->
    valid_machine_state wO s bp ->
    valid_machine_state wO s bq ->
    (forall bO s' wO' bO' bp' s'', step_star s bO s' bO' -> join_wterms bO wO -> join_wterms bO' wO' -> step s' bp s'' bp' -> exists bq', step_star_sans_s_iter s' bq s'' bq' /\ cbisimilar wO' s'' bp' s'' bq') ->
    (forall bO s' wO' bO' bq' s'', step_star s bO s' bO' -> join_wterms bO wO -> join_wterms bO' wO' -> step s' bq s'' bq' -> exists bp', step_star_sans_s_iter s' bp s'' bp' /\ cbisimilar wO' s'' bq' s'' bp') ->
    cbisimilar wO s bp s bq.
  Proof.
    intros.
    apply abisim_cbisim.
    eapply abisim_step; eauto.
    eapply abisim_step; eauto.
  Qed.

  Lemma cbisim_tm_par_intro: forall wO s w P1 t1 P2 t2 P3 t3 P4 t4,
    (forall w1 w2 bO s' bO' wO',
      w1 |= gsep_con P1 TT ->
      w2 |= gsep_con P2 TT ->
      step_star s bO s' bO' ->
      gjoin w1 w2 w ->
      join_wterms bO wO -> join_wterms bO' wO' ->
      cbisimilar wO' s' ((w1: t1 || w2: t2);; tm_par P3 t3 P4 t4) s' (w1: t1;;t3 || w2: t2;; t4)) ->
    valid_machine_state wO s (w:tm_skip) ->
    cbisimilar wO s (w: (tm_par P1 t1 P2 t2);; tm_par P3 t3 P4 t4)
      s (w: tm_par P1 (t1;;;t3) P2 (t2;;; t4)).
  Proof.
    intros.
    eapply cbisim_step.
    - introv [?]; discriminate.
    - introv [?]; discriminate.
    - eapply valid_machine_state_wterm'; eauto.
    - eapply valid_machine_state_wterm'; eauto.
    - intros.
    inverts H4. inverts H9.
    exists (w1: t1;; t3|| w2: t2;; t4); split.
    eapply step_cons_.
    apply st_par; eauto.
    eapply step_cons_.
    apply st_par1; apply st_seq.
    eapply step_cons_.
    apply st_par2; apply st_seq.
    apply step_nil_.
    eapply H; eauto.
    - intros.
    inverts H4.
    exists ((w1: t1 || w2: t2);; tm_par P3 t3 P4 t4); split.
    eapply step_cons_.
    apply st_seq1; apply st_par; eauto.
    apply step_nil_.
    apply cbisim_symm.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    eapply H; eauto.
  Qed.
    



  Definition pPrivateChannels C P2 P1 t := forall s wO w1 w2 w12 wT,
    w1 |= P1 ->
    w2 |= P2 ->
    gjoin w1 w2 w12 ->
    valid_state C s ([wO,w12]%list) wT ->
    private_communications w2 w12 /\
    cannot_communicate_channels s (w12:t) /\
    noninterfering w1 w2.

  Lemma htriple_term_ensures_pred': forall P1 P3 s w1 t1,
    htriple P1 t1 (gsep_con P3 TT) ->
    w1|= gsep_con P1 TT ->
    term_ensures_pred s (w1 : t1) (gsep_con P3 TT).
  Proof.
    intros.
    eapply htriple_term_ensures_pred with (P:=gsep_con P1 TT); auto.
    eapply H_Consequence with (Q:=gsep_con P3 TT).
    reflexivity. reflexivity.
    setoid_rewrite<- TT_gsplit at 2.
    setoid_rewrite gstar_assoc.
    apply H_Frame; auto.
  Qed.

  Lemma valid_state_valid_machine_state: forall C s wO w wT,
    valid_state C s ([wO, w]%list) wT ->
    valid_machine_state wO s (w:tm_skip).
  Proof.
    intros.
    eapply vs_joins in H.
    edestruct join_list_app_inv as [wOw[wS[?[??]]]]; eauto.
    exists wS w wT; splits; auto.
    reflexivity.
    exists wOw; split; auto.
    eapply join_list_perm; eauto.
    permut4.
    apply gjoin_com; auto.
  Qed.

  Lemma T'_SeqPar: forall C P1 P2 P3 P4 t1 t2 t3 t4,
    (* precise resources ensure that resources aren't transferred from t1 to t4 or t2 to t3 in the
         sequential program  *)
    gprecise P3 -> gprecise P4 ->
    (* Hoare triples to ensure that t1/t2 establishes the state for t3/t4 *)
    htriple P1 t1 P3 ->
    htriple P2 t2 P4 ->
    (* frameability should be implied by hoare triples *)
    (forall w3, w3 |= P3 -> frameable_bisim_Cs (w3: t3)) ->
    (forall w4, w4 |= P4 -> frameable_bisim_Cs (w4: t4)) ->
    termination_impl (gsep_con P1 TT) t1 (gsep_con P2 TT) t2 ->
    termination_impl (gsep_con P2 TT) t2 (gsep_con P1 TT) t1 ->
    pPrivateChannels C (gsep_con P2 TT) (gsep_con P1 TT) (tm_par P1 (t1;;; t3) P2 (t2;;; t4)) ->
    (* -------------------------------------- *)
    C |-t {{TT}} tm_par P1 t1 P2 t2;;; tm_par P3 t3 P4 t4 ==== tm_par P1 (t1;;; t3) P2 (t2;;; t4).
  Proof.
    introv HprecP3 HprecP4.
    introv Hh1 Hh2 HframeP3 HframeP4.
    introv Hterm12 Hterm21.
    introv Hprivate.
    intros wO F s w wT Hw Hvs.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply cbisim_tm_par_intro.
    - intros.
    assert (Hvms: valid_machine_state wO s (w1: t1 || w2: t2)).
      eapply valid_machine_state_wterm with (b1:=w:tm_skip); eauto.
      eapply valid_state_valid_machine_state; eauto.
      reflexivity. exists w1 w2; splits; auto; try reflexivity.
    assert (Hvms': valid_machine_state wO' s' (w1: t1 || w2: t2)).
      eapply valid_machine_state_step_starO'; eauto.
    edestruct Hprivate as [?[??]]; eauto.
    eapply cbisim_seq_par; eauto; try reflexivity.
    red; intros; eapply Hh1; eauto.
    red; intros; eapply Hh2; eauto.
    eapply Hterm21; eauto.
    apply valid_machine_state_par_swap; auto.
    exists w1 w2; splits; eauto; try reflexivity.
    eapply cannot_communicate_channels_step_starO; eauto.
    destruct H6 as [w_[?[??]]].
    simpl in H6; subst w_.
    exists w; splits; auto.
    exists w1 w2; splits; auto; reflexivity.
    repeat intro. simpl in H8; subst w0. simpl in H9; subst w3.
    apply H7.
    - eapply valid_state_valid_machine_state; eauto.
  Qed.


  Lemma bisim_Cs_tm_par_assoc: forall P1 P2 P3 t1 t2 t3,
    (forall w1, w1|=P1 -> frameable_bisim_Cs (w1: t1)) ->
    (forall w2, w2|=P2 -> frameable_bisim_Cs (w2: t2)) ->
    (forall w3, w3|=P3 -> frameable_bisim_Cs (w3: t3)) ->
    gprecise P1 -> gprecise P2 -> gprecise P3 ->
    term_bisimilar_Cs (tm_par (gsep_con P1 P2) (tm_par P1 t1 P2 t2) P3 t3) (tm_par P1 t1 (gsep_con P2 P3) (tm_par P2 t2 P3 t3)).
  Proof.
Admitted.


  Lemma T_ParAssoc: forall P P1 P2 P3 t1 t2 t3,
    (forall w1, w1|=P1 -> frameable_bisim_Cs (w1: t1)) ->
    (forall w2, w2|=P2 -> frameable_bisim_Cs (w2: t2)) ->
    (forall w3, w3|=P3 -> frameable_bisim_Cs (w3: t3)) ->
    gprecise P1 -> gprecise P2 -> gprecise P3 ->
    |-t {{P}} tm_par (gsep_con P1 P2) (tm_par P1 t1 P2 t2) P3 t3 === tm_par P1 t1 (gsep_con P2 P3) (tm_par P2 t2 P3 t3).
  Proof.
    intros.
    setoid_rewrite bisim_Cs_tm_par_assoc; auto.
    apply T_Reflexivity.
  Qed.


  Lemma T_While_False_Skip: forall e z t,
    |-t {{ pIsFalse e }} tm_while_max e z t === tm_skip.
  Proof.
    intros.
    apply T_bisim_Cs; intros.
    apply bisim_Cs_false_while_skip.
    eapply pIsFalse_gstar; eauto.
  Qed.

  Lemma T_While_Post_False: forall P e t,
    |-t {{ P }} tm_while_max e i_inf t;;; tm_assert (pIsFalse e) === tm_while_max e i_inf t.
  Proof.
    intros.
    apply T_bisim_Cs; intros.
    symmetry.
    apply bisim_Cs_while_inf_post_false'.
  Qed.


  Lemma T_While_Split: forall P e z (u:iterations) t,
    |-t {{ P }} tm_while_max e z t;;; tm_while_max e u t === tm_while_max e (iter_plus2 z u) t.
  Proof.
    intros.
    apply T_bisim_Cs; intros.
    apply bisim_Cs_while_split_iter_plus2.
  Qed.

  Lemma T_While_Limit_Skip: forall P e t,
    |-t {{ P }} tm_while_max e 0 t === tm_skip.
  Proof.
    intros.
    apply T_bisim_Cs; intros.
    apply bisim_Cs_while_zero_skip.
  Qed.

  Lemma T_Seq1: forall P1 t1 t2 t3,
    |-t {{P1}} t1 === t3 ->
    |-t {{P1}} t1;;; t2 === t3;;; t2.
  Proof.
    intros.
    apply T_bisim; intros.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_seq1; eauto.
  Qed.

  Lemma T_Seq2': forall P1 t1 t2 t3,
    |-t {{TT}} t2 === t3 ->
    |-t {{P1}} t1;;; t2 === t1;;; t3.
  Proof.
    intros.
    apply T_bisim; intros.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_seq2; eauto.
    intros.
    eapply H.
    setoid_rewrite gstar_com.
    apply gstar_intro with (x:=TT) (y:=TT); auto.
    reflexivity.
    apply TT_gstar.
    reflexivity.
  Qed.

  Lemma frame_halted_assert_par_will_join: forall P1 P2 P3 P4 w1 w2 w t1 t2,
    frame_halted_assert P1 t1 P3 ->
    frame_halted_assert P2 t2 P4 ->
    gjoin w1 w2 w ->
    w1|= P1 -> w2|= P2 ->
    w|= gsep_con P1 P2 ->
    par_will_join (w1: t1) (w2: t2).
Admitted.






  (*
    The use of respects_global_state is a blunt instrument used to demonstrate simple parallelization.
    Thus, we believe that the proof of this lemma is not crucial.
  *)
  Lemma no_communication_respects_global_state: forall b,
    wno_channels b ->
    respects_global_state b.
  Proof.
    (* Informal proof: without permission to any channels, all instructions that modify the global
    state will get stuck, and thus not succeed in modifying the global state. *)
  Admitted.

  Lemma no_channels_heap_respects_global_state: forall P t,
    P |-- no_channels ->
    (forall w, w|=P -> respects_global_state (w: t)).
  Proof.
    intros.
    eapply no_communication_respects_global_state.
    unfold wno_channels.
    exists w; split.
    reflexivity.
    apply H; auto.
  Qed.
  

  Lemma T_SeqParSimpl: forall P1 P2 P3 P4 t1 t2 t3 t4,
    (* precise resources ensure that resources aren't transferred from t1 to t4 or t2 to t3 in the
         sequential program  *)
    gprecise P3 -> gprecise P4 ->
    (* Hoare triples to ensure that t1/t2 establishes the state for t3/t4 *)
    frame_halted_assert P1 t1 (gsep_con P3 TT) ->
    frame_halted_assert P2 t2 (gsep_con P4 TT) ->
    (* frameability means that the behavior of t3/t4 will not change as long as they have at least
         the resources specified by P3/P4. *)
    (forall w3, w3|=P3 -> frameable_bisim_Cs (w3: t3)) ->
    (forall w4, w4|=P4 -> frameable_bisim_Cs (w4: t4)) ->
    (* determinism makes the proof work for the current def of 'bisimilar' *)
    (forall w3, w3|=P3 -> deterministic_Cs_wterm (w3: t3)) ->
    (forall w4, w4|=P4 -> deterministic_Cs_wterm (w4: t4)) ->
    (* this is a blunt way of saying that threads do not interfere with each other *)
    (P3 |-- no_channels) ->
    (P4 |-- no_channels) ->
    (* ------------------ *)
    |-t {{TT}} tm_par P1 t1 P2 t2;;; tm_par P3 t3 P4 t4 === tm_par P1 (t1;;; t3) P2 (t2;;; t4).
  Proof.
    introv HprecP3 HprecP4.
    introv Hh1 Hh2.
    introv Hframe3 Hframe4.
    introv Hdet2 Hdet4.
    introv Hresp2 Hresp4.
    apply T_bisim; intros.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_step.
    introv [?]; discriminate.
    introv [?]; discriminate.
    (**)
    intros.
    inverts H1.
    inverts H7.
    clear H2.
    exists ((w1: t1;; t3) || (w2: t2;; t4))%wterm; split.
    eapply step_cons.
    apply st_par; eauto.
    eapply step_cons.
    apply st_par1; apply st_seq.
    eapply step_single.
    apply st_par2; apply st_seq.
    apply bisim_simple_par_dist; auto.
    eapply htriple_term_ensures_pred'; eauto.
    eapply htriple_term_ensures_pred'; eauto.
    eapply frame_halted_assert_par_will_join with (P1:=gsep_con P1 TT) (P2:=gsep_con P2 TT); eauto.
    apply H_Frame; eauto.
    apply H_Frame; eauto.
    exists w1 w2; splits; auto.
    intros; apply no_channels_heap_respects_global_state with P3; auto.
    intros; apply no_channels_heap_respects_global_state with P4; auto.
    (**)
    intros. clear H2.
    inverts H1.
    exists ((w1: t1 || w2: t2);; tm_par P3 t3 P4 t4)%wterm; split.
      apply step_single. apply st_seq1; apply st_par; auto.
      setoid_rewrite bisim_Cs_tm_seq_seq.
      apply bisim_simple_par_dist; auto.
      eapply htriple_term_ensures_pred'; eauto.
      eapply htriple_term_ensures_pred'; eauto.
      eapply frame_halted_assert_par_will_join with (P1:=gsep_con P1 TT) (P2:=gsep_con P2 TT); eauto.
      apply H_Frame; eauto.
      apply H_Frame; eauto.
      exists w1 w2; splits; auto.
      intros; apply no_channels_heap_respects_global_state with P3; auto.
      intros; apply no_channels_heap_respects_global_state with P4; auto.
  Qed.

  Lemma T_Assert_False_While: forall P e z t,
    |-t {{P}} tm_assert (pIsFalse e);;; tm_while_max e z t === tm_assert (pIsFalse e).
  Proof.
    intros.
    apply T_Transitivity with (tm_assert (pIsFalse e);;; tm_skip)%term.
    apply T_Assert_Intro.
    eapply T_Weaken.
    apply T_While_False_Skip.
    setoid_rewrite gsep_overlap_con.
    apply and_elim4.
    apply pIsFalse_gstar.
    apply T_Skip2.
  Qed.


  Lemma simple_logical_loop: forall P (f: iter_schema) e (t: term),
    f = istm_loop_var e 0 t ->
    logical_loop P e 0 f.
  Proof.
    intros.
    subst f; unfold term_subst_iter.
    splits.
    eapply T_Assert_False_While.
    apply T_While_Post_False.
    intros.
    unfold term_subst_iter.
    rewrite i_plus_zero1.
    repeat rewrite iter_expr_minus_zero.
    apply T_Symmetry.
    apply T_While_Split.
  Qed.

  Lemma T_Loop: forall P Q (f g : iter_schema) (x y n m: nat),
    x > 0 -> y > 0 ->
    frame_halted_assert P (f x) P ->
    |-t {{P}} tm_assert Q;;; f i_inf === tm_assert Q ->  |-t {{P}} f i_inf;;; tm_assert Q === f i_inf ->
    |-t {{P}} tm_assert Q;;; g i_inf === tm_assert Q ->  |-t {{P}} g i_inf;;; tm_assert Q === g i_inf ->
    (forall z u, |-t {{P}} f (iter_plus2 z (i_plus n u)) === f z;;; f u) ->
    (forall z u, |-t {{P}} g (iter_plus2 z (i_plus m u)) === g z;;; g u) ->
    |-t {{P}} f x === g y ->    
    |-t {{P}} f i_inf === g i_inf.
  Proof.
    intros.
    apply T_bisim; intros.
    eapply bisim_loop with (x:=x) (y:=y) (n:=n) (m:=m) (P:=gsep_con P F) (Q:=Q); eauto.

    apply bisim_split_prop_impl.
    unfold bisim_split_prop'.
    intros; setoid_rewrite<- bisim_Cs_tm_seq_seq; eauto.
    eapply H6; eauto.

    apply bisim_split_prop_impl.
    unfold bisim_split_prop'.
    intros; setoid_rewrite<- bisim_Cs_tm_seq_seq; eauto.
    eapply H7; eauto.

    splits; intros; eauto.

    split; intros; setoid_rewrite<- bisim_Cs_tm_seq_seq; eauto.

    split; intros; setoid_rewrite<- bisim_Cs_tm_seq_seq; eauto.
  Qed.

  Arguments Scope istm_loop_var [ _ _ term_scope ].
  Arguments Scope istm_seq [term_scope term_scope].

  Lemma T_Guard_Exists_Elim: forall B (P: B->pred world) t1 t2,
    (forall v, |-t {{P v}} t1 === t2) <-> |-t {{existential P}} t1 === t2.
  Proof.
    split; intros.
    apply T_bisim; intros.
    setoid_rewrite ex_gcon in H0.
    destruct H0 as [x ?].
    eapply H. eauto.
    apply T_bisim; intros.
    eapply H; eauto.
    eapply gstar_intro with (P v) F; eauto.
    repeat intro.
    exists v; auto.
  Qed.

  Lemma T_Guard_Forall_Elim: forall B (P: B->pred world) t1 t2 v,
    |-t {{P v}} t1 === t2 ->
    |-t {{universal P}} t1 === t2.
  Proof.
    intros.
    apply T_Weaken with (P v); auto.
  Qed.


  Lemma T_While_IfT: forall P e t,
    |-t {{P}} tm_while_max e 1 t === tm_ift e t.
  Proof.
    intros.
    apply T_bisim_Cs; intros.
    apply bisim_Cs_while_one_if.
  Qed.



(*
  Lemma T_If_Intro: forall P e t1 t2 t3,
    |-t {{P && pIsTrue e}} t1 === t3 ->
    |-t {{P && pIsFalse e}} t2 === t3 ->
    |-t {{P}} tm_if e t1 t2 === t3.
  Proof.
    intros.
    apply T_bisim; intros.
    destruct H1 as [wP[wF[?[??]]]].
    destruct (evalTrueFalse_dec e wP) as [[?|?]|[??]].

    setoid_rewrite bisim_Cs_if_true; auto.
    apply H with (F:=F).
    eapply pIsTrue_gstar_up; eauto.
    split.
    exists wP wF; splits; auto.
    split; auto. 
    apply pIsTrue_evaluable; auto.
    eapply pIsTrue_gframe; eauto.
    eapply pIsTrue_gframe; eauto.

    setoid_rewrite bisim_Cs_if_false; auto.
    apply H0 with (F:=F).
    eapply pIsFalse_gstar_up; eauto.
    split.
    exists wP wF; splits; auto.
    split; auto. 
    apply pIsFalse_evaluable; auto.
    eapply pIsFalse_gframe; eauto.
    eapply pIsFalse_gframe; eauto.

    setoid_rewrite bisim_Cs_if_noeval_assert_FF.
    apply H1 with (F:=F).
    exists wP wF; splits; auto.
    split; auto.
    split; auto.
*)

  Lemma T_Loop_While: forall P e (f: iter_schema) (x n: nat),
    logical_loop P e n f ->
    x > 0 ->
    frame_halted_assert P (f x) P ->
    |-t {{P}} f x === tm_ift e (f x) ->
    |-t {{P}} f i_inf === tm_while_max e i_inf (f x).
  Proof.
    introv Hll Hx Hframe Hfif.
    pose (istm_loop_var e 0 (f x)) as g.
    cut ( |-t {{P}} f i_inf === g i_inf).
    intro H.
    eapply H.
    destruct Hll as [?[??]].

    apply T_Loop with (Q:=pIsFalse e) (x:=x) (y:=1) (n:=n) (m:=0); auto.

    simpl.
    setoid_rewrite bisim_Cs_assert_false_while_nop.
    apply T_Reflexivity.

    simpl.
    apply T_While_Post_False.

    intros.
    subst g. unfold term_subst_iter. fold term_subst_iter.
    rewrite i_plus_zero1.
    repeat rewrite iter_expr_minus_zero.
    apply T_Symmetry.
    apply T_While_Split.

    subst g.
    unfold term_subst_iter. fold term_subst_iter.
    repeat rewrite iter_expr_minus_zero.
    apply T_Symmetry.
    eapply T_Transitivity.
    apply T_While_IfT.
    apply T_Symmetry.
    auto.
  Qed.


  (* Proven using T-Loop *)
  Lemma T_While_Assert_True': forall P e t,
    frame_halted_assert P (tm_ift e t) P ->
    |-t {{P}} tm_while e (tm_assert (pIsTrue e);;; t) === tm_while e t.
  Proof.
    intros.
    pose (istm_loop_var e 0 (tm_assert (pIsTrue e);;; t)) as f.
    pose (istm_loop_var e 0 t) as g.
    cut ( |-t {{P}} f i_inf === g i_inf).
    intros.
    eapply H0.

    apply T_Loop with (Q:=pIsFalse e) (x:=1) (y:=1) (n:=0) (m:=0); auto;
      try solve [subst f g; eapply simple_logical_loop; auto].
    simpl.
    setoid_rewrite bisim_Cs_while_one_if.
    setoid_rewrite<- bisim_Cs_if_assert_true; auto.
    simpl.
    eapply T_Transitivity.
    apply T_While_IfT.
    eapply T_Transitivity.
    apply T_Symmetry.
    apply T_If_Assert_True.
    apply T_Symmetry.
    apply T_While_IfT.
  Qed.


  Lemma T_While_Assert_True: forall P e t,
    |-t {{P}} tm_while e (tm_assert (pIsTrue e);;; t) === tm_while e t.
  Proof.
    intros.
    apply T_term_bisim_Cs.
    apply bisim_Cs_while_assert_true.
  Qed.

  Lemma T_If_Split2: forall P e t1 t2 t3 t4,
    (forall v, |-t {{P}} tm_assert (eval e v);;; t1 === t1;;; tm_assert (eval e v)) ->
    (forall v, |-t {{P}} tm_assert (eval e v);;; t2 === t2;;; tm_assert (eval e v)) ->
    |-t {{P}} tm_if e (t1;;;t3) (t2;;;t4) === (tm_if e t1 t2);;; tm_if e t3 t4.
  Proof.
    intros.
    apply T_bisim; intros.
    destruct (evalTrueFalse_dec e w) as [[?|?]|[??]].

    setoid_rewrite bisim_Cs_tm_seq_seq.
    setoid_rewrite bisim_Cs_if_true; auto.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    setoid_rewrite<- bisim_Cs_seq_skip1'.
    destruct e0 as [b[??]].
    setoid_rewrite<- bisim_Cs_assert_skip with (F:=TT) (P:= eval e (nat_value b)).
    setoid_rewrite<- bisim_Cs_tm_seq_seq.
    eapply bisim_seq1_rewrite'.
    eapply H with F; auto.
    setoid_rewrite<- bisim_Cs_tm_seq_seq.
    setoid_rewrite<- bisim_Cs_tm_seq_assoc.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_seq2.
    intros.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_assert_intro.
    intros.
    setoid_rewrite bisim_Cs_if_true.
    apply bisim_refl.
    apply pIsTrue_gstar with TT; eauto.
    eapply gstar_intro with (x:=eval e (nat_value b)); eauto.
    apply exists_intro.
    exists b; eauto.
    apply and_intro; auto.
    apply TT_gstar; auto.
    apply TT_gstar; auto.

    setoid_rewrite bisim_Cs_tm_seq_seq.
    setoid_rewrite bisim_Cs_if_false; auto.
    setoid_rewrite bisim_Cs_tm_seq_seq.
    setoid_rewrite<- bisim_Cs_seq_skip1'.
    setoid_rewrite<- bisim_Cs_assert_skip with (F:=TT) (P:= pIsFalse e).
    setoid_rewrite<- bisim_Cs_tm_seq_seq.
    eapply bisim_seq1_rewrite'.
    eapply H0 with F; auto.
    setoid_rewrite<- bisim_Cs_tm_seq_seq.
    setoid_rewrite<- bisim_Cs_tm_seq_assoc.
    setoid_rewrite bisim_Cs_assert_false_if.
    apply bisim_refl.
    apply TT_gstar; auto.
    apply TT_gstar; auto.

    setoid_rewrite bisim_Cs_tm_seq_seq.
    apply bisim_Cs_bisim.    
    apply bisim_Cs_blocked.
    introv [?]; discriminate.
    introv [?]; discriminate.
    intros.
    inverts H4; contradiction.
    intros.
    inverts H4. inverts H9; contradiction.
  Qed.


  Lemma frame_halted_assert_seq: forall P1 t1 P2 t2 Q,
    frame_halted_assert P1 t1 P2 ->
    frame_halted_assert P2 t2 Q ->
    frame_halted_assert P1 (t1;;; t2) Q.
  Proof.
    intros.
    repeat intro.
    edestruct cstep_star_tm_seq_resolve as [[??]|?]; eauto; subst.
    inverts H4.
    clear H2.
    edestruct cstep_star_seq_inversion as [[s'0[C'0[w'0[??]]]] | [b1''[??]]]; eauto.
    eapply H0; eauto.
    eapply H; eauto.
    inverts H4.
  Qed.

  Lemma T_While_Unroll: forall P e t,
    frame_halted_assert P (tm_ift e (t;;; tm_ift e t)) P ->
    |-t {{P}} tm_while e t === tm_while e (t;;; tm_ift e t).
  Proof.
    intros.
    pose (istm_loop_var e 0 (istm_term t)) as f.
    change (tm_while e t) with (f i_inf).
    setoid_rewrite<- bisim_Cs_while_assert_true at 2.
    setoid_rewrite<- bisim_Cs_assert_true_if with (t2:=tm_skip).
    setoid_rewrite bisim_Cs_while_assert_true.
    cut (tm_ift e (t;;; tm_ift e t) == f 2)%term; intros.
    setoid_rewrite H0.
    apply T_Loop_While with (n:=0); auto.

    eapply simple_logical_loop; subst f; eauto.

    setoid_rewrite<- H0. auto.

    simpl.
    change (iter_const 2) with (iter_plus2 1 1).
    setoid_rewrite<- bisim_Cs_while_split_iter_plus2.
    setoid_rewrite bisim_Cs_while_one_if.
    setoid_rewrite bisim_Cs_ift_split.
    apply T_Seq1.
    setoid_rewrite bisim_Cs_if_assert_true at 2.
    setoid_rewrite bisim_Cs_assert_true_if.
    setoid_rewrite<- bisim_Cs_if_assert_true.
    apply T_Reflexivity.

    simpl.
    change (iter_const 2) with (iter_plus2 1 1).
    setoid_rewrite<- bisim_Cs_while_split_iter_plus2.
    setoid_rewrite bisim_Cs_while_one_if.
    setoid_rewrite bisim_Cs_ift_split.
    reflexivity.
  Qed.
    
  Lemma T_While_Rotation: forall P e t1 t2,
    frame_halted_assert P (tm_ift e (t1;;; tm_ift e t2)) P ->
    |-t {{P}} tm_while e (t1;;; tm_ift e t2) === (tm_ift e t1);;; tm_while e (t2;;; tm_ift e t1).
  Proof.
    intros.
    pose (istm_loop_var e 0 (t1;;; tm_ift e t2)) as f.
    pose (istm_seq (istm_seq (tm_ift e t1) (istm_loop_var e 0 (t2;;; tm_ift e t1))) (tm_ift e t2)) as g.

    cut ( |-t {{P}} f i_inf === g i_inf).
    intros.
    eapply T_Transitivity.
    apply H0.
    simpl.
    setoid_rewrite bisim_Cs_while_inf_post_false' at 1.
    repeat setoid_rewrite<- bisim_Cs_tm_seq_assoc.
    setoid_rewrite bisim_Cs_assert_false_if_nop with (e:=e) (t1:=t2) (t2:=tm_skip).
    setoid_rewrite bisim_Cs_tm_seq_skip2'.
    setoid_rewrite<- bisim_Cs_while_inf_post_false'.
    apply T_Reflexivity.

    eapply T_Loop with (x:=2) (y:=1) (n:=0) (m:=1) (Q:=pIsFalse e); auto.

    simpl.
    change (iter_const 2) with (iter_plus2 1 1).
    setoid_rewrite<- bisim_Cs_while_split_iter_plus2.
    setoid_rewrite bisim_Cs_while_one_if.
    eapply frame_halted_assert_seq; eauto.

    simpl.
    setoid_rewrite bisim_Cs_assert_false_while_nop.
    apply T_Reflexivity.

    simpl.
    setoid_rewrite bisim_Cs_while_inf_post_false' at 2.
    apply T_Reflexivity.

    simpl.
    repeat setoid_rewrite bisim_Cs_tm_seq_assoc.
    setoid_rewrite bisim_Cs_assert_false_if_nop.
    setoid_rewrite bisim_Cs_tm_seq_skip2'.
    setoid_rewrite bisim_Cs_assert_false_while_nop.
    setoid_rewrite bisim_Cs_assert_false_if_nop.
    setoid_rewrite bisim_Cs_tm_seq_skip2'.
    apply T_Reflexivity.

    simpl.
    setoid_rewrite bisim_Cs_while_inf_post_false'.
    repeat setoid_rewrite<- bisim_Cs_tm_seq_assoc.
    setoid_rewrite bisim_Cs_tm_seq_assoc at 3.
    setoid_rewrite bisim_Cs_ift_assert_false_nop.
    repeat setoid_rewrite bisim_Cs_tm_seq_assoc.
    setoid_rewrite<- bisim_Cs_tm_seq_assoc at 2.
    setoid_rewrite<- bisim_Cs_while_inf_post_false'.
    apply T_Reflexivity.

    intros.
    subst f. unfold term_subst_iter.
    repeat rewrite iter_expr_minus_zero.
    apply T_Symmetry.
    rewrite i_plus_zero1.
    apply T_While_Split.

    intros.
    subst g. unfold term_subst_iter.
    repeat rewrite iter_expr_minus_zero.
    repeat setoid_rewrite bisim_Cs_tm_seq_assoc.
    apply T_Seq1.
    setoid_rewrite<- bisim_Cs_while_split_iter_plus2.
    rewrite<- iter_plus2_iterations.
    setoid_rewrite<- bisim_Cs_while_split_iter_plus2.
    setoid_rewrite bisim_Cs_while_one_if.
    setoid_rewrite bisim_Cs_ift_split.
    repeat setoid_rewrite bisim_Cs_tm_seq_assoc.
    apply T_Reflexivity.

    intros.
    subst f g. unfold term_subst_iter.
    repeat rewrite iter_expr_minus_zero.
    change (iter_const 2) with (iter_plus2 1 1).
    setoid_rewrite<- bisim_Cs_while_split_iter_plus2.
    setoid_rewrite bisim_Cs_while_one_if.
    setoid_rewrite bisim_Cs_ift_split.
    repeat setoid_rewrite bisim_Cs_tm_seq_assoc.
    apply T_Reflexivity.
  Qed.

  Lemma T_Assign_Load_Comm: forall (x i:ident) e1 e2,
    x <> i ->
    not_free_in_expr x e1 ->
    |-t {{TT}} tm_assign i e1;;; tm_load x e2 === tm_load x (expr_subst e2 i e1);;; tm_assign i e1.
  Proof.
    intros; apply T_bisim_Cs; intros.
    apply bisim_Cs_assign_load_commute; auto.
  Qed.

  Lemma T_Assign_Store_Comm: forall (x i:ident) e1 e2 e3,
    |-t {{TT}} tm_assign i e1;;; tm_store e2 e3 === tm_store (expr_subst e2 i e1) (expr_subst e3 i e1);;; tm_assign i e1.
  Proof.
    intros; apply T_bisim_Cs; intros.
    apply bisim_Cs_assign_store_commute; auto.
  Qed.

  Lemma Example_Software_Pipelining: forall (a b: ident) (i ai bi: ident),
    ai<>i -> bi<>i -> a<>i -> b<>i ->
    |-t {{TT}}
      tm_while expr_true ((
        ai::= [expr_plus a i];;;
        expr_plus a i<- expr_plus ai 1;;;
        bi::= [expr_plus b i];;;
        ai::= [expr_plus a i];;;
        expr_plus b i<- expr_plus bi ai;;;
        i::= expr_plus i 1))
    ===
      ai::= [expr_plus a i];;;
      expr_plus a i<- expr_plus ai 1;;;
      bi::= [expr_plus b i];;;
      ai::= [expr_plus a i];;;
      expr_plus b i <- expr_plus bi ai;;;
      ai::= [expr_plus a (expr_plus i 1)];;;
      expr_plus a (expr_plus i 1) <- expr_plus ai 1;;;
      tm_while expr_true ((
        bi::= [expr_plus b (expr_plus i 1)];;;
        ai::= [expr_plus a (expr_plus i 1)];;;
        expr_plus b (expr_plus i 1)<- expr_plus bi ai;;;
        ai::= [expr_plus a (expr_plus i 2)];;;
        expr_plus a (expr_plus i 2)<- expr_plus ai 1;;;
        i::= expr_plus i 1)).
  Proof.
    introv Hai Hbi Ha Hb.
    assert (not_free_in_expr i a).
      intros; apply not_free_in_expr_var; auto.
    assert (not_free_in_expr i b).
      intros; apply not_free_in_expr_var; auto.
    assert (not_free_in_expr i ai).
      intros; apply not_free_in_expr_var; auto.
    assert (not_free_in_expr i bi).
      intros; apply not_free_in_expr_var; auto.
    assert (not_free_in_expr i 1).
      intros; apply not_free_in_expr_nat.
    assert (not_free_in_expr ai (expr_plus i 1)).
      intros; apply not_free_in_expr_expr_plus; auto.
      intros; apply not_free_in_expr_var; auto.
    assert (not_free_in_expr bi (expr_plus i 1)).
      intros; apply not_free_in_expr_expr_plus; auto.
      intros; apply not_free_in_expr_var; auto.

    eapply T_Transitivity.
    apply T_While_Unroll.

admit.

    setoid_rewrite bisim_Cs_if_true_const; auto.
    repeat setoid_rewrite bisim_Cs_tm_seq_assoc.
    setoid_rewrite<- bisim_Cs_tm_seq_assoc at 5.
    setoid_rewrite<- bisim_Cs_tm_seq_assoc at 5.
    setoid_rewrite bisim_Cs_tm_seq_assoc at 2.
    setoid_rewrite bisim_Cs_incr_assign_commute with (i:=i) (y:=ai) (x:=a) (n:=1); auto.
    repeat setoid_rewrite bisim_Cs_tm_seq_assoc.

    remember (ai::= [ _ ];;; _;;;_;;;_;;;_;;;_;;;_)%term as t1.
    do 4 setoid_rewrite<- bisim_Cs_tm_seq_assoc at 1.
    do 3 setoid_rewrite bisim_Cs_tm_seq_assoc at 2.
    remember (i::= _;;;_;;;_;;;_;;;_)%term as t2.
    setoid_rewrite<- bisim_Cs_if_true_const with (n:=1) (t2:=tm_skip) at 4; auto.
    unfold expr_true.
    eapply T_Transitivity.
    apply T_While_Rotation.
admit.
    setoid_rewrite bisim_Cs_if_true_const; auto.
    subst t2.
    rewrite Heqt1 at 2.  
Ltac do_nat n:= match n with
  | S ?n' => setoid_rewrite<- bisim_Cs_tm_seq_assoc at 1; do_nat n'
  | _ => idtac
  end.  
Ltac focus_tm_seq n:=
    repeat setoid_rewrite bisim_Cs_tm_seq_assoc;
    do_nat n;
    repeat setoid_rewrite bisim_Cs_tm_seq_assoc at 2.

    focus_tm_seq 7.
    setoid_rewrite bisim_Cs_incr_assign_commute with (i:=i) (y:=ai) (x:=a) (n:=1); auto.
    focus_tm_seq 5.
    setoid_rewrite bisim_Cs_assign_load_commute at 2; auto.
    focus_tm_seq 4.
    setoid_rewrite bisim_Cs_assign_load_commute at 2; auto.
    focus_tm_seq 3.
    setoid_rewrite bisim_Cs_assign_store_commute at 1; auto.
    focus_tm_seq 2.
    setoid_rewrite bisim_Cs_assign_load_commute at 2; auto.
    focus_tm_seq 1.
    setoid_rewrite bisim_Cs_assign_store_commute at 1; auto.

    repeat setoid_rewrite bisim_Cs_tm_seq_assoc.
    setoid_rewrite bisim_Cs_assign_load_commute at 1; auto.
    focus_tm_seq 10.
    setoid_rewrite bisim_Cs_assign_load_commute at 1; auto.
    focus_tm_seq 9.
    setoid_rewrite bisim_Cs_assign_store_commute at 1; auto.
    focus_tm_seq 8.
    setoid_rewrite bisim_Cs_assign_load_commute at 1; auto.
    focus_tm_seq 7.
    setoid_rewrite bisim_Cs_assign_store_commute at 1; auto.

    focus_tm_seq 5.
    remember (bi::= [ expr_subst (b+i) i (i+1) ];;;_;;;_;;;_;;;_;;;i::= (i+1))%term as t2.
    setoid_rewrite<- bisim_Cs_if_true_const with (n:=1) (t2:=tm_skip) at 6; auto.
    eapply T_Transitivity.
    apply T_Seq2 with (P2:=TT); intros.
admit.
    apply T_Symmetry; apply T_While_Unroll.
admit.
    subst t2.
   apply T_term_bisim_Cs; intros.
    apply Morphisms.bisim_Cs_tm_seq_m.
    reflexivity.
    apply Morphisms.bisim_Cs_tm_while_m.
    reflexivity.
    setoid_rewrite expr_subst_plus.
    setoid_rewrite expr_subst_plus.
    setoid_rewrite expr_subst_id; auto.
    setoid_rewrite expr_subst_not_free; auto.
    setoid_rewrite<- (expr_plus_assoc i 1 1).

    setoid_rewrite expr_plus_nat_const.
    change (1+1) with 2.
    reflexivity.
  Qed.

      
  

End SimpleTransformations.


