Require Export JoinLogic.
Require Import Machine.
Require Import ContextSemantics.
Require Import world_model.
Require Import msl.msl_standard.
Require Import SfLib.
Require Import LibTactics.

Module Notation.
  Export JoinLogic.Notation.

  Lemma gjoin_dec'
     : forall w1 w2 : world,
       {w : world | @join _ gworld_sa w1 w2 w} + {(forall w : world, ~ @join _ gworld_sa w1 w2 w)}.
    apply gjoin_dec.
  Qed.

(*  Definition world_sepalgd : sepalgd world := {| JoinLogic.join_dec := join_dec |}.
  Definition gworld_sepalgd := {| JoinLogic.join_dec := gjoin_dec' |}.
  Existing Instance world_sepalgd.
  Existing Instance gworld_sepalgd.*)

  Notation "X (+) Y" := (@Join world world_sa X%join Y%join) (at level 50, left associativity) : join_scope.
  Notation "X (x) Y" := (@Join world gworld_sa X%join Y%join) (at level 50, left associativity) : join_scope.

  Module U.
    Export JoinLogic.Notation.U.
    Notation "X ⨁ Y" := (@Join world world_sa X%join Y%join) (at level 50, left associativity) : join_scope.
    Notation "X ⊗ Y" := (@Join world gworld_sa X%join Y%join) (at level 50, left associativity) : join_scope.
  End U.
End Notation.


Section Definitions.
  Import Machine.Notation.
  Import ContextSemantics.Notation.
  Import Notation.
  Import Notation.U.

  Open Scope join_scope.
  Open Scope context_scope.

  Lemma world_dec_eq: forall (w1 w2:world), {w1=w2}+{w1<>w2}. Admitted.
  Lemma join_wterms_dec': forall b w, {join_wterms b w}+{~join_wterms b w}.
  Proof.
    intro b.
    induction b; intros.
    simpl; apply world_dec_eq.
    simpl; auto.
    destruct (join_wterms_dec b1) as [[w1 Hw1]|].
    destruct (join_wterms_dec b2) as [[w2 Hw2]|].
    destruct (gjoin_dec w1 w2) as [[w12 Hw12] |].
    destruct (world_dec_eq w12 w). subst w12.
    left. exists w1; exists w2; auto.
    right. intro.
    simpl in H.
    destruct H as [w1'[w2'[?[??]]]].
    replace w1' with w1 in *.
    replace w2' with w2 in *.
    replace w12 with w in *.
    intuition.
    eapply @join_eq with (H:=gworld_sa) (x:=w1) (y:=w2); auto.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
    right. intro.
    simpl in H.
    destruct H as [w1'[w2'[?[??]]]].
    replace w1' with w1 in *.
    replace w2' with w2 in *.
    simpl in n. apply n in H1. contradiction.
    eapply join_wterms_eq; eauto.
    eapply join_wterms_eq; eauto.
    right. intro.
    simpl in H.
    destruct H as [w1'[w2'[?[??]]]].
    apply n in H0. contradiction.
    right. intro.
    simpl in H.
    destruct H as [w1'[w2'[?[??]]]].
    apply n in H. contradiction.
  Qed.

  Definition Join_wterm (b:wterm) : JoinPred :=
    Pred (join_wterms b) (join_wterms_dec' b) (join_wterms_dec b) (join_wterms_eq b).

  Lemma join_state_dec': forall s w, {join_state s w}+{~join_state s w}. Admitted.

  Definition Join_state (s:state) : JoinPred :=
    Pred (join_state s) (join_state_dec' s) (join_state_dec s) (join_state_functional s).

  Lemma join_world_list_dec': forall wl w, {join_world_list wl w}+{~join_world_list wl w}. Admitted.
  Definition Join_list (wl:list world) : JoinPred :=
    Pred (join_world_list wl) (join_world_list_dec' wl) (join_world_list_dec wl) (join_world_list_functional wl).

  Definition Join_context C := Join_wterm C[gwemp:tm_skip].

End Definitions.

Definition JWorld := @World world.
Coercion JWorld : world >-> JoinPred.
Coercion Join_context : context >-> JoinPred.
Coercion Join_wterm : wterm >-> JoinPred.
Coercion Join_state : state >-> JoinPred.

Hint Unfold JWorld.

Module Morphisms.
  Export JoinLogic.Morphisms.
End Morphisms.

Section GWorldDefs.
  Import Notation.
  Import Notation.U.
  Import Morphisms.
  Import Machine.Notation.
  Import ContextSemantics.Notation.

  Open Scope join_scope.
  Open Scope context_scope.

  Lemma Join_state_new: forall R s c,
    Join_state (s_new_ch s c R) == Join_state s.
  Proof.
  Admitted.

  Lemma Join_state_pop: forall R w wl q s c,
    s_get_ch s c = Some (R,(w::wl)%list,q) ->
    Join_state (s_pop_ch s c) (x) w == Join_state s.
  Proof.
  Admitted.

  Lemma Join_state_push: forall s c w v,
    Join_state (s_push_ch s c w v) == Join_state s (x) w.
  Proof.
  Admitted.

  Lemma Join_state_rm: forall s c R wl q,
    s_get_ch s c = Some (R,wl,q) ->
    Join_state (s_rm_ch s c) == Join_state s (x) Join_list wl.
  Proof.
  Admitted.


  Lemma Join_wterm_term: forall w t,
    Join_wterm (w:t) == w.
  Proof. split; intros; inverts H; constructor; simpl; eauto. Qed.

  Lemma Join_wterm_seq: forall b t,
    Join_wterm (b;;t) == b.
  Proof. split; intros; inverts H; constructor; eauto. Qed.

  Hint Unfold equiv Join_equiv.

  Lemma Join_wterm_par: forall b1 b2,
    Join_wterm (b1 || b2) == b1 (x) b2.
  Proof.
    intros; eapply Join_equiv_equiv'.
    split; repeat intro; auto; destruct H as [x[y[?[??]]]]; exists x y; auto.
 Qed.

  Lemma Join_cx_hole:
    Join_context cx_hole == gwemp.
  Proof. intros; apply Join_equiv_equiv'; split; eauto. Qed.

  Lemma Join_cx_seq: forall C t,
    Join_context (cx_seq C t) == C.
  Proof. intros; apply Join_equiv_equiv'; split; eauto. Qed.

  Lemma Join_cx_par1: forall C b2,
    Join_context (cx_par1 C b2) == C (x) b2.
  Proof. intros; apply Join_equiv_equiv'; simpl; split; repeat intro; destruct H as [x[y[?[??]]]]; exists x; exists y; auto. Qed.

  Lemma Join_cx_par2: forall b1 C,
    Join_context (cx_par2 b1 C) == b1 (x) C.
  Proof. intros; apply Join_equiv_equiv'; simpl; split; repeat intro; destruct H as [x[y[?[??]]]]; exists x; exists y; auto. Qed.
 
  Hint Resolve gjoin_bot1 gjoin_bot2.

  Lemma Join_gwemp1: forall X, World gwemp (x) X == X.
  Proof.
    intros.
    eapply Join_bot1.
    apply gjoin_bot2.
  Qed.

  Lemma Join_gwemp2: forall X, X (x) World gwemp == X.
  Proof.
    setoid_rewrite Join_com.
    apply Join_gwemp1.
  Qed.


Lemma Join_plus_context: forall C b,
  Join_wterm (C[b]) == C (x) b.
Proof.
  intros.
  unfold Join_context.
  apply Join_equiv_equiv'.
  induction C; 
    unfold Join_context, set_context; fold set_context.
  apply Join_equiv_equiv'; rewrite Join_cx_hole.
  setoid_rewrite Join_gwemp1.
  reflexivity.
  eauto.
  apply Join_equiv_equiv'.
  setoid_rewrite (Join_wterm_par).
  apply Join_equiv_equiv' in IHC.
  rewrite IHC.
  pattern (b (x) w); setoid_rewrite Join_swap1 at 1.
  reflexivity.
  apply Join_equiv_equiv' in IHC.
  apply Join_equiv_equiv'.
  setoid_rewrite (Join_wterm_par).
  rewrite IHC.
  pattern (b (x) w); setoid_rewrite Join_assoc.
  reflexivity.
Qed.

Lemma Join_compose_contexts: forall C1 C2,
  Join_context (compose_contexts C1 C2) == C1 (x) C2.
Proof.
  intros.
  unfold Join_context at 1.
  rewrite<- context_compose_correct.
  do 2 rewrite Join_plus_context.
  rewrite Join_wterm_term.
  rewrite Join_gwemp2.
  reflexivity.
Qed.

Lemma Join_join_wterms_equiv: forall b w,
  join_wterms b w <-> Join_wterm b == w.
Proof.
  split; intros.
  apply Join_equiv_equiv'.
  split; intros; simpl in *.
  eapply join_wterms_eq; eauto.
  subst; auto.
  apply Join_equiv_equiv' in H.
  destruct (H w).
  simpl in *.
  apply H1.
  auto.
Qed.

Lemma Join_prod_incr: forall c (x z:world) Y,
  can_send c x ->
  World z == x (x) Y ->
  World (w_prod_incr c z) == (w_prod_incr c x) (x) Y.
Proof.
  intros.
  apply Forces_equiv in H0.
  apply Forces_equiv.
  inverts H0.
  inverts H4.
  apply Forces_Forces'_equiv.
  exists (w_prod_incr c x); exists wy; splits.
  destruct H6 as [[?][?]].
  destruct z as [z]; destruct x as [w1]; destruct wy as [w2].
  unfold w_prod in *.
  unfold w_prod_incr, snd, fst in *.
  split; split; intuition.
  simpl.
  apply join_channels_incr1; auto.
  reflexivity.
  apply Forces_Forces'_equiv; auto.
Qed.

Lemma Join_cons_incr: forall c (x z:world) Y,
  can_recv c x ->
  World z == x (x) Y ->
  World (w_cons_incr c z) == (w_cons_incr c x) (x) Y.
Proof.
  intros.
  apply Forces_equiv in H0.
  apply Forces_equiv.
  inverts H0.
  inverts H4.
  apply Forces_Forces'_equiv.
  exists (w_cons_incr c x); exists wy; splits.
  destruct H6 as [[?][?]].
  destruct z as [z]; destruct x as [w1]; destruct wy as [w2].
  unfold w_prod in *.
  unfold w_prod_incr, snd, fst in *.
  split; split; intuition.
  simpl.
  apply join_channels_incr1; auto.
  reflexivity.
  apply Forces_Forces'_equiv; auto.
Qed.

  Lemma Join_upd_var: forall i v (x:world) Y z,
    LocalsMap.map_share i (w_env x) = Share.top ->
    World z == x (x) Y ->
    World (w_env_set z i v) == (w_env_set x i v) (x) Y.
  Proof.
    intros.
  apply Forces_equiv in H0.
  apply Forces_equiv.
    inverts H0.
    inverts H4.
    apply Forces_Forces'_equiv.
    exists (w_env_set x i v); exists wy; splits.
    apply gjoin_w_env_set'; auto.
    reflexivity.
    apply Forces_Forces'_equiv; auto.
  Qed.

  Program Definition pWorld (X: JoinPred) : pred world := exist _ (fun w => (Forces w X)%pred) _.


  Lemma Join_gprecise1: forall (P1 : pred world) (w w1 w1' : world) X Y,
    gprecise P1 ->
    P1 w1 -> P1 w1' -> World w == World w1 (x) X -> World w == World w1' (x) Y -> X == Y /\ w1'=w1.
  Proof.
    intros.
    cut (X==Y); intros.
    split; auto.
    rewrite H4 in H2.
    assert (World w1 == World w1').
      eapply @Join_canc2 with (sa:=gworld_sa); eauto.
    apply Forces_equiv in H5.
    inverts H5; auto.
    edestruct Join_split with (sa:=gworld_sa) as [w1_[wy[??]]]. apply H3.
    apply Forces_equiv in H4. inverts H4.
    edestruct Join_split with (sa:=gworld_sa) as [w1_[wx[??]]]. apply H2.
    apply Forces_equiv in H4. inverts H4.
    rewrite<- H5 in H3; rewrite<- H6 in H2.
    rewrite<- H5; rewrite<- H6.
    clear H5 H6.
    cut (wx=wy); intros.
      subst; reflexivity.
    eapply gprecise_star_eq with (w1:=w1) (w1':=w1') (w2:=wx) (w2':=wy); eauto.
    apply Join_join; eauto.
    apply Join_join; eauto.
  Qed.

  Lemma Joins_gprecise1: forall (P1 : pred world) (w1 w1' : world) X Y,
    gprecise P1 ->
    P1 w1 -> P1 w1' -> {[World w1 (x) X == World w1' (x) Y]} -> {[X == Y]} /\ w1'=w1.
  Proof.
    intros. edestruct H2 as [w[??]].
    edestruct Join_gprecise1 with (w1:=w1) (w1':=w1'); eauto.
    apply Forces_equiv; eauto.
    apply Forces_equiv; eauto.
    split; auto.
    setoid_rewrite H5.
    inverts H4.
    exists wy; auto.
  Qed.

(*
  Program Fixpoint world_prod_incr c (X: World) : World := match X with
    | world_plus X Y => if world_prod_incr c X ***FIX***
    | world_just a => world_just (w_prod_incr c a)
    | world_prop f f_dec f_sat f_fun => world_prop (fun w=> f (w_prod_incr c w)) _ _ _
    end.
  Next Obligation. apply f_dec. Qed.
  Next Obligation.
    destruct f_sat as [[w Hf]|].
    destruct (f_dec (w_prod_incr c w)); eauto.
    destruct (world_dec_eq w (w_prod_incr c w)).
    rewrite e in Hf. contradiction.
    destruct (w_prod_incr_eq_dec c w) as [[w' Hw'] | ].
    left. exists w'; subst; auto.
    right. repeat intro. apply (n1 w0).
    apply f_fun; auto.
    right; auto.
  Qed.
  Next Obligation.
    apply w_prod_incr_fun with c; apply f_fun; auto.
  Qed.
*)


Open Scope context_scope.


End GWorldDefs.


Module Tactics.
  Require Import SetoidClass.
  Export JoinLogic.Tactics.

    Tactic Notation "simplw" :=
      repeat (
          (try setoid_rewrite Join_state_push);
          (try setoid_rewrite Join_state_new)
      );
      (try setoid_rewrite Join_plus_context);
      repeat (
          (try setoid_rewrite Join_cx_hole);
          (try setoid_rewrite Join_cx_seq);
          (try setoid_rewrite Join_cx_par1);
          (try setoid_rewrite Join_cx_par2);
          (try setoid_rewrite Join_compose_contexts)
      ); repeat (
          (try setoid_rewrite Join_wterm_term);
          (try setoid_rewrite Join_wterm_seq);
          (try setoid_rewrite Join_wterm_par);
          (try setoid_rewrite Join_gwemp1);
          (try setoid_rewrite Join_gwemp2)
      );
      repeat (setoid_rewrite Join_assoc).

  Tactic Notation "simplw" "in" hyp(H) :=
      repeat (
          (try setoid_rewrite Join_state_push in H);
          (try setoid_rewrite Join_state_new in H)
      );
      (try setoid_rewrite Join_plus_context in H);
      repeat (
          (try setoid_rewrite Join_cx_hole in H);
          (try setoid_rewrite Join_cx_seq in H);
          (try setoid_rewrite Join_cx_par1 in H);
          (try setoid_rewrite Join_cx_par2 in H);
          (try setoid_rewrite Join_compose_contexts in H)
      ); repeat (
          (try setoid_rewrite Join_wterm_term in H);
          (try setoid_rewrite Join_wterm_seq in H);
          (try setoid_rewrite Join_wterm_par in H);
          (try setoid_rewrite Join_gwemp1 in H);
          (try setoid_rewrite Join_gwemp2 in H)
      );
      repeat (setoid_rewrite Join_assoc in H).

  Tactic Notation "solvew" :=
    match goal with
      | [  |- equiv ?X _ ] => simplw; repeat elimw; solve [reflexivity]
    end.

  Ltac substw_exact w :=
    match goal with
    | [H : World w == World ?w2 |- _] => match w2 with
         | w => fail | _ => rewrite H in * end
    | [H : World ?w1 == World w |- _] => match w1 with
         | w => fail | _ => rewrite<- H in * end
    | [H : World w == ?X |- _] => rewrite H in *
    | [H : ?X == World w |- _] => rewrite<- H in *

    | [H : Join_wterm w == Join_wterm ?w2 |- _] => match w2 with
         | w => fail | _ => rewrite H in * end
    | [H : Join_wterm ?w1 == Join_wterm w |- _] => match w1 with
         | w => fail | _ => rewrite<- H in * end
    | [H : Join_wterm w == ?X |- _] => rewrite H in *
    | [H : ?X == Join_wterm w |- _] => rewrite<- H in *
  end.

  Tactic Notation "substw":=
    try (match reverse goal with
      | [H : World _ == Join _ _ |- _] => rewrite H in *
      | [H : World _ == Join_wterm _ |- _] => rewrite H in *
      | [H : World ?w == World ?w2 |- _] => match w2 with
           | w => fail
           | _ => rewrite H in *
         end; fail
    end).

  Tactic Notation "substw" constr(w):=
    match goal with
      | [H : World w == Join _ _ |- _] => rewrite H in *
      | [H : World w == Join_wterm _ |- _] => rewrite H in *
      | [H : World w == World ?w2 |- _] => match w2 with
          | w => fail | _ => rewrite H in * end
    end.

  Module Tests.
    Import Machine.Notation.
    Import ContextSemantics.Notation.
    Import Notation Notation.U.
    Open Scope join_scope.
    Open Scope context_scope.
    Open Scope wterm_scope.

  Tactic Notation "elimw" :=
  let do_elim w := patternw w; progress (apply Morphisms.Join_m; try reflexivity)
  in let rec find_elim Z := match Z with
(*    | Join_context ?w => do_elim w
    | Join_wterm ?w => do_elim w
    | Join_state ?w => do_elim w
    | World ?w => do_elim w
    | JGWorld ?w => do_elim w *)
    | Join _ ?X ?Y => first [find_elim X | find_elim Y]
    | ?w => do_elim w
    end
  in
  match goal with
  | [  |- equiv ?X  _ ] => find_elim X
  end.

    Lemma test_simplw1: forall {A} {sa:sepalg A} X (C:context) (b b1 b2 b3: wterm) s c w1 w2 v R fch,
      X (x) s_push_ch (s_new_ch (s_push_ch s c w1 v) fch R) c w2 v (x) (cx_par1 C b3)[b] (x) (b1||b2) == X (x) s (x) w1 (x) w2 (x) C (x) b3 (x) b (x) b1 (x) b2.
    Proof. intros; simplw. reflexivity. Qed.
    Lemma test_simplw2: forall {A} {sa:sepalg A} X (C:context) (b b1 b2 b3: wterm) s c w1 w2 v R w fch,
      w ||= X (x) s_push_ch (s_new_ch (s_push_ch s c w1 v) fch R) c w2 v (x) (cx_par1 C b3)[b] (x) (b1||b2) -> w||= X (x) s (x) w1 (x) w2 (x) C (x) b3 (x) b (x) b1 (x) b2.
    Proof. intros; simplw in H. exact H. Qed.
    Lemma test_solvew1: forall {A} {sa:sepalg A} X (C:context) (b b1 b2 b3: wterm) (s:state) c (w1 w2:world) v R fch,
      s (x) w1 (x) (cx_par1 C b3)[b] (x) X (x) w2 (x) (b1||b2) == X (x) s_push_ch (s_new_ch (s_push_ch s c w1 v) fch R) c w2 v (x) C (x) b3 (x) b (x) b1 (x) b2.
    Proof. intros. solvew. Qed.

  End Tests.

End Tactics.
