Require Import msl.base.
Require Import msl.sepalg.
Require Import msl.sepalg_generators.
Require Import msl.shares.

Set Implicit Arguments.

Section SepAlgChannelEndpoint.

  Let t := Share.t.
  Let ja := Share.sa.
  Let P (pn: t*nat) := identity (fst pn) -> (snd pn)=0.

  (* If the share is bot, then the send/recv count through hte channel must be 0 *)
  Definition ch_endpoint := { pn:t*nat | P pn}.

  Definition ce_share (ce : ch_endpoint) : Share.t :=
    let (x, _) := ce in
    let (p,_) := x in p.
  Definition ce_count (ce : ch_endpoint) : nat :=
    let (x, e) := ce in
    let (_,n) := x in n.
  Definition ce_pair (ce : ch_endpoint) : (Share.t * nat) :=
    let (x, e) := ce in x.

  Program Definition ce_make (p : t) (n : nat) (Hni : identity p -> n=0)
     : ch_endpoint := (p,n).
  Program Definition ce_make_top (n : nat)
     : ch_endpoint := (Share.top,n).
  Next Obligation.
    repeat intro; elimtype False; apply top_share_nonidentity; auto.
  Qed.

  Definition join_ce' (x y z: t*nat) :=
    join (fst x) (fst y) (fst z) /\ (snd x)+(snd y)=(snd z).
  Definition join_ce (x y z: { a:t*nat | P a }) :=
    join_ce' (projT1 x) (projT1 y) (projT1 z).

  Lemma saf_ce : sepalgfacts join_ce.
  Proof.
    apply SepAlgFacts.

    (* join_eq *)
    intros.
    unfold join_ce, join_ce' in *.
    destruct x; destruct y; destruct z; destruct z'; simpl in *.
    destruct H; destruct H0.
    revert p2.
    replace x2 with x1; intros.
    replace p2 with p1; auto.
    apply proof_irr.
    destruct x1; destruct x2; simpl in *.
    f_equal; subst; auto.
    eapply join_eq; eauto.

    (* join_assoc *)
    intros. unfold join_ce, join_ce' in *.
    destruct H; destruct H0.
    destruct a; destruct b; destruct c; destruct d; destruct e.
    simpl in *.
    destruct (join_assoc H H0) as [f [? ?]].
    assert (H5: P (f, snd x0 + snd x1)).
    unfold P; simpl; intros.
    assert (identity (fst x0)).
      eapply split_identity; eauto.
    assert (identity (fst x1)).
      eapply split_identity. eapply (join_com); eauto. auto.
    apply p0 in H6.
    apply p1 in H7.
    omega.
    exists (existT (fun x => P x) (f, snd x0 + snd x1) H5).
    simpl. do 2 (split; auto). omega.

    (* join_com *)
    intros. destruct H.
    split. eapply join_com; auto. omega.

    (* join_canc *)
    intros.
    destruct a1; destruct a2; destruct b; destruct c.
    destruct H; destruct H0; simpl in *.
    revert p0.
    replace x0 with x; intros.
    replace p0 with p; auto.
    apply proof_irr.
    destruct x; destruct x0.
    replace t0 with t1. replace n0 with n. trivial.
    simpl in *. omega.
    eapply join_canc; eauto.
    
    (* join_ex_identities *)
    intros.
    destruct a as [[p n] Hpn].
    destruct (join_ex_identities p) as [e He].
    assert (H: P (e, 0)) by (unfold P; auto).
    exists (existT (fun x => P x) (e,0) H).
    red; simpl; split. destruct He; apply identity_unit; simpl; auto.
      simpl; auto.

    (* join_self *)
    intros.
    destruct a; destruct b.
    destruct H; simpl in *.
    revert p0.
    replace x0 with x; intros.
    replace p0 with p; auto.
    apply proof_irr.
    destruct x; destruct x0; simpl in *.
    replace n0 with n. replace t1 with t0. trivial.
    apply join_self; auto.
    assert (identity t0).
      apply unit_identity with t0.
      unfold unit_for.
      replace t1 with t0 in *; auto.
      eapply join_self; eauto.
    apply p in H1. simpl in H1.
    subst; auto.
  Qed.

  Definition sa_ce : sepalg { a:t*nat | P a } := SepAlg saf_ce.

  Lemma join_ce_join: forall a b c, join_ce a b c <-> @join ch_endpoint sa_ce a b c.
  Proof.
    unfold join, sa_ce; destruct saf_ce; split; simpl; auto.
  Qed.

  Existing Instance sa_ce.

  Definition ce_unit: ch_endpoint :=
    existT (fun pn => P pn) (Share.bot, 0)
    (fun _ => eq_refl 0).
  Lemma ce_unit_identity: identity ce_unit.
    apply unit_identity with ce_unit.
    unfold unit_for.
    apply join_ce_join.
    split; simpl; auto.
    generalize bot_identity; intro.
    rewrite identity_unit_equiv in H.
    unfold unit_for in H. auto.
  Qed.


End SepAlgChannelEndpoint.

Section SepAlgChannel.
  Variable cname:Type.

  Definition channels := cname -> option (pjoinable sa_ce).
  Definition channels_unit : channels := fun d=>None.

  Definition sa_channels : sepalg channels := sa_fun _ (sa_lift _).

  Existing Instance sa_channels.

  Lemma channels_unit_identity: identity channels_unit.
  Proof.
    apply unit_identity with channels_unit.
    unfold unit_for.
    unfold join, sa_channels, sa_lift; destruct saf_lift.
    simpl in *.
    unfold m_join. simpl.
    intros.
    apply lj_None1.
  Qed.


End SepAlgChannel.


