(*
 * Copyright (c) 2009-2010, Andrew Appel, Robert Dockins and Aquinas Hobor.
 *
 *)

Require Import base.
Require Import ageable.
Require Import knot.

Module Knot_Lemmas (K : KNOT).
  Import K.TF.
  Import K.

  Lemma unsquash_inj : forall k1 k2,
    unsquash k1 = unsquash k2 ->
    k1 = k2.
  Proof.
    intros.
    rewrite <- (squash_unsquash k1).
    rewrite <- (squash_unsquash k2).
    rewrite H.
    trivial.
  Qed.
  Implicit Arguments unsquash_inj.
  
  Lemma squash_surj : forall k, exists n, exists Fp,
    squash (n, Fp) = k.
  Proof.
  intros.
  remember (unsquash k).
  destruct p.
  exists n.
  exists f.
  rewrite Heqp.
  rewrite squash_unsquash.
  trivial.
  Qed.

  Lemma unsquash_approx : forall k n Fp,
    unsquash k = (n, Fp) ->
    Fp = fmap (approx n) Fp.
  Proof.
    intros.
    generalize H; intro.
    rewrite <- (squash_unsquash k) in H.
    rewrite H0 in H.
    rewrite unsquash_squash in H.
    inversion H.
    rewrite H2.
    symmetry.
    trivial.
  Qed.
  Implicit Arguments unsquash_approx.
  
  Lemma approx_approx1 : forall m n,
    approx n = approx n oo approx (m+n).
  Proof.
    intros.
    extensionality p x; destruct x as [k o].
    unfold approx, compose; simpl.
    destruct (le_gt_dec n (level k)); auto.
    destruct (le_gt_dec (m+n) (level k)); auto.
    elimtype False; omega.
  Qed.

  Lemma approx_approx2 : forall m n,
    approx n = approx (m+n) oo approx n.
  Proof.
    intros.
    extensionality p x; destruct x as [k o].
    unfold approx, compose; simpl.
    destruct (le_gt_dec (m+n) (level k)); auto.
    destruct (le_gt_dec n (level k)); auto.
    elimtype False; omega.
  Qed.

  (* These are provided since sometimes it is tedious to break things out;
      they are not interesting except as engineering artifacts. *)
  Lemma unsquash_squash_unfolded : forall nf,
    unsquash (squash nf) = (fst nf, fmap (approx (fst nf)) (snd nf)).
  Proof.
    intros.
    destruct nf.
    apply unsquash_squash.
  Qed.
	
  Lemma unsquash_approx_unfolded : forall k,
    unsquash k = (fst (unsquash k), fmap (approx (fst (unsquash k))) (snd (unsquash k))).
  Proof.
    intros.
    case_eq (unsquash k); intros.
    simpl.
    apply injective_projections; simpl; trivial.
    apply (unsquash_approx H).
  Qed.

(*
  Lemma unsquash_not_surj : 
    (exists rbot : rhs, rbot <> rhs_top) ->
    (exists Fp : F predicate, True) ->
    forall n, exists Fp, forall k, unsquash k <> (n, Fp).
  Proof.
  intros.
  destruct H as [bot ?].
  destruct H0 as [anF _].
  remember (fun (p : predicate) (w : knot * other) => bot) as badf.
  remember (fmap badf anF) as badF.
  exists badF.
  repeat intro.
  generalize (unsquash_approx H0); intro.
  rewrite HeqbadF in H1.
  replace (fmap (approx n) (fmap badf anF)) with
    ((fmap (approx n) oo fmap badf) anF) in H1 by trivial.
  rewrite fmap_comp in H1.
  XXXXX  (* Gah, annoying *)
*)

End Knot_Lemmas.
