(*
 * Copyright (c) 2009-2010, Andrew Appel, Robert Dockins and Aquinas Hobor.
 *
 *)

Unset Automatic Introduction.
Generalizable All Variables.

Require Import base.
Require Import ageable.
Require Import sepalg.
Require Import sepalg_generators.
Require Import sepalg_functors.
Require Import age_sepalg.

Require Import predicates_hered.
Require Import subtypes.

Import predicates_hered.Notation.
Local Open Scope pred.
Local Open Scope entails_pred.

Definition compareR `{ASA A} : relation A := @comparable A as_alg.
Definition extendR  `{ASA A} : relation A := join_sub.

Lemma valid_rel_compare `{ASA A} : valid_rel compareR.
Proof.
  intros A HA XA XA'.
  split; hnf; intros.

  apply comparable_common_unit in H0.
  destruct H0 as [w [? ?]].
  destruct (age1_join2  _ H1 H)
    as [u [v [? [? ?]]]].
  destruct (age1_join _ H0 H3)
    as [u' [v' [? [? ?]]]].
  assert (u' = v').
  unfold age in *; congruence.
  subst v'.
  exists u'; auto.
  assert (x = v).
  unfold age in *; congruence.
  subst v.
  apply common_unit_comparable.
  exists u; auto.

  apply comparable_common_unit in H.
  destruct H as [w [? ?]].
  destruct (unage_join2 _ H H0)
    as [u [v [? [? ?]]]].
  destruct (unage_join _ H1 H3)
    as [u' [v' [? [? ?]]]].
  exists v'; auto.
  apply common_unit_comparable.
  destruct (join_ex_units u) as [uu Huu].
  red in Huu.
  exists uu; split.
  destruct (join_assoc Huu H2) as [q [? ?]].
  assert (q = z).
  eapply join_eq; eauto.
  subst q; auto.
  destruct (join_assoc Huu H5) as [q [? ?]].
  assert (q = v').
  eapply join_eq; eauto.
  subst q.
  auto.
Qed.

Lemma valid_rel_extend `{ASA A} : valid_rel extendR.
Proof.
  intros A HA XA XA'; intros; split; hnf; intros.
  destruct H0 as [w ?].
  destruct (age1_join2 _ H0 H)
    as [u [v [? [? ?]]]].
  exists u; auto.
  exists v; auto.

  destruct H.
  destruct (unage_join _ H H0)
    as [u [v [? [? ?]]]].
  exists v; auto.
  exists u; auto.
Qed.

Definition compareM `{ASA A} : modality
  := exist _ compareR valid_rel_compare.
Definition extendM `{ASA A} : modality
  := exist _ extendR valid_rel_extend.

(* Definitions of the BI connectives. *)

Program Definition emp `{ASA A} : pred A := identity.
Next Obligation.
  rewrite identity_unit_equiv.
  rewrite identity_unit_equiv in H1.
  unfold unit_for in *.
  destruct (age1_join _ H1 H0) as [y' [z' [? [? ?]]]].
  assert (a' = y' /\ a' = z') by (unfold age in *; split; congruence).
  destruct H5; subst; subst.
  auto.
Qed.

Program Definition sep_con `{ASA A} (p q:pred A) : pred A := fun x:A =>
  exists y:A, exists z:A, join y z x /\ p y /\ q z.
Next Obligation.
  destruct H1 as [y [z [? [? ?]]]].
  destruct (age1_join2 _ H1 H0) as [w [v [? [? ?]]]].
  exists w; exists v; split; auto.
  split.
  apply pred_hereditary with y; auto.
  apply pred_hereditary with z; auto.
Qed.

Program Definition wand `{ASA A} (p q:pred A) : pred A := fun x =>
  forall x' y z, necR x x' -> join x' y z -> p y -> q z.
Next Obligation.
  apply H1 with x' y; auto.
  apply rt_trans with a'; auto.
  apply rt_step; auto.
Qed.

Module Notation.
  Export predicates_hered.Notation.
  Notation "P '*' Q" := (sep_con P%pred Q%pred) : pred.
  Notation "P '-*' Q" := (wand P%pred Q%pred) (at level 60, right associativity) : pred.
  Notation "'%' e"  := (box extendM e)(at level 30, right associativity): pred.
End Notation.
Import Notation.

Lemma extendM_refl `{ASA A}: reflexive _ extendM.
Proof.
intros; intro; simpl; apply join_sub_refl.
Qed.

Lemma compareM_refl `{ASA A}: reflexive _ compareM.
Proof.
intros; intro; simpl.
apply comparable_refl.
Qed.

Hint Resolve @extendM_refl.
Hint Resolve @compareM_refl.


(* Rules for the BI connectives *)

Lemma wand_con_adjoint `{ASA A} : forall (P Q R:pred A),
  ((P * Q) |-- R)%entails_pred = (P |-- (Q -* R)).
Proof.
  intros. apply prop_ext.
  split; intros.
  hnf; intros; simpl; intros.
  apply H0.
  exists x'; exists y.
  intuition.
  apply pred_nec_hereditary with a; auto.
  hnf; intros.
  hnf in H0.
  unfold wand in H0; simpl in H0.
  destruct H1 as [w [v [? [? ?]]]].
  eapply H0; eauto.
Qed.

Lemma con_assoc `{ASA A} : forall (P Q R:pred A),
  (P * Q) * R = P * (Q * R).
Proof.
  intros; apply equiv_eq; hnf; intros.
  destruct H0 as [x [y [? [? ?]]]].
  destruct H1 as [z [w [? [? ?]]]].
  destruct (join_assoc H1 H0) as [q [? ?]].
  exists z; exists q; intuition.
  exists w; exists y; intuition.
  destruct H0 as [x [y [? [? ?]]]].
  destruct H2 as [z [w [? [? ?]]]].
  apply join_com in H0.
  apply join_com in H2.
  destruct (join_assoc H2 H0) as [q [? ?]].
  exists q; exists w; intuition.
  exists x; exists z; intuition.
Qed.

Lemma con_comm `{ASA A} : forall (P Q:pred A),
  P * Q = Q * P.
Proof.
  intros; apply equiv_eq; hnf; intros.
  destruct H0 as [x [y [? [? ?]]]].
  exists y; exists x; intuition; apply join_com; auto.
  destruct H0 as [x [y [? [? ?]]]].
  exists y; exists x; intuition; apply join_com; auto.
Qed.

Lemma split_con `{ASA A} : forall (P Q R S:pred A),
  P |-- Q ->
  R |-- S ->
  (P * R) |-- (Q * S).
Proof.
  intros; hnf; intros.
  destruct H2 as [x [y [? [? ?]]]].
  exists x; exists y; intuition.
Qed.

Lemma con_cut `{ASA A} : forall (P Q R S:pred A),
  P |-- (Q -* R) ->
  S |-- Q ->
  (P * S) |-- R.
Proof.
  intros.
  rewrite wand_con_adjoint.
  hnf; intros.
  simpl; intros.
  eapply H0; eauto.
Qed.

Lemma con_emp1 `{ASA A} : forall (P:pred A),
  emp * P = P.
Proof.
  intros; apply equiv_eq; hnf; intros.
  destruct H0 as [x [y [? [? ?]]]].
  simpl in H1.
  replace a with y; auto.
  destruct (join_ex_units a) as [u ?].
  exists u; exists a; intuition.
  simpl.
  apply unit_identity with a; auto.
Qed.

Lemma con_emp2 `{ASA A} : forall (P:pred A),
  P * emp = P.
Proof.
  intros.
  rewrite con_comm.
  apply con_emp1.
Qed.


Lemma later_wand `{ASA A} : forall P Q,
  |>(P -* Q) = |>P -* |>Q.
Proof.
  intros.
  repeat rewrite later_age.
  apply equiv_eq; hnf; intros.
  simpl; intros.
  simpl in H0.
  case_eq (age1 a); intros. 
  spec H0 a0 H5.
  apply nec_refl_or_later in H1.
  destruct H1; subst.
  destruct (age1_join2 _ H2 H4) as [w [v [? [? ?]]]].
  eapply H0; eauto.
  replace a0 with w; auto.
  congruence.
  assert (necR a0 x').
  eapply age_later_nec; eauto.
  destruct (age1_join2 _ H2 H4) as [w [v [? [? ?]]]].
  apply H0 with w v; auto.
  apply rt_trans with x'; auto.
  apply rt_step; auto.
  apply nec_refl_or_later in H1; destruct H1; subst.
  destruct (age1_join2 _  H2 H4) as [w [v [? [? ?]]]].
  hnf in H6.
  rewrite H5 in H6; discriminate.
  clear -H1 H5.
  elimtype False.
  revert H5; induction H1; auto.
  intros.
  unfold age in H.
  rewrite H in H5; discriminate.

  simpl; intros.
  simpl in H0.
  destruct (valid_rel_nec).
  destruct (H6 _ _ H2 _ H1).
  destruct (unage_join _ H3 H7) as [w [v [? [? ?]]]].
  apply H0 with x w v; auto.
  intros.
  replace a'0 with y; auto.
  congruence.
Qed.

Lemma later_con `{ASA A} : forall P Q,
  |>(P * Q) = |>P * |>Q.
Proof.
  intros.
  repeat rewrite later_age.
  apply equiv_eq; hnf; intros.
  simpl in H0.
  case_eq (age1 a); intros.
  destruct (H0 a0) as [w [v [? [? ?]]]]; auto.
  destruct (unage_join2 _ H2 H1) as [w' [v' [? [? ?]]]].
  exists w'; exists v'; intuition.
  simpl; intros.
  replace a' with w; auto.
  unfold age in *; congruence.
  simpl; intros.
  replace a' with v; auto.
  unfold age in *; congruence.
  destruct (join_ex_units a).
  exists x; exists a.
  intuition.
  hnf; intros.
  red in u.
  simpl in H2.
  destruct (age1_join _ u H2) as [s [t [? [? ?]]]].
  unfold age in H5.
  rewrite H1 in H5; discriminate.
  hnf; intros.
  simpl in H2.
  unfold age in H2.
  rewrite H1 in H2; discriminate.
  
  destruct H0 as [w [v [? [? ?]]]].
  hnf; intros.
  simpl in H3.
  destruct (age1_join2 _ H0 H3) as [w' [v' [? [? ?]]]].
  exists w'; exists v'; intuition.
Qed.


Lemma sub_con `{ASA A} : forall G P P' Q Q',
  G |-- P >=> P' ->
  G |-- Q >=> Q' ->
  G |-- P * Q >=> P' * Q'.
Proof.
  repeat intro.
  destruct H5 as [w1 [w2 [? [? ?]]]].
  exists w1; exists w2; split; auto.
  split.
  spec H0 a H2.
  rewrite <- (box_refl_trans fashionM) in H0.
  spec H0 a' H3.
  eapply pred_nec_hereditary in H0; eauto.
  eapply H0; eauto.
  simpl; apply comparable_fashionR.
  apply comparable_sym.

  eapply join_sub_comparable; eauto.
  simpl; apply fashionR_refl.
  simpl; exact fashionR_trans.
  spec H1 a H2.
  rewrite <- (box_refl_trans fashionM) in H1.
  spec H1 a' H3.
  eapply pred_nec_hereditary in H1; eauto.
  eapply H1; eauto.
  simpl; apply comparable_fashionR.
  apply comparable_sym.
  eapply join_sub_comparable; eauto.
  simpl; apply fashionR_refl.
  simpl; exact fashionR_trans.
Qed.

Lemma sub_wand `{ASA A} : forall G P P' Q Q',
  G |-- P' >=> P ->
  G |-- Q >=> Q' ->
  G |-- (P -* Q) >=> (P' -* Q').
Proof.
  repeat intro.
  spec H1 a H2.
  rewrite <- (box_refl_trans fashionM) in H1.
  spec H1 a' H3.
  eapply pred_nec_hereditary in H1; eauto.
  eapply pred_nec_hereditary in H1; eauto.
  eapply H1; eauto.
  simpl; apply comparable_fashionR.
  eapply join_sub_comparable; eauto.
  eapply H5; eauto.
  spec H0 a H2.
  rewrite <- (box_refl_trans fashionM) in H0.
  spec H0 a' H3.
  eapply pred_nec_hereditary in H0; eauto.
  eapply pred_nec_hereditary in H0; eauto.
  eapply H0; eauto.
  simpl; apply comparable_fashionR.
  apply joins_comparable; eauto.
  simpl; apply fashionR_refl.
  simpl; exact fashionR_trans.
  simpl; apply fashionR_refl.
  simpl; exact fashionR_trans.
Qed.


Lemma FF_con : forall `{ASA A} (P:pred A),
  FF * P = FF.
Proof.
  intros. apply equiv_eq; repeat intro.
  destruct H0 as [? [? [? [? ?]]]].  elim H1.
  elim H0.
Qed.

Lemma ex_con : forall `{ASA A} B (P:B->pred A) Q,
  (existential P) * Q = Ex_ x:B, P x * Q.
Proof.
  intros. apply equiv_eq.
  repeat intro.
  destruct H0 as [w1 [w2 [? [? ?]]]].
  destruct H1 as [x ?]. exists x.
  exists w1. exists w2. intuition.
  repeat intro.
  destruct H0. destruct H0 as [w1 [w2 [? [? ?]]]].
  exists w1. exists w2; intuition.
  exists x; auto.
Qed.
